/**
 * # Step Indicator
 * Set of indicators represent different steps
 *
 * @example
 * new StepIndicator(el, params);
 *
 * @module components/step-indicator.js
 */

(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/outer-height'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/outer-height'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.StepIndicator = factory(root.Spark.BaseComponent, root.Spark.outerHeight);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * StepIndicator constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var StepIndicator = function(el, params) {

    params = params || {};

    if (!el) {
      return;
    }
    this._setParams(this.defaults, true);
    this._setParams(params);
    this._cacheElements(el);
    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._calculateStyle();
  };

  StepIndicator.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _toggleClass: Base.toggleClass,
    _getElementOffset: Base.getElementOffset,
    remove: Base.remove,


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      containerEl: null,
      header: null,
      subtitle: null,
      type: null,
      _body: null,
      _list: null,
      _listPaddingTotal: 0,
      _dropdownExpand: false,
      _dropdownHeader: null,
      _items: null,
      _onResizeBound: null,
      _onVisibleBound: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['type', 'header', 'subtitle'],

    /**
     * Store a reference to the element.
     * @param {Element} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.containerEl = this.containerEl || el.parentNode;
      this._body = this.el.querySelector('.spark-step-indicator__body');
      this._list = this.el.querySelector('.spark-step-indicator__list');


      this.type = this.type !== null ? this.type : (this.el.attributes['data-type'] && this.el.attributes['data-type'].value);
      this.header = this.header !== null ? this.header : (this.el.attributes['data-header'] && true);
      this.subtitle = this.subtitle !== null ? this.subtitle : (this.el.attributes['data-subtitle'] && true);



      this._items = this.el.querySelectorAll('.spark-step-indicator__item');
    },

    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onResizeBound = this._onResize.bind(this);
      this._onLoadBound = this._onLoad.bind(this);
      this._onVisibleBound = this._onVisible.bind(this);
      this._toggleDropdownBound = this._toggleDropdown.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      window.addEventListener('resize', this._onResizeBound);
      document.addEventListener('spark.visible-children', this._onVisibleBound, true);
      window.addEventListener('load', this._onLoadBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      window.removeEventListener('resize', this._onResizeBound);
      document.removeEventListener('spark.visible-children', this._onVisibleBound, true);
      document.removeEventListener('click', this._toggleDropdownBound);
      window.removeEventListener('load', this._onLoadBound);
    },


    /**
     * Change selected step state into incomplete
     * @param {Int} index of selected step
     */
    _resetStepState: function(stepIndex) {
      if( stepIndex < 0 || stepIndex >= this._items.length ) {
        return;
      }
      this._removeClass(this._items[stepIndex], ['spark-step-indicator__item--completed', 'spark-step-indicator__item--current', 'spark-step-indicator__item--disabled']);
    },


    /**
     * Set selected step state: completed/current/disabled/incomplete
     * For current step, will remove current step from other steps
     * @param {Int} index of selected step
     * @param {String} state to be set
     */
    setStepState: function(stepIndex, state) {
      if( stepIndex < 0 || stepIndex >= this._items.length ) {
        return;
      }
      switch(state) {
        case 'completed':
          this._resetStepState(stepIndex);
          this._addClass(this._items[stepIndex], 'spark-step-indicator__item--completed');
          break;
        case 'current':
          var currentStep = this._list.querySelector('.spark-step-indicator__item--current');
          this._removeClass(currentStep, 'spark-step-indicator__item--current');
          this._addClass(this._items[stepIndex], 'spark-step-indicator__item--current');
          break;
        case 'disabled':
          this._resetStepState(stepIndex);
          this._addClass(this._items[stepIndex], 'spark-step-indicator__item--disabled');
          break;
        case 'incomplete':
          this._resetStepState(stepIndex);
          break;
      }
      this._calculateStyle();
    },


    /**
     * Public interface to let use manually update style of step indicator
     * including set different option
     */
    update: function() {
      this._cacheElements(this.el);
      this._calculateStyle();
    },

    /**
     * Provide a method to bind click callback function to certain step
     * @param {Array} a set of step indices represent the target of callback
     * @param {Function} callback function
     */
    bindStepClickCallback: function(indexArr, callback) {
      callback = callback || noop;
      for(var i=0; i < indexArr.length; i++) {
         this._items[indexArr[i]].addEventListener('click', callback);
      }
    },

    /**
     * Remove click callback from steps
     * @param {Array} a set of step indices represent the target of callback
     * @param {Function} callback function
     *
     */
    removeStepClickCallback: function(indexArr, callback) {
      callback = callback || noop;
      for(var i=0; i < indexArr.length; i++) {
         this._items[indexArr[i]].removeEventListener('click', callback);
      }
    },

    /**
     * Work for _calculateStyle.
     * Comparing the width of list and total items, including padding
     * Sometimes after calculating, even itemTotalWidth is bigger 1px than listWidth,
     * it also means items are fit in the list, not over width.
     * So add 2px to list width when comparing to avoid this.
     * @return {Boolean}
     */
    _isOverWidth: function() {
      var listWidth = this._list.offsetWidth;
      var itemTotalWidth = this._listPaddingTotal;
      for (var i = 0; i < this._items.length; i++) {
        itemTotalWidth += this._items[i].offsetWidth;
      }
      if( listWidth + 2 < itemTotalWidth ) {
        return true;
      }
      return false;
    },


    /**
     * Calcuate different style based settings
     */
    _calculateStyle: function() {
      if(!this.header) {
        var titleEl = this.el.querySelector('.spark-step-indicator__title');
        titleEl.style.display = 'none';
        this._addClass(this._body, 'spark-step-indicator__body--no-border');
      }
      if(!this.subtitle) {
        var subtitleEl = this.el.querySelector('.spark-step-indicator__subtitle');
        subtitleEl.style.display = 'none';
      }
      // Remove special DOM and Class of dropdown variation
      this._switchFromDropdown();
      switch(this.type) {
        case 'standard-dropdown':
          this._calculateRespStyle();
          break;
        case 'large':
          this._calculateLargeStyle();
          break;
        case 'condensed':
          this._calculateCondensedStyle();
          break;
        case 'dropdown':
          this._calculateDropdownStyle();
          break;
        default:
          this._calculateStandardStyle();
      }
      // Work for IE11, detect IE11 via userAgent
      // userAgent: Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv 11.0) like Gecko
      var isIE11 = !!(navigator.userAgent.match(/Trident/) && navigator.userAgent.match(/rv[ :]11/));
      if(isIE11) {
        this._addItemWidth();
      }
    },


    /**
     * For IE11 ONLY, Add width: 100% to work around IE11 bug
     * IE11 will have incorrect position with flex-grow max-width and justify-content when items hit max-width
     * Sometimes after calculating, even itemTotalWidth is less 1px than listWidth,
     * it also means items are fit in the list, not hit the max-width
     * So add 2px to list width when comparing to avoid this.
     */
    _addItemWidth: function() {
      if( !this._hasClass(this._list, 'spark-step-indicator__list--condensed') &&
         !this._hasClass(this._body, 'spark-step-indicator__body--dropdown') ) {
        var listWidth = this._list.offsetWidth;
        var itemTotalWidth = this._listPaddingTotal;
        for (var i = 0; i < this._items.length; i++) {
          itemTotalWidth += this._items[i].offsetWidth;
        }
        if(listWidth - itemTotalWidth > 2) {
          this._switchWidthForIe11(0, true);
        }
        else {
          this._switchWidthForIe11(0, false);
        }
      }
      else {
        this._switchWidthForIe11(0, false);
      }
    },


    /**
     * Switch spark-step-indicator__item style width: 100% to work around IE11 bug
     * @param {Number} start index of step
     * @param {Boolean} whether add width: 100%
     */
    _switchWidthForIe11: function(index, val) {
      for (var i = index; i < this._items.length; i++) {
        this._items[i].style.width = val ? '100%' : '';
      }
    },


    /**
     * Fix on large variation
     * Calculate node and list length remove connect line of each end
     * Sometimes after calculating, even itemTotalWidth is bigger 1px than listWidth,
     * it also means items are fit in the list, not over width.
     * So add 2px to list width when comparing to avoid this.
     * Dependency of _calculateCondensedStyle() and _calculateStandardStyle()
     *
     */
    _calculateLargeStyle: function() {
      this._list.style['flex-wrap'] = 'wrap';
      var listWidth = this._list.offsetWidth;
      var itemTotalWidth = this._listPaddingTotal;
      for (var i = 0; i < this._items.length; i++) {
        this._removeClass(this._items[i], 'noline');
        itemTotalWidth += this._items[i].offsetWidth;
        if( listWidth + 2 < itemTotalWidth && i > 0) {
          this._addClass(this._items[i-1], 'noline');
          itemTotalWidth = this._items[i].offsetWidth + this._listPaddingTotal;
        }
      }
    },


    /**
     * Fix on condensed variation
     * Same as fix large variation except CSS class
     */
    _calculateCondensedStyle: function(){
      this._addClass(this._list, 'spark-step-indicator__list--condensed');
      this._calculateLargeStyle();
    },

    /**
     * Fix on dropdown variation
     */
    _calculateDropdownStyle: function(){
      this._switchToDropdown();
    },


    /**
     * Change only between large and condensed variation
     * no dropdown variation
     */
    _calculateStandardStyle: function() {
      this._list.style.visibility = 'hidden';
      for (var i = 0; i < this._items.length; i++) {
        this._removeClass(this._items[i], 'noline');
      }
      this._removeClass(this._list, 'spark-step-indicator__list--condensed');
      if(this._isOverWidth()) {
        this._addClass(this._list, 'spark-step-indicator__list--condensed');
        this._calculateLargeStyle();
      }
      this._list.style.visibility = 'visible';
    },


    /**
     * Make list invisible and change it into large variation at first.
     * Switch variation from top to bottom to find the fit one
     */
    _calculateRespStyle: function() {
      this._list.style.visibility = 'hidden';
      this._removeClass(this._list, 'spark-step-indicator__list--condensed');
      if(this._isOverWidth()) {
        this._addClass(this._list, 'spark-step-indicator__list--condensed');
        if(this._isOverWidth()) {
          this._removeClass(this._list, 'spark-step-indicator__list--condensed');
          this._switchToDropdown();
        }
      }
      this._list.style.visibility = 'visible';
    },


    /**
     * Create header element for dropdown variation
     * Try to find the current step of indicator and show it on the header
     * if not found, show the first step
     * @return {Object} header element of dropdown variation
     */
    _createDropdownHeader: function() {
      var el = document.createElement('a');
      var currentStep = this._list.querySelector('.spark-step-indicator__item--current');
      // Can not find current step
      if(!currentStep) {
        currentStep = this._items[0];
      }
      el.innerHTML = currentStep.innerHTML;
      el.innerHTML += '<span class="spark-step-indicator__notice">Select a Step</span>';

      // Create label element like '2 of 7'
      var stepNotice = document.createElement('span');
      var currentIndex = Array.prototype.indexOf.call(this._items, currentStep);
      stepNotice.innerHTML = (currentIndex + 1) + ' of ' + this._items.length;
      stepNotice.className = 'spark-step-indicator__label--dropdown';

      el.appendChild(stepNotice);
      el.className = currentStep.className;
      this._addClass(el,'spark-step-indicator__item--dropdown__header');
      this._dropdownHeader = el;
    },


    /**
     * Some extra event binder and DOM of dropdown variation need to be set
     */
    _switchToDropdown: function() {
      this._addClass(this._body, 'spark-step-indicator__body--dropdown');
      this._createDropdownHeader();
      this._list.parentNode.insertBefore(this._dropdownHeader, this._list);
      document.removeEventListener('click', this._toggleDropdownBound);
      document.addEventListener('click', this._toggleDropdownBound);
      this._addClass(this._list, 'collapse');
    },


    /**
     * Remove extra event and DOM when switch variation from dropdown
     */
    _switchFromDropdown: function() {
      this._removeClass(this._body, 'spark-step-indicator__body--dropdown');
      var header = this._list.parentNode.querySelector('.spark-step-indicator__item--dropdown__header');
      document.removeEventListener('click', this._toggleDropdownBound);
      if(header) {
        this._list.parentNode.removeChild(header);
      }
      this._removeClass(this._list, 'collapse');
    },

    /**
     * Toggle collapse/expand state of step list in dropdown variation
     * If there is a scroll in dropdown variation, scroll to 'current' node
     * @param {Object} Reference of DOM obj
     * @param {Object} Event of click
     */
    _toggleDropdown: function(e) {
      var target = e.target || e.srcElement;
      if( this._dropdownHeader.contains(target) ) {
        if(!this._dropdownExpand) {
          this._removeClass(this._list, 'collapse');
          this._addClass(this._dropdownHeader , 'expand');
          this._dropdownExpand = true;
        }
        else {
          this._addClass(this._list, 'collapse');
          this._removeClass(this._dropdownHeader , 'expand');
          this._dropdownExpand = false;
        }
      }
      else {
        this._addClass(this._list, 'collapse');
        this._removeClass(this._dropdownHeader , 'expand');
        this._dropdownExpand = false;
      }
      if(this._list.offsetHeight > 0) {
        var scrollMove = 0;
        for( var i = 0; i < this._items.length; i ++) {
          if( this._hasClass(this._items[i], 'spark-step-indicator__item--current') ){
            this._list.scrollTop = scrollMove;
            break;
          }
          scrollMove += this._items[i].offsetHeight;
        }
      }
    },


    /**
     * When the window finish loading
     */
    _onLoad: function() {
      // Cache list left + right padding for width calculating
      var listStyles = getComputedStyle(this._list);
      var listPaddingLeft = parseInt(listStyles.getPropertyValue('padding-left'), 10);
      var listPaddingRight = parseInt(listStyles.getPropertyValue('padding-right'), 10);
      this._listPaddingTotal = listPaddingLeft + listPaddingRight;
      this._calculateStyle();
    },


    /**
     * When the window is resized, base on params make some reponsive change.
     */
    _onResize: function() {
      this._calculateStyle();
    },


    /**
     * When a parent container shows its children and our element
     * is inside of it, resize
     * @param  {[type]} e [description]
     * @return {[type]}   [description]
     */
    _onVisible: function(e) {
      if(e.target.contains(this.el)) {
        window.setTimeout(function() {
          this._calculateStyle();
        }.bind(this),0);
      }
    },
  };

  Base.exportjQuery(StepIndicator, 'StepIndicator');

  return StepIndicator;
}));
