/**
 * # DateSelect
 * Create a select list of days, months or years.
 *
 * @example
 * new DateSelect(el);
 *
 * @module helper/date-select.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/select-input', 'spark/helpers/date'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'), require('../components/select-input'), require('./date'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.DateSelect = factory(root.Spark.BaseComponent, root.Spark.SelectInput, root.Spark.dateHelper);
  }
}(this, function(Base, SelectInput, dateHelper) {

  'use strict';

  var noop = function() {};

  /**
   * DateSelect constructor
   * @param {Element} el Optional
   * @param {Object} params Optional
   */
  function DateSelect(el, params) {

    // If the first argument is a plain object, create a default element
    // since the user MUST provide additional params but the element
    // is optional. Doing it this way to keep the arity the same
    // as other components.
    if (!(el instanceof HTMLElement)) {
      params = el || {};
      el = this._createDefaultElement();
    }

    this._setParams(this.defaults, true);
    this._setParams(params);
    this._bindEventListenerCallbacks();
    this._createSelect(el);
  }

  DateSelect.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['type', 'monthNames', 'min', 'max', 'onChange'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      type: null,
      monthNames: null,
      min: null,
      max: null,
      select: null,
      onChange: null,
      _onChangeBound: null
    },


    /**
     * Get the value.
     * @param {Boolean} asInt Get the value as a parsed integer.
     * @return {Mixed}
     */
    getValue: function(asInt) {
      return asInt ? parseInt(this.select.getValue(), 10) : this.select.getValue();
    },


    /**
     * Set the value.
     * @param {Mixed} val
     */
    setValue: function(val) {
      return this.select.setValue(val);
    },


    /**
     * Update the date select's options.
     * @param {Object|Array} params
     */
    setOptions: function(params) {

      params = params || {};

      this.min = params.min;
      this.max = params.max;

      // Default days
      if (this.type === 'day') {
        this.min = this.min || 1;
        this.max = this.max || 31;
      }
      // Default months
      else if (this.type === 'month') {

        // No monthNames yet and no min or max
        if ((!this.monthNames && !this.min && !this.max) || params.monthNames) {
          this.monthNames = params.monthNames || this._getDefaultMonthNames();
        } else {
          this.min = this.min || 1;
          this.max = this.max || 12;
        }
      }
      // Default years
      else if (this.type === 'year') {
        var date = new Date();
        this.min = this.min || date.getFullYear() - 100;
        this.max = this.max || (this.min || date.getFullYear()) + 100;
      }

      var i = this.min ? this.min - 1 : 0;
      var len = this.max || this.monthNames.length;
      var opts = [{}];

      for (; i < len; i++) {
        opts.push({
          value: i + 1,
          text: this.monthNames ? this.monthNames[i] : i + 1
        });
      }

      this.select.setOptions(opts);
    },


    /**
     * Set the label text for the select input.
     * @param {String} text Optional
     */
    setLabel: function(text) {
      this.select.setLabel(text !== undefined ? text : this._getTypeText());
    },


    /**
     * Create the default input element.
     * @param {String} type
     * @return {Element}
     */
    _createDefaultElement: function() {
      var el = document.createElement('span');
      el.className = 'spark-select';
      el.innerHTML = '<select class="spark-select__input"></select><span class="spark-label"></span>';
      return el;
    },


    /**
     * Create a select input helper.
     * @param {Object} el
     */
    _createSelect: function(el) {
      this.select = new SelectInput(el, {
        onChange: this._onSelectChangeBound
      });
      this.setOptions();
      this.setLabel();
    },


    /**
     * Make a list of month options.
     * @todo : make this work with languages. perhaps either by passing an option
     * or detecting the language attribute on the `<html>` tag.
     * @return {Array}
     */
    _getDefaultMonthNames: function() {
      return dateHelper.getMonthNamesShort();
    },


    /**
     * Get the text for this type of date select.
     * @todo make this work with languages.
     * @return {String}
     */
    _getTypeText: function() {
      return this.type.charAt(0).toUpperCase() + this.type.slice(1);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onSelectChangeBound = this._onSelectChange.bind(this);
    },


    /**
     * When the typeahead changes, make sure the value is valid. This
     * is very basic validation. More complex validation like the number
     * of days in a specific month should be handled by the callback.
     * And run our callback.
     * @param {String} val The value of the input
     * @param {String} oldVal The previous value
     */
    _onSelectChange: function(val) {
      (this.onChange || noop)(val, this);
    }
  };

  Base.exportjQuery(DateSelect, 'DateSelect');

  return DateSelect;
}));
