/**
 * # Store Form Data
 * Store the data of form fields inline so it can be reverted.
 *
 * @example
 * formData.store(el);
 * formData.revert(el);
 *
 * @module helper/store-form-data.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define([], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory();
  } else {
    root.Spark = root.Spark || {};
    root.Spark.storeFormData = factory();
  }
}(this, function() {

  'use strict';

  /**
   * Find all the form elements inside a given element and run a callback on each.
   * @param  {Element} el
   * @param  {Function} cb
   */
  function findAll(el, cb) {

    var inputs = el.querySelectorAll('input, select, textarea');
    var i = 0;
    var len = inputs.length;

    for (; i < len; i++) {
      cb(inputs[i]);
    }
  }

  return {

    /**
     * Find all the form elements inside a given element and store their current
     * value as a data attribute.
     * @param {Element} el
     */
    store: function(el) {

      findAll(el, function(input) {

        var name = input.nodeName.toLowerCase();
        var value;

        switch (name) {
          case 'select':
            value = input.selectedIndex;
            break;
          default:
            value = encodeURI(input.value);
            break;
        }

        input.setAttribute('data-stored-value', value);
      });
    },


    /**
     * Revert all the form elements inside of a given element.
     * @param {Element} el
     */
    restore: function(el) {

      findAll(el, function(input) {

        var name = input.nodeName.toLowerCase();
        var value = input.getAttribute('data-stored-value');

        // No stored value
        if (!value && value !== '') {
          return;
        }

        switch (name) {
          case 'select':
            input.options[value].selected = true;
            break;
          default:
            input.value = decodeURI(value);
            break;
        }

        input.removeAttribute('data-stored-value');
      });
    },


    /**
     * Clear the stored data on all the form elements inside of a given element.
     * @param {Element} el
     */
    clear: function(el) {
      findAll(el, function(input) {
        input.removeAttribute('data-stored-value');
      });
    }
  };
}));
