/**
 * # NumberSelector
 * An increment/decrement for number inputs.
 *
 * @example
 * new NumberSelector(el);
 *
 * @module components/number-selector.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.NumberSelector = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  /**
   * NumberSelector constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var NumberSelector = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  NumberSelector.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _getElementMatchingParent: Base.getElementMatchingParent,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: [],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      inputEl: null,
      buttonUpEl: null,
      buttonDownEl: null,
      _onChangeBound: null,
      _onClickBound: null
    },


    /**
     * Get or set the current value of the number selector.
     * @param {Number} val @optional
     * @return {Number} The current value
     */
    value: function(val) {

      if (val !== undefined) {
        this.inputEl.value = this._getConformedNumber(val);
      }

      return parseInt(this.inputEl.value, 10);
    },


    /**
     * Increment by the step value.
     */
    increment: function() {
      this.value(this.value() + (this._getStep() || 1));
    },


    /**
     * Decrement by the step value.
     */
    decrement: function() {
      this.value(this.value() - (this._getStep() || 1));
    },

    /**
     * Disable number selector
     */
    disable: function() {
      this.inputEl.disabled = true;
      this.buttonUpEl.disabled = true;
      this.buttonDownEl.disabled = true;
    },

    /**
     * Enable number selector
     */
    enable: function() {
      this.inputEl.disabled = false;
      this.buttonUpEl.disabled = false;
      this.buttonDownEl.disabled = false;
    },


    /**
     * Store a reference to the whole number-selector, as well as the
     * input element. Also, get some default values from the input
     * element (min, max, steps).
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.inputEl = el.querySelector('input');
      this.buttonUpEl = el.querySelector('.spark-number-selector__up, .spark-number-picker__up');
      this.buttonDownEl = el.querySelector('.spark-number-selector__down, .spark-number-picker__down');

      if (!this.inputEl) {
        throw new Error('NumberSelector must include an `<input>` element!');
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onChangeBound = this._onChange.bind(this);
    },


    /**
     * Add event listeners for touchstart and mouse click.
     */
    _addEventListeners: function() {
      this.el.addEventListener('click', this._onClickBound);
      this.el.addEventListener('change', this._onChangeBound);
    },


    /**
     * Remove event listeners for touchstart and mouse click.
     */
    _removeEventListeners: function() {
      this.el.removeEventListener('click', this._onClickBound);
      this.el.removeEventListener('change', this._onChangeBound);
    },


    /**
     * Get the current value of the min property.
     * @return {Number}
     */
    _getMin: function() {
      return this._getInputPropAsNumber('min');
    },


    /**
     * Get the current value of the max property.
     * @return {Number}
     */
    _getMax: function() {
      return this._getInputPropAsNumber('max');
    },


    /**
     * Get the current value of the step property.
     * @return {Number}
     */
    _getStep: function() {
      return this._getInputPropAsNumber('step');
    },


    /**
     * Get a property as a number.
     * @param {String} key
     * @return {Number}
     */
    _getInputPropAsNumber: function(key) {
      return parseFloat(this.inputEl.getAttribute(key));
    },


    /**
     * Get the given number within the min/max range of the input element.
     * @param {Number} num
     * @return {Number}
     */
    _getConformedNumber: function(num) {

      var max = this._getMax();
      var min = this._getMin();
      var step = this._getStep();

      // Move in increments if we have a defined step size
      if (step) {

        var diff = num % step;
        var halfStep = step / 2;
        var overHalf = diff >= halfStep;

        num = overHalf ? num + (step - diff) : num - diff;
      }

      return max !== undefined && num > max ? max : (min !== undefined && num < min ? min : num);
    },


    /**
     * When either of the buttons are clicked, update the value.
     * @param {Object} e
     */
    _onClick: function(e) {

      // The increment button is clicked
      // @todo Remove .spark-number-picker in 2.0.0
      if (this._getElementMatchingParent(e.target, '.spark-number-selector__up, .spark-number-picker__up', this.el)) {
        e.preventDefault();
        this.increment();
      }
      // The decrement button is clicked
      // @todo Remove .spark-number-picker in 2.0.0
      else if (this._getElementMatchingParent(e.target, '.spark-number-selector__down, .spark-number-picker__down', this.el)) {
        e.preventDefault();
        this.decrement();
      }
    },


    /**
     * When the input value changes, max sure we are in bounds.
     * @param {Object} e
     */
    _onChange: function() {
      this.value(parseFloat(this.inputEl.value));
    }
  };

  Base.exportjQuery(NumberSelector, 'NumberSelector');

  return NumberSelector;
}));
