/**
 * # ToggleSwitch
 * ToggleSwitch and collapse an element.
 *
 * @example
 * new ToggleSwitch(el);
 *
 * @module components/ToggleSwitch.js
 */

/**
 * @ignore
 */
(function(root, factory) {

    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['spark/components/base'], factory);
    } else if (typeof module === 'object' && module.exports) {
        module.exports = factory(require('./base'));
    } else {
        root.Spark = root.Spark || {};
        root.Spark.ToggleSwitch = factory(root.Spark.BaseComponent);
    }
}(this, function(Base) {

    'use strict';

    /**
     * ToggleSwitch constructor.
     * @param {Element} el
     * @param {Object} params
     */
    var ToggleSwitch = function(el) {

        if (!el) {
            return;
        }

        this._cacheElements(el);
        this._bindEventListenerCallbacks();
        this._addEventListeners();
    };

    ToggleSwitch.prototype = {

        /**
         * Include common functionality.
         */
        _getElementMatchingParent: Base.getElementMatchingParent,
        _getMatchingChild: Base.getMatchingChild,
        remove: Base.remove,


        /**
         * Default values for internal properties we will be setting.
         * These are set on each construction so we don't leak properties
         * into the prototype chain.
         * @type {Object}
         */
        defaults: {
            el: null,
            activate: null,
            deactivate: null,
            toggle: null,
            input: null,
            _onKeydownBound: null
        },

        /**
         * Toggle the ToggleSwitch state.
         */
        toggle: function() {
            // Check if input exists
            if (!this.input) return;

            // toggle checked property
            if (this.input.checked)
                this.deactivate();
            else
                this.activate();
        },

        /**
         * Activate toggle state
         */
        activate: function() {
            // Check if input exists
            if (!this.input) return;
            this.input.checked = true;
        },

        /**
         * Deactivate toggle state
         */
        deactivate: function() {
            // Check if input exists
            if (!this.input) return;
            this.input.checked = false;
        },

        /**
         * Store a reference to the element.
         * @param {Element} el
         */
        _cacheElements: function(el) {
            this.el = el;
            this.input = el.querySelector('.spark-toggle__input');
        },


        /**
         * Create bound versions of event listener callbacks and store them.
         * Otherwise we can't unbind from these events later because the
         * function signatures won't match.
         */
        _bindEventListenerCallbacks: function() {
            this._onKeydownBound = this._onKeydown.bind(this);
        },


        /**
         * Add event listeners for DOM events.
         */
        _addEventListeners: function() {
            this.el.addEventListener('keydown', this._onKeydownBound);
        },


        /**
         * Remove event listeners for DOM events..
         */
        _removeEventListeners: function() {
            this.el.removeEventListener('keydown', this._onKeydownBound);
        },


        /**
         * When the space or enter key is pressed on the toggle, toggle!
         * @param {Object} e
         */
        _onKeydown: function(e) {

            if (!this._getElementMatchingParent(e.target, '.spark-toggle-switch, spark-toggle-switch__handle', this.el)) {
                return;
            }

            var code = e.keyCode || e.which;

            switch (code) {
                case 32:
                    // space
                    // Skip, native works as expected
                    break;
                case 13:
                    // enter
                    e.preventDefault();
                    this.toggle();
                    break;
                case 39:
                case 40:
                    // right
                    // down
                    e.preventDefault();
                    this.activate();
                    break;
                case 37:
                case 38:
                    // left
                    // up
                    e.preventDefault();
                    this.deactivate();
                    break;
            }

        }
    };

    Base.exportjQuery(ToggleSwitch, 'ToggleSwitch');

    return ToggleSwitch;
}));
