'use strict';

var test = require('tape');
var BaseComponent = require('../../components/base');

// Create a new class composed of the BaseComponent methods
var Cls = function() {};
Cls.prototype = {
  exportComponent: BaseComponent.exportComponent,
  setParams: BaseComponent.setParams,
  unsetParams: BaseComponent.unsetParams,
  remove: BaseComponent.remove,
  toggleClass: BaseComponent.toggleClass,
  addClass: BaseComponent.addClass,
  removeClass: BaseComponent.removeClass,
  hasClass: BaseComponent.hasClass,
  getChildIndex: BaseComponent.getChildIndex,
  elementHasParent: BaseComponent.elementHasParent,
  elementMatches: BaseComponent.elementMatches,
  getElementMatchingParent: BaseComponent.getElementMatchingParent,
  getElementMatchingParents: BaseComponent.getElementMatchingParents,
  getElementOffset: BaseComponent.getElementOffset,
  getNodeListIndex: BaseComponent.getNodeListIndex,
  trim: BaseComponent.trim,
  round: BaseComponent.round,
  appendChildren: BaseComponent.appendChildren,
  insertBefore: BaseComponent.insertBefore,
  getBreakpoint: BaseComponent.getBreakpoint
};
var inst = new Cls();

// @todo: is there a good way to test for this?
test('should export a jQuery module', function(t) {
  t.end();
});

// @todo: is there a good way to test for this?
test('should return a cached jquery instance', function(t) {
  t.end();
});

// @todo: is there a good way to test for this?
test('should call internal public methods on an instannce, but error on non-existent or private methods', function(t) {
  t.end();
});

test('should set only whitelisted params unless forced to', function(t) {

  inst._whitelistedParams = ['a', 'b', 'd'];

  inst.setParams({
    a: true,
    b: true,
    c: true
  });

  t.equal(inst.a, true, 'sets a parameter to true');
  t.equal(inst.b, true, 'sets b parameter to true');
  t.equal(inst.c, undefined, 'does not set c to true');

  inst.setParams({
    d: true
  }, true);

  t.equal(inst.d, true, 'sets d parameter to true when forced');

  t.end();
});

test('should unset params from an object', function(t) {

  inst.testKey = 42;

  inst.unsetParams(['testKey']);
  t.notOk(inst.testKey, 'key is deleted when passed as an array');

  inst.anotherTest = 'string';
  inst.test3 = 24;
  inst.unsetParams({
    'anotherTest': null,
    'test3': null
  });
  t.notOk(inst.anotherTest && inst.test3, 'keys are deleted when an object is passed');

  var obj = {
    a: 1,
    b: true,
    c: 'name',
    d: {}
  };
  inst.unsetParams(['a', 'b', 'c'], obj);
  t.notOk(obj.a && obj.b && obj.c, 'keys are deleted from a passed scope object');
  t.ok(obj.d, 'key remains untouched if not passed');

  t.end();
});

test('should toggle the class on an element', function(t) {

  var el = document.createElement('div');

  t.ok(el.className.indexOf('test-class') === -1, 'does not have a class set at first');
  inst.toggleClass(el, 'test-class', true);
  t.ok(el.className.indexOf('test-class') !== -1, 'adds the test class');
  inst.toggleClass(el, 'test-class', true);
  t.ok(el.className.match(/test\-class/gi).length === 1, 'doesn\'t add the test class more than once');
  inst.toggleClass(el, 'test-class', false);
  t.ok(el.className.indexOf('test-class') === -1, 'removes the test class');

  el.innerHTML = '<span></span><span></span>';
  var spans = el.querySelectorAll('span');

  inst.toggleClass(spans, 'span-test-class', true);

  for (var j = 0; j < spans.length; j++) {
    t.ok(spans[j].className.indexOf('span-test-class') !== -1, 'adds the span test class to item ' + (j + 1) + '/' + spans.length + ' in a list of nodes');
  }

  var spansArr = [document.createElement('span'), document.createElement('span')];

  inst.toggleClass(spansArr, 'span-test-class', true);

  for (j = 0; j < spansArr.length; j++) {
    t.ok(spansArr[j].className.indexOf('span-test-class') !== -1, 'adds the span test class to item ' + (j + 1) + '/' + spansArr.length + ' in an array of elements');
  }

  t.end();
});

test('should add a class to an element', function(t) {

  var el = document.createElement('div');

  inst.addClass(el, 'added-class');
  t.ok(el.className.indexOf('added-class') !== -1, 'adds the class');
  inst.addClass(el, 'added-class');
  t.ok(el.className.match(/added\-class/gi).length === 1, 'doesn\'t add the test class more than once');

  t.end();
});

test('should remove a class from an element', function(t) {

  var el = document.createElement('div');

  inst.addClass(el, 'added-class');
  t.ok(el.className.indexOf('added-class') !== -1, 'has the class');
  inst.removeClass(el, 'added-class');
  t.ok(el.className.indexOf('added-class') === -1, 'removes the class');

  t.end();
});

test('should determine if a class exists on an element or string', function(t) {

  var el = document.createElement('div');
  el.className = 'test-class another-test-class';

  t.equal(inst.hasClass(el, 'test-class'), true, 'has a class of "test-class"');
  t.equal(inst.hasClass(el, 'another-test-class'), true, 'has a class of "another-test-class"');
  t.equal(inst.hasClass(el, 'third-test-class'), false, 'has no class of "third-test-class"');

  t.equal(inst.hasClass('string-class another-class', 'string-class'), true, 'has a class of "string-class"');
  t.equal(inst.hasClass('string-class another-class', 'another-string-class'), false, 'has no class of "another-string-class"');

  t.end();
});

test('should get the index of a child element', function(t) {

  var el = document.createElement('div');
  el.innerHTML = '<span id="first"></span><span id="second"></span>';
  var spans = el.querySelectorAll('span');
  var firstSpan = el.querySelector('#first');
  var secondSpan = el.querySelector('#second');
  var thirdSpan = document.createElement('span');

  t.equal(inst.getChildIndex(spans, firstSpan), 0, 'finds the first child element');
  t.equal(inst.getChildIndex(spans, secondSpan), 1, 'finds the second child element');
  t.equal(inst.getChildIndex(spans, thirdSpan), -1, 'does not find a third child element');

  t.end();
});

test('should determine if a given element has another for a parent', function(t) {

  var el = document.createElement('div');
  var child1 = document.createElement('span');
  var child2 = document.createElement('span');
  var child1child1 = document.createElement('span');

  child1.appendChild(child1child1);
  el.appendChild(child1);

  t.ok(inst.elementHasParent(child1, el), 'finds an immediate parent');
  t.ok(inst.elementHasParent(child1child1, el), 'finds a grandparent');
  t.notOk(inst.elementHasParent(child2, el), 'does not find an non-existent parent');

  t.end();
});

test('should determine if a given element matches a query selector', function(t) {

  var el = document.createElement('div');

  t.ok(inst.elementMatches(el, 'div'), 'matches the proper query selector');
  t.notOk(inst.elementMatches(el, 'span'), 'does not match an incorrect query selector');

  t.end();
});

test('should get a parent based on a query string', function(t) {

  var el = document.createElement('div');
  var child1 = document.createElement('span');
  var sibling1 = document.createElement('span');
  var child1child1 = document.createElement('span');

  child1.appendChild(child1child1);
  el.appendChild(child1);
  el.appendChild(sibling1);

  t.ok(inst.getElementMatchingParent(child1child1, 'div'), 'matches the proper query selector');
  t.notOk(inst.getElementMatchingParent(child1child1, 'section'), 'does not match an incorrect query selector');
  t.notOk(inst.getElementMatchingParent(child1child1, 'div', child1), 'optionally limits the search for a matching parent');
  t.ok(inst.getElementMatchingParent(child1child1, 'span', [child1, sibling1]), 'optionally limits the search for a matching parent');

  t.end();
});

test('should get an array of parents based on a query string', function(t) {

  var el = document.createElement('div');
  var child1 = document.createElement('div');
  var child1child1 = document.createElement('span');

  child1.appendChild(child1child1);
  el.appendChild(child1);

  t.equal(inst.getElementMatchingParents(child1child1, 'div').length, 2, 'finds a list of parents');
  t.equal(inst.getElementMatchingParents(child1, 'div').length, 1, 'finds a list of parents');
  t.notOk(inst.getElementMatchingParents(child1child1, 'section').length, 'does not match an incorrect query selector');

  t.end();
});

test('should get the offset pixel values of an element', function(t) {

  var el = document.createElement('div');
  var offsets = inst.getElementOffset(el);

  t.equal(offsets.left, 0, 'has a correct left offset');
  t.equal(offsets.top, 0, 'has a correct top offset');

  t.end();
});

test('should get the index of an item in a NodeList', function(t) {

  var el = document.createElement('div');
  var child1 = document.createElement('span');
  var child2 = document.createElement('span');

  el.appendChild(child1);
  el.appendChild(child2);

  t.equal(inst.getNodeListIndex(el.children, child1), 0, 'finds the index of the first child');
  t.equal(inst.getNodeListIndex(el.children, child2), 1, 'finds the index of the second child');
  t.equal(inst.getNodeListIndex(el.children, document.createElement('span')), -1, 'does not find the index when an element is not a child');

  t.end();
});

test('should trim the whitespace at the start and end of a string', function(t) {
  t.equal(inst.trim('   space at front'), 'space at front', 'trims space at the front');
  t.equal(inst.trim('space at back   '), 'space at back', 'trims space at the back');
  t.equal(inst.trim('   space at front and back   '), 'space at front and back', 'trims space at the front and back');
  t.end();
});

test('should round a number to a given number of decimal places', function(t) {
  t.equal(inst.round(34.2342124, 0), 34, 'rounds to zero places');
  t.equal(inst.round(34.2352124, 2), 34.24, 'rounds to two places');
  t.equal(inst.round(34.2352124, 10), 34.2352124000, 'rounds to ten places');
  t.end();
});

test('should append multiple children to a node', function(t) {

  var defaultHTML = '<span></span><p></p><p></p>';
  var el = document.createElement('div');
  el.innerHTML = defaultHTML;

  var children = [document.createElement('span'), document.createElement('ul')];

  inst.appendChildren(el, children);

  t.notEqual(el.children[0], children[0], 'first child of node is not the first element in the appended list');
  t.equal(el.children[3], children[0], 'forth child of node is the first element in the appended list');

  el.innerHTML = defaultHTML;

  inst.appendChildren(el, children, true);
  t.equal(el.children[0], children[0], 'first child of node is the first element in the appended list because clear was passed');

  var anotherEl = el.cloneNode(true);
  inst.appendChildren(el, anotherEl.children);
  t.notEqual(el.children[0], anotherEl.children[0], 'moves elements of an HTMLCollection');

  t.end();
});

test('should insert multiple children to a node before a given node', function(t) {

  var defaultHTML = '<span></span><p></p><p></p>';
  var el = document.createElement('div');
  el.innerHTML = defaultHTML;

  var beforeEl = el.children[1];

  var children = [document.createElement('span'), document.createElement('ul')];

  inst.insertBefore(el, beforeEl, children);

  t.equal(el.children[1], children[0], 'appended elements are in front of element to insert before');
  t.equal(el.children[3], beforeEl, 'element to insert before is now after the appended elements');

  t.end();
});

test('get the name of a breakpoint for a given width', function(t) {

  t.equal(inst.getBreakpoint(100), 'xs', 'extra small breakpoint matched');
  t.equal(inst.getBreakpoint(641), 'sm', 'small breakpoint matched');
  t.equal(inst.getBreakpoint(801), 'md', 'medium breakpoint matched');
  t.equal(inst.getBreakpoint(1049), 'lg', 'large breakpoint matched');
  t.equal(inst.getBreakpoint(1801), 'xl', 'extra large breakpoint matched');

  var breakpoints = {
    a: {
      min: 0,
      max: 1000
    },
    b: {
      min: 1001,
      max: 1523
    }
  };

  t.equal(inst.getBreakpoint(12, breakpoints), 'a', 'custom breakpoint matched');
  t.equal(inst.getBreakpoint(1200, breakpoints), 'b', 'custom breakpoint matched');
  t.notOk(inst.getBreakpoint(5000, breakpoints), 'out of range breakpoint not found');

  t.end();
});

test('should clean itself up on remove', function(t) {

  var el = document.createElement('div');
  var inst = new Cls(el);
  inst._removeEventListeners = function() {};
  inst.el = el;
  inst._unsetParams = inst.unsetParams;
  inst.defaults = {
    el: null
  };

  t.equal(inst.el, el, 'element is the passed element');
  inst.remove();
  t.notOk(inst.el, 'element is unset');

  t.end();
});

test('should trigger an event on an element', function(t) {

  var el = document.createElement('div');
  el.addEventListener('change', function(e) {
    t.equal(e.target, el, 'target of the event matches the element it was called upon');
    t.end();
  });
  BaseComponent.triggerEvent(el, 'change');
});

test('should scroll the window to a position', function(t) {

  var el = document.createElement('div');
  BaseComponent.scrollWindowTo(el, {
    callback: function() {

      t.equal(window.scrollY, el.clientTop, 'scrolls to the position of a given element');

      BaseComponent.scrollWindowTo({
        x: 10,
        y: 50,
        callback: function() {
          // t.equal(window.scrollY, 50);
          t.end();
        }
      });
    }
  });
});

test('should tween a parameter from one value to another', function(t) {

  var obj = {
    val: 0
  };

  BaseComponent.tween({
    target: obj,
    prop: 'val',
    start: 0,
    end: 100,
    duration: 10,
    callback: function() {

      t.equal(obj.val, 100, 'tweens a value');

      BaseComponent.tween({
        target: obj,
        prop: 'val',
        start: [50, 100, 600],
        end: [0, -150, 800],
        duration: 10,
        callback: function() {
          t.equal(obj.val[0], 0, 'tweens a value');
          t.equal(obj.val[1], -150, 'tweens a value');
          t.equal(obj.val[2], 800, 'tweens a value');
          t.end();
        }
      });
    }
  });
});

test('should copy attributes from one element to another', function(t) {

  var a = document.createElement('div');
  var b = document.createElement('div');

  a.id = 'test';
  a.className = 'one two three six';
  a.setAttribute('data-test', 'a test value');

  BaseComponent.copyAttributes(a, b);

  t.equal(b.id, 'test', 'copies an id');
  t.equal(b.className, 'one two three six', 'copies a class');
  t.equal(a.getAttribute('data-test'), 'a test value', 'copies an arbitrary attribute');

  t.end();
});

test('should get a sibling before which matches a query selector', function(t) {

  var el = document.createElement('div');
  el.innerHTML = '<span></span><input><textarea></textarea><span></span>';

  t.equal(el.children[0], BaseComponent.getSiblingBefore(el.children[2], 'span'));

  t.end();
});

test('should get a sibling after which matches a query selector', function(t) {

  var el = document.createElement('div');
  el.innerHTML = '<span></span><input><textarea></textarea><span></span>';

  t.equal(el.children[2], BaseComponent.getSiblingAfter(el.children[0], 'textarea'));

  t.end();
});

test('should get a child node that matches a query', function(t) {

  var el = document.createElement('div');
  el.innerHTML = '<span><input></span><textarea></textarea><span></span>';

  t.equal(el.children[0], BaseComponent.getMatchingChild(el, 'span'), 'finds a matching child');
  t.notOk(BaseComponent.getMatchingChild(el, 'input'), 'does not find a matching child');

  t.end();
});
