'use strict';

var test = require('tape');
var DateInput = require('../../components/date-input');

var markup = '<input class="spark-date__field" type="date" name="example1" value="2015-02-02" role="datebox"><span class="spark-label">DOB</span>';
var toggleMarkup = '<div class="spark-date__toggle"></div>';
var messageMarkup = '<div class="spark-input__message"></div>';

test('should create date input component', function(t) {
  var el = document.createElement('label');

  t.throws(function() {
    new DateInput(el);
  }, 'needs a date input element to be constructed');

  el.innerHTML = markup;
  new DateInput(el);

  t.end();
});

test('should cache elements passed to it', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup + toggleMarkup + messageMarkup;

  var inputEl = el.querySelector('input');

  var inst = new DateInput(el);

  t.equal(inst.el, el, 'element is cached');
  t.equal(inst.inputEl, inputEl, 'input element is cached');
  t.equal(inst.toggleEl, el.querySelector('.spark-date__toggle'), 'toggle is cached');
  t.equal(inst.messageEl, el.querySelector('.spark-input__message'), 'message container is cached');

  t.end();
});

test('should set whitelisted parameters', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup;

  var onChange = function() {};
  var onFocus = function() {};
  var onBlur = function() {};

  var inst = new DateInput(el, {
    onChange: onChange,
    onFocus: onFocus,
    onBlur: onBlur,
    isTypeahead: true
  });

  t.equal(inst.onChange, onChange, 'onChange is set');
  t.equal(inst.onFocus, onFocus, 'onFocus is set');
  t.equal(inst.onBlur, onBlur, 'onBlur is set');
  t.ok(inst.isTypeahead, 'isTypeahead is set');

  inst = new DateInput(el, {
    isSelect: true
  });

  t.ok(inst.isSelect, 'isSelect is set');
  t.notOk(inst.isTypeahead, 'isTypeahead is false if isSelect is set to true');

  t.end();
});

test('should update the active class when shown/hidden', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup;
  el.querySelector('input').value = '';

  var inst = new DateInput(el);

  t.notOk(inst.isActive, 'isActive is false');
  t.equal(el.className.indexOf('active'), -1, 'active class is not set');

  inst.show();

  t.ok(inst.isActive, 'isActive is true');
  t.notEqual(el.className.indexOf('active'), -1, 'active class is set');

  inst.hide();

  t.notOk(inst.isActive, 'isActive is false');
  t.equal(el.className.indexOf('active'), -1, 'active class is not set');

  t.end();
});

test('should parse params from the markup or passed in the config', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup;

  var inst = new DateInput(el);
  t.ok(inst.isActive, 'active by default because of an existing value');
  t.notOk(inst.isSelect, 'no select inputs without passing a special class');
  t.ok(inst.isTypeahead, 'defaults to being a typeahead');

  el = document.createElement('label');
  el.innerHTML = markup;

  el.querySelector('input').removeAttribute('value');
  el.innerHTML += '<span class="spark-date__toggle"></span>';
  inst = new DateInput(el);
  t.notOk(inst.isActive, 'not active by default because of no existing value');
  t.ok(inst.isTypeahead, 'defaults to being a typeahead');

  el = document.createElement('label');
  el.innerHTML = markup;

  el.innerHTML = markup;
  el.className += ' spark-date--select';
  inst = new DateInput(el);
  t.ok(inst.isSelect, 'is a select input because of an existing select class');
  t.notOk(inst.isTypeahead, 'not a typeahead by default because it is a select');

  el = document.createElement('label');
  el.innerHTML = markup;

  el.querySelector('input').setAttribute('data-format', 'DD-MM-YY');
  inst = new DateInput(el);
  t.equal(inst.parsedFormat.parts[0].name, 'day', 'first format part is a day');

  t.end();
});

test('should set the value of the input', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup;

  var inst = new DateInput(el);
  inst.setValue({
    day: 12,
    month: 8,
    year: 1999
  });

  t.equal(inst.inputEl.value, '1999-08-12', 'sets the value properly');
  t.equal(inst.typeaheads.day.getValue(), '12', 'sets the day typeahead value properly');
  t.equal(inst.typeaheads.month.getValue(), '08', 'sets the month typeahead value properly');
  t.equal(inst.typeaheads.year.getValue(), '1999', 'sets the year typeahead value properly');

  t.end();
});

test('should validate day values', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup;

  var inst = new DateInput(el);
  inst.setValue({
    day: 30,
    month: 2,
    year: 1999
  });

  t.equal(inst.typeaheads.day.getValue(true), 28, 'constrains day values');

  inst.setValue({
    year: 2016
  });
  inst.setValue({
    day: 30
  });

  t.equal(inst.typeaheads.day.getValue(true), 29, 'constrains day values for a leap year');

  t.end();
});

test('should pad short values', function(t) {

  var el = document.createElement('label');
  el.innerHTML = markup;

  var inst = new DateInput(el);
  inst.setValue({
    day: 3,
    month: 2,
    year: 1999
  });

  t.equal(inst.typeaheads.day.getValue(), '03', 'pads a short day');
  t.equal(inst.typeaheads.month.getValue(), '02', 'pads a month day');

  t.end();
});

test('should run onBlur, onFocus and onChange callbacks', function(t) {

  var count = 0;
  var el = document.createElement('label');
  el.innerHTML = markup;

  function cb() {
    count++;
    if (count === 3) {
      t.end();
    }
  }

  var inst = new DateInput(el, {
    onChange: cb,
    onBlur: cb,
    onFocus: cb
  });

  inst.inputEl.value = '2014-02-02';
  inst._onTypeaheadFocus(10, inst.typeaheads.month);
  inst._onPieceChange();
  inst._onTypeaheadBlur(10, inst.typeaheads.month);
});
