/*! sabre-spark - v1.9.12 - 2018-05-01 DO NOT MODIFY THIS FILE DIRECTLY OR YOU WILL NOT BE ABLE TO UPDATE YOUR VERSION OF SPARK. */
/**
 * # BaseComponent
 * The base class for Spark JS components. Contains methods used across multiple components.
 *
 * @module components/base.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define([], function() {
      return factory(root);
    });
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(root);
  } else {
    root.Spark = root.Spark || {};
    root.Spark.BaseComponent = factory(root);
  }
}(this, function(root) {

  'use strict';

  // See if there is an existing implementation of query selector matching.
  var vendorMatch = typeof Element !== 'undefined' && (Element.prototype.matches || Element.prototype.matchesSelector || Element.prototype.webkitMatchesSelector || Element.prototype.mozMatchesSelector || Element.prototype.msMatchesSelector || Element.prototype.oMatchesSelector);
  var trimRE = /^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g;
  var ws = /\s+/;
  var cleanup = /\s{2,}/g;
  var noop = function() {};
  var raf = window.requestAnimationFrame ||
    window.msRequestAnimationFrame ||
    window.mozRequestAnimationFrame ||
    window.webkitRequestAnimationFrame ||
    (function() {

      var fps = 60;
      var del = 1000 / fps;
      var start = Date.now();
      var prev = start;

      return function requestAnimationFrame(callback) {

        var requestTime = Date.now();
        var timeout = Math.max(0, del - (requestTime - prev));
        var timeToCall = requestTime + timeout;

        prev = timeToCall;

        return window.setTimeout(function onAnimationFrame() {
          callback(timeToCall - start);
        }, timeout);
      };
    })();
  var caf = window.cancelAnimationFrame = window.mozCancelAnimationFrame ||
    window.webkitCancelAnimationFrame ||
    window.cancelRequestAnimationFrame ||
    window.msCancelRequestAnimationFrame ||
    window.mozCancelRequestAnimationFrame ||
    window.webkitCancelRequestAnimationFrame ||
    function cancelAnimationFrame(id) {
      window.clearTimeout(id);
    };

  /**
   * BaseComponent object. This isn't a method because it can't be instantiated directly
   * like other components. No fancy "extends" method or anything because extension happens
   * happens through composition not inheritance.
   * @type {Object}
   */
  var BaseComponent = {

    /**
     * Request animation frame
     * @type {Function}
     */
    requestAnimationFrame: raf,


    /**
     * Cancel anmation frame
     * @type {Function}
     */
    cancelAnimationFrame: caf,


    /**
     * Breakpoints being used in the CSS.
     * @type {Object}
     */
    breakpoints: {
      xs: {
        min: 0,
        max: 543
      },
      sm: {
        min: 544,
        max: 795
      },
      md: {
        min: 796,
        max: 1047
      },
      lg: {
        min: 1048,
        max: 1799
      },
      xl: {
        min: 1800,
        max: Infinity
      }
    },

    /**
     * Export as a jQuery plugin.
     * @param {Mixed} Obj
     * @param {String} name
     */
    exportjQuery: function(Obj, name) {

      if (!root.jQuery) {
        return;
      }

      var fnName = 'spark' + name;

      root.jQuery.fn[fnName] = function(method, options) {
        return this.each(function() {
          return BaseComponent.loadOrCreateJQueryInstance(fnName, Obj, this, method, options);
        });
      };
    },


    /**
     * Load a cached instance of a component for use with jQuery. This way, calling `$('...').sparkTextInput()`
     * doesn't create a new instance each time, but rather returns an existing instance (if signatures match).
     * If the instance doesn't already exist, create it and cache it. To remove a cached instance of a jQuery
     * plugin, use `$('...').removeData()`.
     * @param {String} fnName The name of the jQuery function.
     * @param {Function} Ctor The constructor
     * @param {Element} el
     * @param {String|Object} method The method to invoke on the instance. This can be omitted and options will take its place.
     * @param {Object} options
     * @return {Object}
     */
    loadOrCreateJQueryInstance: function(fnName, Ctor, el, method, options) {

      var methodIsString = typeof method === 'string';
      options = methodIsString ? options : method;
      method = methodIsString ? method : null;

      // Find a cached instance
      var dataName = 'spark.' + fnName;
      var $el = $(el);
      var cachedInstance = $el.data(dataName);
      var instance;

      // We have a cached instance
      if (cachedInstance) {
        instance = cachedInstance;
      }
      // Create and cache
      else {
        instance = new Ctor(el, options);
        $el.data(dataName, instance);
      }

      // If we have a method to call, do so.
      if (method) {

        // Pitch a fit if this is a private method.
        if (method[0] === '_') {
          throw new Error('Cannot access private method "' + method + '" on the ' + fnName + ' class.');
        }

        // Fail if this method doesn't exist.
        if (typeof instance[method] !== 'function') {
          throw new Error('No method "' + method + '" is defined on the ' + fnName + ' class.');
        }

        instance[method](options);
      }

      return instance;
    },


    /**
     * Set a hash of parameters if they're whitelisted or we're told to force the set.
     * This is used to set initial values as well as set passed parameters.
     * @param {Object} params
     * @param {Boolean} force Force setting even if the param is not whitelisted.
     */
    setParams: function(params, force) {

      for (var i in params) {
        if (this._whitelistedParams.indexOf(i) !== -1 || force) {
          this[i] = params[i];
        }
      }
    },


    /**
     * Unset all parameters.
     * @param {Array|Object} keys
     * @param {Object} scope The object to unset the params from. Defaults to `this`.
     */
    unsetParams: function(keys, scope) {

      // If passed an object just get the keys.
      keys = keys instanceof Array ? keys : Object.keys(keys);

      scope = scope || this;

      var i = 0;
      var len = keys.length;

      for (; i < len; i++) {
        delete scope[keys[i]];
      }
    },


    /**
     * Toggle a class on an element given a condition.
     * @param {Element|Array} el An element or array of elements to update.
     * @param {String} name
     * @param {Boolean} enable
     */
    toggleClass: function(el, name, enable) {

      if (!el) {
        return;
      }

      // If we're passed an array, toggle the class on each.
      if (el instanceof NodeList || el instanceof Array) {

        for (var i = 0, len = el.length; i < len; i++) {
          BaseComponent.toggleClass(el[i], name, enable);
        }

        return;
      }

      var action;
      if (enable !== undefined) {
        enable = typeof enable === 'function' ? enable.call(null, el) : enable;
        action = enable ? 'add' : 'remove';
      } else {
        action = BaseComponent.hasClass(el, name) ? 'remove' : 'add';
      }

      return BaseComponent[action + 'Class'](el, name);
    },


    /**
     * Add a class on an element.
     * @param {Element|Array} el An element or array of elements to update.
     * @param {String} name
     */
    addClass: function(el, name) {

      if (arguments.length === 2 && typeof name === 'string') {
        name = BaseComponent.trim(name).split(ws);
      } else {
        name = name instanceof Array ? name : Array.prototype.slice.call(arguments, 1);
      }

      // optimize for best, most common case
      if (name.length === 1 && el.classList) {
        if (name[0]) el.classList.add(name[0]);
        return el;
      }

      var toAdd = [];
      var i = 0;
      var l = name.length;
      var item;
      var clsName = typeof el.className === 'string' ? el.className : (el.getAttribute ? el.getAttribute('class') : '');

      // see if we have anything to add
      for (; i < l; i++) {
        item = name[i];
        if (item && !BaseComponent.hasClass(clsName, item)) {
          toAdd.push(item);
        }
      }

      if (toAdd.length) {
        if (typeof el.className === 'string') {
          el.className = BaseComponent.trim((clsName + ' ' + toAdd.join(' ')).replace(cleanup, ' '));
        } else if (el.setAttribute) {
          el.setAttribute('class', BaseComponent.trim((clsName + ' ' + toAdd.join(' ')).replace(cleanup, ' ')));
        }
      }

      return el;
    },


    /**
     * Remove a class on an element.
     * @param {Element|Array} el An element or array of elements to update.
     * @param {String} name
     */
    removeClass: function(el, name) {

      if (arguments.length === 2 && typeof name === 'string') {
        name = BaseComponent.trim(name).split(ws);
      } else {
        name = name instanceof Array ? name : Array.prototype.slice.call(arguments, 1);
      }

      // optimize for best, most common case
      if (name.length === 1 && el.classList) {
        if (name[0]) el.classList.remove(name[0]);
        return el;
      }

      // store two copies
      var clsName = ' ' + (typeof el.className === 'string' ? el.className : (el.getAttribute ? el.getAttribute('class') : '')) + ' ';
      var result = clsName;
      var current;
      var start;
      for (var i = 0, l = name.length; i < l; i++) {
        current = name[i];
        start = current ? result.indexOf(' ' + current + ' ') : -1;
        if (start !== -1) {
          start += 1;
          result = result.slice(0, start) + result.slice(start + current.length);
        }
      }

      // only write if modified
      if (clsName !== result) {
        if (typeof el.className === 'string') {
          el.className = BaseComponent.trim(result.replace(cleanup, ' '));
        } else if (el.setAttribute) {
          el.setAttribute('class', BaseComponent.trim(result.replace(cleanup, ' ')));
        }
      }

      return el;
    },


    /**
     * See if an element has a class.
     * @param {Element|String} el
     * @param {String} name
     * @return {Boolean}
     */
    hasClass: function(el, name) {
      var cName = (typeof el === 'object' ? el.className || ((el.getAttribute && el.getAttribute('class')) || '') : el || '').replace(/[\t\r\n\f]/g, ' ');
      return (' ' + cName + ' ').indexOf(' ' + name + ' ') !== -1;
    },


    /**
     * Get the index of a child element.
     * @param {NodeList} nodes A nodelist.
     * @param {Element} child A node.
     */
    getChildIndex: function(nodes, child) {
      return Array.prototype.indexOf.call(nodes, child);
    },


    /**
     * See if an element has another element for a parent.
     * @param {Element} child
     * @param {Element} possibleParent
     * @return {Boolean}
     */
    elementHasParent: function(child, possibleParent) {

      var parent = child.parentNode;

      while (parent) {

        if (parent === possibleParent) {
          return true;
        }

        parent = parent.parentNode;
      }

      return false;
    },


    /**
     * See if an element matches a query selector.
     * @param {Element} el
     * @param {String} query
     * @return {Boolean}
     */
    elementMatches: function(el, query) {

      if (vendorMatch) return vendorMatch.call(el, query);

      var nodes = el.parentNode.querySelectorAll(query);

      for (var i = 0; i < nodes.length; i++) {
        if (nodes[i] === el) return true;
      }

      return false;
    },


    /**
     * See if an element has another element for a parent.
     * @param {Element} parent
     * @param {String} query
     * @param {Array|Element} limitEl The last element we should check.
     * @return {Boolean|Element}
     */
    getElementMatchingParent: function(parent, query, limitEl) {

      limitEl = limitEl instanceof Array ? limitEl : [limitEl || document.body];

      while (parent) {

        if (BaseComponent.elementMatches(parent, query)) {
          return parent;
        }

        if (limitEl.indexOf(parent) !== -1) {
          return false;
        }

        parent = parent.parentNode;
      }

      return false;
    },


    /**
     * See if an element has parents which match a query.
     * @param {Element} parent
     * @param {String} query
     * @param {Element} limitEl The last element we should check.
     * @return {Boolean|Array}
     */
    getElementMatchingParents: function(parent, query, limitEl) {

      var list = [];

      while ((parent = BaseComponent.getElementMatchingParent(parent.parentNode, query, limitEl))) {
        list.push(parent);
      }

      return list;
    },


    /**
     * Get the offset position of the element.
     * @param {Element} el
     * @param {Boolean} lessScroll Subtract the scroll value
     * @return {Object}
     */
    getElementOffset: function(el, lessScroll) {

      var rect = {
        top: 0,
        left: 0
      };

      // Native implementation
      if (el.getBoundingClientRect) {

        var bounding = el.getBoundingClientRect();
        rect.left = bounding.left;
        rect.top = bounding.top;

        if (!lessScroll) {
          rect.left += typeof window.scrollX !== 'undefined' ? window.scrollX : window.pageXOffset;
          rect.top += typeof window.scrollY !== 'undefined' ? window.scrollY : window.pageYOffset;
        }
      } else {
        var x = 0,
          y = 0;
        do {
          x += el.offsetLeft + (lessScroll ? el.scrollLeft : 0);
          y += el.offsetTop + (lessScroll ? el.scrollTop : 0);
        }
        while ((el = el.offsetParent));

        rect.left = x;
        rect.top = y;
      }

      return rect;
    },


    /**
     * Get the index of an element in a nodelist.
     * @param {NodeList} els
     * @param {Node} el
     */
    getNodeListIndex: function(els, el) {
      return Array.prototype.indexOf.call(els, el);
    },


    /**
     * Trim whitespace on a string.
     * @param {String} str
     */
    trim: function(str) {
      return str.replace(trimRE, '');
    },


    /**
     * Round to a given number of decimal places.
     * @param {Number} num
     * @param {Number} len
     */
    round: function(num, len) {
      len = len !== undefined ? len : 2;
      var x = Math.pow(10, len);
      return Math.round(num * x) / x;
    },


    /**
     * Append an array of children to a node.
     * @param {Element} el
     * @param {Array} children
     * @param {Boolean} empty Empty the node before adding children?
     */
    appendChildren: function(el, children, empty) {

      empty = empty === undefined ? false : empty;

      if (empty) {
        el.textContent = '';
      }

      var domList = children instanceof window.HTMLCollection;

      if (domList) {
        while (children.length) {
          el.appendChild(children[0]);
        }
      } else {

        var i = 0;
        var len = children.length;

        for (; i < len; i++) {
          if (children[i]) {
            el.appendChild(children[i]);
          }
        }
      }
    },


    /**
     * Insert an array of elements before a node.
     * @param {Element} el
     * @param {Element} beforeEl
     * @param {Array} children
     */
    insertBefore: function(el, beforeEl, children) {

      var i = 0;
      var len = children.length;

      for (; i < len; i++) {
        el.insertBefore(children[i], beforeEl);
      }
    },


    /**
     * Find the active breakpoint.
     * @param {Number} width
     */
    getBreakpoint: function(width, breakpoints) {

      breakpoints = breakpoints || BaseComponent.breakpoints;

      var i;

      for (i in breakpoints) {
        if (width >= breakpoints[i].min && width <= breakpoints[i].max) {
          return i;
        }
      }
    },


    /**
     * Remove the component from the DOM and prepare for garbage collection by dereferencing values.
     * @param {Boolean} leaveElement Leave the element intact.
     */
    remove: function(leaveElement) {

      (this._removeEventListeners || noop)();

      if (!leaveElement && this.el.parentNode) {
        this.el.parentNode.removeChild(this.el);
      }

      this._unsetParams(this.defaults);
    },


    /**
     * Trigger a DOM event on an element.
     * @param {Element} el
     * @param {String} name
     */
    triggerEvent: function(el, name) {

      var event;

      if (document.createEvent) {
        event = document.createEvent('HTMLEvents');
        event.initEvent(name, true, true);
        event.eventName = name;
        el.dispatchEvent(event);
      } else {
        event = document.createEventObject();
        event.eventType = name;
        event.eventName = name;
        el.fireEvent('on' + event.eventType, event);
      }
    },


    /**
     * Scroll the window to a specific element or position.
     * @param {Object} params
     */
    scrollWindowTo: function(params) {

      params = params || {};

      var offset;
      var x;
      var y;

      if (params instanceof HTMLElement) {
        offset = BaseComponent.getElementOffset(params);
        x = offset.left;
        y = offset.top;
        params = arguments[1] || {};
      } else {
        x = params.x || 0;
        y = params.y || 0;
      }

      BaseComponent.tween({
        target: window,
        prop: 'scrollTo',
        start: [window.scrollX, window.scrollY],
        end: [x, y],
        duration: params.duration,
        callback: params.callback
      });
    },


    /**
     * Tween from one value to another.
     * @param {Object} params
     * @return {Long}
     */
    tween: function(params) {

      params = params || {};

      var begin;
      var obj = params.target;

      if (!obj) {
        throw new Error('Cannot tween without a target!');
      }

      var prop = typeof params.prop === 'string' ? [params.prop] : params.prop;
      var start = typeof params.start === 'number' ? [params.start] : params.start;
      var end = typeof params.end === 'number' ? [params.end] : params.end;
      var duration = params.duration || 250;
      var callback = params.callback || noop;

      // Ensure we have the same number of start and end properties.
      if (start.length !== end.length) {
        throw new Error('Cannot tween two different sets of parameters!');
      }

      var f = function(ts) {

        // Keep track of when we start
        if (!begin)
          begin = ts;

        // Progress
        var prog = ts - begin;

        // Percentage complete
        var per = Math.min(prog / duration, 1);

        // Adjust the values for the percentage complete.
        var args = [];
        var i = 0;
        var len = start.length;
        for (; i < len; i++) {
          args[i] = start[i] + ((end[i] - start[i]) * per);
        }

        // Apply the values for each property.
        i = 0;
        len = prop.length;
        var arg;
        for (; i < len; i++) {

          // If this is the last property but we have more arguments, set them all.
          arg = i + 1 === len && args.length - 1 > i ? args.slice(i) : args[i];

          if (typeof obj[prop[i]] === 'function') {
            obj[prop[i]].apply(obj, arg);
          } else {
            obj[prop[i]] = arg;
          }
        }

        // Keep going if we have more to do.
        if (prog < duration)
          raf(f);
        else
          callback();
      };

      return raf(f);
    },


    /**
     * Copy all of the attributes from one element to another.
     * @param {Element} a
     * @param {Element} b
     */
    copyAttributes: function(a, b) {

      var i = 0;
      var len = a.attributes.length;

      for (; i < len; i++) {
        b.setAttribute(a.attributes[i].name, a.attributes[i].value);
      }
    },

    /**
     * Wrap an element with another element
     * @param {Element} el
     * @param {Element} wrapper
     * @return {Element}
     */
    wrapElement: function(el, wrapper) {
      wrapper = wrapper || document.createElement('div');
      if (el.nextSibling) {
        el.parentNode.insertBefore(wrapper, el.nextSibling);
      } else {
        el.parentNode.appendChild(wrapper);
      }
      return wrapper.appendChild(el);
    },


    /**
     * Create a range of numbers.
     * @param {Number} start
     * @param {Number} stop
     * @param {Number} step Optional
     */
    range: function(start, stop, step) {
      if (stop == null) {
        stop = start || 0;
        start = 0;
      }
      if (!step) {
        step = stop < start ? -1 : 1;
      }

      var length = Math.max(Math.ceil((stop - start) / step), 0);
      var range = new Array(length);

      for (var idx = 0; idx < length; idx++, start += step) {
        range[idx] = start;
      }

      return range;
    },


    /**
     * Get a nearest sibling before the given element which matches
     * the given query selector.
     * @param {Element} el
     * @param {String} query
     * @return {Element|Null}
     */
    getSiblingBefore: function(el, query) {

      while ((el = el.previousElementSibling)) {
        if (BaseComponent.elementMatches(el, query)) {
          return el;
        }
      }

      return null;
    },


    /**
     * Get a nearest sibling after the given element which matches
     * the given query selector.
     * @param {Element} el
     * @param {String} query
     * @return {Element|Null}
     */
    getSiblingAfter: function(el, query) {

      while ((el = el.nextElementSibling)) {
        if (BaseComponent.elementMatches(el, query)) {
          return el;
        }
      }

      return null;
    },


    /**
     * Get a child that matches the selector.
     * @param {Element} el
     * @param {String} query
     * @return {Element|Null}
     */
    getMatchingChild: function(el, query) {

      var i = 0;
      var len = el.children.length;

      for (; i < len; i++) {
        if (BaseComponent.elementMatches(el.children[i], query)) {
          return el.children[i];
        }
      }

      return null;
    },

    /**
     * See if an element has children which match a query.
     * @param {Element} el
     * @param {String} query
     * @return {List}
     */
    getElementMatchingChildren: function(el, query) {

      var list = [];
      var i = 0;
      var len = el.children.length;

      for (; i < len; i++) {
        if (BaseComponent.elementMatches(el.children[i], query)) {
          list.push(el.children[i]);
        }
      }

      return list;
    },
  };

  return BaseComponent;
}));

/**
 * # Each
 * Apply a callback function to a list of any kind: Array, NodeList, HTMLCollection or Object.
 *
 * @example
 * each([], callback);
 * each(el.children, callback);
 * each(el.childNodes, callback);
 * each({}, callback);
 *
 * @module helper/each.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define([], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory();
  } else {
    root.Spark = root.Spark || {};
    root.Spark.each = factory();
  }
}(this, function() {

  'use strict';

  /**
   * @param {Array|NodeList|HTMLCollection|Object} list
   * @param {Function} cb
   */
  function each(list, cb) {

    if (!list) {
      return;
    }

    if (typeof cb !== 'function') {
      throw new Error('Cannot invoke `each` without a callback!');
    }

    var i = 0;
    var len = list.length;

    // Object
    if (len === undefined) {
      for (i in list) {
        if (i !== 'prototype' && list.hasOwnProperty(i)) {
          cb(i, list[i]);
        }
      }
    }
    // Array-like
    else {
      for (; i < len; i++) {
        cb(list[i]);
      }
    }
  }

  return each;
}));

/**
 * # Set Caret
 * Set the caret position in an input.
 *
 * @example
 * setCaret(el, -1);
 * setCaret(el, 2);
 *
 * @module helper/set-caret.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define([], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory();
  } else {
    root.Spark = root.Spark || {};
    root.Spark.setCaret = factory();
  }
}(this, function() {

  'use strict';

  return function setCaret(el, start, end) {

    var originalActiveElement = document.activeElement;

    start = start < 0 ? el.value.length + start + 1 : start;
    end = end < 0 ? el.value.length + end + 1 : end;

    // IE support
    if (document.selection) {
      el.focus();
      var sel = document.selection.createRange();
      sel.moveStart('character', -el.value.length);
      sel.moveStart('character', start);
      sel.moveEnd('character', end !== undefined ? end : start);
      sel.select();
    } else if (el.selectionStart || el.selectionStart === 0) {
      el.selectionStart = start;
      el.selectionEnd = end !== undefined ? end : start;
    }

    // If we didn't have focus, go back to focusing on the original
    if (originalActiveElement !== el) {
      el.blur();
      originalActiveElement.focus();
    }

    return {
      start: start,
      end: end
    };
  };
}));

/**
 * # Store Form Data
 * Store the data of form fields inline so it can be reverted.
 *
 * @example
 * formData.store(el);
 * formData.revert(el);
 *
 * @module helper/store-form-data.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define([], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory();
  } else {
    root.Spark = root.Spark || {};
    root.Spark.storeFormData = factory();
  }
}(this, function() {

  'use strict';

  /**
   * Find all the form elements inside a given element and run a callback on each.
   * @param  {Element} el
   * @param  {Function} cb
   */
  function findAll(el, cb) {

    var inputs = el.querySelectorAll('input, select, textarea');
    var i = 0;
    var len = inputs.length;

    for (; i < len; i++) {
      cb(inputs[i]);
    }
  }

  return {

    /**
     * Find all the form elements inside a given element and store their current
     * value as a data attribute.
     * @param {Element} el
     */
    store: function(el) {

      findAll(el, function(input) {

        var name = input.nodeName.toLowerCase();
        var value;

        switch (name) {
          case 'select':
            value = input.selectedIndex;
            break;
          default:
            value = encodeURI(input.value);
            break;
        }

        input.setAttribute('data-stored-value', value);
      });
    },


    /**
     * Revert all the form elements inside of a given element.
     * @param {Element} el
     */
    restore: function(el) {

      findAll(el, function(input) {

        var name = input.nodeName.toLowerCase();
        var value = input.getAttribute('data-stored-value');

        // No stored value
        if (!value && value !== '') {
          return;
        }

        switch (name) {
          case 'select':
            input.options[value].selected = true;
            break;
          default:
            input.value = decodeURI(value);
            break;
        }

        input.removeAttribute('data-stored-value');
      });
    },


    /**
     * Clear the stored data on all the form elements inside of a given element.
     * @param {Element} el
     */
    clear: function(el) {
      findAll(el, function(input) {
        input.removeAttribute('data-stored-value');
      });
    }
  };
}));

/**
 * # Build Select
 * Build a select input.
 *
 * @example
 * buildSelect({
 *   attributes: {
 *     name: 'test',
 *     class: 'spark-select__field'
 *   },
 *   options: [1, 2, 3, 4, 5, 6]
 * });
 *
 * @example
 * buildSelect({
 *   attributes: {
 *     name: 'test',
 *     class: 'spark-select__field',
 *     'data-attr': 'thing',
 *      multiple: true
 *   },
 *   selected: [2, 4],
 *   options: [
 *     {text: 'One', value: 1},
 *     {text: 'Two', value: 2},
 *     {text: 'Three', value: 3},
 *     {text: 'Four', value: 4},
 *     {text: 'Five', value: 5},
 *   ]
 * });
 *
 * @module helper/build-select.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.buildSelect = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  /**
   * @param {Object} params
   */
  function buildSelect(params) {

    var el = document.createElement('select');
    var html = '';
    var attrs = params.attributes;
    var selected = params.selected instanceof Array ? params.selected : (params.selected ? [params.selected] : []);
    var opts = params.options;

    var i;
    var len = opts.length;

    // Set attributes
    for (i in attrs) {
      el.setAttribute(i, attrs[i]);
    }

    // Add options
    for (i = 0; i < len; i++) {
      if (typeof opts[i] === 'object') {
        html += '<option value="' + opts[i].value + '" ' + (selected.indexOf(opts[i].value) !== -1 ? 'selected' : '') + '>' + opts[i].text + '</option>';
      } else {
        html += '<option value="' + opts[i] + '" ' + (selected.indexOf(opts[i]) !== -1 ? 'selected' : '') + '>' + opts[i] + '</option>';
      }
    }

    el.innerHTML = html;

    return el;
  }

  Base.exportjQuery(buildSelect, 'buildSelect');

  return buildSelect;
}));

/**
 * # Date helper
 * General helpers for working with dates.
 *
 * @module helper/date.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.dateHelper = factory(root.Spark.BaseComponent);
  }
}(this, function() {

  'use strict';

  var monthNames = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];
  var monthNamesShort = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'June', 'July', 'Aug', 'Sept', 'Oct', 'Nov', 'Dec'];
  var dayNames = ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'];
  var dayNamesShort = ['Sun', 'Mon', 'Tues', 'Wed', 'Thur', 'Fri', 'Sat'];
  var adjustedDayNames = [];
  var adjustedDayNamesShort = [];
  var weekStartsOn = 0;

  var dateHelper = {

    /**
     * Transform a date into an object of date values.
     * @param {Date} date
     * @return {Object}
     */
    create: function(date) {

      date = date instanceof Date ? date : new Date(date.year, date.month - 1, date.day);

      var inst = Object.create(dateHelper);
      inst._date = date;
      inst._cache = {};

      return inst;
    },


    /**
     * Get a year.
     * @return {Number}
     */
    get year() {
      this._instanceCheck('year');
      return this._date.getFullYear();
    },


    /**
     * Set a year.
     * @param {Number} y
     */
    set year(y) {
      this._instanceCheck('year');
      this._clearCache();
      return this._date.setFullYear(y);
    },


    /**
     * Get a month.
     * @return {Number} 1-12
     */
    get month() {
      this._instanceCheck('month');
      return this._date.getMonth() + 1;
    },


    /**
     * Set a month.
     * @param {Number} m 1-12
     */
    set month(m) {
      this._instanceCheck('month');
      this._clearCache();
      return this._date.setMonth(m - 1);
    },


    /**
     * Get a day.
     * @return {Number} 1-31
     */
    get day() {
      this._instanceCheck('day');
      return this._date.getDate();
    },


    /**
     * Set a day.
     * @param {Number} d 1-31
     */
    set day(d) {
      this._instanceCheck('day');
      this._clearCache();
      return this._date.setDate(d);
    },


    /**
     * Sets the day, month and year values at once.
     * @param {Object} params
     */
    set: function(params) {
      params = params || {};
      this.year = params.year || this.year;
      this.month = params.month || this.month;
      this.day = params.day || this.day;
    },


    /**
     * Get the full name of the month.
     * @param {Number} num
     * @return {String}
     */
    getMonthName: function(num) {
      return monthNames[num - 1];
    },


    /**
     * Get the month name.
     * @return {String}
     */
    get monthName() {
      this._instanceCheck('monthName');
      return dateHelper.getMonthName(this.month);
    },


    /**
     * Get the list of month names.
     * @return {Array}
     */
    getMonthNames: function() {
      return monthNames;
    },


    /**
     * Get the short name of the month.
     * @param {Number} num
     * @return {String}
     */
    getMonthNameShort: function(num) {
      return monthNamesShort[num - 1];
    },


    /**
     * Get the month name.
     * @return {String}
     */
    get monthNameShort() {
      this._instanceCheck('monthName');
      return dateHelper.getMonthNameShort(this.month);
    },


    /**
     * Get the list of short month names.
     * @return {Array}
     */
    getMonthNamesShort: function() {
      return monthNamesShort;
    },


    /**
     * Set the month names.
     * @param {Array} names
     */
    setMonthNames: function(names) {
      if (names.length === 12) monthNames = names;
    },


    /**
     * Set the short month names.
     * @param {Array} names
     */
    setMonthNamesShort: function(names) {
      if (names.length === 12) monthNamesShort = names;
    },


    /**
     * Get the day of the week for a given day.
     * @param {Object} date
     * @return {Number} 1-7
     */
    getDayOfWeek: function(date) {
      var day = (date instanceof Date ? date : new Date(date.year, date.month - 1, date.day)).getDay() - weekStartsOn;
      return (day < 0 ? 7 - Math.abs(day) : day) + 1;
    },


    /**
     * Get the day of the week.
     * @return {Number}
     */
    get dayOfWeek() {
      return dateHelper.getDayOfWeek(this._date);
    },


    /**
     * Get the full name of a day of the week.
     * @param {Number} num
     * @return {String}
     */
    getDayName: function(num) {
      return dayNames[num - 1 + weekStartsOn] || dayNames[dayNames.length - num - 1 + weekStartsOn];
    },


    /**
     * Get the day name.
     * @return {String}
     */
    get dayName() {
      this._instanceCheck('dayName');
      return dateHelper.getDayName(this.dayOfWeek);
    },


    /**
     * Get the full name of the days of the week.
     * @return {Array}
     */
    getDayNames: function() {
      return adjustedDayNames.length ? adjustedDayNames : dayNames;
    },


    /**
     * Get the short name of the day.
     * @param {Number} num
     * @return {String}
     */
    getDayNameShort: function(num) {
      return dayNamesShort[num - 1 + weekStartsOn] || dayNames[dayNames.length - num - 1 + weekStartsOn];
    },


    /**
     * Get the short day name.
     * @return {String}
     */
    get dayNameShort() {
      this._instanceCheck('dayNameShort');
      return dateHelper.getDayNameShort(this.dayOfWeek);
    },


    /**
     * Get the full name of the days of the week.
     * @return {Array}
     */
    getDayNamesShort: function() {
      return adjustedDayNamesShort.length ? adjustedDayNamesShort : dayNamesShort;
    },


    /**
     * Set the day names.
     * @param {Array} names
     */
    setDayNames: function(names) {
      if (names.length === 7) dayNames = names;
    },


    /**
     * Set the short day names.
     * @param {Array} names
     */
    setDayNamesShort: function(names) {
      if (names.length === 7) dayNamesShort = names;
    },


    /**
     * Get the index of the first day of the week.
     * @return {Number}
     */
    getWeekStartsOn: function() {
      return weekStartsOn;
    },


    /**
     * Set the index of the first day of the week.
     * @param {Number} index
     * @return {String}
     */
    setWeekStartsOn: function(number) {

      weekStartsOn = number;

      if (number) {
        adjustedDayNames = dayNames.slice(weekStartsOn);
        adjustedDayNames = adjustedDayNames.concat(dayNames.slice(0, weekStartsOn));
        adjustedDayNamesShort = dayNamesShort.slice(weekStartsOn);
        adjustedDayNamesShort = adjustedDayNamesShort.concat(dayNamesShort.slice(0, weekStartsOn));
      } else {
        adjustedDayNames = [];
        adjustedDayNamesShort = [];
      }
    },


    /**
     * Get the current date.
     * @return {Object}
     */
    now: function() {
      return dateHelper.create(new Date());
    },


    /**
     * Get the next year after the given date.
     * This obviously isn't very complicated, but it exists
     * for parity with how we get the week, day and month.
     * @param {Object} date
     * @return {Object}
     */
    getNextYear: function(date) {
      return dateHelper.create(new Date(date.year + 1, date.month - 1, date.day));
    },


    /**
     * Get the year following this.
     * @return {Object}
     */
    get nextYear() {
      this._instanceCheck('nextYear');
      return this._cache.nextYear || (this._cache.nextYear = dateHelper.getNextYear(this));
    },


    /**
     * Get the first day of the week for a given date.
     * @param {Object} date
     * @return {Object}
     */
    getWeekStart: function(date) {
      var inst = dateHelper.create(new Date(date.year, date.month - 1, date.day - dateHelper.getDayOfWeek(date) + 1));
      inst.weekStartsOn = weekStartsOn;
      return inst;
    },


    /**
     * Get the start of the week for this date.
     * @return {Object}
     */
    get weekStart() {
      this._instanceCheck('weekStart');
      return this._cache.weekStart && this._cache.weekStart.weekStartsOn === weekStartsOn ? this._cache.weekStart : (this._cache.weekStart = dateHelper.getWeekStart(this));
    },


    /**
     * Get the first day of the month for a given date.
     * @param {Object} date
     * @return {Object}
     */
    getMonthStart: function(date) {
      var inst = dateHelper.create(new Date(date.year, date.month - 1, 1));
      return inst;
    },


    /**
     * Get the start of the month for this date.
     * @return {Object}
     */
    get monthStart() {
      this._instanceCheck('monthStart');
      return this._cache.monthStart || (this._cache.monthStart = dateHelper.getMonthStart(this));
    },


    /**
     * Get the next week after the given date.
     * @param {Object} date
     * @return {Object}
     */
    getNextWeek: function(date) {
      var start = dateHelper.getWeekStart(date);
      return dateHelper.create(new Date(start.year, start.month - 1, start.day + 7));
    },


    /**
     * Get the week following this.
     * @return {Object}
     */
    get nextWeek() {
      this._instanceCheck('nextWeek');
      return this._cache.nextWeek || (this._cache.nextWeek = dateHelper.getNextWeek(this));
    },


    /**
     * Get the next day after the given date.
     * @param {Object} date
     * @return {Object}
     */
    getNextDay: function(date) {
      return dateHelper.create(new Date(date.year, date.month - 1, date.day + 1));
    },


    /**
     * Get the day following this.
     * @return {Object}
     */
    get nextDay() {
      this._instanceCheck('nextDay');
      return this._cache.nextDay || (this._cache.nextDay = dateHelper.getNextDay(this));
    },


    /**
     * Get the next month after the given date.
     * @param {Object} date
     * @return {Object}
     */
    getNextMonth: function(date) {
      // Date() has a bug if last day of month is 31 when calculating the next month.
      // Need to account for that so that it doesn't round up the date/month.
      return date.day === 31 ? dateHelper.create(new Date(date.year, date.month, 1)) : dateHelper.create(new Date(date.year, date.month, date.day));
    },


    /**
     * Get the month following this.
     * @return {Object}
     */
    get nextMonth() {
      this._instanceCheck('nextMonth');
      return this._cache.nextMonth || (this._cache.nextMonth = dateHelper.getNextMonth(this));
    },


    /**
     * Get the previous year after the given date.
     * This obviously isn't very complicated, but it exists
     * for parity with how we get the week, day and month.
     * @param {Object} date
     * @return {Object}
     */
    getPreviousYear: function(date) {
      return dateHelper.create(new Date(date.year - 1, date.month - 1, date.day));
    },


    /**
     * Get the year preceding this.
     * @return {Object}
     */
    get previousYear() {
      this._instanceCheck('previousYear');
      return this._cache.previousYear || (this._cache.previousYear = dateHelper.getPreviousYear(this));
    },


    /**
     * Get the previous week after the given date.
     * @param {Object} date
     * @return {Object}
     */
    getPreviousWeek: function(date) {
      var start = dateHelper.getWeekStart(date);
      var inst = dateHelper.create(new Date(start.year, start.month - 1, start.day - 7));
      inst.weekStartsOn = weekStartsOn;
      return inst;
    },


    /**
     * Get the week preceding this.
     * @return {Object}
     */
    get previousWeek() {
      this._instanceCheck('previousWeek');
      return this._cache.previousWeek || (this._cache.previousWeek = dateHelper.getPreviousWeek(this));
    },


    /**
     * Get the previous day after the given date.
     * @param {Object} date
     * @return {Object}
     */
    getPreviousDay: function(date) {
      return dateHelper.create(new Date(date.year, date.month - 1, date.day - 1));
    },


    /**
     * Get the day preceding this.
     * @return {Object}
     */
    get previousDay() {
      this._instanceCheck('previousDay');
      return this._cache.previousDay || (this._cache.previousDay = dateHelper.getPreviousDay(this));
    },


    /**
     * Get the previous month after the given date.
     * @param {Object} date
     * @return {Object}
     */
    getPreviousMonth: function(date) {
      return dateHelper.create(new Date(date.year, date.month - 2, date.day));
    },


    /**
     * Get the month preceding this.
     * @return {Object}
     */
    get previousMonth() {
      this._instanceCheck('previousMonth');
      return this._cache.previousMonth || (this._cache.previousMonth = dateHelper.getPreviousMonth(this));
    },


    /**
     * Get the last day of the month.
     * @param {Object} date
     * @return {Object}
     */
    getMonthEnd: function(date) {
      return dateHelper.create(new Date(date.year, date.month, 0));
    },


    /**
     * Get the last day of the month.
     * @return {Object}
     */
    get monthEnd() {
      this._instanceCheck('monthEnd');
      return this._cache.monthEnd || (this._cache.monthEnd = dateHelper.getMonthEnd(this));
    },


    /**
     * Does a given day equal another? Or is it present in a list of others?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    equal: function(date, compare, full) {
      return this.equalDay(date, compare, full);
    },


    /**
     * Does a given day equal another? Or is it present in a list of others?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    equalDay: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (compare[i] && date.year === compare[i].year && date.month === compare[i].month && date.day === compare[i].day)
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a week equal to another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    equalWeek: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || !date.weekStart.equalDay(compare[i].weekStart))
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a month equal to another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    equalMonth: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || (date.year !== compare[i].year) || (date.year === compare[i].year && date.month !== compare[i].month))
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a year equal to another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    equalYear: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || date.year !== compare[i].year)
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a date before another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    before: function(date, compare, full) {
      return this.beforeDay(date, compare, full);
    },


    /**
     * Is a given date before another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    beforeDay: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || date._date >= compare[i]._date)
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a week before another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    beforeWeek: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || !date.weekStart.beforeDay(compare[i].weekStart))
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a month before another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    beforeMonth: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || (date.year > compare[i].year) || (date.year === compare[i].year && date.month >= compare[i].month))
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a year before another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    beforeYear: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || date.year >= compare[i].year)
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a date after another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    after: function(date, compare, full) {
      return this.afterDay(date, compare, full);
    },


    /**
     * Is a given date after another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    afterDay: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || date._date <= compare[i]._date)
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a week after another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    afterWeek: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || !date.weekStart.afterDay(compare[i].weekStart))
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a month after another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    afterMonth: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || (date.year < compare[i].year) || (date.year === compare[i].year && date.month <= compare[i].month))
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Is a year after another?
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full Return a successful match only if all matches are found.
     * @return {Boolean}
     */
    afterYear: function(date, compare, full) {

      var args = this._checkComparisonArgs(date, compare, full);
      date = args[0];
      compare = args[1];
      full = args[2];

      var i = 0;
      var len = compare.length;
      var matches = 0;

      for (; i < len; i++) {
        if (!compare[i] || date.year <= compare[i].year)
          continue;
        else
          matches++;
      }

      return full ? matches === len : !!matches;
    },


    /**
     * Get the earliest date in an array.
     * @return {Object}
     */
    earliest: function(arr) {

      var i = 0;
      var len = arr.length;
      var e;

      for (; i < len; i++) {
        if (!e || arr[i].before(e))
          e = arr[i];
      }

      return e;
    },


    /**
     * Get the latest date in an array.
     * @return {Object}
     */
    latest: function(arr) {

      var i = 0;
      var len = arr.length;
      var l;

      for (; i < len; i++) {
        if (!l || arr[i].after(l))
          l = arr[i];
      }

      return l;
    },


    /**
     * Clone a date instance.
     * @param {Object} date
     * @return {Object}
     */
    clone: function(date) {

      // If we weren't passed a date, use this instance.
      if (!date && this._date && this._date instanceof Date && dateHelper.isPrototypeOf(this)) {
        date = this;
      }

      // No date, can't clone.
      if (!date) {
        throw new Error('Must pass a date to clone or call on an instance.');
      }

      return dateHelper.create(new Date(date._date.valueOf()));
    },


    /**
     * If a comparison function is called on an instance, properly
     * assign the vars.
     * @param {Object} date
     * @param {Object|Array} compare
     * @param {Boolean} full
     */
    _checkComparisonArgs: function(date, compare, full) {

      if (compare === undefined || typeof compare === 'boolean') {

        if (!dateHelper.isPrototypeOf(this)) {
          throw new Error('Cannot compare only one date!');
        }

        full = compare;
        compare = date;
        date = this;
      }

      compare = compare instanceof Array ? compare : [compare];

      return [date, compare, full];
    },


    /**
     * Check to see if we have an instance of the date object.
     * @param  {String} prop
     */
    _instanceCheck: function(prop) {
      if (!this._date || !(this._date instanceof Date) || !dateHelper.isPrototypeOf(this))
        throw new Error('Cannot access the property "' + prop + '" of the date helper with creating an instance!');
    },


    /**
     * Clear the cache.
     */
    _clearCache: function() {
      this._instanceCheck('clearCache');
      this._cache = {};
    }
  };

  return dateHelper;
}));

/**
 * # Outer Height
 * Get the outer height of an element (including margin and border)
 *
 * @example
 * outerHeight(el, computedStyles);
 *
 * @module helper/outer-height.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.outerHeight = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  /**
   * @param {Element} el
   * @param {Object} styles Optional Already have computed styles? Pass them in.
   */
  function outerHeight(el, styles) {

    styles = styles || window.getComputedStyle(el);

    var props = ['marginTop', 'marginBottom', 'borderTop', 'borderBottom'];
    var height = el.clientHeight;

    props.forEach(function(prop) {
      height += parseInt(styles[prop] || 0, 10);
    });

    return height;
  }

  Base.exportjQuery(outerHeight, 'outerHeight');

  return outerHeight;
}));

/**
 * # Pad Number
 * Pad a number with leading zeros
 *
 * @example
 * pad(4, 2);
 *
 * @module helper/pad-number.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define([], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory();
  } else {
    root.Spark = root.Spark || {};
    root.Spark.padNumber = factory();
  }
}(this, function() {

  'use strict';

  function pad(n, width, c) {
    c = c || '0';
    n = n + '';
    return n.length >= width ? n : new Array(width - n.length + 1).join(c) + n;
  }

  return pad;
}));

/**
 * # Animate Height
 * Animate the height of an element since we can't do this w/ pure CSS. Sigh.
 *
 * @example
 * animateHeight({
 *   el: ...,
 *   toggleEl: ...,
 *   // Optional params
 *   action: 'collapse'|'expand',
 *   heightAnimationClass: 'spark-animate-height',
 *   opacityAnimationClass: 'spark-animate-opacity',
 *   toggleProperty: 'display'|'overflow'|'visibility',
 *   toggleValue: 'block'|'none'|'visible'|'hidden',
 *   animationDuration: 250
 * });
 *
 * @module helper/animate-height.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/outer-height'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'), require('./outer-height'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.animateHeight = factory(root.Spark.BaseComponent, root.Spark.outerHeight);
  }
}(this, function(Base, outerHeight) {

  'use strict';

  var noop = function() {};
  var runningAnimations = {
    els: [],
    completeCallbacks: []
  };

  /**
   * Get the inverse toggle value
   * @param  {String} property
   * @param  {String} originalValue
   * @return {String}
   */
  function getInverseToggleValue(el, property, originalValue) {

    // Get the value to toggle to for the given property
    switch (property) {
      case 'overflow':
      case 'visibility':
        return originalValue === 'visible' ? 'hidden' : 'visible';
      default:
        return originalValue === 'block' || originalValue === 'inline-block' ? 'none' : 'block';
    }
  }

  function onComplete(params) {

    // Reset toggle el visibility
    params.toggleEl.style[params.toggleProperty] = '';

    // Remove the height property
    params.el.style.height = '';
    params.toggleEl.style.height = '';
    params.toggleEl.style.marginBottom = '';
    params.toggleEl.style.marginTop = '';

    // Remove the spark-animate-height class so the transitions no longer apply
    Base.removeClass(params.el, params.heightAnimationClass);
    Base.removeClass(params.toggleEl, params.heightAnimationClass);

    // Run the callback
    params.onComplete();

    // Remove the element and callback from their respective arrays
    var runningIndex = runningAnimations.els.indexOf(params.el);
    runningAnimations.els.splice(runningIndex, 1);
    runningAnimations.completeCallbacks.splice(runningIndex, 1);
  }

  /**
   * @param {Object} params
   */
  function animateHeight(params) {

    params = params || {};

    var el = params.el;

    if (!el) {
      return;
    }

    var collapse = params.action && params.action === 'collapse';
    var heightAnimationClass = params.heightAnimationClass || 'spark-animate-height';

    // Allow for elements to be passed or selector strings
    var toggleEl = typeof params.toggleEl === 'string' ? el.querySelector(params.toggleEl) : params.toggleEl;

    // No element to be switching with toggling so we can't determine the desired height to animate to.
    if (!toggleEl || Base.hasClass(el, 'spark-no-animate')) {
      return;
    }

    // The style property to use when toggling visibility
    var toggleProperty = params.toggleProperty || 'display';
    var toggleStyles = window.getComputedStyle(toggleEl);
    var originalToggleValue = toggleStyles[toggleProperty];
    var toggleValue = params.toggleValue || getInverseToggleValue(toggleProperty, originalToggleValue);

    // If we are already animating, stop now.
    var runningIndex = runningAnimations.els.indexOf(el);
    if (runningIndex !== -1) {

      var completeCallback = runningAnimations.completeCallbacks[runningIndex];
      if (completeCallback) {
        clearTimeout(completeCallback);
      }

      onComplete({
        el: el,
        toggleEl: toggleEl,
        onComplete: params.onComplete || noop,
        collapse: collapse,
        toggleProperty: toggleProperty,
        toggleValue: toggleValue,
        heightAnimationClass: heightAnimationClass
      });
    }

    // Store the current height
    var originalHeight = outerHeight(el);

    // Toggle the visible property
    toggleEl.style[toggleProperty] = toggleValue;

    // Now that the toggle el is taking up space, get the new height which we will be animating to
    var targetElHeight = outerHeight(el);

    // We need to store the original and target toggle element heights. They differ depending on
    // whether we are going to expand or collapse.
    var targetToggleElHeight;
    var originalToggleElHeight;

    // If we are collapsing, reset the toggle style and set it when we're done. Set the height so
    // that we can animate down to 0 or up to the target height.
    if (collapse) {
      toggleEl.style[toggleProperty] = originalToggleValue;
      originalToggleElHeight = outerHeight(toggleEl, toggleStyles);
      targetToggleElHeight = 0;
    } else {
      targetToggleElHeight = outerHeight(toggleEl, toggleStyles);
      originalToggleElHeight = 0;
    }

    // Set the original height
    el.style.height = originalHeight + 'px';
    toggleEl.style.height = originalToggleElHeight + 'px';
    toggleEl.style.marginBottom = '0px';
    toggleEl.style.marginTop = '0px';

    // Add the spark-animate-height class which will setup the transition-property and duration
    Base.addClass(el, heightAnimationClass);
    Base.addClass(toggleEl, heightAnimationClass);

    runningAnimations.els.push(el);

    // We need to wait a tick to toggle the height or else the animation class won't function
    setTimeout(function() {

      // Set the height to the target height
      el.style.height = targetElHeight + 'px';
      toggleEl.style.height = targetToggleElHeight + 'px';

      // Remove inline styles after the animation is complete
      runningAnimations.completeCallbacks.push(setTimeout(function() {
        onComplete({
          el: el,
          toggleEl: toggleEl,
          onComplete: params.onComplete || noop,
          collapse: collapse,
          toggleProperty: toggleProperty,
          toggleValue: toggleValue,
          heightAnimationClass: heightAnimationClass
        });
      }, params.animationDuration !== undefined ? params.animationDuration : 250));
    }, 0);
  }

  Base.exportjQuery(animateHeight, 'animateHeight');

  return animateHeight;
}));

/**
 * # Parse Date Format
 * Given a date format string, break it down into pieces.
 *
 * @example
 * parseDateFormat('MM-DD-YYYY');
 *
 * @module helper/parse-date-format.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/pad-number', 'spark/helpers/date'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'), require('./pad-number'), require('./date'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.parseDateFormat = factory(root.Spark.BaseComponent, root.Spark.padNumber, root.Spark.dateHelper);
  }
}(this, function(Base, padNumber, dateHelper) {

  'use strict';

  /**
   * Map characters to their special meanings.
   * @type {Object}
   */
  var map = {
    d: 'day',
    m: 'month',
    y: 'year',
    '-': 'divider',
    '/': 'divider',
    ' ': 'space'
  };

  /**
   * Given a format and a string, get the day, month and year values from that string.
   * @param {Object} format
   * @return {Function}
   */
  function getValues(format) {

    return function(value) {

      var i = 0;
      var index = 0;
      var len = format.length;
      var values = {};

      // Loop through all format pieces
      for (; i < len; i++) {

        // Only worry about date values
        if (['day', 'month', 'year'].indexOf(format[i].name) !== -1) {

          // If the passed value doesn't contain a format piece, it's invalid.
          if (value.length < index + format[i].length)
            return;

          values[format[i].name] = parseInt(value.substr(index, format[i].length), 10);
        }
        index += format[i].length;
      }

      return dateHelper.create(values);
    };
  }

  /**
   * Create a formatted date string given an object of values.
   * @param {Object} format
   * @return {Function}
   */
  function getString(format) {

    /**
     * @param {Object} vals
     * @return {String}
     */
    return function(vals) {

      var i = 0;
      var len = format.length;
      var str = '';

      for (; i < len; i++) {

        // Numbers
        if (vals[format[i].name]) {
          str += padNumber(vals[format[i].name], format[i].length);
        }
        // Dividers
        else {
          str += format[i].value;
        }
      }

      return str;
    };
  }

  /**
   * @param {String} format
   * @return {Object}
   */
  function parseDateFormat(format) {

    var f = format.toLowerCase();
    var i = 0;
    var len = f.length;

    var obj = {
      parts: []
    };

    for (; i < len; i++) {

      // If there is a matching character mapping and the last part was of the same name, increment its length
      // and add to its content.
      if (map[f[i]] && obj.parts.length && obj.parts[obj.parts.length - 1].name === map[f[i]]) {
        obj.parts[obj.parts.length - 1].length++;
        obj.parts[obj.parts.length - 1].value += format[i];
        continue;
      }

      obj.parts.push({
        name: map[f[i]] ? map[f[i]] : '',
        value: format[i],
        length: 1
      });
    }

    // Add a way to convert the parsed date into a regex-ish string that works with the Typeahead implementation.
    obj.getValues = getValues(obj.parts);
    obj.getString = getString(obj.parts);

    return obj;
  }

  Base.exportjQuery(parseDateFormat, 'parseDateFormat');

  return parseDateFormat;
}));

/**
 * # Transform
 * Apply a cross-browser transform style.
 *
 * @example
 * transform('translateX', '-100px');
 *
 * @module helper/transform.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.transform = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  /**
   * @param {String} type
   * @param {String} val
   */
  function transform(type, val) {

    var prefixes = ['-webkit-', '-moz-', '-o-', '-ms-', ''];
    var i = 0;
    var len = prefixes.length;
    var str = '';

    for (; i < len; i++) {
      if (typeof val === 'object') {
        str += prefixes[i] + 'transform: ';

        for (var j in val) {
          str += j + '(' + val[j] + '); ';
        }
      } else {
        str += prefixes[i] + 'transform: ' + type + '(' + val + '); ';
      }
    }

    return str;
  }

  Base.exportjQuery(transform, 'transform');

  return transform;
}));

/**
 * # Typeahead
 * Listen to an input element and format it as the user types.
 *
 * @example
 * new Typeahead(el);
 *
 * @module helper/typeahead.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Typeahead = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * Typeahead constructor
   * @param {Element} el
   * @param {Object} params
   */
  function Typeahead(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._setParams(params || {});
    this._cacheElements(el);

    this._maintainFocus(function() {
      this._parseParams();
      this._bindEventListenerCallbacks();
      this._addEventListeners();
    });
  }

  Typeahead.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _triggerEvent: Base.triggerEvent,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['actionCodes', 'format', 'placeholder', 'matchPlaceholderSize', 'onChange', 'onFocus', 'onBlur', 'onBackspace', 'onEnd'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      inputEl: null,
      placeholderEl: null,
      placeholder: null,
      characters: null,
      format: null,
      ignoreCodes: [
        9, // Tab
        16, // Shift
        17, // Ctrl
        18, // Alt
        20, // CAPS
        91, // Meta
        93, // Alt
        112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123 // F1-F12
      ],
      actionCodes: {
        BACKSPACE: 8,
        DELETE: 46,
        LEFT: 37,
        RIGHT: 39
      },
      pasteCode: 86, // v
      pauseBlurFocus: 0,
      matchPlaceholderSize: false,
      isFocused: false,
      isRunning: false,
      onChange: null,
      onFocus: null,
      onBlur: null,
      onBackspace: null,
      onEnd: null,
      _atEnd: false,
      _oldVal: null,
      _onInputBound: null,
      _onKeydownBound: null,
      _onFocusBound: null,
      _onBlurBound: null,
      _onPlaceholderClickBound: null
    },


    /**
     * Run the formatting. @todo: rename this.
     * @param {Number} cursorIndex
     * @param {Object} params
     */
    run: function(cursorIndex, params) {

      params = params || {};

      if (this.isRunning) return;

      this.isRunning = true;

      var oldVal = this.inputEl.value;
      var val = '';
      var placeholder = '';
      var i = 0;
      var len = this.format.length;
      var skipCount = 0;
      var valDone = false;

      for (; i < len; i++) {

        // Add numbers
        if (this.format[i] === '\\d') {

          if (this.characters[i - skipCount]) {
            val += this.characters[i - skipCount];
          }
          else {
            valDone = true;
          }

          placeholder += valDone ? this.placeholder[i] : this.characters[i - skipCount];
        }
        // Placeholder characters
        else {

          if (!valDone) {
            val += this.format[i];
          }

          placeholder += this.format[i];

          skipCount++;
        }
      }

      if (this.isFocused) {
        cursorIndex = cursorIndex === undefined ? this._getCaretEnd() : cursorIndex;
      }

      // If there are no characters, set the cursorIndex to be the last placeholder entry.
      if (this.isFocused && !this.characters.length) {
        cursorIndex = val.length;
      }

      // No characters and we shouldn't use just placeholder values
      if (!this.characters.length && params.notOnlyPlaceholders) {
        val = '';
      }

      this.inputEl.value = val;
      this.placeholderEl.innerHTML = placeholder;

      this._updateWidth();

      if (this.isFocused) {
        this._setCaretPositionTranslated(cursorIndex);
      }

      if (val !== oldVal) {
        this._triggerEvent(this.inputEl, 'input');
      }

      this.isRunning = false;

      if (val !== oldVal) {
        (this.onChange || noop)(val, oldVal, this.inputEl);
      }

      if (!this._atEnd && this.isFocused && this.characters.length === this.maxLength && this._caretIsAtEnd()) {
        this._atEnd = true;
        (this.onEnd || noop)(this);
      } else {
        this._atEnd = false;
      }
    },


    /**
     * Add a character to the characters array at a given index.
     * @param {String} character
     * @param {Number} start
     * @param {Number} end
     * @param {Boolean} skipCheck
     */
    addCharacterAtIndex: function(character, start, end, skipCheck) {

      // Don't add at an index beyond what we can support.
      if (this.maxLength && start >= this.maxLength) {
        return;
      }

      if (!skipCheck) {

        var re;

        // Try to build a regex for this format character.
        try {
          re = new RegExp(this.format[start]);
        } catch (e) {}

        if (!re || !re.exec(character)) {
          return;
        }
      }

      this.characters.splice(start, end - start, character);

      // If we've added at an index that pushes the length beyond what we support,
      // remove the trailing characters.
      if (this.maxLength && this.characters.length > this.maxLength) {
        this.characters.splice(this.maxLength, this.characters.length);
      }

      this.run(start + 1);
    },


    /**
     * Add a character at the position of the caret.
     * @param {String} character
     */
    addCharacterAtCaret: function(character) {

      var pos = this._getCaretStart();
      var re;

      // If we're beyond the bounds of the format, stop.
      if (this.format[pos] === undefined) {
        (this.onEnd || noop)(this, character);
        return;
      }

      // Try to build a regex for this format character.
      try {
        re = new RegExp(this.format[pos]);
      } catch (e) {}

      // We couldn't build a regex (so it's invalid) or the regex failed (so it's invalid)
      if (!re || !re.exec(character)) {
        if (this._moveCaret('right')) {
          this.addCharacterAtCaret(character);
        }
        return;
      }

      this.addCharacterAtIndex(character, this._getCaretStartTranslated(), this._getCaretEndTranslated(), true);
    },


    /**
     * Remove a character from the character array by index.
     * @param {Number} index
     * @param {Number} length Optional
     * @param {Number} offset Optional
     */
    removeCharacterAtIndex: function(index, length, offset) {

      // Don't want a negative splice length or else we start
      // removing characters from the end.
      if (index + offset < 0) {
        return;
      }

      length = length !== undefined ? length : 1;
      this.characters.splice(index + offset, length);
      this.run(index + (offset || 1));
    },


    /**
     * Remove the character at the caret.
     * @param {Number} offset Optional An offset from the current position.
     */
    removeCharacterAtCaret: function(offset) {

      var start = this._getCaretStartTranslated();
      var end = this._getCaretEndTranslated();
      var length = 1;
      var tmp;

      if (start !== end) {

        // If the end is less than the start, the user dragged from right to left.
        // Just swap them to make it easier to handle.
        if (end < start) {
          tmp = start;
          start = end;
          end = tmp;
        }

        // The length of characters removed
        length = end - start;

        // Bump the start position @todo: haven't thought through why this is, but it's needed.
        start++;
      }

      this.removeCharacterAtIndex(start, length, offset);
    },


    /**
     * Remove the character in the current range.
     */
    removeCharactersInRange: function() {
      this.removeCharacterAtIndex(this._getCaretStartTranslated(), this._getCaretEndTranslated());
    },


    /**
     * Set the value of the typeahead. Maintain the position of the caret.
     * @param {String} value
     */
    setValue: function(value) {

      this.settingValue = true;
      this.pause();

      this.characters = (value + '').split('');
      this.run();

      if (this.isFocused) this._setCaretPosition(this._getCaretStart());

      this.resume();
      this.settingValue = false;
    },


    /**
     * Get the value of the typeahead.
     * @param {Boolean} asInt Get the value as a parsed integer.
     * @return {String|Number}
     */
    getValue: function(asInt) {
      return asInt && this.inputEl.value ? parseInt(this.inputEl.value, 10) : this.inputEl.value;
    },


    /**
     * Move the caret position.
     * @param {Number} pos
     */
    moveCaret: function(pos) {
      this._setCaretPositionTranslated(pos);
    },


    /**
     * Move the caret to the end of the input.
     */
    moveCaretToEnd: function() {
      this.moveCaret(this.characters.length);
    },


    /**
     * Move the caret to the start of the input.
     */
    moveCaretToStart: function() {
      this.moveCaret(0);
    },


    /**
     * Pause events.
     */
    pause: function() {
      this.pauseBlurFocus++;
    },


    /**
     * Resume events.
     */
    resume: function() {
      this.pauseBlurFocus--;
    },


    /**
     * Clear the value.
     */
    clear: function() {
      this.pause();
      this.characters = [];
      this.run(0, {notOnlyPlaceholders: true});
      this.resume();
    },


    /**
     * Store a reference to the needed elements.
     * @param {Object} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.inputEl = this.el.querySelector('[type="text"], [type="email"], [type="phone"], textarea') || this._createDefaultInputElement();
      this.placeholderEl = this.el.querySelector('.spark-input__placeholder') || this._createDefaultPlaceholderElement();
    },


    /**
     * Parse parameters from the elements.
     */
    _parseParams: function() {

      // Store the value characters
      this.characters = this._parseCharacters(this.inputEl.value);

      // Store format
      this.format = this._parseFormat(this.format ? this.format : this.inputEl.getAttribute('data-typeahead-format'));

      // Store the original placeholder
      this.placeholder = this.placeholder ? this.placeholder : this.inputEl.getAttribute('placeholder').split('');

      // Get the total number of characters we can have
      this.maxLength = this._getCharactersAllowedCount(this.format);
    },


    /**
     * Parse the format string into an array.
     * @param  {String} format
     * @return {Array}
     */
    _parseFormat: function(format) {

      var i = 0;
      var len = format.length;
      var arr = [];
      var lastWasEscape = false;

      for (; i < len; i++) {
        if (format[i] === '\\' && !lastWasEscape) {
          lastWasEscape = true;
        } else {
          arr.push((lastWasEscape ? '\\' : '') + format[i]);
          lastWasEscape = false;
        }
      }

      return arr;
    },


    /**
     * Parse the characters string into an array, ignoring characters which don't
     * match the format requirements.
     * @param {String} characters
     * @return {Array}
     */
    _parseCharacters: function(characters) {

      var chars = characters.split('');
      var i = 0;
      var len = characters.length;
      var regexes = [];
      var arr = [];

      for (; i < len; i++) {

        // Try to build a regex for this format character.
        try {
          // Make sure this format starts with an escape character. @todo: this is pretty limiting, but
          // it's necessary or else '-' or '+' matches properly.
          regexes[i] = this.format[i][0] === '\\' ? new RegExp(this.format[i]) : null;
        } catch (e) {}

        // If we were able to create a regex and our char passes, add it to the array
        // of characters to return.
        if (regexes[i] && regexes[i].exec(chars[i])) {
          arr.push(chars[i]);
        }
      }

      return arr;
    },


    /**
     * Create the default input element.
     * @return {Element}
     */
    _createDefaultInputElement: function() {

      var el = document.createElement('input');
      el.className = 'spark-input__field';
      el.setAttribute('data-typeahead', '');
      el.setAttribute('type', 'tel');

      this.el.appendChild(el);

      return el;
    },


    /**
     * Create the default input element.
     * @return {Element}
     */
    _createDefaultPlaceholderElement: function() {
      var el = document.createElement('span');
      el.className = 'spark-input__placeholder';
      this.el.appendChild(el);
      return el;
    },


    /**
     * Get the maximum number of characters allowed.
     * @param {Array} format
     * @return {Number}
     */
    _getCharactersAllowedCount: function(format) {

      var i = 0;
      var len = format.length;
      var allowed = 0;

      for (; i < len; i++) {
        if (format[i] === '\\d') {
          allowed++;
        }
      }

      return allowed;
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onKeydownBound = this._onKeydown.bind(this);
      this._onInputBound = this._onInput.bind(this);
      this._onFocusBound = this._onFocus.bind(this);
      this._onBlurBound = this._onBlur.bind(this);
      this._onPlaceholderClickBound = this._onPlaceholderClick.bind(this);
    },


    /**
     * Add event listeners to keypress and keydown.
     */
    _addEventListeners: function() {
      this.inputEl.addEventListener('keydown', this._onKeydownBound, false);
      this.inputEl.addEventListener('input', this._onInputBound, false);
      this.inputEl.addEventListener('focus', this._onFocusBound, false);
      this.placeholderEl.addEventListener('click', this._onPlaceholderClickBound, false);
    },


    /**
     * Add event listeners to keypress and keydown.
     */
    _removeEventListeners: function() {
      this.inputEl.removeEventListener('keydown', this._onKeydownBound);
      this.inputEl.removeEventListener('focus', this._onFocusBound);
      this.placeholderEl.removeEventListener('click', this._onPlaceholderClickBound);
    },


    /**
     * Get the position of the caret in the element.
     * @return {Number} The index
     */
    _getCaretStart: function() {

      return this._maintainFocus(function() {

        var caretPosition;

        // IE support
        if (document.selection) {
          this.inputEl.focus();
          var sel = document.selection.createRange();
          sel.moveStart('character', -this.inputEl.value.length);
          caretPosition = sel.text.length;
        } else if (this.inputEl.selectionStart || this.inputEl.selectionStart === 0) {
          caretPosition = this.inputEl.selectionStart;
        }

        return caretPosition;
      });
    },


    /**
     * Get the end position of the caret in the element.
     * @return {Number} The index
     */
    _getCaretEnd: function() {

      return this._maintainFocus(function() {

        var caretPosition;

        // IE support - @todo: this doesn't work in IE
        if (document.selection) {
          this.inputEl.focus();
          var sel = document.selection.createRange();
          sel.moveStart('character', -this.inputEl.value.length);
          caretPosition = sel.text.length;
        } else if (this.inputEl.selectionEnd || this.inputEl.selectionEnd === 0) {
          caretPosition = this.inputEl.selectionEnd;
        }

        return caretPosition;
      });
    },


    /**
     * Is the caret at the end of the input?
     * @return {Boolean}
     */
    _caretIsAtEnd: function() {
      return this._getCaretStart() === this.maxLength;
    },


    /**
     * Set the position of the caret in the element.
     * @return {Number} The index
     */
    _setCaretPosition: function(pos) {

      return this._maintainFocus(function() {

        // IE support
        if (document.selection) {
          this.inputEl.focus();
          var sel = document.selection.createRange();
          sel.moveStart('character', -this.inputEl.value.length);
          sel.moveStart('character', pos);
          sel.moveEnd('character', 0);
          sel.select();
        } else if (this.inputEl.selectionStart || this.inputEl.selectionStart === 0) {
          this.inputEl.selectionStart = pos;
          this.inputEl.selectionEnd = pos;
        }
      });
    },


    /**
     * Get the position of the caret translated to the corresponding index in the
     * characters array. This means ignoring format characters.
     * @param {Number} pos
     * @return {Number}
     */
    _getCaretPositionTranslated: function(pos) {

      var i = 0;
      var skipCount = 0;

      for (; i < pos; i++) {

        // Count non-numbers as a skip. @todo: this needs to work with more than numbers.
        if (this.format[i] !== '\\d') {
          skipCount++;
        }
      }

      return pos - skipCount;
    },


    /**
     * Get the starting position of the caret translated.
     * @return {Number}
     */
    _getCaretStartTranslated: function() {
      return this._getCaretPositionTranslated(this._getCaretStart());
    },


    /**
     * Get the ending position of the caret translated.
     * @return {Number}
     */
    _getCaretEndTranslated: function() {
      return this._getCaretPositionTranslated(this._getCaretEnd());
    },


    /**
     * Set the position of the caret translated to the corresponding index in the
     * characters array. This means ignoring format characters.
     * @param {Number} pos
     */
    _setCaretPositionTranslated: function(pos) {

      var i = 0;
      var skipCount = 0;

      for (; i < pos + skipCount; i++) {

        // Count non-numbers as a skip. @todo: this needs to work with more than numbers.
        if (this.format[i] !== undefined && this.format[i] !== '\\d') {
          skipCount++;
        }
      }

      this._setCaretPosition(pos + skipCount);
    },


    /**
     * Move the caret position
     * @param  {String} direction The direction of the movement
     * @return {Boolean}           Was the caret actually moved?
     */
    _moveCaret: function(direction) {

      var curPos = this._getCaretStart();

      if (direction === 'left') {
        this._setCaretPosition(curPos - 1);
      } else if (direction === 'right') {
        this._setCaretPosition(curPos + 1);
      }

      return curPos !== this._getCaretStart();
    },


    /**
     * Empty the input when we only have placeholders.
     */
    _emptyWhenOnlyPlaceholders: function() {
      if (!this.characters.length) {
        this.clear();
      }
    },


    /**
     * Run a callback function that may change the focus of the document, but
     * make sure focus goes back to where it needs to be. Also, set the state
     * so that blur/focus events don't fire from this instance.
     * @param {Function} callback
     */
    _maintainFocus: function(callback) {

      this.pause();

      var originalActiveElement = document.activeElement;

      //For IE
      if(!originalActiveElement) {
        originalActiveElement = document.body;
      }

      var output = (callback || noop).call(this);

      // If we didn't have focus, go back to focusing on the original
      if (originalActiveElement !== this.inputEl) {
        this.inputEl.blur();
        originalActiveElement.focus();
      }

      this.resume();

      return output;
    },


    /**
     * Update the width of the typeahead. If we should be matching the width
     * of the placeholder, do so. Otherwise, take no action.
     */
    _updateWidth: function() {

      if (this.matchPlaceholderSize) {
        this.placeholderEl.style.width = 'auto';
        // Add 2px to account for caret width in IE... @todo: better possible fix?
        this.inputEl.style.width = 'auto';
        this.inputEl.style.width = this.placeholderEl.offsetWidth + 2 + 'px';
        this.placeholderEl.style.width = '';
      }
    },


    /**
     * Listen for delete and arrows.
     * @param  {Object} e
     */
    _onKeydown: function(e) {

      var code = e.keyCode || e.which;

      if (code === this.pasteCode && (e.metaKey || e.ctrlKey)) {
        return;
      }

      if (code === this.actionCodes.BACKSPACE) {
        this.removeCharacterAtCaret(-1);
        this._onBackspace();
        e.preventDefault();
      } else if (code === this.actionCodes.DELETE) {
        this.removeCharacterAtCaret(0);
        e.preventDefault();
      } else if (code === this.actionCodes.LEFT) {
        if (!this._getCaretStart()) {
          (this.onBackspace || noop)();
        }
      } else if (code === this.actionCodes.RIGHT) {
        if (this._getCaretStart() === this.characters.length) {
          (this.onEnd || noop)();
        }
      } else {
        if (this.ignoreCodes.indexOf(code) === -1) {
          e.preventDefault();
          this.addCharacterAtCaret(String.fromCharCode(code));
        }
      }
    },


    /**
     * When the input event fires, validate. This happens
     * with a copy+paste.
     * @param {Object} e
     */
    _onInput: function(e) {
      e.preventDefault();
      this.characters = this._parseCharacters(this.inputEl.value);
      this.run();
    },


    /**
     * When we focus, run the formatting.
     * @param {Object} e
     */
    _onFocus: function() {

      window.removeEventListener('blur', this._onBlurBound, false);
      window.addEventListener('blur', this._onBlurBound, false);
      this.inputEl.removeEventListener('blur', this._onBlurBound, false);
      this.inputEl.addEventListener('blur', this._onBlurBound, false);

      if (this.isFocused || this.pauseBlurFocus || this.isRunning) return;

      this.run();
      (this.onFocus || noop)(this.getValue());
      this.isFocused = true;
      this._oldVal = this.inputEl.value;
    },


    /**
     * When we blur, if we have no characters, remove the placeholders.
     * @param {Object} e
     */
    _onBlur: function() {

      window.removeEventListener('blur', this._onBlurBound);
      this.inputEl.removeEventListener('blur', this._onBlurBound);

      this.isFocused = false;

      if (this.pauseBlurFocus || this.isRunning) return;

      this._emptyWhenOnlyPlaceholders();

      // preventDefault will not dispatch change event
      // manually dispatch change event
      if(this._oldVal !== this.inputEl.value) {
        this._triggerEvent(this.inputEl, 'change');
      }

      (this.onBlur || noop)(this.getValue());
    },


    /**
     * When the placeholder receives a click event, focus on the input. This happens in IE10 for some
     * reason that I cannot fully fathom, but it has something to do with the explicit width being
     * set on an empty element.
     * @param {Object} e
     */
    _onPlaceholderClick: function(e) {
      e.preventDefault();
      e.stopPropagation();
      this.inputEl.focus();
    },


    /**
     * When we backspace, if we have no characters left let listeners know.
     * @param {Object} e
     */
    _onBackspace: function() {
      if (!this._getCaretStart())(this.onBackspace || noop)();
    }
  };

  Base.exportjQuery(Typeahead, 'Typeahead');

  return Typeahead;
}));

/**
 * # DateTypeahead
 * Listen to an input element and format it as the user types.
 *
 * @example
 * new DateTypeahead(el);
 *
 * @module helper/date-typeahead.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/typeahead'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'), require('./typeahead'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.DateTypeahead = factory(root.Spark.BaseComponent, root.Spark.Typeahead);
  }
}(this, function(Base, Typeahead) {

  'use strict';

  var noop = function() {};

  /**
   * DateTypeahead constructor
   * @param {Element} el Optional
   * @param {Object} params Optional
   */
  function DateTypeahead(el, params) {

    // If only one arg passed, assume it was a parameters
    // object since the user MUST provide those but the element
    // is optional. Doing it this way to keep the arity the same
    // as other components.
    if (arguments.length < 2) {
      params = el || {};
      el = this._createDefaultElement();
    }

    this._setParams(this.defaults, true);
    this._setParams(params);
    this._bindEventListenerCallbacks();
    this._createTypeahead(el, params);
  }

  DateTypeahead.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['type', 'onChange', 'onFocus', 'onBlur', 'onBackspace', 'onEnd'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      type: null,
      typeahead: null,
      onChange: null,
      onFocus: null,
      onBlur: null,
      onBackspace: null,
      onEnd: null,
      _onTypeaheadChangeBound: null,
      _onTypeaheadFocusBound: null,
      _onTypeaheadBlurBound: null,
      _onTypeaheadBackspaceBound: null,
      _onTypeaheadEndBound: null
    },


    /**
     * Get the value.
     * @param {Boolean} asInt Get the value as a parsed integer.
     * @return {Mixed}
     */
    getValue: function(asInt) {
      return this.typeahead.getValue(asInt);
    },


    /**
     * Set the value.
     * @param {Mixed} val
     */
    setValue: function(val) {
      return this.typeahead.setValue(val);
    },


    /**
     * Run the typeahead calculations.
     */
    run: function() {
      return this.typeahead.run();
    },


    /**
     * Pause the typeahead events.
     */
    pause: function() {
      return this.typeahead.pause();
    },


    /**
     * Reseume typeahead events.
     */
    resume: function() {
      return this.typeahead.resume();
    },


    /**
     * Create the default input element.
     * @return {Element}
     */
    _createDefaultElement: function() {
      var el = document.createElement('span');
      el.className = 'spark-input';
      return el;
    },


    /**
     * Create a typeahead with the given format.
     * @param {Object} el
     * @param {Object} params
     */
    _createTypeahead: function(el, params) {

      params = params || {};

      if (!params.placeholder) {
        throw new Error('You must provide a placeholder value for a DateTypeahead.');
      }

      if (params.length !== undefined) {
        params.format = this._lengthToFormat(params.length);
      }

      if (!params.format) {
        throw new Error('You must provide a format value for a DateTypeahead.');
      }

      this.typeahead = new Typeahead(el, {
        placeholder: params.placeholder,
        format: params.format,
        matchPlaceholderSize: true,
        onChange: this._onTypeaheadChangeBound,
        onFocus: this._onTypeaheadFocusBound,
        onBlur: this._onTypeaheadBlurBound,
        onBackspace: this._onTypeaheadBackspaceBound,
        onEnd: this._onTypeaheadEndBound
      });

      // Ensure we have an ARIA label
      var input = el.querySelector('input');
      if (input && !input.getAttribute('aria-label')) {
        input.setAttribute('aria-label', this.type);
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onTypeaheadChangeBound = this._onTypeaheadChange.bind(this);
      this._onTypeaheadFocusBound = this._onTypeaheadFocus.bind(this);
      this._onTypeaheadBlurBound = this._onTypeaheadBlur.bind(this);
      this._onTypeaheadBackspaceBound = this._onTypeaheadBackspace.bind(this);
      this._onTypeaheadEndBound = this._onTypeaheadEnd.bind(this);
    },


    /**
     * Take a length and return a format string with that many digits.
     * @param {Number} length
     * @return {String}
     */
    _lengthToFormat: function(length) {

      var i = 0;
      var ret = '';

      for (; i < length; i++) {
        ret += '\\d';
      }

      return ret;
    },


    /**
     * Check to see if an input value is valid.
     * @param {Mixed} val
     * @param {Boolean} allowEmpty All the value to be empty instead of 0.
     */
    _checkValidity: function(val, allowEmpty) {

      val = parseInt(val, 10);

      var origVal = val;
      var isNumber = !isNaN(val);

      // If we were passed an empty string or something, don't try to validate.
      // Treat zeros as a non-entry for days and months.
      if (isNumber) {

        if (this.type === 'day') {
          val = val ? Math.min(Math.max(val, 1), 31) : (allowEmpty ? '' : 0);
        } else if (this.type === 'month') {
          val = val ? Math.min(Math.max(val, 1), 12) : (allowEmpty ? '' : 0);
        } else {
          val = val === 0 ? (allowEmpty ? '' : 0) : Math.max(val, 0);
        }
      }

      // Need to make sure we aren't looping forever on these updates.
      if (isNumber && val !== origVal) {
        this.typeahead.setValue(val + '');
        return false;
      }

      return true;
    },


    /**
     * When the typeahead changes, make sure the value is valid. This
     * is very basic validation. More complex validation like the number
     * of days in a specific month should be handled by the callback.
     * And run our callback.
     * @param {String} val The value of the input
     * @param {String} oldVal The previous value
     */
    _onTypeaheadChange: function(val) {
      if (this._checkValidity(val)) {
        (this.onChange || noop)(val, this);
      }
    },


    /**
     * When the typeahead gains focus, let anyone who is interested know.
     * @param {String} val
     */
    _onTypeaheadFocus: function(val) {
      (this.onFocus || noop)(val, this);
    },


    /**
     * When the typeahead loses focus, let anyone who is interested know.
     * @param {String} val
     */
    _onTypeaheadBlur: function(val) {
      this._checkValidity(val, true);
      (this.onBlur || noop)(val, this);
    },


    /**
     * When the typeahead fires a backspace event because it's empty and
     * the user is hitting backspace, let anyone who is interested know.
     * @param {String} val
     */
    _onTypeaheadBackspace: function(val) {
      (this.onBackspace || noop)(val, this);
    },


    /**
     * When the typeahead is full and at its end, let anyone who is interested know.
     * @param {Object} typeahead
     * @param {String} character Optional The character to pass to the next input.
     */
    _onTypeaheadEnd: function(typeahead, character) {
      (this.onEnd || noop)(this, character);
    }
  };

  Base.exportjQuery(DateTypeahead, 'DateTypeahead');

  return DateTypeahead;
}));

/**
 * # Debounce
 * Debounce a function call
 *
 *
 * @module helper/debounce.js
 */

/**
 * @ignore
 */
(function (root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  }
  else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'));
  }
  else {
    root.Spark = root.Spark || {};
    root.Spark.debounce = factory(root.Spark.BaseComponent);
  }
}(this, function (Base) {

  'use strict';

  /**
   * @param {Function} func
   * @param {Integer} delay
   */
   function debounce(func, delay) {
     var timer;
     return function() {
       var args = arguments;
       if (timer) window.clearTimeout(timer);
       timer = window.setTimeout(function() {
         func.apply(this, args);
       }, delay);
     };
   }
  Base.exportjQuery(debounce, 'debounce');

  return debounce;
}));

/**
 * # Expand
 * Expand and collapse an element.
 *
 * @example
 * new Expand(el);
 *
 * @module components/expand.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/animate-height'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/animate-height'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Expand = factory(root.Spark.BaseComponent, root.Spark.animateHeight);
  }
}(this, function(Base, animateHeight) {

  'use strict';

  var noop = function() {};

  /**
   * Expand constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Expand = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  Expand.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _hasClass: Base.hasClass,
    _getElementMatchingParent: Base.getElementMatchingParent,
    remove: Base.remove,


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      isExpanded: false,
      onBeforeExpand: null,
      onAfterExpand: null,
      onBeforeCollapse: null,
      onAfterCollapse: null,
      _onClickBound: null,
      _onKeydownBound: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['onBeforeExpand', 'onAfterExpand', 'onBeforeCollapse', 'onAfterCollapse'],


    /**
     * Expand
     */
    expand: function() {

      (this.onBeforeExpand || noop)();

      animateHeight({
        el: this.el,
        toggleEl: '.spark-expand__content, .spark-panel__content'
      });

      this.isExpanded = true;
      this._updateClasses();
      var e = document.createEvent('Event');
      e.initEvent('spark.visible-children', true, true);
      this.el.dispatchEvent(e);

      // If the expand element have input, focus on the first one.
      if(this.el.querySelector('input')) {
        this.el.querySelector('input').focus();
      }

      (this.onAfterExpand || noop)();
    },


    /**
     * Collapse
     */
    collapse: function() {

      (this.onBeforeCollapse || noop)();

      animateHeight({
        el: this.el,
        toggleEl: '.spark-expand__content, .spark-panel__content',
        toggleValue: 'none',
        action: 'collapse'
      });

      this.isExpanded = false;
      this._updateClasses();

      (this.onAfterCollapse || noop)();
    },


    /**
     * Toggle the expand state.
     */
    toggle: function() {
      this[this.isExpanded ? 'collapse' : 'expand']();
    },


    /**
     * Store a reference to the element.
     * @param {Element} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.isExpanded = this._hasClass(this.el, 'expanded');
    },


    /**
     * Update classes for the expand or collapse state.
     */
    _updateClasses: function() {
      this._toggleClass(this.el, 'expanded', this.isExpanded);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onKeydownBound = this._onKeydown.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      this.el.addEventListener('click', this._onClickBound);
      this.el.addEventListener('keydown', this._onKeydownBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      this.el.removeEventListener('click', this._onClickBound);
      this.el.removeEventListener('keydown', this._onKeydownBound);
    },


    /**
     * When we are clicked, toggle the expanded state.
     * @param {Object} e
     */
    _onClick: function(e) {

      if (!this._getElementMatchingParent(e.target, '.spark-expand__toggle, [data-role="toggle"], [role="heading"]', this.el)) {
        return;
      }

      e.preventDefault();
      this.toggle();
    },


    /**
     * When the space or enter key is pressed on the toggle, toggle!
     * @param {Object} e
     */
    _onKeydown: function(e) {

      if (!this._getElementMatchingParent(e.target, '.spark-expand__toggle, [data-role="toggle"], [role="heading"]', this.el)) {
        return;
      }

      var code = e.keyCode || e.which;

      // Space or enter
      if (code === 32 || code === 13) {
        e.preventDefault();
        this.toggle();
      }
    }
  };

  Base.exportjQuery(Expand, 'Expand');

  return Expand;
}));

/**
 * # Menu
 *
 * @example
 * new Menu(el, {
 *   // Optional. Callback method for when the menu toggles.
 *   onToggle: function(){}
 * });
 *
 * @module components/menu.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/animate-height', 'spark/helpers/transform'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/animate-height'), require('../helpers/transform'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Menu = factory(root.Spark.BaseComponent, root.Spark.animateHeight, root.Spark.transform);
  }
}(this, function(Base, animateHeight, transform) {

  'use strict';

  /**
   * Menu constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Menu = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._checkAnimation();
  };

  Menu.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _appendChildren: Base.appendChildren,
    _elementHasParent: Base.elementHasParent,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _getElementMatchingParents: Base.getElementMatchingParents,
    _getMatchingChild: Base.getMatchingChild,
    _wrapElement: Base.wrapElement,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['onToggle'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      cachedList: null,
      el: null,
      toggleEl: null,
      wrapperEl: null,
      _onClickBound: null,
      _onFocusBound: null,
      _onBlurBound: null,
      onToggle: function() {}
    },


    /**
     * Store a reference to the tabs list, each tab and each panel.
     * Set which tab is active, or use the first.
     * @param {Element} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.toggleEl = this.el.querySelector('.spark-menu__toggle');
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onFocusBound = this._onFocus.bind(this);
      this._onBlurBound = this._onBlur.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      this.el.addEventListener('click', this._onClickBound);
      this.el.addEventListener('focus', this._onFocusBound, true);
      this.el.addEventListener('blur', this._onBlurBound, true);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      this.el.removeEventListener('click', this._onClickBound);
      this.el.removeEventListener('focus', this._onFocusBound);
      this.el.removeEventListener('blur', this._onBlurBound);
    },

    /**
     * Toggle the open state of an item.
     * @param {Element} item
     */
    _toggleItem: function(item) {

      if (this._hasClass(item, 'open')) {
        this._closeItem(item);
      } else {
        this._openItem(item);
      }
    },

    /**
     * Check for a nested list and create the wrappers needed
     * for animating the lists
     *
     */
    _checkAnimation: function() {
      if (this.el.querySelector('.spark-menu__list-next')) {
        this.cachedList = this.cachedList || [];
        this._createMenuAnimationWrapper();
        this._animateListChange();
      }
    },

    /**
     * Create wrapper class to help with animation of sliding lists
     *
     */
    _createMenuAnimationWrapper: function() {
      if (this.wrapperEl) {
        return;
      }

      var wrapperEl = document.createElement('div');
      this._addClass(wrapperEl, 'spark-menu__animation-wrapper');
      this._wrapElement(this.el.querySelector('.spark-menu__list'), wrapperEl);
      this.wrapperEl = wrapperEl;
    },


    /**
     * Animate the position of the animation wrapper. Optionally, do
     * so immediately without waiting for an animation.
     * @param {Boolean} noAnimate
     */
    _animateListChange: function(noAnimate) {

      if (noAnimate) {
        this._addClass(this.wrapperEl, 'no-animate');
      }

      this.wrapperEl.setAttribute('style', transform('translateX', '-' + (this.cachedList.length * 100) + '%'));

      if (noAnimate) {
        setTimeout(function() {
          this._removeClass(this.wrapperEl, 'no-animate');
        }.bind(this), 1);
      }
    },

    /**
     * Append list to menu element
     * @param {Element} list
     * @param {Boolean} noAnimate
     */
    _appendList: function(item, noAnimate) {

      // Create wrapper
      this._createMenuAnimationWrapper();

      var newList = item.cloneNode(true);
      this._addClass(newList, 'nestedList');
      newList.setAttribute('data-nested-list-id', newList.getAttribute('id'));
      newList.removeAttribute('id');

      if (this.wrapperEl) {
        // Add child node to wrapper
        this.wrapperEl.appendChild(newList);
        // Add to cached Array to keep track of all added lists
        this.cachedList.push(newList);
        // Slide navigation
        this._animateListChange(noAnimate);
      }
    },

    /**
     * Remove list to nav
     *
     */
    _removeLastList: function() {
      // If there are any items to remove
      if (this.cachedList.length) {
        // Retrieve last item from list
        var removeElement = this.cachedList.pop();
        if (this.wrapperEl) {
          // Slide navigation
          this._animateListChange();
        }
        window.setTimeout(function() {
          // Remove itself from DOM
          removeElement.parentNode.removeChild(removeElement);
        }, 250);
      }
    },

    /**
     * Remove all lists from panel menu
     *
     */
    _removeAllCachedLists: function() {
      if (this.cachedList) {
        // If there are any items to remove
        while (this.cachedList.length) {
          // While there are still items, remove them
          this._removeLastList();
        }
      }
    },

    /**
     * Finds and returns the next nested list
     * @param {Object} item
     * @return {Object}
     */
    _getNextList: function(item) {
      return item.querySelector('.spark-menu__list-next') ? document.querySelector(item.querySelector('.spark-menu__list-next').getAttribute('data-menu')) : null;
    },


    /**
     * Open an item by animating it.
     * @param {Object} item
     */
    _openItem: function(item) {

      // Item is already open
      if (this._hasClass(item, 'open')) {
        return;
      }

      animateHeight({
        el: item,
        toggleEl: '.spark-menu__list'
      });

      this._addClass(item, 'open');
    },


    /**
     * Close an item by animating it shut.
     * @param {Object} item
     */
    _closeItem: function(item) {

      // Item is already closed
      if (!this._hasClass(item, 'open')) {
        return;
      }

      animateHeight({
        el: item,
        toggleEl: '.spark-menu__list',
        toggleValue: 'none',
        action: 'collapse'
      });

      this._removeClass(item, 'open');
    },


    /**
     * Make an item active.
     * @param {Element} item
     */
    _activateItem: function(item) {

      // Item is already active
      if (this._hasClass(item, 'active')) {
        return;
      }

      // Deactivate any active items
      var parents = this._getElementMatchingParents(item, '.spark-menu__list', this.el);
      this._deactivateItems(parents[parents.length - 1]);
      this._deactivateItemSiblings(item);

      // Add the active class
      this._addClass(item, 'active');

      // If there is a parent that is also a list item, open it.
      this._activateItemParents(item, this.el);
    },


    /**
     * Activate parent items.
     * @param {Element} el
     * @param {Element} limitEl
     */
    _activateItemParents: function(el, limitEl) {

      var parents = this._getElementMatchingParents(el.parentNode, '[class*="list-item"]', limitEl);

      var i = 0;
      var len = parents.length;

      // Add the active class
      for (; i < len; i++) {
        this._openItem(parents[i]);
        this._addClass(parents[i], 'child-active');
      }
    },


    /**
     * Deactivate items.
     * @param {Element} el
     */
    _deactivateItems: function(el) {

      var actives = el.querySelectorAll('[class*="list-item"].active');
      var i = 0;
      var len = actives.length;

      // Remove the active class
      for (; i < len; i++) {
        this._removeClass(actives.item(i), 'active');
      }
    },

    /**
     * Deactivate siblings items.
     * @param {Element} el
     */
    _deactivateItemSiblings: function(el) {

      var actives = el.parentNode.querySelectorAll('[class*="list-item"].child-active');
      var i = 0;
      var len = actives.length;

      // Remove the active class
      for (; i < len; i++) {
        this._removeClass(actives[i], 'child-active');
        this._removeClass(actives[i], 'open');
      }
    },


    /**
     * Open the parents of the active item.
     *
     */
    _openActiveParents: function() {

      var activeItem = this.el.querySelector('.active');
      if (activeItem) {
        var parentItems = this._getElementMatchingParents(activeItem, '.spark-menu__list-item', this.el);
        var itemLinks;
        var nextList;

        for (var i = parentItems.length - 1; i >= 0; i--) {
          itemLinks = this._getMatchingChild(parentItems[i], '.spark-menu__list-links');
          if (itemLinks && itemLinks.querySelector('.spark-menu__list-next')) {
            nextList = this._getNextList(parentItems[i]);
            if (nextList && !this._cachedListContainsID(nextList.getAttribute('id'))) {
              this._appendList(nextList, true);
            }
          } else {
            this._addClass(parentItems[i], 'open');
          }
        }
      }
    },


    /**
     * Check if the cached list contains a certain ID
     * @param {String} id
     * @return {Boolean}
     */
    _cachedListContainsID: function contains(id) {
      var i = this.cachedList.length;
      while (i--) {
        if (this.cachedList[i].getAttribute('data-nested-list-id') === id) {
          return true;
        }
      }
      return false;
    },

    /**
     * When an item is clicked, make it active. Determine if the click was on an expand
     * button and open the list if so.
     * @todo: It should be possible to opt out of this behavior.
     * @param {Object} e
     */
    _onClick: function(e) {

      // Don't make forms active
      if (this._getElementMatchingParent(e.target, 'form', this.el)) {
        return;
      }

      // Toggle the visibility of the menu?
      var toggle = e.target === this.toggleEl || this._elementHasParent(e.target, this.toggleEl);
      if (toggle) {
        return this.onToggle(e, this);
      }

      // Is there a parent to open and an item?
      var open = this._getElementMatchingParent(e.target, '.spark-menu__list-expand', this.el);
      var item = this._getElementMatchingParent(e.target, '.spark-menu__list-item', this.el);

      // If we have no item or have been told to ignore the item
      if (!item || this._getElementMatchingParent(e.target, '.spark-menu__ignore', this.el)) {
        return;
      }
      if (open) {
        return this._toggleItem(item);
      }

      // Check if we have a valid item and we aren't inside the expanded header
      if (item && !this._elementHasParent(e.target, document.querySelector('.spark-header--visible'))) {

        var next = this._getNextList(item);

        if (next && this._hasClass(e.target, 'spark-menu__list-next')) {
          // Active item
          this._activateItem(item);
          this._appendList(next);
          return;
        }

        var back = this._getElementMatchingParent(e.target, '.spark-menu__list-back', item);

        if (back && this._hasClass(e.target, 'spark-menu__list-back')) {
          this._removeLastList();
          return;
        }
      }

      // Active item
      this._activateItem(item);
    },


    /**
     * Keep track of when items have focus.
     * @param {Object} e
     */
    _onFocus: function(e) {

      var parent = e.target;
      var lastParent = parent;

      while(parent) {
        parent = this._getElementMatchingParent(lastParent.parentNode, '.spark-menu__list-item', this.el);
        if (!parent || parent === lastParent) break;
        this._addClass(parent, 'has-focus');
        lastParent = parent;
      }
    },


    /**
     * Keep track of when items lose focus.
     * @param {Object} e
     */
    _onBlur: function(e) {

      var parent = e.target;
      var lastParent = parent;

      while(parent) {
        parent = this._getElementMatchingParent(lastParent.parentNode, '.spark-menu__list-item', this.el);
        if (!parent || parent === lastParent) break;
        this._removeClass(parent, 'has-focus');
        lastParent = parent;
      }
    }
  };

  Base.exportjQuery(Menu, 'Menu');

  return Menu;
}));

/**
 * # Modal
 * Show content in a modal.
 *
 * @example
 * new Modal(el);
 *
 * @module components/modal.js
 */

(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Modal = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  /**
   * Modal constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Modal = function(el, params) {

    params = params || {};

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el, params.modalEl);
    this._setParams(params);
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  Modal.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _getElementMatchingParent: Base.getElementMatchingParent,
    remove: Base.remove,


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      modalEl: null,
      scrollEl: null,
      isActive: false,
      _onClickBound: null,
      _onKeyupBound: null,
      _onModalClickBound: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: [],


    /**
     * Open
     */
    open: function() {
      this.isActive = true;
      this._addWindowEventListeners();
      this._updateClasses();
      this._addClass(document.body, 'spark-modal-open');
    },


    /**
     * Close
     */
    close: function() {
      this.isActive = false;
      this._removeWindowEventListeners();
      this._updateClasses();
      this._removeClass(document.body, 'spark-modal-open');
    },


    /**
     * Store a reference to the element. Either a modal itself
     * or a button referencing a modal may be passed.
     * @param {Element} el
     * @param {Element} modalEl @optional
     */
    _cacheElements: function(el, modalEl) {

      var modalPassed = this._hasClass(el, 'spark-modal');

      if (modalPassed) {
        this.modalEl = el;
      } else {
        this.el = el;
        this.modalEl = modalEl || document.querySelector(el.getAttribute('data-modal'));
      }

      this.scrollEl = this.modalEl.querySelector('.spark-modal__scroll') || this.modalEl;

      this.isActive = this._hasClass(this.el || this.modalEl, 'active');
    },


    /**
     * Update classes for the open or close state.
     */
    _updateClasses: function() {
      this._toggleClass(this.modalEl, 'active', this.isActive);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onKeyupBound = this._onKeyup.bind(this);
      this._onModalClickBound = this._onModalClick.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      if (this.el) this.el.addEventListener('click', this._onClickBound);
      if (this.modalEl) this.modalEl.addEventListener('click', this._onModalClickBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      if (this.el) this.el.removeEventListener('click', this._onClickBound);
      if (this.modalEl) this.modalEl.removeEventListener('click', this._onModalClickBound);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addWindowEventListeners: function() {
      this._removeWindowEventListeners();
      window.addEventListener('keyup', this._onKeyupBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeWindowEventListeners: function() {
      window.removeEventListener('keyup', this._onKeyupBound);
    },


    /**
     * When the button is clicked.
     * @param {Object} e
     */
    _onClick: function() {
      this.open();
    },


    /**
     * When we are clicked, toggle the opened state.
     * @param {Object} e
     */
    _onModalClick: function(e) {

      // The close button is clicked or the actual modal (gray area)
      if (this._getElementMatchingParent(e.target, '.spark-modal__close, .spark-modal__dismiss', this.modalEl) || e.target === this.scrollEl || e.target === this.modalEl) {
        e.preventDefault();
        this.close();
      }
    },


    /**
     * When a key is pressed on the window and it's an ESC, close the modal.
     * @param {Object} e
     */
    _onKeyup: function(e) {
      if (e.keyCode === 27) {
        this.close();
      }
    }
  };

  Base.exportjQuery(Modal, 'Modal');

  return Modal;
}));

/**
 * # Header
 * A primary page header/navigation.
 *
 * @example
 * new Header(el, {
 *   // Optional. Alternate breakpoint values.
 *   breakpoints: {
 *    xs: {
 *      min: 0,
 *      max: 639
 *    },
 *    // ...
 *  }
 * });
 *
 * @module components/header.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/menu'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('./menu'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Header = factory(root.Spark.BaseComponent, root.Spark.Menu);
  }
}(this, function(Base, Menu) {

  'use strict';

  /**
   * Header constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Header = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._bindEventListenerCallbacks();
    this._cacheElements(el);
    this._setParams(params || {});
    this._addEventListeners();

    // Determine the initial menu size now and when we get a ready state change
    this._determineInitialSize();
    this._listenForReadyStateChange();
  };

  Header.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _addClass: Base.addClass,
    _hasClass: Base.hasClass,
    _removeClass: Base.removeClass,
    _getChildIndex: Base.getChildIndex,
    _elementHasParent: Base.elementHasParent,
    _appendChildren: Base.appendChildren,
    _insertBefore: Base.insertBefore,
    _getBreakpoint: Base.getBreakpoint,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _getElementMatchingParents: Base.getElementMatchingParents,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['breakpoints', 'fixed', 'fixedDistance'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      fixed: false,
      fixedDistance: 10,
      navEl: null,
      menuEl: null,
      listEl: null,
      listEls: null,
      listMoreEl: null,
      listMoreListEl: null,
      placeholder: null,
      toggleEl: null,
      lastBreakpoint: null,
      currentBreakpoint: null,
      isActive: false,
      isCollapsed: null,
      moreSwapIndex: -1,
      menu: null,
      breakpoints: null,
      _onResizeBound: null,
      _onScrollBound: null,
      _onMoreClickBound: null,
      _onToggleClickBound: null,
      _onNavClickBound: null
    },


    /**
     * Update the elements used.
     */
    update: function() {

      this._removeEventListeners();
      this._removePlaceholder();
      this._cacheElements(this.el);
      this._addEventListeners();
      this._ensureActiveAtMoreSwapIndex();
      this.checkFixed();

      // Run on the next frame so sizes have updated
      setTimeout(function() {
        this._determineMenuSize();
      }.bind(this), 0);
    },


    /**
     * Check of we should be fixed.
     */
    checkFixed: function() {

      if (!this.fixed) {
        return;
      }

      var scrollTop = window.pageYOffset !== undefined ? window.pageYOffset : window.document.body.scrollTop;
      var isCondensed = scrollTop > this.fixedDistance;
      this._toggleClass(this.el, 'spark-header--condensed', isCondensed);
      this._toggleClass(document.body, 'spark-header-condensed', isCondensed);
    },


    /**
     * Store a reference to the tabs list, each tab and each panel.
     * Set which tab is active, or use the first.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.navEl = this.el.querySelector('.spark-header__nav');
      this.menuEl = this.navEl && this.navEl.querySelector('.spark-header__menu');
      this.listEl = this.menuEl && this.menuEl.querySelector('.spark-header__list, [data-role="list"]'); // @todo: remove data-role in v2.x.x
      this.toggleEl = this.el.querySelector('.spark-header__toggle');

      if (!this.fixed && this._hasClass(this.el, 'spark-header--fixed')) {
        this.fixed = true;
        this.checkFixed();
      }

      // Create a new instance of the menu component
      if (this.menuEl) {
        this.menu = new Menu(this.menuEl, {
          onToggle: this._onToggleClickBound
        });
      }

      // The items in the list need to show/hide based on the width of the container.
      // Cache these items so we can manipulate their display independent of what is
      // currently in the DOM. Also, create the "More" dropdown which will be shown
      // and hidden based on availabile space.
      if (this.listEl && this.listEl.children.length) {
        this.listEls = Array.prototype.slice.call(this.listEl.children, 0);
        this._createListMore();
      }

      // Create a clone of the header which will NOT be affected by changes in breakpoint.
      // This lets us continue to measure how many list elements will fit. Since we go to the
      // "condensed" view when we are at the sm/xs breakpoint OR only one item will fit in the nav,
      // we can't rely on breakpoints alone to determine what to show. Without a cloned placeholder
      // it is impossible to continue to measure the available space once we show the condensed view.
      if (this.listEl) {
        this._createPlaceholder();
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {

      this._onResizeBound = this._onResize.bind(this);
      this._onScrollBound = this._onScroll.bind(this);
      this._onMoreClickBound = this._onMoreClick.bind(this);
      this._onToggleClickBound = this._onToggleClick.bind(this);
      this._onNavClickBound = this._onNavClick.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {

      if (!this.listEl) {
        return;
      }

      window.addEventListener('resize', this._onResizeBound);
      window.addEventListener('orientationchange', this._onResizeBound);

      if (this.listMoreEl) {
        this.listMoreListEl.addEventListener('click', this._onMoreClickBound);
      }

      if (this.toggleEl) {
        this.toggleEl.addEventListener('click', this._onToggleClickBound);
      }

      if (this.navEl) {
        this.navEl.addEventListener('click', this._onNavClickBound);
      }

      if (this.fixed) {
        window.addEventListener('scroll', this._onScrollBound);
      }
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {

      if (!this.listEl) {
        return;
      }

      window.removeEventListener('resize', this._onResizeBound);
      window.removeEventListener('orientationchange', this._onResizeBound);

      if (this.listMoreEl) {
        this.listMoreListEl.removeEventListener('click', this._onMoreClickBound);
      }

      if (this.toggleEl) {
        this.toggleEl.removeEventListener('click', this._onToggleClickBound);
      }

      if (this.navEl) {
        this.navEl.removeEventListener('click', this._onNavClickBound);
      }

      if (this.fixed) {
        window.removeEventListener('scroll', this._onScrollBound);
      }
    },


    /**
     * Get the current breakpoint for the header.
     */
    _getCurrentBreakpoint: function() {
      this.lastBreakpoint = this.currentBreakpoint;
      this.currentBreakpoint = this._getBreakpoint(this.el.clientWidth, this.breakpoints);
      this.el.setAttribute('data-breakpoint', this.currentBreakpoint);
    },


    /**
     * Create a placeholder for the whole header so that we can keep track
     * of the width of each child element regardless of whether or not we're
     * condensed. Condensed styles do not apply to instances of the element
     * with the placeholder class.
     */
    _createPlaceholder: function() {

      var div = document.createElement('div');
      div.innerHTML = this.navEl.outerHTML;

      var el = div.children[0];
      el.setAttribute('aria-hidden', true);

      this._addClass(el, 'spark-header__placeholder');

      this._disablePlaceholderLinkTab(el);

      this.el.appendChild(el);

      // Cache the common elements
      this.placeholder = {
        el: el,
        menuEl: el.querySelector('.spark-header__menu'),
        listEl: el.querySelector('.spark-header__list, [data-role="list"]') // @todo: remove data-role in v2.x.x
      };

      // Add a copy of the "more" button to the list so we always know what size it would be
      if (this.listMoreEl) {
        this.placeholder.listEl.innerHTML += this.listMoreEl.outerHTML;
        this.placeholder.listMoreEl = this.placeholder.listEl.querySelector('.spark-header__more');
      }
    },


    /**
     * Remove the placeholder
     */
    _removePlaceholder: function() {

      if (this.placeholder) {
        this.placeholder.el.parentNode.removeChild(this.placeholder.el);
        this.placeholder.menuEl.parentNode.removeChild(this.placeholder.menuEl);
        this.placeholder.listEl.parentNode.removeChild(this.placeholder.listEl);
      }

      if (this.listMoreEl) {
        this.placeholder.listMoreEl.parentNode.removeChild(this.placeholder.listMoreEl);
      }
    },


    /**
     * Disable tabbing for items in the placeholder.
     * @param {Element} el
     */
    _disablePlaceholderLinkTab: function(el) {

      // Set a negative tab index on each link in the placeholder
      var links = el.querySelectorAll('.spark-menu__list-link, .spark-menu__list-expand');
      var i = 0;
      var len = links.length;

      for (; i < len; i++) {
        links.item(i).setAttribute('tabindex', -1);
      }
    },


    /**
     * Create a place to store overflow items of the list.
     * Also add this item to the placeholder element so we always know
     * which size it would be.
     */
    _createListMore: function() {

      var div = document.createElement('div');
      div.innerHTML = '<li><a class="spark-menu__list-link spark-menu__ignore" title="More Items" tabindex="0"><i class="spark-icon-menu-ellipsis-horizontal spark-icon--fill"></i></a><ul class="spark-menu__list"></ul></li>';

      var li = div.children[0];
      this._addClass(li, 'spark-menu__list-item spark-header__more');

      this.listMoreEl = li;
      this.listMoreListEl = li.querySelector('ul');
    },


    /**
     * Determine the menu size.
     */
    _determineInitialSize: function() {
      this._addClass(this.el, 'spark-header--visible');
      this._ensureActiveAtMoreSwapIndex();
      this._determineMenuSize();
    },


    /**
     * Determine how many nav items can fit.
     * @param {Boolean} isSwap Optional Is this a swapping event? If so, ignore redundancy checks.
     */
    _determineMenuSize: function(isSwap) {

      // Don't do anything w/o primary nav.
      if (!this.listEls || !this.listEls.length) {
        return;
      }

      // If we're at the XS or SM breakpoint, don't worry about this stuff.
      if (this._isMenuBreakpoint(['xs', 'sm'])) {
        this._removeListMore();
        return this._toggleCollapsed(true);
      }

      // Get the items to show and hide
      var items = this._getItemsToShowAndHide();

      // Add a class saying that the size has been determined. This removes the overflow:hidden
      // so that dropdowns will appear.
      this._addClass(this.el, 'spark-header--overflow-checked');

      // If there are less than two elements to show and we have hidden elements, collapse the nav.
      if (items.show.length < 2 && items.hide.length) {
        this._removeListMore();
        return this._toggleCollapsed(true);
      }

      // We aren't at the XS breakpoint and there aren't too few items to show, so disable collapsing
      this._toggleCollapsed(false);

      // If the number of children to hide is the same as those already hidden, stop.
      if (items.hide.length === this.listMoreListEl.children.length && !isSwap) {

        if (!items.hide.length) {
          this._removeListMore();
        }

        return;
      }

      // Add the elements we're supposed to show before the "more element"
      this._insertBefore(this.listEl, this.listMoreEl, items.show);

      // If we have items to hide, append them to the more element
      if (items.hide.length) {
        this._appendChildren(this.listMoreListEl, items.hide);
      }
      // Otherwise, remove the more element
      else {
        this._removeListMore();
      }
    },


    /**
     * Listen for the ready state change and rerun the menu size determination.
     */
    _listenForReadyStateChange: function() {

      // Already loaded
      if (document.readyState === 'complete' || document.readyState === 'loaded') {
        return;
      }

      // Bound listener
      var run = function() {
        if (document.readyState === 'complete' || document.readyState === 'loaded') {
          this._determineMenuSize();
          document.removeEventListener('readystatechange', run);
        }
      }.bind(this);

      // Only run once
      document.addEventListener('readystatechange', run);
    },


    /**
     * Check the primary nav breakpoint.
     * @param {String|Array} name A string or array of string names of breakpoints to check for
     */
    _isMenuBreakpoint: function(name) {
      this._getCurrentBreakpoint();
      return name instanceof Array ? name.indexOf(this.currentBreakpoint) !== -1 : this.currentBreakpoint === 'xs';
    },


    /**
     * Get the items to show and hide.
     * @return {Object}
     */
    _getItemsToShowAndHide: function() {

      var width = this.placeholder.listEl.clientWidth;
      var children = this.placeholder.listEl.children;
      var i = 0;
      var len = children.length;
      var hideIndex = -1;

      this._addListMore();

      // Always include the width of the more button.
      var childrenWidth = this.placeholder.listMoreEl.clientWidth || 0;

      // Loop through the children until we hit a point where they don't fit anymore
      for (; i < len && hideIndex === -1; i++) {
        childrenWidth += children[i].clientWidth;
        if (childrenWidth > width) {
          hideIndex = i;
        }
      }

      // Find all the children that fit and don't fit
      var items = {
        show: hideIndex !== -1 ? Array.prototype.slice.call(this.listEls, 0, hideIndex) : this.listEls,
        hide: hideIndex !== -1 ? Array.prototype.slice.call(this.listEls, hideIndex) : []
      };

      // If we have an index to swap for the last "show" element, replace that element
      if (this.moreSwapIndex > -1 && this.moreSwapIndex >= items.show.length) {

        // Remove the last element from the show array
        var toHide = items.show.splice(items.show.length - 1, 1)[0];

        // Get the index to remove from the hide array. Account for the offset.
        var toShowIndex = this.moreSwapIndex - hideIndex;

        // Remove the desired element from the hide array
        var toShow = items.hide.splice(toShowIndex, 1)[0];

        // Add the toShow element to the end of the show array
        items.show.push(toShow);

        // Insert the toHide element into the hide array at the position of
        // the element we just removed from the hide array.
        items.hide.splice(toShowIndex, 0, toHide);
      }

      return items;
    },


    /**
     * Ensure that any active item is set to the more swap index. This ensures
     * that the active item is always visible on the screen.
     */
    _ensureActiveAtMoreSwapIndex: function() {

      if (!this.listEls || !this.listEls.length) {
        return;
      }

      var el = this.el.querySelector('[class*="list-item"].active');
      if (el) {
        var parents = this._getElementMatchingParents(el, '.spark-menu__list-item', this.el);

        if (parents && parents[parents.length - 1]) {
          el = parents[parents.length - 1];
        }

        var index = this._getChildIndex(this.listEls, el);

        if (index !== this.moreSwapIndex) {
          this.moreSwapIndex = index;
        }
      }
    },

    /**
     * Add a placeholder for overflow items to the list.
     */
    _addListMore: function() {
      if (this.listMoreEl.parentNode !== this.listEl) {
        this.listEl.appendChild(this.listMoreEl);
      }
    },


    /**
     * Remove a placeholder for overflow items from the primary nav.
     */
    _removeListMore: function() {
      if (this.listMoreEl.parentNode) {
        this.listMoreEl.parentNode.removeChild(this.listMoreEl);
      }
    },


    /**
     * Reset the children of the primary navigation.
     */
    _resetMenuChildren: function() {
      this.moreSwapIndex = -1;
      this._removeClass(this.el, 'spark-header--overflow-checked');
      this._appendChildren(this.listEl, this.listEls);
    },


    /**
     * Toggle the collapsed nav style.
     * @param {Boolean} enable
     */
    _toggleCollapsed: function(enable) {

      // Same collapsed state is already set
      if (enable === this.isCollapsed) {
        return;
      }

      // Reset children and remove a special no-animate class to top-level items when we collapse
      if (enable) {
        this._enableTopLevelToggling();
        this._resetMenuChildren();
      } else {
        if (this.menu) {
          this.menu._removeAllCachedLists();
        }
        this._disableTopLevelToggling();
      }

      this.isCollapsed = enable;
      this._toggleClass(this.el, 'spark-header--collapsed', enable);
      this._toggleClass(this.el, 'spark-header--visible', !enable);
    },


    /**
     * Enable toggling on top-level items.
     */
    _enableTopLevelToggling: function() {

      var i = 0;
      var len = this.listEls.length;

      for (; i < len; i++) {
        this._removeClass(this.listEls[i], 'spark-no-animate');
      }
    },


    /**
     * Disable toggling on top-level items.
     */
    _disableTopLevelToggling: function() {

      var i = 0;
      var len = this.listEls.length;

      for (; i < len; i++) {
        this._addClass(this.listEls[i], 'spark-no-animate');
      }
    },


    /**
     * When the window resizes, redetermine the size of the primary nav elements.
     */
    _onResize: function() {

      // Ensure that any active item we may have is at the swap index
      this._ensureActiveAtMoreSwapIndex();
      this._determineMenuSize();

      // If we are fixed, do the scroll check
      if (this.fixed) {
        this.checkFixed();
      }
    },


    /**
     * Check to see if the header should be fixed.
     * @param {Object} e
     */
    _onScroll: function() {
      this.checkFixed();
    },


    /**
     * When a link in the more list is clicked, swap it with the last element in the visible list.
     * @param {Object} e
     */
    _onMoreClick: function(e) {

      // Don't do any swapping if we're in a collapsed state
      if (this.isCollapsed) {
        return;
      }

      // Get the index of the clicked element
      var li = this._getElementMatchingParent(e.target, 'li', this.listMoreListEl);

      // Save the index of the element to be swapped
      this.moreSwapIndex = this._getChildIndex(this.listEls, li);

      // Redetermine the primary nav size
      this._determineMenuSize(true);
    },


    /**
     * When the toggle is clicked, toggle the active state on the nav
     * @param {Object} e
     */
    _onToggleClick: function(e) {
      e.preventDefault();
      this.isActive = !this.isActive;
      this._toggleClass(this.navEl, 'active', this.isActive);
      this.menu._openActiveParents();
    },


    /**
     * When the nav is clicked, set to inactive.
     * @param {Object} e
     */
    _onNavClick: function(e) {
      if (e.target === this.navEl && this.isCollapsed) {
        this.isActive = !this.isActive;
        this._removeClass(this.navEl, 'active');
      }
    }
  };

  Base.exportjQuery(Header, 'Header');

  return Header;
}));

/**
 * # NumberSelector
 * An increment/decrement for number inputs.
 *
 * @example
 * new NumberSelector(el);
 *
 * @module components/number-selector.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.NumberSelector = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  /**
   * NumberSelector constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var NumberSelector = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  NumberSelector.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _getElementMatchingParent: Base.getElementMatchingParent,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: [],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      inputEl: null,
      buttonUpEl: null,
      buttonDownEl: null,
      _onChangeBound: null,
      _onClickBound: null
    },


    /**
     * Get or set the current value of the number selector.
     * @param {Number} val @optional
     * @return {Number} The current value
     */
    value: function(val) {

      if (val !== undefined) {
        this.inputEl.value = this._getConformedNumber(val);
      }

      return parseInt(this.inputEl.value, 10);
    },


    /**
     * Increment by the step value.
     */
    increment: function() {
      this.value(this.value() + (this._getStep() || 1));
    },


    /**
     * Decrement by the step value.
     */
    decrement: function() {
      this.value(this.value() - (this._getStep() || 1));
    },

    /**
     * Disable number selector
     */
    disable: function() {
      this.inputEl.disabled = true;
      this.buttonUpEl.disabled = true;
      this.buttonDownEl.disabled = true;
    },

    /**
     * Enable number selector
     */
    enable: function() {
      this.inputEl.disabled = false;
      this.buttonUpEl.disabled = false;
      this.buttonDownEl.disabled = false;
    },


    /**
     * Store a reference to the whole number-selector, as well as the
     * input element. Also, get some default values from the input
     * element (min, max, steps).
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.inputEl = el.querySelector('input');
      this.buttonUpEl = el.querySelector('.spark-number-selector__up, .spark-number-picker__up');
      this.buttonDownEl = el.querySelector('.spark-number-selector__down, .spark-number-picker__down');

      if (!this.inputEl) {
        throw new Error('NumberSelector must include an `<input>` element!');
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onChangeBound = this._onChange.bind(this);
    },


    /**
     * Add event listeners for touchstart and mouse click.
     */
    _addEventListeners: function() {
      this.el.addEventListener('click', this._onClickBound);
      this.el.addEventListener('change', this._onChangeBound);
    },


    /**
     * Remove event listeners for touchstart and mouse click.
     */
    _removeEventListeners: function() {
      this.el.removeEventListener('click', this._onClickBound);
      this.el.removeEventListener('change', this._onChangeBound);
    },


    /**
     * Get the current value of the min property.
     * @return {Number}
     */
    _getMin: function() {
      return this._getInputPropAsNumber('min');
    },


    /**
     * Get the current value of the max property.
     * @return {Number}
     */
    _getMax: function() {
      return this._getInputPropAsNumber('max');
    },


    /**
     * Get the current value of the step property.
     * @return {Number}
     */
    _getStep: function() {
      return this._getInputPropAsNumber('step');
    },


    /**
     * Get a property as a number.
     * @param {String} key
     * @return {Number}
     */
    _getInputPropAsNumber: function(key) {
      return parseFloat(this.inputEl.getAttribute(key));
    },


    /**
     * Get the given number within the min/max range of the input element.
     * @param {Number} num
     * @return {Number}
     */
    _getConformedNumber: function(num) {

      var max = this._getMax();
      var min = this._getMin();
      var step = this._getStep();

      // Move in increments if we have a defined step size
      if (step) {

        var diff = num % step;
        var halfStep = step / 2;
        var overHalf = diff >= halfStep;

        num = overHalf ? num + (step - diff) : num - diff;
      }

      return max !== undefined && num > max ? max : (min !== undefined && num < min ? min : num);
    },


    /**
     * When either of the buttons are clicked, update the value.
     * @param {Object} e
     */
    _onClick: function(e) {

      // The increment button is clicked
      // @todo Remove .spark-number-picker in 2.0.0
      if (this._getElementMatchingParent(e.target, '.spark-number-selector__up, .spark-number-picker__up', this.el)) {
        e.preventDefault();
        this.increment();
      }
      // The decrement button is clicked
      // @todo Remove .spark-number-picker in 2.0.0
      else if (this._getElementMatchingParent(e.target, '.spark-number-selector__down, .spark-number-picker__down', this.el)) {
        e.preventDefault();
        this.decrement();
      }
    },


    /**
     * When the input value changes, max sure we are in bounds.
     * @param {Object} e
     */
    _onChange: function() {
      this.value(parseFloat(this.inputEl.value));
    }
  };

  Base.exportjQuery(NumberSelector, 'NumberSelector');

  return NumberSelector;
}));

/**
 * # NumberPicker
 * An increment/decrement for number inputs.
 *
 * @example
 * new NumberPicker(el);
 *
 * @deprecated Remove in 2.0.0 in favor of number selector naming
 * @module components/number-picker.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/number-selector'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('./number-selector'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.NumberPicker = factory(root.Spark.BaseComponent, root.Spark.NumberSelector);
  }
}(this, function(Base, NumberSelector) {

  'use strict';

  Base.exportjQuery(NumberSelector, 'NumberPicker');

  return NumberSelector;
}));

/**
 * # TextInput
 * A text input container.
 *
 * @example
 * new TextInput(el, {
 *   // Optional. Validation function, will override default one
 *   validate: function() {},
 *   // Optional. Regular expression of validation
 *   validatePattern: '([^@]{1,})(@)([^\.]{1,})(\.)([^\.]{1,})',
 *   // Only if set validatePattern work. Optional. Callback for when validate.
 *   onValidate: function(inputInstance, regResult, value) {},
 *   // Optional. Callback for when the input value changes.
 *   onChange: function(inputInstance, value) {},
 *   // Optional. Callback for when input focus.
 *   onFocus: function(inputInstance, value) {},
 *   // Optional. Callback for when the input blur.
 *   onBlur: function(inputInstance, value) {}
 * });
 *
 * @module components/text-input.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/animate-height', 'spark/helpers/typeahead'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/animate-height'), require('../helpers/typeahead'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.TextInput = factory(root.Spark.BaseComponent, root.Spark.animateHeight, root.Spark.Typeahead);
  }
}(this, function(Base, animateHeight, Typeahead) {

  'use strict';

  var noop = function() {};

  /**
   * TextInput constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var TextInput = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._bindEventListenerCallbacks();
    this._cacheElements(el);
    this._setParams(params || {});
    this._addEventListeners();
  };

  TextInput.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    remove: Base.remove,
    _toggleClass: Base.toggleClass,
    _hasClass: Base.hasClass,
    _getChildIndex: Base.getChildIndex,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['validate', 'validatePattern', 'onValidate', 'onChange', 'onFocus', 'onBlur'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      inputEl: null,
      passwordToggleEl: null,
      clearEl: null,
      isActive: false,
      isTextarea: false,
      validatePattern: false,
      type: null,
      showCharacters: false,
      showCharactersRemaining: false,
      maxlength: null,
      typeahead: null,
      onChange: noop,
      onValidate: noop,
      onFocus: noop,
      onBlur: noop,
      _onFocusBound: null,
      _onBlurBound: null,
      _onInputBound: null,
      _onTogglePasswordViewHideBound: null,
      _onClearClickBound: null
    },


    /**
     * Show the input by adding the active state and setting character counts (if necessary).
     */
    show: function() {
      this.isActive = true;
      this._updateClass();
      this._setCharactersCount();
    },


    /**
     * Hide the input by removing the active state.
     */
    hide: function() {
      this.isActive = false;
      this._updateClass();
    },


    /**
     * Run the validation.
     * @todo: this needs some work to be made more robust. Should be able to pass validation pattern,
     * should store a valid state, etc.
     */
    validate: function() {

      var validate = this.validatePattern;

      // Nothing to validate.
      if (!validate) {
        return;
      }

      var re = new RegExp(validate);

      // Passes the component instance, true or false for valid, and the current value.
      (this.onValidate || noop)(this, re.test(this.inputEl.value), this.inputEl.value);
    },


    /**
     * Set the error state.
     * @param {String} message Optional
     */
    setError: function(message) {

      // Animate down
      if (!this._isMessageVisible()) {
        this._showMessage();
      }

      this.clearWarning();
      this.clearSuccess();

      this.el.setAttribute('data-error', true);

      if (message) {
        this.setMessage(message);
      }
    },


    /**
     * Set the error state.
     */
    clearError: function() {
      this.el.removeAttribute('data-error', true);
    },


    /**
     * Set the warning state.
     * @param {String} message Optional
     */
    setWarning: function(message) {

      // Animate down
      if (!this._isMessageVisible()) {
        this._showMessage();
      }

      this.clearError();
      this.clearSuccess();

      this.el.setAttribute('data-warning', true);

      if (message) {
        this.setMessage(message);
      }
    },


    /**
     * Set the error state.
     */
    clearWarning: function() {
      this.el.removeAttribute('data-warning', true);
    },


    /**
     * Set the success state.
     * @param {String} message Optional
     */
    setSuccess: function(message) {

      // Animate down
      if (!this._isMessageVisible()) {
        this._showMessage();
      }

      this.clearError();
      this.clearWarning();

      this.el.setAttribute('data-success', true);

      if (message) {
        this.setMessage(message);
      }
    },


    /**
     * Set the success state.
     */
    clearSuccess: function() {
      this.el.removeAttribute('data-success', true);
    },


    /**
     * Clear all messages.
     */
    clearMessages: function() {
      this._hideMessage(function() {
        this.clearError();
        this.clearWarning();
        this.clearSuccess();
      }.bind(this));
    },


    /**
     * Set the message text.
     * @param {String} message
     */
    setMessage: function(message) {
      this.messageEl.innerHTML = message;
    },

    setValue: function(value) {
      this.inputEl.value = value;
      this._onFocus();
      this._onBlur();
    },


    /**
     * Show the message
     */
    _showMessage: function() {

      animateHeight({
        el: this.el,
        toggleEl: '.spark-input__message'
      });
    },


    /**
     * Hide the message.
     * @param {Function} callback
     */
    _hideMessage: function(callback) {

      animateHeight({
        el: this.el,
        toggleEl: '.spark-input__message',
        toggleValue: 'none',
        action: 'collapse',
        onComplete: callback
      });
    },


    /**
     * Is the message currently visible?
     * @return {Boolean}
     */
    _isMessageVisible: function() {
      return this.el.getAttribute('data-error') || this.el.getAttribute('data-warning') || this.el.getAttribute('data-success');
    },


    /**
     * Store a reference to the needed elements.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.inputEl = this.el.querySelector('input, textarea');
      this.passwordToggleEl = this.el.querySelector('.spark-input__password-toggle');

      if (!this.inputEl) {
        throw new Error('No <input> or <textarea> element present in input container!', this.el);
      }

      this.messageEl = this.el.querySelector('.spark-input__message') || document.createElement('span');
      this.clearEl = this.el.querySelector('.spark-input__clear');

      this._parseParams();

      if (this.inputEl.value) {
        this.show();
        this._onInput();
      }
    },


    /**
     * Parse parameters from the elements.
     */
    _parseParams: function() {

      this.validatePattern = this.validatePattern || this.inputEl.getAttribute('data-validate');
      this.type = this.inputEl.getAttribute('type') || 'text';
      this.showCharacters = this.el.getAttribute('data-characters') !== null ? true : false;
      this.showCharactersRemaining = this.el.getAttribute('data-characters-remaining') !== null ? true : false;
      this.maxlength = this.inputEl.getAttribute('maxlength') || this.inputEl.getAttribute('data-maxlength-soft') || null;
      this.isTextarea = this.inputEl.nodeName.toLowerCase() === 'textarea' ? true : false;
      this.typeahead = this.inputEl.getAttribute('data-typeahead') !== null ? new Typeahead(this.el, {
        onBlur: this._onBlurBound
      }) : null;
      this.isActive = this.inputEl.value ? true : false;
    },


    /**
     * Set the characters count attribute.
     */
    _setCharactersCount: function() {

      if (this.showCharacters) {
        this.el.setAttribute('data-characters', this.inputEl.value.length);
      } else if (this.showCharactersRemaining) {

        var remaining = this.maxlength - this.inputEl.value.length;

        this.el.setAttribute('data-characters-remaining', remaining);

        if (remaining < 1) {
          this.el.setAttribute('data-characters-remaining-danger', true);
        } else {
          this.el.removeAttribute('data-characters-remaining-danger');
        }
      }
    },


    /**
     * Set the height of the textarea so that it doesn't scroll.
     */
    _setTextareaHeight: function() {

      var style = window.getComputedStyle(this.inputEl);
      var borders = parseInt(style.borderTopWidth, 10) + parseInt(style.borderBottomWidth, 10);

      this.inputEl.style.height = null;

      var height = this.inputEl.scrollHeight;
      var lines;

      // No height, most likely the element is invisible. Get a rough
      // approximation of height so we have something.
      if (!height) {
        lines = this.inputEl.innerHTML.split('\n');
        height = (Math.max(parseFloat(style.lineHeight)) * (lines.length)) + parseFloat(style.paddingTop) + parseFloat(style.paddingBottom);
      }

      this.inputEl.style.height = (height + borders) + 'px';
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onFocusBound = this._onFocus.bind(this);
      this._onBlurBound = this._onBlur.bind(this);
      this._onInputBound = this._onInput.bind(this);
      this._onTogglePasswordViewHideBound = this._onTogglePasswordViewHide.bind(this);
      this._onClearClickBound = this._onClearClick.bind(this);
    },


    /**
     * Add event listeners for focus, blur, input, and click.
     */
    _addEventListeners: function() {
      this.inputEl.addEventListener('focus', this._onFocusBound);
      this.inputEl.addEventListener('blur', this._onBlurBound);
      this.inputEl.addEventListener('input', this._onInputBound);
      if (this.passwordToggleEl) {
        this.passwordToggleEl.addEventListener('click', this._onTogglePasswordViewHideBound);
      }

      if (this.clearEl) {
        this.clearEl.addEventListener('click', this._onClearClickBound);
      }
    },


    /**
     * Remove event listeners for focus, blur and input.
     */
    _removeEventListeners: function() {

      this.inputEl.removeEventListener('focus', this._onFocusBound);
      this.inputEl.removeEventListener('blur', this._onBlurBound);
      this.inputEl.removeEventListener('input', this._onInputBound);
      this.passwordToggleEl.removeEventListener('click', this._onTogglePasswordViewHideBound);

      if (this.clearEl) {
        this.clearEl.removeEventListener('click', this._onClearClickBound);
      }
    },


    /**
     * Update the active class.
     */
    _updateClass: function() {
      this._toggleClass(this.el, 'active', this.isActive);
    },


    /**
     * When the input element gains focus.
     * @param {Object} e
     */
    _onFocus: function() {
      this.show();
      (this.onFocus || noop)(this, this.inputEl.value);
    },


    /**
     * When the input element loses focus.
     * @param {Object} e
     */
    _onBlur: function() {
      if (!this.inputEl.value) {
        this.hide();
      }
      (this.onBlur || noop)(this, this.inputEl.value);
    },


    /**
     * When the value is about to change, run the validation, set the characters count
     * and resize if we're a textarea.
     * @param {Object} e
     */
    _onInput: function() {

      this.validate();
      this._setCharactersCount();

      if (this.isTextarea) {
        this._setTextareaHeight();
      }

      (this.onChange || noop)(this, this.inputEl.value);
    },


    /**
     * When a clear button is clicked, empty the field.
     * @param {Object} e
     */
    _onClearClick: function() {
      this.inputEl.value = '';
      this.hide();
      (this.onChange || noop)(this, this.inputEl.value);
    },

    /**
     * Toggle the current type value (text/password) of password input.
     * @param {Object} e
     */
    _onTogglePasswordViewHide: function(e) {
      e.preventDefault();
      this.inputEl.setAttribute('type', this.inputEl.getAttribute('type') === 'password' ? 'text' : 'password');
    }
  };

  Base.exportjQuery(TextInput, 'TextInput');

  return TextInput;
}));

/**
 * # Popover
 * Show and hide a popover. Should do some sanity checks on positioning as well.
 *
 * @example
 * new Popover(el, {
 *   // Optional. Default anchoring of the content's x and y-axis relative to the button.
 *   defaultAnchorX: 'center', // 'left', 'center', 'right'
 *   defaultAnchorY: 'center' // 'left', 'center', 'right'
 * });
 *
 * @module components/popover.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Popover = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * Popover constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Popover = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._setParams(params || {});
    this._cacheElements(el);
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  Popover.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _hasClass: Base.hasClass,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _elementHasParent: Base.elementHasParent,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _getElementOffset: Base.getElementOffset,
    _appendChildren: Base.appendChildren,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['defaultAnchorX', 'defaultAnchorY', 'direction', 'contentEl', 'onOpen', 'onClose'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      toggleEl: null,
      contentEl: null,
      caretEl: null,
      anchorTo: null,
      isActive: false,
      isPaused: false,
      direction: 'bottom',
      defaultAnchorX: 'center',
      defaultAnchorY: 'center',
      closeTimer: null,
      onOpen: null,
      onClose: null,
      _onClickBound: null,
      _onContentClickBound: null,
      _onWindowClickBound: null,
      _onWindowResizeBound: null
    },


    /**
     * Open.
     * @param {Object} params Optional
     */
    open: function(params) {

      params = params || {};

      // If there is a timer running for the close event, clear it so it doesn't close stuff during open.
      if (this.closeTimer) {
        clearTimeout(this.closeTimer);
        this.closeTimer = null;
      }

      this.anchorTo = params.anchorTo || this.anchorTo;

      this._addWindowEventListeners();
      this.isActive = true;
      this._moveContent();
      this._updatePosition(params);
      this._addClass(this.contentEl, 'animate');
      var e = document.createEvent('Event');
      e.initEvent('spark.visible-children', true, true);
      this.contentEl.dispatchEvent(e);
      this._updateAttributes();

      (params.complete || noop)();
      (this.onOpen || noop)();
    },


    /**
     * Close.
     * @param {Object} params Optional
     */
    close: function(params) {

      params = params || {};

      // If there is a timer running for the close even, clear it so we don't run close stuff twice.
      if (this.closeTimer) {
        clearTimeout(this.closeTimer);
        this.closeTimer = null;
      }

      this._removeWindowEventListeners();
      this.isActive = false;
      this._updateAttributes();

      this.closeTimer = setTimeout(function() {
        this._clearPosition();
        this._moveContent();
        this._removeClass(this.contentEl, 'animate');
        (params.complete || noop)();
        (this.onClose || noop)();
      }.bind(this), 250);
    },


    /**
     * Toggle the open state.
     */
    toggle: function() {
      this[this.isActive ? 'close' : 'open']();
    },


    /**
     * Set the content. Optionally append instead of replacing.
     * @param {Element|Array|NodeList} content
     * @param {Object} params Optional
     */
    setContent: function(content, params) {
      params = params || {};
      this._appendChildren(this.contentEl, content.length ? content : [content], !(params.append || false));
    },


    /**
     * Pause the popover. This stops is from positioning itself.
     */
    pause: function() {
      this.isPaused = true;
      this._clearPosition();
    },


    /**
     * Resume the popover positioning itself.
     */
    resume: function() {
      this.isPaused = false;
    },


    /**
     * Store a reference to the tabs list, each tab and each panel.
     * Set which tab is active, or use the first.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      // If a content element was already passed, make sure it has a popover content class
      if (this.contentEl) {
        this._addClass(this.contentEl, 'spark-popover__content');
        this._addClass(this.contentEl, 'spark-popover__content--' + this.direction);
      }

      this.el = el;
      this.toggleEl = this.el.querySelector('.spark-popover__toggle, [data-role="toggle"]') || this.el;
      this.contentEl = this.contentEl || this.el.querySelector('[class*="spark-popover__content--"]') || this._createContentEl();
      this.caretEl = this.el.querySelector('.spark-popover__caret') || this._createCaretEl();
      this.isActive = this._hasClass(this.toggleEl, 'popover-active');

      //check if we're on an asp forms page
      var a = document.querySelectorAll('body > form');
      for (var i = 0; i < a.length; i++) {
        if (a[i].contains(this.el)) {
          this.targetForm = a[i];
          break;
        }
      }

      this._determineDirectionAndAnchor();
    },


    /**
     * Move the content node to be at the root of the page.
     */
    _moveContent: function() {
      if (this.targetForm) {
        if (this.isActive) {
          this.targetForm.appendChild(this.contentEl);
        } else if (this.contentEl.parentNode !== this.el && !this.isActive) {
          this.el.appendChild(this.contentEl);
        }
      } else {
        // Move the popover to the root of the body.
        if (this.isActive && (this.contentEl.parentNode !== document.body)) {
          document.body.appendChild(this.contentEl);
        } else if (this.contentEl.parentNode !== this.el && !this.isActive) {
          this.el.appendChild(this.contentEl);
        }
      }
    },


    /**
     * Determine the direction of of the popover.
     */
    _determineDirectionAndAnchor: function() {

      if (this.contentEl.className.indexOf('popover__content--left') !== -1) {
        return (this.direction = 'left');
      } else if (this.contentEl.className.indexOf('popover__content--right') !== -1) {
        return (this.direction = 'right');
      } else if (this.contentEl.className.indexOf('popover__content--top') !== -1) {
        return (this.direction = 'top');
      }

      this.defaultAnchorX = this.contentEl.getAttribute('data-anchor-x') || this.defaultAnchorX;
      this.defaultAnchorY = this.contentEl.getAttribute('data-anchor-y') || this.defaultAnchorY;
    },


    /**
     * Update classes for the open or close state.
     */
    _updateAttributes: function() {

      this._toggleClass(this.el, 'popover-active', this.isActive);
      this._toggleClass(this.contentEl, 'popover-active', this.isActive);
      this._toggleClass(this.toggleEl, 'active', this.isActive);
    },


    /**
     * Update the position of the popover.
     */
    _updatePosition: function() {

      if (!this.isActive) {
        return;
      }

      if (this.isPaused) {
        return this._clearPosition();
      }

      this._addClass(this.contentEl, 'measure');

      var toggleEl = this.anchorTo || this.toggleEl;
      var caretEl = this.caretEl;
      var toggleOffset = this._getElementOffset(toggleEl);
      var toggleHeight = toggleEl.offsetHeight;
      var toggleWidth = toggleEl.offsetWidth;
      var caretHeight = caretEl.offsetHeight;
      var caretWidth = caretEl.offsetWidth;
      var caretAdjustedWidth = Math.sqrt(Math.pow(caretWidth, 2) + Math.pow(caretHeight, 2)); // Accounts for the 90deg rotation in the CSS
      var contentHeight = this.contentEl.offsetHeight;
      var contentWidth = this.contentEl.offsetWidth;
      var docHeight = document.documentElement.scrollHeight;
      var docWidth = document.documentElement.scrollWidth;
      var xAxis = this.direction === 'bottom' || this.direction === 'top';
      var top = 0;
      var left = 0;
      var anchor = xAxis ? this.defaultAnchorX : this.defaultAnchorY;
      var anchorPos = xAxis ? (toggleWidth / 2) : (toggleHeight / 2);

      // Initial values
      if (this.direction === 'left') {
        top = toggleOffset.top - (contentHeight / 2) + (toggleHeight / 2);
        left = toggleOffset.left - contentWidth;
      } else if (this.direction === 'right') {
        top = toggleOffset.top - (contentHeight / 2) + (toggleHeight / 2);
        left = toggleOffset.left + toggleWidth;
      } else if (this.direction === 'top') {
        top = toggleOffset.top - contentHeight;
        left = toggleOffset.left - (contentWidth / 2) + (toggleWidth / 2);
      } else {
        top = toggleOffset.top + toggleHeight;
        left = toggleOffset.left - (contentWidth / 2) + (toggleWidth / 2);
      }

      // Check boundaries
      if (xAxis) {
        if (left + contentWidth > docWidth) {
          anchor = 'right';
          left = toggleOffset.left + toggleWidth - contentWidth;
          anchorPos = ((toggleWidth - caretAdjustedWidth) / 2 - caretWidth / 2);
        }
        if (left < 0) {
          anchor = 'left';
          left = 0;
          anchorPos = toggleOffset.left + ((toggleWidth - caretAdjustedWidth) / 2 + caretWidth / 2);
        }
      } else {
        if (top + contentHeight > docHeight) {
          anchor = 'bottom';
          anchorPos = anchorPos - (this.caretEl.offsetWidth * 1.5);
          top = toggleOffset.top;
        } else if (top < 0) {
          anchor = 'top';
          top = toggleOffset.top;
        }
      }

      this.contentEl.style.top = top + 'px';
      this.contentEl.style.left = left + 'px';
      this.contentEl.setAttribute('data-anchor', anchor);
      this._removeClass(this.contentEl, 'measure');

      this._updateCaretAnchorPosition({
        anchor: anchor,
        anchorPos: anchorPos
      });
    },


    /**
     * Clear position styling.
     */
    _clearPosition: function() {
      this.contentEl.removeAttribute('style');
      this.contentEl.removeAttribute('data-anchor-x');
      this.contentEl.removeAttribute('data-anchor-y');
      this.caretEl.removeAttribute('style');
    },


    /**
     * Update the position of the caret to be centered on the toggle element.
     */
    _updateCaretAnchorPosition: function(params) {

      if (!this.caretEl) {
        return;
      }

      this.caretEl.style.left = '';
      this.caretEl.style.right = '';
      this.caretEl.style.top = '';
      this.caretEl.style.bottom = '';
      this.caretEl.style[params.anchor] = params.anchorPos + 'px';
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onContentClickBound = this._onContentClick.bind(this);
      this._onWindowClickBound = this._onWindowClick.bind(this);
      this._onWindowResizeBound = this._onWindowResize.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      this.el.addEventListener('click', this._onClickBound);
      this.contentEl.addEventListener('click', this._onContentClickBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      this.el.removeEventListener('click', this._onClickBound);
      this.contentEl.removeEventListener('click', this._onContentClickBound);
    },


    /**
     * Add event listeners to the window.
     */
    _addWindowEventListeners: function() {
      this._removeWindowEventListeners();
      window.addEventListener('click', this._onWindowClickBound);
      window.addEventListener('resize', this._onWindowResizeBound);
      window.addEventListener('orientationchange', this._onWindowResizeBound);
    },


    /**
     * Remove window event listeners.
     */
    _removeWindowEventListeners: function() {
      window.removeEventListener('click', this._onWindowClickBound);
      window.removeEventListener('resize', this._onWindowResizeBound);
      window.removeEventListener('orientationchange', this._onWindowResizeBound);
    },


    /**
     * Create a content element.
     * @return {Element}
     */
    _createContentEl: function() {
      var el = document.createElement('div');
      this._addClass(el, 'spark-popover__content');
      this._addClass(el, 'spark-popover__content--' + this.direction);
      el.setAttribute('role', 'tooltip');
      return el;
    },


    /**
     * Create the caret element.
     * @return {Element}
     */
    _createCaretEl: function() {
      var el = document.createElement('div');
      el.className = 'spark-popover__caret';
      this.contentEl.appendChild(el);
      return el;
    },


    /**
     * When we are clicked, toggle the popover-active state.
     * @param {Object} e
     */
    _onClick: function(e) {

      // If this is the toggle element, toggle.
      if (e.target === this.toggleEl || this._elementHasParent(e.target, this.toggleEl)) {
        e.preventDefault();
        this.toggle();
        return;
      }
    },


    /**
     * When the toggle is clicked, close if it's a link. If it's content, don't do anything but stop
     * the event from bubbling.
     * @param {Object} e
     */
    _onContentClick: function(e) {

      // If this is a link, close.
      if (this._getElementMatchingParent(e.target, '.spark-popover__list-link', this.contentEl) || this._getElementMatchingParent(e.target, '.spark-popover__close', this.contentEl)) {
        this.close();
        return;
      }
    },


    /**
     * When the window is clicked and it's not part of the popover, close the popover.
     * @param {Objec} e
     */
    _onWindowClick: function(e) {
      if (e.target !== this.el && !this._elementHasParent(e.target, this.el) && !this._elementHasParent(e.target, this.contentEl)) {
        this.close();
      }
    },


    /**
     * When the window is resized, ensure the proper position of the popover.
     */
    _onWindowResize: function() {
      this._updatePosition();
    }
  };

  Base.exportjQuery(Popover, 'Popover');

  return Popover;
}));

/**
 * # SelectInput
 * A select input container.
 *
 * @example
 * new SelectInput(el);
 *
 * @module components/select-input.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.SelectInput = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * SelectInput constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var SelectInput = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  SelectInput.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['onChange', 'onFocus', 'onBlur'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      selectEl: null,
      labelEl: null,
      hasValue: false,
      isActive: false,
      onChange: null,
      onFocus: null,
      onBlur: null,
      previousValue: null,
      _onFocusBound: null,
      _onBlurBound: null,
      _onInputBound: null
    },


    /**
     * Get the value.
     * @return {String}
     */
    getValue: function() {
      return this.selectEl.value;
    },


    /**
     * Set the value.
     * @param {String|Number} val
     */
    setValue: function(val) {

      // Cast to a string for comparison
      val = val + '';

      var i = 0;
      var len = this.selectEl.children.length;

      for (; i < len; i++) {
        if (this.selectEl.children[i].value === val) {
          this.selectEl.children[i].selected = true;
          this._updateClass();
          return;
        }
      }
    },


    /**
     * Set the options.
     * @param {Array} opts
     */
    setOptions: function(opts) {

      var i = 0;
      var len = opts.length;
      var str = '';

      // Store the index of the currently selected option so we can set
      // it when we're all done.
      var curIndex = this.selectEl.selectedIndex;

      for (; i < len; i++) {
        str += '<option ' + (opts[i].value !== undefined ? 'value="' + (opts[i].value || '') + '"' : '') + '>' + (opts[i].text || '') + '</option>';
      }

      this.selectEl.innerHTML = str;
      this.selectEl.selectedIndex = Math.min(len - 1, curIndex);
    },


    /**
     * Set the value of the label.
     * @param {String} text
     */
    setLabel: function(text) {
      if (!this.labelEl) return;
      this.labelEl.innerHTML = text;
    },


    /**
     * Store a reference to the needed elements.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.selectEl = this.el.querySelector('select');
      this.labelEl = this.el.querySelector('.spark-label');

      if (!this.selectEl) {
        throw new Error('A <select> element must be present!', this.el);
      }

      this._updateClass();
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onFocusBound = this._onFocus.bind(this);
      this._onBlurBound = this._onBlur.bind(this);
      this._onInputBound = this._onInput.bind(this);
    },


    /**
     * Add event listeners for focus, blur and input.
     */
    _addEventListeners: function() {
      this.selectEl.addEventListener('focus', this._onFocusBound);
      this.selectEl.addEventListener('blur', this._onBlurBound);
      this.selectEl.addEventListener('input', this._onInputBound);
    },


    /**
     * Remove event listeners for focus, blur and input.
     */
    _removeEventListeners: function() {
      this.selectEl.removeEventListener('focus', this._onFocusBound);
      this.selectEl.removeEventListener('blur', this._onBlurBound);
      this.selectEl.removeEventListener('input', this._onInputBound);
    },


    /**
     * Update the active class.
     */
    _updateClass: function() {
      this.hasValue = this.selectEl.value ? true : false;
      this._toggleClass(this.el, 'has-value', this.hasValue);
      this._toggleClass(this.el, 'active', this.isActive);
    },


    /**
     * When the input element gains focus.
     * @param {Object} e
     */
    _onFocus: function() {
      this.isActive = true;
      this._updateClass();
      var value = this.getValue();
      (this.onFocus || noop)(value);
    },


    /**
     * When the input element loses focus.
     * @param {Object} e
     */
    _onBlur: function() {
      this.isActive = false;
      this._updateClass();
      var value = this.getValue();
      (this.onBlur || noop)(value);
    },


    /**
     * When the value is about to change, run the validation, set the characters count
     * and resize if we're a textarea.
     * @param {Object} e
     */
    _onInput: function() {

      this._updateClass();

      var value = this.getValue();

      if (value !== this.previousValue) {
        this.previousValue = value;
        (this.onChange || noop)(value);
      }
    }
  };

  Base.exportjQuery(SelectInput, 'SelectInput');

  return SelectInput;
}));

/**
 * # DateSelect
 * Create a select list of days, months or years.
 *
 * @example
 * new DateSelect(el);
 *
 * @module helper/date-select.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/select-input', 'spark/helpers/date'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('../components/base'), require('../components/select-input'), require('./date'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.DateSelect = factory(root.Spark.BaseComponent, root.Spark.SelectInput, root.Spark.dateHelper);
  }
}(this, function(Base, SelectInput, dateHelper) {

  'use strict';

  var noop = function() {};

  /**
   * DateSelect constructor
   * @param {Element} el Optional
   * @param {Object} params Optional
   */
  function DateSelect(el, params) {

    // If the first argument is a plain object, create a default element
    // since the user MUST provide additional params but the element
    // is optional. Doing it this way to keep the arity the same
    // as other components.
    if (!(el instanceof HTMLElement)) {
      params = el || {};
      el = this._createDefaultElement();
    }

    this._setParams(this.defaults, true);
    this._setParams(params);
    this._bindEventListenerCallbacks();
    this._createSelect(el);
  }

  DateSelect.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['type', 'monthNames', 'min', 'max', 'onChange'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      type: null,
      monthNames: null,
      min: null,
      max: null,
      select: null,
      onChange: null,
      _onChangeBound: null
    },


    /**
     * Get the value.
     * @param {Boolean} asInt Get the value as a parsed integer.
     * @return {Mixed}
     */
    getValue: function(asInt) {
      return asInt ? parseInt(this.select.getValue(), 10) : this.select.getValue();
    },


    /**
     * Set the value.
     * @param {Mixed} val
     */
    setValue: function(val) {
      return this.select.setValue(val);
    },


    /**
     * Update the date select's options.
     * @param {Object|Array} params
     */
    setOptions: function(params) {

      params = params || {};

      this.min = params.min;
      this.max = params.max;

      // Default days
      if (this.type === 'day') {
        this.min = this.min || 1;
        this.max = this.max || 31;
      }
      // Default months
      else if (this.type === 'month') {

        // No monthNames yet and no min or max
        if ((!this.monthNames && !this.min && !this.max) || params.monthNames) {
          this.monthNames = params.monthNames || this._getDefaultMonthNames();
        } else {
          this.min = this.min || 1;
          this.max = this.max || 12;
        }
      }
      // Default years
      else if (this.type === 'year') {
        var date = new Date();
        this.min = this.min || date.getFullYear() - 100;
        this.max = this.max || (this.min || date.getFullYear()) + 100;
      }

      var i = this.min ? this.min - 1 : 0;
      var len = this.max || this.monthNames.length;
      var opts = [{}];

      for (; i < len; i++) {
        opts.push({
          value: i + 1,
          text: this.monthNames ? this.monthNames[i] : i + 1
        });
      }

      this.select.setOptions(opts);
    },


    /**
     * Set the label text for the select input.
     * @param {String} text Optional
     */
    setLabel: function(text) {
      this.select.setLabel(text !== undefined ? text : this._getTypeText());
    },


    /**
     * Create the default input element.
     * @param {String} type
     * @return {Element}
     */
    _createDefaultElement: function() {
      var el = document.createElement('span');
      el.className = 'spark-select';
      el.innerHTML = '<select class="spark-select__input"></select><span class="spark-label"></span>';
      return el;
    },


    /**
     * Create a select input helper.
     * @param {Object} el
     */
    _createSelect: function(el) {
      this.select = new SelectInput(el, {
        onChange: this._onSelectChangeBound
      });
      this.setOptions();
      this.setLabel();
    },


    /**
     * Make a list of month options.
     * @todo : make this work with languages. perhaps either by passing an option
     * or detecting the language attribute on the `<html>` tag.
     * @return {Array}
     */
    _getDefaultMonthNames: function() {
      return dateHelper.getMonthNamesShort();
    },


    /**
     * Get the text for this type of date select.
     * @todo make this work with languages.
     * @return {String}
     */
    _getTypeText: function() {
      return this.type.charAt(0).toUpperCase() + this.type.slice(1);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onSelectChangeBound = this._onSelectChange.bind(this);
    },


    /**
     * When the typeahead changes, make sure the value is valid. This
     * is very basic validation. More complex validation like the number
     * of days in a specific month should be handled by the callback.
     * And run our callback.
     * @param {String} val The value of the input
     * @param {String} oldVal The previous value
     */
    _onSelectChange: function(val) {
      (this.onChange || noop)(val, this);
    }
  };

  Base.exportjQuery(DateSelect, 'DateSelect');

  return DateSelect;
}));

/**
 * # Calendar Popover
 * Create a calendar from which dates can be selected.
 *
 * @example
 * new CalendarPopover(el);
 *
 * @module components/calendar-popover.js
 */

/**
 * @ignore
 */
(function(root, factory) {
  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/popover', 'spark/helpers/parse-date-format', 'spark/helpers/date', 'spark/helpers/build-select'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('./popover'), require('../helpers/parse-date-format'), require('../helpers/date'), require('../helpers/build-select'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.CalendarPopover = factory(root.Spark.BaseComponent, root.Spark.Popover, root.Spark.parseDateFormat, root.Spark.dateHelper, root.Spark.buildSelect);
  }
}(this, function(Base, Popover, parseDateFormat, dateHelper, buildSelect) {

  'use strict';

  var noop = function() {};
  var domDateFormat = 'YYYY-MM-DD';
  var parsedDomFormat = parseDateFormat(domDateFormat);

  /**
   * Calendar Popover constructor
   * @param {Element} el Optional
   * @param {Object} params Optional
   */
  function CalendarPopover(el, params) {

    params = params || {};

    // If the first element is an array or array like (NodeList),
    // we will be working with a range.
    if (el && el.hasOwnProperty('length')) {
      params.els = el;
    }
    // If the first argument is a plain object, create a default element
    // since the user MUST provide additional params but the element
    // is optional. Doing it this way to keep the arity the same
    // as other components.
    else if (!(el instanceof HTMLElement)) {
      params = el || {};
      params.els = [this._createDefaultElement()];
    }
    // A single element is passed.
    else {
      params.els = [el];
    }

    params.visibleCounts = params.visibleCounts || (params.visibleCount ? [params.visibleCount] : [1]);
    params.mins = params.mins || (params.min ? [params.min] : []);
    params.maxes = params.maxes || (params.max ? [params.max] : []);
    params.values = params.values || (params.value ? [params.value] : []);

    this._setParams(this.defaults, true);
    this._setParams(params);
    this._cacheElements();
    this._parseParams();

    // Create a calendar element if we weren't passed one.
    (this.calendarEl ? noop : this._createCalendar).call(this);
    this._cacheCalendarElements();

    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._initDatesToShow();
  }

  CalendarPopover.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _toggleClass: Base.toggleClass,
    _appendChildren: Base.appendChildren,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _getElementOffset: Base.getElementOffset,
    _triggerEvent: Base.triggerEvent,
    _getBreakpoint: Base.getBreakpoint,
    _scrollWindowTo: Base.scrollWindowTo,
    _getSiblingBefore: Base.getSiblingBefore,
    _getSiblingAfter: Base.getSiblingAfter,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['els', 'visibleCounts', 'autoAdvance', 'isRange', 'autoClose', 'closeDelay', 'mins', 'maxes', 'values', 'daysDisabled', 'daysInfo', 'quickJump', 'calendarEl', 'viewRange', 'animate', 'animationDuration', 'showOnFocus', 'onChange'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      els: null,
      inputEls: null,
      calendarEl: null,
      calendarContentEl: null,
      calendarOverflowEl: null,
      visibleCounts: null,
      activeIndex: null,
      mins: null,
      minVisible: null,
      maxes: null,
      maxVisible: null,
      values: null,
      daysDisabled: null,
      daysInfo: null,
      viewRange: 'month',
      autoAdvance: true,
      autoClose: true,
      quickJump: false,
      closeDelay: 500,
      isRange: null,
      nextButtonEl: null,
      previousButtonEl: null,
      animate: true,
      animationDuration: 100,
      showOnFocus: false,
      onChange: null,
      _hoverStarts: null,
      _hoverEnds: null,
      _currentBreakpoint: null,
      _animationQueue: null,
      _currentContent: null,
      _datesToShow: null,
      _monthsShowing: null,
      _atMin: false,
      _atMax: false,
      _inputsChanging: false,
      _openCloseTimer: null,
      _renderCache: null,
      _onClickBound: null,
      _onCalendarMouseOverBound: null,
      _onCalendarMouseOutBound: null,
      _onCalendarClickBound: null,
      _onSelectChangeBound: null,
      _onInputChangeBound: null,
      _onInputFocusBound: null,
      _onResizeBound: null
    },


    /**
     * Open the calendar widget.
     * @param {Number} index Optional The index of the element to tie actions to
     * @param {Object} params Optional
     */
    open: function(index, params) {

      this._unfillToggle();

      // Allow for only params to be passed
      if (typeof index === 'object') {
        params = index;
        index = 0;
      }

      index = index || 0;
      params = params || {};

      // Already open
      if (this.activeIndex === index) {
        return;
      }

      // Clear any timers which may be running to close or open this
      clearTimeout(this._openCloseTimer);

      // Set the new index
      this.activeIndex = index;

      // If we don't have a popover yet, create it.
      if (!this.popover) {
        this._createPopover();
      }

      // Get the list of dates to show for this view.
      this._initDatesToShow();

      // Ensure the cached values are in line with the input values.
      this.updateValues();

      // Create a calendar or calendars inside the popover
      this.render({
        clearCache: true
      });

      // Finally, show the popover after a delay or on the next tick.
      this._openCloseTimer = setTimeout(function() {
        this.popover.open({
          anchorTo: this.els[index].querySelector('.spark-date__calendar-toggle')
        });
      }.bind(this), params.delay || 0);

      // Activate the corresponding element
      this._activateElement(index);
    },


    /**
     * Close the calendar widget.
     * @param {Object} params
     */
    close: function(params) {

      this._unfillToggle();

      params = params || {};

      // Can't close if we aren't or haven't yet opened.
      if (this.activeIndex === null || !this.popover) {
        return;
      }

      // Deactive the corresponding element
      this._deactivateElement(this.activeIndex);

      // Clear any existing close or open events which are queued
      clearTimeout(this._openCloseTimer);

      // Close after a delay, or on the next tick
      this.activeIndex = null;
      this._openCloseTimer = setTimeout(function() {
        this.popover.close();
      }.bind(this), params.delay || 0);
    },


    /**
     * Get the value.
     * @param {Number|Element} index Optional
     * @return {Mixed}
     */
    getValue: function(index) {
      return this.values[index || 0];
    },


    /**
     * Set the date for a given element.
     * @param {String|Object} value
     * @param {Number|Element} index
     * @param {Boolean} skipRangeCheck Optional Don't check for sequential range values.
     */
    setValue: function(value, index, skipRangeCheck) {

      var obj = typeof value === 'object' ? value : parsedDomFormat.getValues(value);
      var el;

      // If we are passed an element instead of an index, use that.
      if (index instanceof HTMLElement) {
        el = index;
        index = this.inputEls.indexOf(el);
      }
      // Otherwise, find the element in our list.
      else {
        index = index || 0;
        el = this.inputEls[index];
      }

      // Update the stored value
      this.values[index] = obj;

      // If we are working with a range, make sure that the values are in order.
      if (value) {
        this._checkValues(index, skipRangeCheck);
      }

      // Set the value on the input element. Make sure we don't get into an infinite
      // loop since we listen to the change event on the input. We have to trigger it
      // so that other components listening to it get the update.
      el.value = value && obj === value ? parsedDomFormat.getString(value) : value || '';
      this._inputsChanging = this._inputsChanging || [];
      if (this._inputsChanging.indexOf(index) === -1) {
        this._inputsChanging.push(index);
        this._triggerEvent(el, 'change');
        (this.onChange || noop)(el, el.value, this);
        var pos = this._inputsChanging.indexOf(index);
        this._inputsChanging.splice(pos, 1);
      }

      this.render({
        clearCache: true
      });
    },


    /**
     * Take the date values from the inputs and set them as dates on the calendar.
     * @param {Number} index Optional A single index to update
     */
    updateValues: function(index) {

      this.values = this.values || [];

      var els = index === undefined ? this.inputEls : [this.inputEls[index]];
      var i = index || 0;
      var len = (index || 0) + els.length;

      for (; i < len; i++) {
        this.values[i] = this.inputEls[i].value ? parsedDomFormat.getValues(this.inputEls[i].value) : null;
      }
    },


    /**
     * Render the calendar or calendars into the popover.
     * @param {Object} params
     */
    render: function(params) {

      params = params || {};

      // If we don't have a popover yet, create it.
      if (!this.popover) {
        this._createPopover();
      }

      var content;

      // Clear the cache so that we don't show out-of-date values.
      if (params.clearCache) {
        this._renderCache = {};
      }

      // Create the visible days, weeks, months or years
      if (this.viewRange === 'year')
        content = this._renderYears();
      else if (this.viewRange === 'week')
        content = this._renderWeeks();
      else if (this.viewRange === 'day')
        content = this._renderDays();
      else
        content = this._renderMonths();

      this._insertContent(content, params);

      // Update attributes
      this._updateAttributes();
    },


    /**
     * Cleans up event listeners and removes helpers.
     */
    remove: function() {
      this.popover.remove();
      Base.remove.apply(this, arguments);
    },


    /**
     * Move to the next set of dates.
     */
    next: function() {
      if (this._atMax) return;
      this._setDatesToShow(1);
      this.render({
        append: 1
      });
    },


    /**
     * Move to the previous set of dates.
     */
    previous: function() {
      if (this._atMin) return;
      this._setDatesToShow(-1);
      this.render({
        prepend: 1
      });
    },


    /**
     * Show a specific date on the calendar.
     * @param {Object} params
     */
    showDate: function(params) {

      var month = params.month || null;
      var year = params.year || null;
      var day = params.day || null;

      // Nothing to do.
      if (!month && !year && !day) {
        return;
      }

      var showing = this._datesToShow[this.activeIndex].clone();
      var noun = this.viewRange.charAt(0).toUpperCase() + this.viewRange.slice(1);

      // Only change the values we were given
      if (year) {
        showing.year = year;
      }
      if (month) {
        showing.month = month;
      }
      if (day) {
        showing.day = day;
      }

      // Check min
      var min = this.mins[this.activeIndex];
      if (min && showing['before' + noun](min)) {
        showing = min.clone();
      }

      // Check max
      var max = this.maxes[this.activeIndex];
      if (max && showing['after' + noun](max)) {
        showing = max.clone();
      }
      this._datesToShow[this.activeIndex] = showing;

      // Populate the rest of the dates
      this._setDatesToShow();

      this.render();
    },


    /**
     * Render the appropriate number of years.
     * @return {Array}
     */
    _renderYears: function() {
      throw new Error('Displaying years in the calendar is not yet supported!');
    },


    /**
     * Render the appropriate number of weeks.
     * @return {Array}
     */
    _renderWeeks: function() {
      throw new Error('Displaying weeks in the calendar is not yet supported!');
    },


    /**
     * Render the appropriate number of days.
     * @return {Array}
     */
    _renderDays: function() {
      throw new Error('Displaying days in the calendar is not yet supported!');
    },


    /**
     * Render the appropriate number of months.
     * @return {Array}
     */
    _renderMonths: function() {

      var months = [];
      var i = 0;
      var len = this._datesToShow.length;
      var current = dateHelper.now();

      for (; i < len; i++) {
        months.push(this._renderMonth(this._datesToShow[i], current));
      }

      return months;
    },


    /**
     * Render a month.
     * @param {Object} date
     * @param {Object} current
     * @return {Element}
     */
    _renderMonth: function(date, current) {

      // A unique key for this month used for caching
      var key = date.year + '-' + date.month;

      // Ensure we have a cache
      this._renderCache = this._renderCache || {};

      // Return a cached instance
      if (this._renderCache[key]) {
        return this._renderCache[key];
      }

      // Create the element
      var el = document.createElement('div');
      var html = '';
      el.className = 'spark-calendar__month ' + this._getMonthClassNames(date);

      // Add the title or dropdowns.
      html += '<div class="spark-calendar__month-title">';
      if (this.quickJump && this.visibleCounts[this.activeIndex] === 1) {
        html += this._renderMonthQuickJump(date);
      } else {
        html += date.monthName + ' ' + date.year;
      }
      html += '</div>';

      // Add the days of the week headings
      html += this._renderMonthDaysOfWeek();

      // Add the days
      html += this._renderMonthDays(date, current);

      el.innerHTML = html;

      // Store in the cache and return
      return (this._renderCache[key] = el);
    },


    /**
     * Render the select inputs used for jumping ahead/backward several months.
     * @param {Object} date
     */
    _renderMonthQuickJump: function(date) {

      var el = document.createElement('div');
      var years;
      var months;

      // Min and max year to show
      var min = this.mins[this.activeIndex] ? this.mins[this.activeIndex].clone() : null;
      var max = this.maxes[this.activeIndex] ? this.maxes[this.activeIndex].clone() : null;

      // No min, go 50 years back
      if (!min) {
        min = date.clone();
        min.year = min.year - 50;
      }

      // No max, go 50 years forward
      if (!max) {
        max = date.clone();
        max.year = max.year + 50;
      }

      // Get the months to show.
      var monthOpts = dateHelper.getMonthNames().map(function(val, i) {

        // Don't show a month that is out of the valid range.
        if (min.year === max.year && (i + 1 < min.month || i + 1 > max.month)) {
          return null;
        }

        return {
          value: i + 1,
          text: val
        };
      }).filter(function(i) {
        return i;
      });


      // Min and max are the same month, so just show text.
      if (!monthOpts.length || monthOpts.length === 1) {
        months = document.createElement('span');
        months.innerHTML = date.monthName + ' ';
      } else {

        // Build a select list of months
        months = document.createElement('label');
        this._addClass(months, 'spark-select spark-calendar__select');
        months.appendChild(buildSelect({
          selected: date.month,
          attributes: {
            class: 'spark-select__input',
            name: 'month'
          },
          options: monthOpts
        }));
      }

      // Min and max year are the same, so just show text.
      if (min.year === max.year) {
        years = document.createElement('span');
        years.innerHTML = date.year;
      } else {

        // Build a select list of years
        years = document.createElement('label');
        this._addClass(years, 'spark-select spark-calendar__select');
        years.appendChild(buildSelect({
          selected: date.year,
          attributes: {
            class: 'spark-select__input',
            name: 'year'
          },
          options: Base.range(max.year, min.year - 1)
        }));
      }

      this._appendChildren(el, [months, years]);

      return el.innerHTML;
    },


    /**
     * Render the days of week row for a month calendar.
     * @return {String}
     */
    _renderMonthDaysOfWeek: function() {

      var daysOfWeek = dateHelper.getDayNames();
      var i = 0;
      var len = daysOfWeek.length;
      var str = '<div class="spark-calendar__days-of-week">';

      for (; i < len; i++) {
        str += '<span class="spark-calendar__day-of-week">' + daysOfWeek[i][0] + '</span>';
      }

      return str += '</div>';
    },


    /**
     * Render the days of week row for a month calendar.
     * @param {Object} date
     * @param {Object} current
     * @return {String}
     */
    _renderMonthDays: function(date, current) {

      var dayOfWeek = date.monthStart.dayOfWeek;
      var startOfWeek = dayOfWeek > 1 ? date.monthStart.weekStart : null;
      var monthEnd = date.monthEnd;
      var weeks = 6;
      var i = 0;
      var j = 0;
      var str = '<div class="spark-calendar__days">';
      var day = 0;
      var month = startOfWeek ? startOfWeek.month : date.month;
      var year = startOfWeek ? startOfWeek.year : date.year;
      var isCurrentMonth = current.year === date.year && current.month === date.month;

      // If we have days that come before the first of the month, the days will start as
      // inactive. We use a 1 here to indicate the date is inactive and _before_
      // the start of the month.
      var inactive = startOfWeek ? 1 : null;

      for (; i < weeks; i++) {
        for (; j < 7; j++) {

          day = startOfWeek ? startOfWeek.day + j : day + 1;
          str += this._renderMonthDay(day, month, year, isCurrentMonth, current, inactive);

          if (startOfWeek && j + 1 >= dayOfWeek - 1) {
            startOfWeek = null;
            inactive = null;
            day = 0;
            month = date.month;
            year = date.year;
            isCurrentMonth = date.month ===  month && date.year === year && inactive === null;
          } else if (day >= monthEnd.day) {

            // Inactive days that come _after_ the last of the month are denoted with a 2.
            // _renderMonthDay handles 1s and 2s differently, but both add the inactive class.
            inactive = 2;
            isCurrentMonth = false;

            day = 0;
            month++;

            if (month > 12) {
              year++;
              month = 1;
            }
          }
        }
        j = 0;
      }

      return str += '</div>';
    },


    /**
     * Render a day of the month.
     * @param {Number} day
     * @param {Number} month
     * @param {Number} year
     * @param {Boolean} isCurrentMonth
     * @param {Object} current
     * @param {Boolean} inactive
     */
    _renderMonthDay: function(day, month, year, isCurrentMonth, current, inactive) {

      var date = dateHelper.create({
        year: year,
        month: month,
        day: day
      });
      var str = '<a';
      var disabled = inactive || this._isDayDisabled(date);

      str += !disabled ? ' data-date="' + parsedDomFormat.getString(date) + '"' : '';
      str += inactive ? ' data-direction="' + (inactive === 2 ? 'next' : 'previous') + '"' : '';
      str += ' class="spark-calendar__day';
      str += (isCurrentMonth && current.day === day && current.month === month ? ' spark-calendar__day--today' : '');
      str += (inactive ? ' spark-calendar__day--inactive' : '');
      str += (disabled ? ' spark-calendar__day--disabled' : '');
      str += (this._isDaySelected(date) ? ' spark-calendar__day--selected' : '');
      str += (this._isDayRangeStart(date) ? ' spark-calendar__range-start' : '');
      str += (this._isDayRangeMiddle(date) ? ' spark-calendar__range-middle' : '');
      str += (this._isDayRangeEnd(date) ? ' spark-calendar__range-end' : '');
      str += (this._isDayRangeLast(date) ? ' spark-calendar__range-last' : '');
      str += '"';
      str += ' href="#"><span>';
      str += day;
      str += this._getDayInfo(date);
      str += '</span></a>';

      return str;
    },


    /**
     * Render the children into the content.
     * @param {Array} content
     */
    _insertContent: function(content, params) {

      this._currentContent = this._currentContent || [];

      params = params || {};

      var keep;
      var i = 0;
      var len;

      // If we've been asked to prepend or append, add the new elements in front,
      // save the number of old elements we're sliding out, animate,
      // then clean up.
      if (this.animate) {

        if (params.prepend) {

          // Get the last elements to keep
          keep = this._currentContent.slice(-params.prepend);
          len = keep.length;
          for (; i < len; i++) {
            content.push(keep[i]);
          }

          this.calendarEl.setAttribute('data-prepend-count', len);

          this._animateContent(function() {

            this.calendarEl.removeAttribute('data-prepend-count');

            this._animateContent(function() {
              content = content.slice(0, -params.prepend);
              this._insertContent(content);
            }, this.animationDuration);
          }, 10);
        }
        // Same but appending
        else if (params.append) {

          // Get the first elements to keep
          keep = this._currentContent.slice(0, params.append);
          len = keep.length;
          for (; i < len; i++) {
            content.unshift(keep[i]);
          }

          this.calendarEl.setAttribute('data-append-count', len);

          this._animateContent(function() {

            this._addClass(this.calendarEl, 'no-animate');
            content = content.slice(params.append);
            this._insertContent(content);
            this.calendarEl.removeAttribute('data-append-count');

            this._animateContent(function() {
              this._removeClass(this.calendarEl, 'no-animate');
            }, 10);
          }, this.animationDuration);
        }
      }

      this._appendChildren(this.calendarContentEl, this._currentContent = content, true);
    },


    /**
     * Is a given day selected?
     * @param {Object} date
     * @return {Boolean}
     */
    _isDaySelected: function(date) {
      return this.values && date.equal(this.values);
    },


    /**
     * Is a given day the start of a range?
     * @param {Object} date
     * @return {Boolean}
     */
    _isDayRangeStart: function(date) {
      return this.inputEls.length > 1 && this.values && date.equal(this.values[0]);
    },


    /**
     * Is a given day the middle of a range?
     * @param {Object} date
     * @return {Boolean}
     */
    _isDayRangeMiddle: function(date) {
      return this.inputEls.length > 1 && this.values && this.values.length > 1 && date.equal(this.values.slice(1, -1));
    },


    /**
     * Is a given day the end of a range?
     * @param {Object} date
     * @return {Boolean}
     */
    _isDayRangeEnd: function(date) {
      return this.inputEls.length > 1 && this.values && this.values.length > 1 && date.equal(this.values[this.values.length - 1]);
    },


    /**
     * Is a given day currently the last
     * @param {Object} date
     * @return {Boolean}
     */
    _isDayRangeLast: function(date) {

      var i = this.values.length - 1;

      for (; i > 0; i--) {

        // We have a value and it's not the same as the date.
        if (this.values[i]) {
          if (!date.equal(this.values[i])) {
            return false;
          }
          break;
        }
      }

      return true;
    },


    /**
     * Is a given day disabled?
     * @param {Object} date
     * @return {Boolean}
     */
    _isDayDisabled: function(date) {
      return (this.daysDisabled &&
          this.daysDisabled[date.year] &&
          this.daysDisabled[date.year][date.month] &&
          this.daysDisabled[date.year][date.month].indexOf(date.day) !== -1) ||
        (this.mins[this.activeIndex] && date.before(this.mins[this.activeIndex], true)) ||
        (this.maxes[this.activeIndex] && date.after(this.maxes[this.activeIndex], true));
    },


    /**
     * Get any "info" for a given day.
     * @param {Object} date
     * @return {String}
     */
    _getDayInfo: function(date) {
      return this.daysInfo && this.daysInfo[date.year] && this.daysInfo[date.year][date.month] && this.daysInfo[date.year][date.month][date.day] ? '<span class="spark-calendar__day-note">' + this.daysInfo[date.year][date.month][date.day] + '</span>' : '';
    },


    /**
     * Get the class names for a month.
     * @param {Object} date
     * @return {String}
     */
    _getMonthClassNames: function(date) {

      var cls = [];

      // Do we have a value in this month?
      if (date.equalMonth(this.values)) {
        cls.push('has-value');
      }

      // Does this month have the start, middle or end of a range?
      if (this.isRange && this.els.length > 1) {

        var start = date.equalMonth(this.values[0]);
        var end = date.equalMonth(this.values[this.values.length - 1]);
        var middle = date.equalMonth(this.values.slice(1, -1));
        var valBefore = date.afterMonth(this.values);
        var valAfter = date.beforeMonth(this.values);
        var afterEnd = this.values[this.values.length - 1] && date.after(this.values[this.values.length - 1]);

        if (start) {
          cls.push('range-start');
        }
        if (end) {
          cls.push('range-end');
        }
        if (middle) {
          cls.push('range-middle');
        }
        if (valBefore) {
          cls.push('value-before');
        }
        if (valAfter) {
          cls.push('value-after');
        }
        if (afterEnd) {
          cls.push('after-range-end');
        }
      }

      return cls.join(' ');
    },


    /**
     * Create the default input element.
     * @param {String} type
     * @return {Element}
     */
    _createDefaultElement: function() {
      var el = document.createElement('span');
      el.className = 'spark-input spark-date';
      el.innerHTML = '<input class="spark-input__field" type="date"><span class="spark-label"></span>';
      return el;
    },


    /**
     * Create the calendar.
     */
    _createCalendar: function() {
      var el = document.createElement('div');
      this._addClass(el, 'spark-calendar');
      el.innerHTML = '<nav class="spark-calendar__nav"><button class="spark-calendar__previous spark-icon-chevron-left" title="Previous"></button><button class="spark-calendar__next spark-icon-chevron-right" title="Next"></button></nav><div class="spark-calendar__overflow"><div class="spark-calendar__content"></div></div>';
      this.calendarEl = el;
      this.calendarContentEl = el.querySelector('.spark-calendar__content');
      this.calendarOverflowEl = el.querySelector('.spark-calendar_overflow');
    },


    /**
     * Create the popover.
     */
    _createPopover: function() {

      this.popover = new Popover(document.createElement('div'), {
        direction: 'bottom',
        contentEl: this.calendarEl,
        onClose: this._onPopoverClose.bind(this),
        onOpen: this._onPopoverOpen.bind(this)
      });
    },


    /**
     * Cache elements.
     */
    _cacheElements: function() {

      this.inputEls = [];

      var i = 0;
      var len = this.els.length;
      var input;

      for (; i < len; i++) {
        input = this.els[i].nodeName.toLowerCase() === 'input' ? this.els[i] : this.els[i].querySelector('input[type="date"]');
        this.inputEls[i] = input;
      }
    },


    /**
     * Cache elements specific to the calendar.
     */
    _cacheCalendarElements: function() {
      this.nextButtonEl = this.calendarEl.querySelector('.spark-calendar__next');
      this.previousButtonEl = this.calendarEl.querySelector('.spark-calendar__previous');
    },


    /**
     * Parse parameters from the elements.
     */
    _parseParams: function() {
      this._parseInputElsParams();
    },


    /**
     * Parse the min, max, value and visible counts from the elements if we can.
     * @return {Number|Boolean}
     */
    _parseInputElsParams: function() {

      var els = this.inputEls;

      if (!els) {
        return;
      }

      var i = 0;
      var len = els.length;
      var mins = [];
      var maxes = [];
      var values = [];

      for (; i < len; i++) {

        if (!els[i]) {
          continue;
        }

        if (this.mins && this.mins[i])
          mins[i] = typeof this.mins[i] === 'object' ? this.mins[i] : parsedDomFormat.getValues(this.mins[i]);
        else if (els[i].getAttribute('min'))
          mins[i] = parsedDomFormat.getValues(els[i].getAttribute('min'));

        if (this.maxes && this.maxes[i])
          maxes[i] = typeof this.maxes[i] === 'object' ? this.maxes[i] : parsedDomFormat.getValues(this.maxes[i]);
        else if (els[i].getAttribute('max'))
          maxes[i] = parsedDomFormat.getValues(els[i].getAttribute('max'));

        if (this.values && this.values[i])
          values[i] = typeof this.values[i] === 'object' ? this.values[i] : parsedDomFormat.getValues(this.values[i]);
        else if (els[i].value)
          values[i] = parsedDomFormat.getValues(els[i].value);
      }

      this.isRange = this.isRange !== null ? this.isRange : (this.els.length > 1 ? true : false);
      this.mins = mins;
      this.minVisible = dateHelper.earliest(mins);
      this.maxes = maxes;
      this.maxVisible = dateHelper.latest(maxes);
      this.values = values;
    },


    /**
     * Get the dates we should be showing. Start with the first value or today's date.
     */
    _initDatesToShow: function() {

      var arr = [];
      var i = 0;
      var len = this.values.length;

      // Get the date for the current index
      if (this.values[this.activeIndex]) {
        arr.push(this.values[this.activeIndex].clone());
      }

      // Get the first date
      for (; i < len && !arr.length; i++) {
        if (this.values[i]) {
          arr.push(this.values[i].clone());
        }
      }

      // If we didn't get a date, use the min
      for (i = 0, len = this.mins.length; i < len && !arr.length; i++) {
        if (this.mins[i]) {
          arr.push(this.mins[i].clone());
        }
      }

      // If we still didn't get a date, add today's date.
      if (!arr.length) {
        arr.push(dateHelper.now());
      }

      this._datesToShow = arr;

      this._setDatesToShow();
    },


    /**
     * Set the dates to show.
     * @param {Number} change Optional The direction to change
     */
    _setDatesToShow: function(change) {

      var arr = this._datesToShow;
      var visibleCount = this._currentBreakpoint === 'xs' ? 1 : (this.visibleCounts[this.activeIndex] || this.visibleCounts[0] || 1);
      var i = 0;
      var noun = this.viewRange.charAt(0).toUpperCase() + this.viewRange.slice(1);

      // If we have more dates to show than we're supposed to show, remove the excess.
      // This happens when visible count changes between renders
      if (arr.length > visibleCount) {
        arr.splice(visibleCount);
      }

      // If we are incrementing or decrementing the starting month
      if (change) {

        // Empty the rest of the dates from the array
        arr = this._datesToShow = arr.splice(0, 1);

        // Decrement
        if (change < 0) {
          for (; i > change; i--) {
            arr[0] = arr[0]['previous' + noun];
          }
        } else {
          for (; i < change; i++) {
            arr[0] = arr[0]['next' + noun];
          }
        }
      }

      // Check bounds
      this._atMin = this._checkMinDateVisible(noun, arr);
      this._atMax = this._checkMaxDateVisible(noun, arr);

      var addDate;
      var lastSelected;
      var action;
      i = 0;

      // Add additional dates
      while (arr.length < visibleCount) {

        // If at the max, prepend
        // If we're in the last input and it has a value and it's a different month than the first value, prepend
        // If at the min, append
        // If an item we're adding is before the min, discard it and append an item after the last in the arr
        // If an item we're adding is after the max, discard it and prepend an item before the first in the arr

        lastSelected = !change &&
          this.activeIndex === this.values.length - 1 &&
          this.values[this.values.length - 1] &&
          this.values[0] &&
          !this.values[this.values.length - 1]['equal' + noun](this.values, true);

        // If we're showing the max date or we have the last input selected and it has a value, add dates before.
        if (this._atMax || lastSelected) {

          addDate = arr[i]['previous' + noun];
          action = 'unshift';

          if (addDate['before' + noun](this.minVisible)) {
            addDate = arr[i]['next' + noun];
            action = 'push';
          }
        } else {

          addDate = arr[i]['next' + noun];
          action = 'push';

          if (addDate['after' + noun](this.maxVisible)) {
            addDate = arr[i]['previous' + noun];
            action = 'unshift';
          }
        }

        if (action === 'push') {
          arr.push(addDate);
          i = arr.length - 1;
        } else {
          arr.unshift(addDate);
          i = 0;
        }
      }

      // Check bounds again
      this._atMin = this._checkMinDateVisible(noun, arr);
      this._atMax = this._checkMaxDateVisible(noun, arr);

      // Update the navigation to reflect the _atMin or _atMax state
      this._updateNav();
    },


    /**
     * Check for the min value in an array of values.
     * @param {Object} min
     * @param {String} noun The type of date to check
     * @param {Array} arr
     * @return {Boolean}
     */
    _checkMinDateVisible: function(noun, arr) {

      if (!this.minVisible) {
        return;
      }

      var min = this.minVisible;

      if (min && (arr[0]['equal' + noun](min) || arr[0]['before' + noun](min))) {
        arr[0] = min;
        return true;
      }

      return false;
    },


    /**
     * Check for the max value in an array of values.
     * @param {Object} max
     * @param {String} noun The type of date to check
     * @param {Array} arr
     * @return {Boolean}
     */
    _checkMaxDateVisible: function(noun, arr) {

      if (!this.maxVisible) {
        return;
      }

      var max = this.maxVisible;

      if (max && arr[arr.length - 1] && (arr[arr.length - 1]['equal' + noun](max) || arr[arr.length - 1]['after' + noun](max))) {
        arr[arr.length - 1] = max;
        return true;
      }

      return false;
    },


    /**
     * Check that the values are in bounds and, optinoally, in sequential order.
     * If checking for sequence, remove those which aren't.
     * @param {Number} setIndex The index of the value most recently set. This shouldn't be removed.
     * @return {Boolean} Did any values change?
     */
    _checkValues: function(setIndex, skipRangeCheck) {

      // Check boundaries
      var changed = this._checkMinMaxValues();

      // Sequential range items check
      if (this.isRange && !skipRangeCheck) {

        var i = this.values.length - 1;

        for (; i >= 0; i--) {
          if (this.values[i] && this.values[i - 1] && this.values[i].beforeDay(this.values[i - 1], true)) {
            if (i === setIndex) {
              this.setValue(null, i - 1, true);
            } else {
              this.setValue(null, i, true);
            }
            changed = true;
          }
        }
      }

      return changed;
    },


    /**
     * Check minimum/maximum values.
     * @return {Boolean}
     */
    _checkMinMaxValues: function() {

      var i = 0;
      var len = this.values.length;
      var changed = false;

      for (; i < len; i++) {
        if ((this.maxes && this.maxes[i] && this.values[i] && this.values[i].after(this.maxes[i], true))) {
          this.values[i] = this.maxes[i].clone();
          changed = true;
        } else if ((this.mins && this.mins[i] && this.values[i] && this.values[i].before(this.mins[i], true))) {
          this.values[i] = this.mins[i].clone();
          changed = true;
        }
      }

      return changed;
    },


    /**
     * Enqueue animations to be run. (Not really animations since those happen w/ CSS. More
     * of a manager of timeouts).
     * @param {Function} cb
     * @param {Number} duration
     */
    _animateContent: function(cb, duration) {
      this._animationQueue = this._animationQueue || [];
      this._animationQueue.push({
        cb: cb,
        d: duration
      });
      this._runAnimation();
    },


    /**
     * Run the first queued animation. When complete, run the next animation.
     * @return {[type]} [description]
     */
    _runAnimation: function() {

      if (!this._animationTimer) {

        var a = this._animationQueue.shift();

        if (a) {

          this._animationTimer = setTimeout(function() {
            a.cb.call(this);
            this._animationTimer = null;
            this._runAnimation();
          }.bind(this), a.d);
        }
      }
    },


    /**
     * Update attributes on the element and its children.
     */
    _updateAttributes: function() {
      this.calendarEl.setAttribute('data-visible-count', this._currentBreakpoint === 'xs' ? 1 : (this.visibleCounts[this.activeIndex] || this.visibleCounts[0] || 1));
      this._updateNav();
    },


    /**
     * Update the navigation to reflect the ability to move forward and backward.
     */
    _updateNav: function() {

      if (this.previousButtonEl) {
        if (this._atMin)
          this.previousButtonEl.setAttribute('disabled', true);
        else
          this.previousButtonEl.removeAttribute('disabled');
      }

      if (this.nextButtonEl) {
        if (this._atMax)
          this.nextButtonEl.setAttribute('disabled', true);
        else
          this.nextButtonEl.removeAttribute('disabled');
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {

      this._onInputChangeBound = this._onInputChange.bind(this);
      this._onSelectChangeBound = this._onSelectChange.bind(this);
      this._onInputFocusBound = this._onInputFocus.bind(this);
      this._onClickBound = this._onClick.bind(this);
      this._onCalendarClickBound = this._onCalendarClick.bind(this);

      this._onCalendarMouseOverBound = this._onCalendarMouseOver.bind(this);
      this._onCalendarMouseOutBound = this._onCalendarMouseOut.bind(this);

      this._onResizeBound = this._onResize.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {

      var i = 0;
      var len = this.els.length;

      for (; i < len; i++) {
        this.els[i].addEventListener('click', this._onClickBound);
        this.inputEls[i].addEventListener('change', this._onInputChangeBound);
        this.inputEls[i].addEventListener('focus', this._onInputFocusBound);
      }

      this.calendarEl.addEventListener('mouseover', this._onCalendarMouseOverBound);
      this.calendarEl.addEventListener('mouseout', this._onCalendarMouseOutBound);
      this.calendarEl.addEventListener('click', this._onCalendarClickBound);

      this.calendarEl.addEventListener('change', this._onSelectChangeBound);

      window.addEventListener('resize', this._onResizeBound);
      window.addEventListener('orientationchange', this._onResizeBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {

      var i = 0;
      var len = this.els.length;

      for (; i < len; i++) {
        this.els[i].removeEventListener('click', this._onClickBound);
        this.inputEls[i].removeEventListener('change', this._onInputChangeBound);
        this.inputEls[i].removeEventListener('focus', this._onInputFocusBound);
      }

      this.calendarEl.removeEventListener('mouseover', this._onCalendarMouseOverBound);
      this.calendarEl.removeEventListener('mouseout', this._onCalendarMouseOutBound);
      this.calendarEl.removeEventListener('click', this._onCalendarClickBound);

      this.calendarEl.removeEventListener('change', this._onSelectChangeBound);

      window.removeEventListener('resize', this._onResizeBound);
      window.removeEventListener('orientationchange', this._onResizeBound);
    },


    /**
     * Focus on the next input if there is one.
     */
    _focusNext: function() {

      var index;

      // If we don't have any null values, don't do anything.
      if ((index = this.values.indexOf(null)) === -1) {
        return;
      }

      if (this.activeIndex < this.els.length - 1) {
        this.open(this.activeIndex + 1);
      } else {
        this.open(index);
      }
    },


    /**
     * Close if we're on the last input and we have values.
     */
    _closeOnLast: function() {
      if (this.values.indexOf(null) === -1) {
        this.close({
          delay: this.closeDelay
        });
      }
    },


    /**
     * Make the toggle icon a fill icon.
     */
    _fillToggle: function() {

      var toggle;

      if (this.els[this.activeIndex] && (toggle = this.els[this.activeIndex].querySelector('.spark-date__calendar-toggle [class*="spark-icon"]'))) {
        this._addClass(toggle, 'spark-icon--fill');
      }
    },


    /**
     * Make the toggle icon a line icon.
     */
    _unfillToggle: function() {

      var toggle;

      if (this.els[this.activeIndex] && (toggle = this.els[this.activeIndex].querySelector('.spark-date__calendar-toggle [class*="spark-icon"]'))) {
        this._removeClass(toggle, 'spark-icon--fill');
      }
    },


    /**
     * Activate the element associated with the calendar.
     * @param {Number} index
     */
    _activateElement: function(index) {
      var el = this.els[index];
      if (el) this._addClass(el, 'active');
    },


    /**
     * Deactivate the element associated with the calendar.
     * @param {Number} index
     */
    _deactivateElement: function(index) {
      var el = this.els[index];
      if (el && !this.values[index] && !this._hasClass(el, 'focus') && !this._hasClass(el, 'has-partial-value')) {
        this._removeClass(el, 'active');
      }
    },


    /**
     * Check the size of the popover and see if we should be showing the XS treatment.
     */
    _checkSize: function() {

      var bp = this._getBreakpoint(window.outerWidth);

      // Don't do anything if the breakpoint hasn't changed.
      if (this._currentBreakpoint === bp) return;

      // Store the breakpoint
      this._currentBreakpoint = bp;

      // Re-render the date range
      this._setDatesToShow();
      this.render(true);
      this.popover.open();
    },


    /**
     * Scroll the active input element into view.
     */
    _scrollToInput: function() {
      var el = this.els[this.activeIndex];
      if (el) this._scrollWindowTo(el);
    },


    /**
     * Update the hover classes.
     * @param {Element} hoverStarts
     * @param {Element} hoverEnds
     */
    _updateHoverClasses: function(hoverStarts, hoverEnds) {

      this._hoverStarts = this._hoverStarts || [];
      this._hoverEnds = this._hoverEnds || [];

      hoverStarts = hoverStarts instanceof Array ? hoverStarts : (hoverStarts ? [hoverStarts] : []);
      hoverEnds = hoverEnds instanceof Array ? hoverEnds : (hoverEnds ? [hoverEnds] : []);

      var allStarts = [];
      var curStarts = [];
      var newStarts = [];
      var allEnds = [];
      var curEnds = [];
      var newEnds = [];

      hoverStarts.forEach(function(el) {

        var index = this._hoverStarts.indexOf(el);

        // Already hovered.
        if (index !== -1) {
          curStarts.push(el);
        }
        // Not already hovered. Ready to add the class.
        else if (el) {
          newStarts.push(el);
          this._addClass(el, 'hover-start');
        }
      }, this);

      allStarts = Array.prototype.concat.call([], curStarts, newStarts);

      this._hoverStarts.forEach(function(el) {
        if (allStarts.indexOf(el) === -1) {
          this._removeClass(el, 'hover-start');
        }
      }, this);


      hoverEnds.forEach(function(el) {

        var index = this._hoverEnds.indexOf(el);

        // Already hovered.
        if (index !== -1) {
          curEnds.push(el);
        }
        // Not already hovered. Ready to add the class.
        else if (el) {
          newEnds.push(el);
          this._addClass(el, 'hover-end');
        }
      }, this);

      allEnds = Array.prototype.concat.call([], curEnds, newEnds);

      this._hoverEnds.forEach(function(el) {
        if (allEnds.indexOf(el) === -1) {
          this._removeClass(el, 'hover-end');
        }
      }, this);

      this._hoverStarts = allStarts;
      this._hoverEnds = allEnds;
    },


    /**
     * When an element is clicked, if the toggle was the target, open the popover.
     * @param {Object} e
     */
    _onClick: function(e) {

      if (this._getElementMatchingParent(e.target, '.spark-date__calendar-toggle', this.els)) {

        var el = this._getElementMatchingParent(e.target, '.spark-date', this.els);
        var index = this.els.indexOf(el);

        e.preventDefault();

        // Open on the next tick. Otherwise we also receive the window click close event.
        this.open(index, {
          delay: 1
        });
      }
    },


    /**
     * When the calendar is hovered, do some highlighting if we're showing a range.
     * @param {Object} e
     */
    _onCalendarMouseOver: function(e) {
      if (this.viewRange === 'month')
        this._onCalendarMouseOverMonths(e.target);
    },


    /**
     * Set hover states for days.
     * @param  {Element} target
     */
    _onCalendarMouseOverMonths: function(target) {

      if (!this.isRange) {
        return;
      }

      var day = this._getElementMatchingParent(target, '.spark-calendar__day');

      if (!day) {
        this._updateHoverClasses();
        return;
      }

      var month = this._getElementMatchingParent(target, '.spark-calendar__month');
      var mHasClass = function(c) {
        return this._hasClass(month, c);
      }.bind(this);
      var newStart = [];
      var newEnd = [];
      var daySel;

      // Only do highlights if we don't already have a value for this index.
      if (!this.values[this.activeIndex]) {

        // A month with a value before it but no value of its own, hover starts
        // from the first day to the hovered day.
        if (mHasClass('value-before') && !mHasClass('after-range-end') && !mHasClass('has-value') && !mHasClass('value-after')) {

          newStart.push(month.querySelector('.spark-calendar__day:not(.spark-calendar__day--inactive):not(.spark-calendar__day--disabled)'));
          newEnd.push(day);

          // Add a hover range to a previous month.
          var prevMonth = month;
          while ((prevMonth = this._getSiblingBefore(prevMonth, '.spark-calendar__month')) && !this._hasClass(prevMonth, 'value-after') && (this._hasClass(prevMonth, 'has-value') || this._hasClass(prevMonth, 'value-before'))) {

            daySel = prevMonth.querySelectorAll('.spark-calendar__day--selected');
            daySel = daySel[daySel.length - 1];
            daySel = daySel || prevMonth.querySelector('.spark-calendar__day');

            if (daySel) {
              newStart.push(daySel);
              daySel = prevMonth.querySelectorAll('.spark-calendar__day');
              daySel = daySel[daySel.length - 1];
              newEnd.push(daySel);
            }
          }
        }
        // A month with a value, highlight either from the hovered day to
        // the selection or from the selection to the day.
        else if (mHasClass('has-value') && !mHasClass('value-before') && !mHasClass('value-after') && ((newStart = this._getSiblingBefore(day, '.spark-calendar__day--selected')) || (newEnd = this._getSiblingAfter(day, '.spark-calendar__day--selected')))) {

          if (newStart) {
            newStart = [newStart];
            newEnd = [day];
          } else {
            newEnd = [newEnd];
            newStart = [day];
          }

          if (this._hasClass(newEnd[0], 'spark-calendar__range-start') || this._hasClass(newStart[0], 'spark-calendar__range-end')) {
            newStart = [];
            newEnd = [];
          }
        }
        // A month with a middle range should highlight before
        else if (mHasClass('has-value') && mHasClass('range-middle') && mHasClass('value-before')) {
          daySel = this._getSiblingBefore(day, '.spark-calendar__day--selected');
          if (daySel) {
            newStart = [daySel];
            newEnd = [day];
          }
        }
      }

      this._updateHoverClasses(newStart, newEnd);
    },


    /**
     * When the calendar is hovered, undo some highlighting if we're showing a range.
     * @param {Object} e
     */
    _onCalendarMouseOut: function(e) {
      if (this.viewRange === 'month')
        this._onCalendarMouseOutMonths(e.target);
    },


    /**
     * Clear hover states for days.
     * @param  {Element} target
     */
    _onCalendarMouseOutMonths: function(target) {

      var day = this._getElementMatchingParent(target, '.spark-calendar__day');
      var days = this._getElementMatchingParent(target, '.spark-calendar__days');
      var month = this._getElementMatchingParent(target, '.spark-calendar__month');

      this._removeClass(day, 'hover');
      this._removeClass(days, 'hover');
      this._removeClass(month, 'hover');
      this._removeClass(this.calendarEl, 'hover');
    },


    /**
     * When the calendar is clicked, handle navigation clicks and date selections.
     * @param {Object} e
     */
    _onCalendarClick: function(e) {

      var nav = this._getElementMatchingParent(e.target, '.spark-calendar__nav', this.calendarEl);
      var day;
      var dir;

      // Navigation clicks
      if (nav) {

        // Previous
        if (this._getElementMatchingParent(e.target, '.spark-calendar__previous', nav)) {
          this.previous();
        }
        // Next
        else if (this._getElementMatchingParent(e.target, '.spark-calendar__next', nav)) {
          this.next();
        }
      }
      // Day click
      else if ((day = this._getElementMatchingParent(e.target, '.spark-calendar__day'))) {

        e.preventDefault();
        e.stopPropagation();

        // Disabled day moves can move us to the next month
        if (this._hasClass(day, 'spark-calendar__day--disabled')) {
          dir = day.getAttribute('data-direction');
          if (dir === 'next') this.next();
          else if (dir === 'previous') this.previous();
        }
        // Enabled day sets the value
        else {
          this.setValue(day.getAttribute('data-date'), this.activeIndex);
          if (this.autoAdvance) this._focusNext();
          if (this.autoClose) this._closeOnLast();
        }
      }
    },


    /**
     * When the popover opens, set the toggle state.
     */
    _onPopoverOpen: function() {
      this._fillToggle();
      this._checkSize();
      if (this._currentBreakpoint === 'xs') this._scrollToInput();
    },


    /**
     * When the popover closes, reset the active state.
     */
    _onPopoverClose: function() {
      this._unfillToggle();
      this._updateHoverClasses();
      this._deactivateElement(this.activeIndex);
      this.activeIndex = null;
    },


    /**
     * When the input that corresponds to this instance changes. Allows us to listen
     * and respond to changes made by other components (Calendar Popover, for example).
     * @param {Object} e
     */
    _onInputChange: function(e) {

      var index = this.inputEls.indexOf(e.target);

      if (this._inputsChanging && this._inputsChanging.indexOf(index) !== -1)
        return;

      this.setValue(e.target.value, index);
    },


    /**
     * When an input receives focus, if we are supposed to automatically show
     * on focus do so.
     * @param {Object} e
     */
    _onInputFocus: function(e) {
      if (!this.showOnFocus) return;
      var index = this.inputEls.indexOf(e.target);
      if (index !== -1) this.open(index);
    },


    /**
     * When one of the calendar quick jump select inputs changes.
     * We have to set the value of the select input back to its original
     * value or else it will be out of sync when the currently active
     * month is shown again.
     * @param {Object} e
     */
    _onSelectChange: function(e) {

      var name = e.target.name;
      var val = e.target.value;
      var curVal = this._datesToShow[this.activeIndex][name];
      var obj = {};

      obj[name] = val;
      this.showDate(obj);
      e.target.value = curVal;
    },


    /**
     * When the window resizes, determine if we're at the XS breakpoint so we
     * can do some mobile-esque stuff!
     * @param {Object} e
     */
    _onResize: function() {
      if (this.activeIndex !== null) this._checkSize();
    }
  };

  Base.exportjQuery(CalendarPopover, 'CalendarPopover');

  return CalendarPopover;
}));

/**
 * # DateInput
 * A date input container.
 *
 * @example
 * new DateInput(el, {
 *   // Optional. Callback for when the input value changes.
 *   onChange: function(inputInstance, value) {}
 * });
 *
 * @module components/date-input.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/text-input', 'spark/helpers/date-typeahead', 'spark/helpers/date-select', 'spark/helpers/parse-date-format', 'spark/helpers/pad-number', 'spark/helpers/date'], factory);
  }
  else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../components/text-input'), require('../helpers/date-typeahead'), require('../helpers/date-select'), require('../helpers/parse-date-format'), require('../helpers/pad-number'), require('../helpers/date'));
  }
  else {
    root.Spark = root.Spark || {};
    root.Spark.DateInput = factory(root.Spark.BaseComponent, root.Spark.TextInput, root.Spark.DateTypeahead, root.Spark.DateSelect, root.Spark.parseDateFormat, root.Spark.padNumber, root.Spark.dateHelper);
  }
}(this, function(Base, TextInput, DateTypeahead, DateSelect, parseDateFormat, padNumber, date) {

  'use strict';

  var domDateFormat = 'YYYY-MM-DD';
  var parsedDomFormat = parseDateFormat(domDateFormat);
  var noop = function() {};

  /**
   * DateInput constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var DateInput = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._bindEventListenerCallbacks();
    this._setParams(params || {});
    this._parseParams();
    this._convertLabel();
    this._initializeInputs();
    this._updateClass();
    this._addEventListeners();
  };

  DateInput.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _removeClass: Base.removeClass,
    _addClass: Base.addClass,
    _hasClass: Base.hasClass,
    _getChildIndex: Base.getChildIndex,
    _appendChildren: Base.appendChildren,
    _triggerEvent: Base.triggerEvent,
    _copyAttributes: Base.copyAttributes,
    _getElementMatchingParent: Base.getElementMatchingParent,


    /**
     * Inherit functionality from TextInput.
     */
    setError: TextInput.prototype.setError,
    clearError: TextInput.prototype.clearError,
    setWarning: TextInput.prototype.setWarning,
    clearWarning: TextInput.prototype.clearWarning,
    setSuccess: TextInput.prototype.setSuccess,
    clearSuccess: TextInput.prototype.clearSuccess,
    clearMessages: TextInput.prototype.clearMessages,
    setMessage: TextInput.prototype.setMessage,
    _showMessage: TextInput.prototype._showMessage,
    _hideMessage: TextInput.prototype._hideMessage,
    _isMessageVisible: TextInput.prototype._isMessageVisible,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['onChange', 'onFocus', 'onBlur', 'isTypeahead', 'isSelect', 'format', 'textFormat', 'showDateAsText'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      inputEl: null,
      messageEl: null,
      toggleEl: null,
      inFocus: null,
      isActive: null,
      isSelect: null,
      isTypeahead: null,
      typeaheads: null,
      typeaheadEls: null,
      selects: null,
      selectEls: null,
      format: null,
      parsedFormat: null,
      showDateAsText: null,
      runningTypeaheads: false,
      textFormat: null,
      onChange: null,
      onFocus: null,
      onBlur: null,
      _hasFocus: false,
      _pauseInputChange: false,
      _onClickBound: null,
      _onPieceChangeBound: null,
      _onTypeaheadFocusBound: null,
      _onTypeaheadBlurBound: null,
      _onTypeaheadBackspaceBound: null,
      _onTypeaheadEndBound: null,
      _onInputChangeBound: null
    },


    /**
     * Show the input by adding the active state and setting character counts (if necessary).
     */
    show: function() {

      if (!this.isActive) {
        this._runTypeaheads();
        this.isActive = true;
        this._updateClass();
      }
    },


    /**
     * Hide the input by removing the active state.
     */
    hide: function() {
      this.isActive = false;
      this._updateClass();
    },


    /**
     * Augment default remove call w/ helper cleanup.
     */
    remove: function() {

      Base.remove.apply(this, arguments);

      // Remove text input
      this.textInput.remove();
      delete this.textInput;

      // Remove typeaheads
      if (this.typeaheads) {
        for (var i in this.typeaheads) {
          this.typeaheads[i].remove();
        }
      }

      // Remove select inputs
      if (this.selectEls) {
        this.selectEls.forEach(this._removeSelectPiece.bind(this));
      }
    },


    /**
     * Given an object with day, month and year, set the value of the input.
     * @param {Object} values
     */
    setValue: function(values) {

      values = values || {
        day: '',
        month: '',
        year: ''
      };
      var i;
      var hadValue;

      for (i in this.typeaheads) {
        if (values[i] !== undefined) {
          this.typeaheads[i].setValue(values[i]);
          hadValue = hadValue || (values[i] ? true : false);
        }
      }

      for (i in this.selects) {
        if (values[i] !== undefined) {
          this.selects[i].setValue(values[i]);
          hadValue = hadValue || (values[i] ? true : false);
        }
      }

      if (!this.isActive && hadValue) {
        this.isActive = true;
      }

      this._padTypeaheads();
      this._updateClass();
      this.updateInput();
    },


    /**
     * Validate the date values.
     */
    validate: function() {

      if (this.isTypeahead) {
        this._validateTypeaheads();
      }
      else if (this.isSelect) {
        this._validateSelects();
      }
    },


    /**
     * Update the input values to match the typeaheads.
     */
    updateInput: function() {

      var inputs;

      if (this.isTypeahead && this.typeaheads) {
        inputs = this.typeaheads;
      }
      else if (this.isSelect) {
        inputs = this.selects;
      }

      if (inputs) {

        var day = (inputs.day && inputs.day.getValue(true)) || 0;
        var month = (inputs.month && inputs.month.getValue(true)) || 0;
        var year = (inputs.year && inputs.year.getValue(true)) || 0;

        var val = this.inputEl.value;

        this.inputEl.value = [day, month, year].indexOf(0) === -1 ? padNumber(year, 4) + '-' + padNumber(month, 2) + '-' + padNumber(day, 2) : '';

        if (val !== this.inputEl.value) {
          this._pauseInputChange = true;
          this._triggerEvent(this.inputEl, 'change');
          (this.onChange || noop)(this.inputEl.value, this.inputEl);
          this._pauseInputChange = false;
        }

      }
    },


    /**
     * Move the focus to a typeahead element.
     * @todo : prepend the character if it's passed
     * @param {Number} i
     * @param {String} character Optional A character to add
     */
    focus: function(i, character) {

      if (!this.isActive || !this.inFocus) {
        return;
      }

      var index = this.typeaheadEls.indexOf(this.inFocus.typeahead.el);
      var sib = this.typeaheadEls[index + i];
      var typeahead;

      // If we were passed a character to prepend, find the typeahead for this element
      if (character) {
        typeahead = this._getTypeaheadByElement(sib);
        if (typeahead) {
          typeahead.typeahead.addCharacterAtIndex(character, 0);
        }
      }

      if (!sib) {
        return false;
      }

      var sibInput = sib.querySelector('input');

      if (sibInput) {
        sibInput.focus();

        // If we have a typeahead (because we needed to prepend a character), move the caret.
        if (typeahead) {
          typeahead.typeahead.moveCaret(1);
        }
      }

      return true;
    },


    /**
     * Move the focus to the next element.
     * @param {String} character Optional A character to add
     */
    focusNext: function(character) {
      if (this.focus(1, character)) {
        if (this.inFocus && !character)
          this.inFocus.typeahead.moveCaretToStart();
      }
    },


    /**
     * Move the focus to the next element.
     * @param {String} character Optional A character to add
     */
    focusPrevious: function(character) {
      if (this.focus(-1, character)) {
        if (this.inFocus)
          this.inFocus.typeahead.moveCaretToEnd();
      }
    },


    /**
     * Do we have any values?
     * @return {Boolean}
     */
    hasPartialValue: function() {

      var i;

      for (i in this.typeaheads) {
        if (this.typeaheads[i].getValue()) {
          return true;
        }
      }

      for (i in this.selects) {
        if (this.selects[i].getValue()) {
          return true;
        }
      }

      return false;
    },


    /**
     * Store a reference to the needed elements.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;

      this.inputEl = this.el.querySelector('[type="date"]');

      if (!this.inputEl) {
        throw new Error('No <input type="date"> element present in date input container!', this.el);
      }

      this.toggleEl = this.el.querySelector('.spark-date__toggle');

      this.messageEl = this.el.querySelector('.spark-input__message') || document.createElement('span');
    },


    /**
     * Parse parameters from the elements.
     */
    _parseParams: function() {

      this.isActive = this.isActive === null ? (this.inputEl.value ? true : false) : this.isActive;
      this.isSelect = this.isSelect === null ? (this._hasClass(this.el, 'spark-date--select') ? true : false) : this.isSelect;
      this.isTypeahead = this.isTypeahead === null ? (!this.isSelect ? true : false) : this.isTypeahead;
      this.format = this.format === null ? (this.inputEl.getAttribute('data-format') ? this.inputEl.getAttribute('data-format') : 'MM-DD-YYYY') : this.format;
      this.textFormat = this.textFormat === null ? (this.inputEl.getAttribute('data-text-format') ? this.inputEl.getAttribute('data-text-format') : 'MM DD YYYY') : this.textFormat;
      this.showDateAsText = this.showDateAsText === null ? this.inputEl.getAttribute('data-show-date-as-text') !== null : this.showDateAsText;

      this.parsedFormat = parseDateFormat(this.format);
      this.parsedTextFormat = parseDateFormat(this.textFormat);
      this.min = this.inputEl.getAttribute('min') && parsedDomFormat.getValues(this.inputEl.getAttribute('min'));
      this.max = this.inputEl.getAttribute('max') && parsedDomFormat.getValues(this.inputEl.getAttribute('max'));
    },


    /**
     * Setup the proper inputs. This could mean creating a typeahead, or creating selects.
     */
    _initializeInputs: function() {

      // @todo: remove this when Android fixes its keyup/keypress/keydown bug
      // http://stackoverflow.com/questions/17139039/keycode-is-always-zero-in-chrome-for-android
      if (this.isTypeahead) {
        this._initializeInputPieces();
        this._runTypeaheads();
      }
      else if (this.isSelect) {
        this._removeClass(this.el, 'spark-input');
        this._initializeInputPieces();
      }
    },


    /**
     * Replace the date input with a group of typeaheads or select inputs.
     * Keep the date input around and store the typeahead data in there in an ISO date format.
     */
    _initializeInputPieces: function() {

      // Hide the original element. This will be updated as the typeahead values change
      this.inputEl.style.display = 'none';

      var els = [];
      var label;

      // Create a new typeahead for each part of the parsed format. Also add placeholder elements.
      this.parsedFormat.parts.forEach(function(part) {
        // Something weird with Node that makes us have to specify what `this` is here.
        (this.isTypeahead ? this._initializeTypeaheadPiece : this._initializeSelectPiece).call(this, els, part);
      }.bind(this));

      // Create a holder for all the pieces
      var piecesEl = document.createElement('span');
      piecesEl.className = this.isTypeahead ? 'spark-input__fields' : 'spark-select-group';

      // Add all the necessary elements
      this._appendChildren(piecesEl, els);

      // If this is a select group, move the label element.
      if (this.isSelect && (label = this.el.querySelector('.spark-label'))) {
        piecesEl.appendChild(label);
      }

      // Add the pieces holder
      this.el.insertBefore(piecesEl, this.inputEl);

      // Set the value
      if (this.inputEl.value) {
        this.setValue(parsedDomFormat.getValues(this.inputEl.value));
        this.isActive = true;
      }
    },


    /**
     * Create a typeahead or placeholder piece.
     * @param {Array} els
     * @param {Object} part
     */
    _initializeTypeaheadPiece: function(els, part) {

      this.typeaheads = this.typeaheads || {};
      this.typeaheadEls = this.typeaheadEls || [];

      var el;

      switch (part.name) {
        case 'day':
        case 'month':
        case 'year':
          this.typeaheads[part.name] = new DateTypeahead({
            type: part.name,
            length: part.length,
            placeholder: part.value,
            onFocus: this._onTypeaheadFocusBound,
            onBlur: this._onTypeaheadBlurBound,
            onChange: this._onPieceChangeBound,
            onBackspace: this._onTypeaheadBackspaceBound,
            onEnd: this._onTypeaheadEndBound
          });
          el = this.typeaheads[part.name].typeahead.el;
          this.typeaheadEls.push(el);
          break;
        default:
          el = document.createElement('span');
          el.innerHTML = part.value;
          el.className = 'spark-input__divider';
          break;
      }

      els.push(el);
    },


    /**
     * Replace the date input with three date dropdowns. Keep the date input around and store the
     * select data in there.
     */
    _initializeSelectPiece: function(els, part) {

      this.selects = this.selects || {};
      this.selectEls = this.selectEls || [];

      if (['day', 'month', 'year'].indexOf(part.name) === -1) {
        return;
      }

      var el;

      this.selects[part.name] = new DateSelect({
        type: part.name,
        onChange: this._onPieceChangeBound
      });
      el = this.selects[part.name].select.el;

      els.push(el);
      this.selectEls.push(el);
    },


    /**
     * If our element is a label, convert it to a div so that
     * we are semantically correct. Can't have more than one
     * input inside of a label!
     */
    _convertLabel: function() {

      if (this.isTypeahead || this.el.nodeName.toLowerCase() !== 'label') {
        return;
      }

      var newEl = document.createElement('fieldset');

      this._copyAttributes(this.el, newEl);
      this._appendChildren(newEl, this.el.children);

      if (this.el.parentNode) {
        this.el.parentNode.replaceChild(newEl, this.el);
      }

      this.el = newEl;
    },


    /**
     * Validate the typeahead values.
     */
    _validateTypeaheads: function() {

      if (!this.typeaheads) {
        return;
      }

      var month = this.typeaheads.month ? this.typeaheads.month.getValue(true) : null;
      var year = this.typeaheads.year ? this.typeaheads.year.getValue(true) : null;

      var maxDay = (month && new Date(year !== null ? year : new Date().getFullYear(), month, 0).getDate()) || this._getMaxDaysInMonth(month);
      var day = this.typeaheads.day ? this.typeaheads.day.getValue(true) : null;

      if (maxDay < day) {
        this.typeaheads.day.setValue(maxDay);
        this.updateInput();
      }
    },


    /**
     * Validate the boundaries of the typeahead values relative to the min and max values.
     */
    _validateTypeaheadBounds: function() {

      var year = this.typeaheads.year ? this.typeaheads.year.getValue(true) : null;
      var month = this.typeaheads.month ? this.typeaheads.month.getValue(true) : null;
      var day = this.typeaheads.day ? this.typeaheads.day.getValue(true) : null;

      if (!year || !month || !day) {
        return;
      }

      var date = new Date(year, month - 1, day);
      var set = '';

      if (this.min && date < new Date(this.min.year, this.min.month - 1, this.min.day)) {
        set = 'min';
      }
      else if (this.max && date > new Date(this.max.year, this.max.month - 1, this.max.day)) {
        set = 'max';
      }

      if (set) {
        this.typeaheads.year.setValue(padNumber(this[set].year, this.typeaheads.year.typeahead.format.length));
        this.typeaheads.month.setValue(padNumber(this[set].month, this.typeaheads.month.typeahead.format.length));
        this.typeaheads.day.setValue(padNumber(this[set].day, this.typeaheads.day.typeahead.format.length));
        this.updateInput();
      }
    },


    /**
     * Pad the typeahead input values.
     */
    _padTypeaheads: function() {

      if (this._pauseInputChange) return;

      this._pauseInputChange = true;

      for (var i in this.typeaheads) {
        this._padTypeahead(this.typeaheads[i]);
      }

      this._pauseInputChange = false;
    },

    /**
     * Pad the typeahead input values.
     * @param {Typeahead} typeahead
     */
    _padTypeahead: function(typeahead) {

      var value = typeahead.getValue();

      if (value) {
        var padded = padNumber(value, typeahead.typeahead.format.length);
        if (value !== padded) typeahead.setValue(padNumber(value, typeahead.typeahead.format.length));
      }
    },


    /**
     * Do any of the typeaheads have a value?
     * @return {Boolean}
     */
    _hasTypeaheadValue: function() {

      for (var i in this.typeaheads) {
        if (this.typeaheads[i].getValue(true)) {
          return true;
        }
      }

      return false;
    },


    /**
     * Validate select input values.
     */
    _validateSelects: function() {

      if (!this.selects) {
        return;
      }

      var month = this.selects.month ? this.selects.month.getValue(true) : null;
      var year = this.selects.year ? this.selects.year.getValue(true) : null;

      var maxDay = (month && new Date(year !== null ? year : new Date().getFullYear(), month, 0).getDate()) || this._getMaxDaysInMonth(month);
      var day = this.selects.day ? this.selects.day.getValue(true) : null;

      this.selects.day.setOptions({
        max: maxDay
      });
      if (maxDay < day) {
        this.selects.day.setValue(maxDay);
      }

      this.updateInput();
    },


    /**
     * Get the maximum number of days for a given month.
     * @param {Number} month The month's number. 1-12.
     * @return {Number} The maximum number of days. 28-31.
     */
    _getMaxDaysInMonth: function(month) {
      if (month === 2) return 29;
      else if ([4, 6, 9, 11].indexOf(month) !== -1) return 30;
      return 31;
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onTypeaheadFocusBound = this._onTypeaheadFocus.bind(this);
      this._onTypeaheadBlurBound = this._onTypeaheadBlur.bind(this);
      this._onPieceChangeBound = this._onPieceChange.bind(this);
      this._onTypeaheadBackspaceBound = this._onTypeaheadBackspace.bind(this);
      this._onTypeaheadEndBound = this._onTypeaheadEnd.bind(this);
      this._onInputChangeBound = this._onInputChange.bind(this);
      this._onVisibleChildrenBound = this._onVisibleChildren.bind(this);
    },


    /**
     * Add event listeners.
     */
    _addEventListeners: function() {
      this.el.addEventListener('click', this._onClickBound);
      this.inputEl.addEventListener('change', this._onInputChangeBound);
      document.addEventListener('spark.visible-children', this._onVisibleChildrenBound, true);
    },


    /**
     * Remove event listeners.
     */
    _removeEventListeners: function() {
      this.el.removeEventListener('click', this._onClickBound);
      this.inputEl.removeEventListener('change', this._onInputChangeBound);
      document.removeEventListener('spark.visible-children', this._onVisibleChildrenBound, true);
    },

    /**
     * Handle the spark.visible-children event
     * @param {Object} e
     */
    _onVisibleChildren: function(e) {
      if (e.target.contains(this.el)) {
        window.setTimeout(function() {
          this.update();
        }.bind(this), 0);
      }
    },

    /**
     * Resize the elements, to account for any changed display property
     */
    update: function() {
      this._runTypeaheads();
    },

    /**
     * Run all typeaheads so they have placeholder values.
     */
    _runTypeaheads: function() {

      // Make sure we don't get into an infinite loop. Even though the logic
      // in the typeaheads should be stopping this from happening, there is
      // something in Safari where the focus and blur events fire in different
      // order than other browsers so those failsafes do not work.
      if (this.runningTypeaheads) {
        return;
      }

      this.runningTypeaheads = true;

      if (this.inFocus) {
        this.inFocus.pause();
      }

      for (var i in this.typeaheads) {
        if (this.typeaheads[i] !== this.inFocus) {
          this.typeaheads[i].run();
        }
      }

      if (this.inFocus) {
        this.inFocus.resume();
        this.inFocus.run();
      }

      this.runningTypeaheads = false;
    },


    /**
     * Update the active and focus classes.
     */
    _updateClass: function() {
      this._toggleClass(this.el, 'active', this.isActive);
      this._toggleClass(this.el, 'has-partial-value', this.hasPartialValue());
      this._toggleClass(this.el, 'focus', this.inFocus ? true : false);
    },


    /**
     * Get the typeahead that corresponds to the given element.
     * @param {Element} el
     * @return {Object}
     */
    _getTypeaheadByElement: function(el) {
      for (var i in this.typeaheads) {
        if (this.typeaheads[i].typeahead.el === el) {
          return this.typeaheads[i];
        }
      }
    },


    /**
     * Show the date as text.
     */
    _showDateText: function() {

      var text = this._getDateText();

      if (!text || !this.showDateAsText) {
        return;
      }

      if (!this.dateTextEl) {
        this._createDateTextEl();
      }

      this.dateTextEl.innerHTML = text;
      this.dateTextEl.style.display = '';
    },


    /**
     * Hide the date as text.
     */
    _hideDateText: function() {

      if (!this.showDateAsText || !this.dateTextEl) {
        return;
      }

      this.dateTextEl.style.display = 'none';
    },


    /**
     * Create the date text element.
     */
    _createDateTextEl: function() {

      var el = document.createElement('div');
      el.className = 'spark-input__overlay';
      el.style.display = 'none';

      this.el.appendChild(el);
      this.dateTextEl = el;
    },


    /**
     * Get the date as text.
     */
    _getDateText: function() {

      var parts = this.parsedTextFormat.parts;
      var i = 0;
      var len = parts.length;
      var str = '';
      var isValid = true;
      var val;

      for(; i < len; i++) {

        val = this.typeaheads[parts[i].name] && this.typeaheads[parts[i].name].getValue();

        switch(parts[i].name) {
          case 'month':
            str += date.getMonthNameShort(val);
            if (!val) {
              isValid = false;
              break;
            }
            break;
          case 'day':
          case 'year':
            str += val;
            if (!val) {
              isValid = false;
              break;
            }
            break;
          default:
            str += parts[i].value;
            break;
        }
      }

      return (isValid ? str : false);
    },


    /**
     * When the value of a typeahead or select changes, validate.
     * @param {Number} val
     * @param {Object} typeahead
     */
    _onPieceChange: function() {

      this.validate();

      if (this.isTypeahead && this.showDateAsText && !this._hasFocus) {

        if (this._showTextTimer) {
          clearTimeout(this._showTextTimer);
        }

        this._showTextTimer = setTimeout(function() {
          this._showDateText();
        }.bind(this), 0);
      }
    },


    /**
     * When the typeahead gains focus.
     * @param {Number} val
     * @param {Object} typeahead
     */
    _onTypeaheadFocus: function(val, typeahead) {

      if (this.runningTypeaheads) return;

      this._hideDateText();

      if (!this._hasFocus) {
        this._hasFocus = true;
        (this.onFocus || noop)(this, this.inputEl.value);
      }

      this._triggerEvent(this.inputEl, 'focus');
      this.inFocus = typeahead;
      this.show();
      this._updateClass();

      if (this._blurTimer) {
        clearTimeout(this._blurTimer);
        this._blurTimer = null;
      }
    },


    /**
     * When the typeahead loses focus, make sure numbers are padded properly.
     * @param {Number} val
     * @param {Object} typeahead
     */
    _onTypeaheadBlur: function(val, typeahead) {

      if (this.runningTypeaheads) return;

      this.inFocus = null;

      this._padTypeahead(typeahead);
      this.updateInput();
      this._updateClass();

      if (!this.inputEl.value && !this._hasTypeaheadValue()) {
        this.hide();
      }
      else {
        this._validateTypeaheadBounds();
      }

      this._blurTimer = setTimeout(function() {
        this._hasFocus = false;
        (this.onBlur || noop)(this, this.inputEl.value);
        this._showDateText();
      }.bind(this), 1);
    },


    /**
     * When the typeahead fires a backspace event, move back to the previous input.
     * @param {Number} val
     * @param {Object} typeahead
     */
    _onTypeaheadBackspace: function() {
      this.focusPrevious();
    },


    /**
     * When the typeahead is at its maximum length and the caret is at the end,
     * focus on the next input field.
     * @param {Object} typeahead
     * @param {String} character Optional
     */
    _onTypeaheadEnd: function(typeahead, character) {
      this.focusNext(character);
    },


    /**
     * When the input that corresponds to this instance changes. Allows us to listen
     * and respond to changes made by other components (Calendar Popover, for example).
     * @param {Object} e
     */
    _onInputChange: function(e) {

      if (this.isTypeahead) {
        this.isActive = e.target.value ? true : false;
        this._updateClass();
      }

      if (this._pauseInputChange) return;
      this.setValue(parsedDomFormat.getValues(e.target.value));
      (this.onChange || noop)(this.inputEl.value, this.inputEl);
    },


    /**
     * When the input group is clicked, focus on the first typeahead
     * if we don't already have focus.
     * @todo : we should still replace the label on typeaheads and use this,
     * but we need a way to focus on the closest element to a click.
     * @param {Object} e
     */
    _onClick: function(/*e*/) {

      /*if (this.isTypeahead || this.inFocus || this.isActive || this._getElementMatchingParent(e.target, '.spark-input__addon')) {
        return;
      }

      var input = this.typeaheadEls[0].querySelector('input');

      if (input) {
        input.focus();
      }*/
    }
  };

  Base.exportjQuery(DateInput, 'DateInput');

  return DateInput;
}));

/**
 * # RangeSlider
 * A dual slider for number inputs.
 *
 * @example
 * new RangeSlider(el, {
 *   // Optional. Slide along the x or y-axis?
 *   isX: true,
 *   // onChange callback
 *   onChange: function(inst, index, value){},
 * });
 *
 * @module components/range-slider.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.RangeSlider = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * RangeSlider constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var RangeSlider = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._updateDisabledClasses();
  };

  RangeSlider.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _getElementOffset: Base.getElementOffset,
    _getNodeListIndex: Base.getNodeListIndex,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['isX', 'onChange', 'onWillChange'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      controlsEl: null,
      inputEls: null,
      handleEls: null,
      trackEl: null,
      trackFillEl: null,
      isActive: null,
      isX: true,
      onChange: null,
      onWillChange: null,
      position: 0,
      width: 0,
      height: 0,
      mins: null,
      maxes: null,
      steps: null,
      values: null,
      percentages: null,
      offsetLeft: 0,
      offsetTop: 0,
      handleSizePercentage: 0,
      currentIndex: null,
      lastIndex: null,
      _oldVal: null,
      _onTouchStartBound: null,
      _onTouchMoveBound: null,
      _onTouchEndBound: null,
      _onMouseDownBound: null,
      _onMouseMoveBound: null,
      _onMouseUpBound: null,
      _onMouseOutBound: null,
      _onFocusBound: null,
      _onKeydownBound: null,
      _onBlurBound: null,
      _onChangeBound: null,
      _onResizeBound: null,
      _onClickBound: null,
      _onVisibleChildrenBound: null
    },


    /**
     * Start the slider moving.
     * @param {Number} index The index of the handle or input element.
     * @param {Number} position The position of the pointer.
     * @param {String} type Optional Which type of events to listen for.
     */
    start: function(index, position, type) {

      // Noop if we're disabled or an invalid index was passed
      if (index < 0 || this.inputEls[index].getAttribute('disabled') !== null) {
        return;
      }

      this._addMoveEventListeners(type || 'mouse');
      this._cacheSize();

      this.isActive = this.isActive || [];
      this.isActive[index] = true;

      this.currentIndex = index;
      this._updateActiveClasses(index);
      this._updateDisabledClasses();
      this._oldVal = this.values[index];
      this.move(position);
    },


    /**
     * Move the value to a given position
     * @param {Number} position
     * @param {Boolean} force Force the move Optional
     */
    move: function(position, force) {

      // Noop if an invalid index was passed we haven't yet started dragging
      if ((!position || !this.isActive || !this.isActive[this.currentIndex]) && !force) {
        return;
      }

      // Treat positions beyond the boundaries as the boundaries
      if (this.isX) {

        // Too far left
        if (position < this.offsetLeft) {
          position = this.offsetLeft;
        }
        // Too far right
        else if (position > this.offsetLeft + this.width) {
          position = this.offsetLeft + this.width;
        }
      } else {

        // Too far top
        if (position < this.offsetTop) {
          position = this.offsetTop;
        }
        // Too far bottom
        else if (position > this.offsetTop + this.height) {
          position = this.offsetTop + this.height;
        }
      }

      // The percentage of the new position relative to slider-container width or height.
      var percentage = this.isX ? (position - this.offsetLeft) / (this.width - this.handleSize) : (position - this.offsetTop) / this.height;

      // The value of the input as a percentage of the value range.
      this.setValue(this.currentIndex, Math.round((percentage - this.handleSizePercentage / 2) * (this.highestMax - this.lowestMin)) + this.lowestMin);
    },


    /**
     * Stop listening to movements.
     * @param {Number} index Optional The index of the handle or input element.
     * @param {String} type Optional Which type of events to listen for.
     */
    stop: function(index, type) {

      if (index !== null && index !== undefined && this.currentIndex !== index) {
        return;
      }

      this.isActive[this.currentIndex] = false;
      this.lastIndex = this.currentIndex;

      if(this._oldVal !== this.values[this.currentIndex]) {
        (this.onChange || noop)(this, this.currentIndex, this.values[this.currentIndex]);
      }

      this.currentIndex = null;
      this._updateActiveClasses(index);
      this._removeMoveEventListeners(type || 'mouse');
    },


    /**
     * Set the value of the handle.
     * @param {Number} index The index of the input element.
     * @param {Number} value
     */
    setValue: function(index, value) {

      // We don't have an input element at that index, so something went wrong.
      if (!this.inputEls[index]) {
        throw new Error('Cannot set value on input element with an index of ' + index + '. That element does not exist.');
      }

      // Move in increments if we have a defined step size
      if (this.steps[index]) {
        value = value - (value % this.steps[index]);
      }

      this.values = this.values || [];

      // Check bounds of the new value
      if (value > this.maxes[index]) {
        value = this.maxes[index];
      } else if (value < this.mins[index]) {
        value = this.mins[index];
      }

      // If there is an input that comes after this, make sure we aren't going beyond it
      if (this.values[index + 1] !== undefined && value >= this.values[index + 1]) {
        value = this.values[index + 1] - (this.steps[index] || 1);
      }
      // If there is an input that comes before this, make sure we aren't going below it
      else if (this.values[index - 1] !== undefined && value <= this.values[index - 1]) {
        value = this.values[index - 1] + (this.steps[index] || 1);
      }

      // If there is an onWillChange callback, run it. If it returns
      // false, then this new value should be considered invalid.
      if (typeof this.onWillChange === 'function') {
        var change = this.onWillChange(this, index, value);
        if (typeof change === 'number') {
          value = change;
        }
      }

      // Store value
      this.values[index] = value;

      // Update elements
      this.inputEls[index].value = this.values[index];
      this.handleEls[index].setAttribute('data-value', this.values[index]);

      // Set the percentage
      this.percentages = this.percentages || [];
      this.percentages[index] = (this.values[index] - this.lowestMin) / (this.highestMax - this.lowestMin);

      // Update the position of the handle
      this._updateHandlePosition(index);
    },


    /**
     * Increment the value by the step size.
     * @param {Boolean} useMultiplier Optional Increment by a multiplied version of the step
     */
    increment: function(useMultiplier) {
      this.setValue(this.currentIndex, this.values[this.currentIndex] + this.steps[this.currentIndex] * (useMultiplier ? 10 : 1));
    },


    /**
     * Decrement the value by the step size.
     * @param {Boolean} useMultiplier Optional Increment by a multiplied version of the step
     */
    decrement: function(useMultiplier) {
      this.setValue(this.currentIndex, this.values[this.currentIndex] - this.steps[this.currentIndex] * (useMultiplier ? 10 : 1));
    },


    /**
     * Remove the element from the DOM and prepare for garbage collection by dereferencing values.
     * @param {Boolean} leaveElement Leave the element intact.
     */
    remove: function(leaveElement) {
      this._removeMoveEventListeners('touch');
      this._removeMoveEventListeners('mouse');
      this._removeMoveEventListeners('keyboard');
      Base.remove.call(this, leaveElement);
    },


    /**
     * Store a reference to the whole slider, as well as the
     * input element. Also, get some default values from the input
     * element (min, max, steps).
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.controlsEl = this.el.querySelector('.spark-slider__controls');
      this.inputEls = this.el.querySelectorAll('input[type="number"]');
      this.handleEls = this.el.querySelectorAll('.spark-slider__handle');
      this.trackEl = this.el.querySelector('.spark-slider__track');
      this.trackFillEl = this.trackEl.querySelector('.spark-slider__track-fill');

      if (!this.inputEls || this.inputEls.length <= 1) {
        throw new Error('Tried to create a slider instance without two number inputs.');
      }

      if (!this.handleEls || this.handleEls.length <= 1) {
        throw new Error('Tried to create a slider instance without two handle buttons.');
      }

      var lowestMin = Infinity;
      var highestMax = -Infinity;
      var i = 0;
      var len = this.inputEls.length;
      var values = [];
      this.mins = [];
      this.maxes = [];
      this.steps = [];

      // Cache the size of the element so that we can properly set values on handles.
      this._cacheSize();

      // Set the minimum and max values for each element. Also set any predefined value.
      for (; i < len; i++) {

        this.mins[i] = parseInt(this.inputEls[i].getAttribute('min'), 10) || null;
        this.maxes[i] = parseInt(this.inputEls[i].getAttribute('max'), 10) || null;
        this.steps[i] = parseInt(this.inputEls[i].getAttribute('step'), 10) || 1;

        if (this.mins[i] < lowestMin) {
          lowestMin = this.mins[i];
        }

        if (this.maxes[i] > highestMax) {
          highestMax = this.maxes[i];
        }
      }

      this.lowestMin = lowestMin;
      this.highestMax = highestMax;

      i = 0;

      // If we have a default value, set it.
      for (; i < len; i++) {

        values[i] = parseInt(this.inputEls[i].getAttribute('value'), 10);

        // It's a number
        if (!isNaN(values[i])) {
          this.setValue(i, values[i]);
        } else {

          // Set as the minimum unless this is the last handle.
          if (i + 1 === len) {
            this.setValue(i, this.maxes[i] !== null ? this.maxes[i] : 0);
          } else {
            this.setValue(i, this.mins[i] !== null ? this.mins[i] : 0);
          }
        }
      }
    },


    /**
     * Save the element dimensions.
     */
    _cacheSize: function() {

      this.width = this.trackEl.clientWidth;
      this.height = this.trackEl.clientHeight;

      this.handleSize = this.isX ? this.handleEls[0].clientWidth : this.handleEls[0].clientHeight;
      this.handleSizePercentage = this.isX ? this.handleEls[0].clientWidth / this.width : this.handleEls[0].clientHeight / this.height;

      var offset = this._getElementOffset(this.controlsEl);
      this.offsetLeft = offset.left;
      this.offsetTop = offset.top;
    },


    /**
     * Set the position of the handle.
     * @param {Number} index The index of the handle element to update.
     */
    _updateHandlePosition: function(index) {

      // Adjust the percentage of the total to be the percentage of the total less the size of the handle.
      var percentage = this.percentages[index] * (1 - this.handleSizePercentage) + (this.handleSizePercentage / 2);

      percentage = Math.round(Math.min(percentage, 1) * 100);

      this.handleEls[index].setAttribute('style', 'left: ' + percentage + '%;');

      var firstPercentage = this.percentages[0];
      var lastPercentage = this.percentages[this.percentages.length - 1];
      this.trackFillEl.setAttribute('style', 'width: ' + ((lastPercentage - firstPercentage) * 100) + '%; left: ' + (firstPercentage * 100) + '%;');
    },


    /**
     * Update the active class on the handle.
     * @param {Number} index The index of the handle element to update.
     */
    _updateActiveClasses: function(index) {

      this._toggleClass(this.handleEls, 'active', false);
      this._toggleClass(this.handleEls[index], 'active', this.isActive[index]);

      if (this.isActive.indexOf(true) !== -1) {
        this.el.setAttribute('data-active-index', this.isActive.indexOf(true));
      } else {
        this.el.removeAttribute('data-active-index');
      }
    },


    /**
     * Update which handles are disabled.
     */
    _updateDisabledClasses: function() {

      var disabledCount = 0;

      for (var i = 0, len = this.inputEls.length; i < len; i++) {

        if (this.inputEls[i].getAttribute('disabled') !== null) {
          this._toggleClass(this.handleEls[i], 'disabled', true);
          disabledCount++;
        } else {
          this._toggleClass(this.handleEls[i], 'disabled', false);
        }
      }

      this._toggleClass(this.el, 'all-disabled', disabledCount === this.handleEls.length);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {

      this._onTouchStartBound = this._onTouchStart.bind(this);
      this._onTouchMoveBound = this._onTouchMove.bind(this);
      this._onTouchEndBound = this._onTouchEnd.bind(this);

      this._onClickBound = this._onClick.bind(this);
      this._onMouseDownBound = this._onMouseDown.bind(this);
      this._onMouseMoveBound = this._onMouseMove.bind(this);
      this._onMouseUpBound = this._onMouseUp.bind(this);

      this._onFocusBound = this._onFocus.bind(this);
      this._onKeydownBound = this._onKeydown.bind(this);
      this._onBlurBound = this._onBlur.bind(this);

      this._onChangeBound = this._onChange.bind(this);

      this._onResizeBound = this._onResize.bind(this);

      this._onVisibleChildrenBound = this._onVisibleChildren.bind(this);
    },


    /**
     * Add event listeners for touchstart and mouse click.
     */
    _addEventListeners: function() {

      this.controlsEl.addEventListener('touchstart', this._onTouchStartBound);
      this.controlsEl.addEventListener('mousedown', this._onMouseDownBound);

      for (var i = 0, len = this.inputEls.length; i < len; i++) {
        this.inputEls[i].addEventListener('change', this._onChangeBound);
      }

      for (var j = 0, len2 = this.handleEls.length; j < len2; j++) {
        this.handleEls[j].addEventListener('focus', this._onFocusBound);
        this.handleEls[j].addEventListener('click', this._onClickBound);
      }
      document.addEventListener('spark.visible-children', this._onVisibleChildrenBound, true);
    },


    /**
     * Remove event listeners for touchstart and mouse click.
     */
    _removeEventListeners: function() {

      this.controlsEl.removeEventListener('touchstart', this._onTouchStartBound);
      this.controlsEl.removeEventListener('mousedown', this._onMouseDownBound);

      document.removeEventListener('spark.visible-children', this._onVisibleChildrenBound);

      for (var i = 0, len = this.inputEls.length; i < len; i++) {
        this.inputEls[i].removeEventListener('change', this._onChangeBound);
      }

      for (var j = 0, len2 = this.handleEls.length; i < len2; i++) {
        this.handleEls[j].removeEventListener('focus', this._onFocusBound);
        this.handleEls[j].removeEventListener('click', this._onClickBound);
      }
    },


    /**
     * Add event listeners for touchmove, touchend, mousemove and mouseup.
     * We add these to the window so that the user can move off of the element
     * but keep dragging the slider handle. Otherwise it's really hard to
     * use the slider unless it's massive.
     * @param {String} type Which type of listeners to add
     */
    _addMoveEventListeners: function(type) {

      // Only listen for events of the type we asked for.
      switch (type) {
        case 'mouse':
          window.addEventListener('mousemove', this._onMouseMoveBound);
          window.addEventListener('mouseout', this._onMouseOutBound);
          window.addEventListener('mouseup', this._onMouseUpBound);
          break;
        case 'touch':
          window.addEventListener('touchmove', this._onTouchMoveBound);
          window.addEventListener('touchend', this._onTouchEndBound);
          break;
        case 'keyboard':
          window.addEventListener('keydown', this._onKeydownBound);
          for (var i = 0, len = this.handleEls.length; i < len; i++) {
            this.handleEls[i].addEventListener('blur', this._onBlurBound);
          }
          break;
      }

      window.addEventListener('resize', this._onResizeBound);
      window.addEventListener('orientationchange', this._onResizeBound);
    },


    /**
     * Remove event listeners for move events.
     * @param {String} type Which type of listeners to add
     */
    _removeMoveEventListeners: function(type) {

      // Only unbind events of the type we asked for.
      switch (type) {
        case 'mouse':
          window.removeEventListener('mousemove', this._onMouseMoveBound);
          window.removeEventListener('mouseup', this._onMouseUpBound);
          break;
        case 'touch':
          window.removeEventListener('touchmove', this._onTouchMoveBound);
          window.removeEventListener('touchend', this._onTouchEndBound);
          break;
        case 'keyboard':
          window.removeEventListener('keydown', this._onKeydownBound);
          for (var i = 0, len = this.handleEls.length; i < len; i++) {
            this.handleEls[i].removeEventListener('blur', this._onBlurBound);
          }
          break;
      }

      window.removeEventListener('resize', this._onResizeBound);
      window.removeEventListener('orientationchange', this._onResizeBound);
    },


    /**
     * When the touch starts, start the slider.
     * @param {Object} e
     */
    _onTouchStart: function(e) {
      this.start(this._getNodeListIndex(this.handleEls, e.target), this.isX ? e.touches[0].pageX : e.touches[0].pageY, 'touch');
    },


    /**
     * When the window fires a touchmove event, adjust our value accordingly
     * @param {Object} e
     */
    _onTouchMove: function(e) {

      if (!this.isActive[this.currentIndex]) {
        return;
      }

      e.preventDefault();

      this.move(this.isX ? e.touches[0].pageX : e.touches[0].pageY);
    },


    /**
     * When the window fires a touchend event, stop tracking touches
     * @param {Object} e
     */
    _onTouchEnd: function(e) {

      if (!this.isActive[this.currentIndex]) {
        return;
      }

      e.preventDefault();

      this.stop(this._getNodeListIndex(this.handleEls, e.target), 'touch');
    },


    /**
     * When the mouse presses down, start the slider.
     * @param {Object} e
     */
    _onMouseDown: function(e) {
      this.start(this._getNodeListIndex(this.handleEls, e.target), this.isX ? e.pageX : e.pageY, 'mouse');
    },


    /**
     * When the window fires a mousemove event, adjust our value accordingly
     * @param {Object} e
     */
    _onMouseMove: function(e) {

      if (!this.isActive[this.currentIndex]) {
        return;
      }

      e.preventDefault();

      this.move(this.isX ? e.pageX : e.pageY);
    },


    /**
     * When the window fires a mouseup event, stop tracking
     * @param {Object} e
     */
    _onMouseUp: function() {

      if (!this.isActive[this.currentIndex]) {
        return;
      }

      this.stop(null, 'mouse');
    },

    /**
     * Handle the spark.visible-children event
     * @param {Object} e
     */
    _onVisibleChildren: function(e) {
      if(e.target.contains(this.el)) {
        window.setTimeout(function() {
          this._onResize();
        }.bind(this),0);
      }
    },


    /**
     * When the window resizes, cache size values for the slider.
     * @param {Object} e
     */
    _onResize: function() {
      this._cacheSize();
      this._updateDisabledClasses();
      for(var i = 0; i < this.handleEls.length; i++) {
        this._updateHandlePosition(i);
      }
    },


    /**
     * When the element receives focus, start listening for keyboard events
     * @param {Object} e
     */
    _onFocus: function(e) {
      this.start(this._getNodeListIndex(this.handleEls, e.target), null, 'keyboard');
    },


    /**
     * When a key is pressed, see if it's a left or right arrow and move the
     * handle accordingly. If the shift key is pressed, we'll increment and
     * decrement by bigger values.
     * @param {Object} e
     */
    _onKeydown: function(e) {

      if (this._getNodeListIndex(this.inputEls, e.target) !== -1) {
        return;
      }

      // Left for x or up for y
      var increment = (this.isX && e.keyCode === 39) || (!this.isX && e.keyCode === 38);

      // Right for x or down for y
      var decrement = (this.isX && e.keyCode === 37) || (!this.isX && e.keyCode === 40);

      if (increment) {
        this.increment(e.shiftKey);
      } else if (decrement) {
        this.decrement(e.shiftKey);
      }
    },


    /**
     * When the element loses focus, stop listening for keyboard events
     * @param {Object} e
     */
    _onBlur: function(e) {
      this.stop(this._getNodeListIndex(this.handleEls, e.target), 'keyboard');
    },


    /**
     * When the input value changes, set our interal value if it's not already our value.
     * @param {Object} e
     */
    _onChange: function(e) {

      var index = this._getNodeListIndex(this.inputEls, e.target);

      this._updateDisabledClasses();

      if (e.target.value !== this.values[index]) {
        this.setValue(index, e.target.value);
      }
      (this.onChange || noop)(this, index, this.values[index]);
    },


    /**
     * Prevent click events on the button. This way we don't accidentally submit the form.
     * @param {Object} e
     */
    _onClick: function(e) {
      e.preventDefault();
    }
  };

  Base.exportjQuery(RangeSlider, 'RangeSlider');

  return RangeSlider;
}));

/**
 * # ProgressIndicator
 *
 * @example
 * new ProgressIndicator(el, {
 *   // Optional. The precision of the progress percentage.
 *   precision: 10 // 10 decimal places
 * });
 *
 * @module components/progress-indicator.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.ProgressIndicator = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  var sizes = {
    large: {
      diameter: 264,
      stroke: 12,
      track: 'M132 6c69.588 0 126 56.412 126 126s-56.412 126-126 126S6 201.588 6 132 62.412 6 132 6z',
      fill: '',
      border: 'M132 252c-66.274 0-120-53.726-120-120S65.726 12 132 12s120 53.726 120 120-53.726 120-120 120z'
    },
    small: {
      diameter: 60,
      stroke: 6,
      track: 'M30 3c14.912 0 27 12.088 27 27S44.912 57 30 57 3 44.912 3 30 15.088 3 30 3z',
      fill: '',
      border: 'M30 54C16.745 54 6 43.255 6 30S16.745 6 30 6s24 10.745 24 24-10.745 24-24 24z'
    },
    extraSmall: {
      diameter: 24,
      stroke: 3,
      track: 'M22.5 12c0 5.8-4.7 10.5-10.5 10.5S1.5 17.8 1.5 12 6.2 1.5 12 1.5 22.5 6.2 22.5 12z',
      fill: '',
      border: ''
    }
  };



  /**
   * ProgressIndicator constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var ProgressIndicator = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  ProgressIndicator.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _hasClass: Base.hasClass,
    _round: Base.round,
    remove: Base.remove,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['precision'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      progressEl: null,
      statusEl: null,
      noteEls: null,
      meterEl: null,
      fillEl: null,
      meterHeight: 0,
      meterWidth: 0,
      notes: null,
      isDeterminate: false,
      value: null,
      precision: 0,
      lastDOMUpdateTime: 0,
      _onResizeBound: null
    },


    /**
     * Set the value of the indicator.
     * @param {Number} val
     */
    set: function(val) {

      if (val === this.value) {
        return;
      }

      if (val > 1) {
        val = 1;
      }

      this.value = val;

      if (this.isDeterminate && this.progressEl) {
        this.progressEl.setAttribute('value', Math.round(val * 100) / 100);
      }

      this._updateDOM();
    },


    /**
     * Store a reference to all the needed elements.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.progressEl = this.el.querySelector('progress');
      this.statusEl = this.el.querySelector('.spark-progress__value-status, [role="status"]');
      this.noteEl = this.el.querySelector('.spark-progress__states');
      this.meterEl = this.el.querySelector('.spark-progress__meter');

      this.isDeterminate = this.progressEl.getAttribute('value') !== null;
      this.size = this._determineSize();

      // If this is a determinate value, replace the meter with the SVG.
      if (this.isDeterminate) {

        var svg = this._buildSVG();
        svg.setAttribute('class', this.meterEl.className);

        this.meterEl.parentNode.replaceChild(svg, this.meterEl);
        this.meterEl = svg;
        this.fillEl = this.meterEl.querySelector('.spark-progress__fill');
      }

      if (this.noteEl) {
        this._parseNotes(this.noteEl);
      }

      if (this.progressEl) {
        this.value = this.progressEl.value;
      }

      this._cacheSize();

      this._updateDOM();
    },


    /**
     * Cache the size of the meter.
     */
    _cacheSize: function() {
      this.meterHeight = this.meterEl.clientHeight;
      this.meterWidth = this.meterEl.clientWidth;
    },


    /**
     * Determine the size of the indicator.
     * @return {String}
     */
    _determineSize: function() {

      if (this.el.className.indexOf('progress--sm') !== -1) {
        return 'small';
      } else if (this.el.className.indexOf('progress--xs') !== -1) {
        return 'extraSmall';
      }

      return 'large';
    },


    /**
     * Build the proper SVG element for this size indicator.
     * @return {Element}
     */
    _buildSVG: function() {
      var size = sizes[this.size];
      var template = '<svg viewBox="0 0 ' + size.diameter + ' ' + size.diameter + '" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="' + size.track + '" stroke-width="' + size.stroke + '" class="spark-progress__track"/><path d="' + (size.fill || size.track) + '" stroke-width="' + size.stroke + '" class="spark-progress__fill"/><path d="' + size.border + '" class="spark-progress__border"/></g></g></svg>';
      var div = document.createElement('div');
      div.innerHTML = template;
      return div.children[0];
    },


    /**
     * Take an unordered list of notes and determine the ranges for
     * when to show a given note.
     * @param  {Element} el
     * @return {Array}
     */
    _parseNotes: function(el) {

      this.notes = this.notes || [];

      var len = el.children.length;
      var i = len - 1;

      for (; i >= 0; i--) {
        this.notes.push({
          min: parseInt(el.children[i].getAttribute('data-value'), 10),
          max: el.children[i + 1] ? parseInt(el.children[i + 1].getAttribute('data-value'), 10) - 1 : 100,
          el: el.children[i]
        });
      }
    },


    /**
     * Update the text visible based on the value. Also adjust the SVG.
     */
    _updateDOM: function() {

      if (!this.isDeterminate) {
        return;
      }

      var updateTime = Date.now();
      var val = this._round(this.value * 100, this.precision);

      // Don't animate if we're animating back to 0 or it's been less than 150ms since our last update.
      var noAnimation = val === 0 || this.lastDOMUpdateTime + 150 > updateTime;
      this._toggleClass(this.fillEl, 'no-animation', noAnimation);

      this.statusEl.innerHTML = val + '%';

      var dashArray = (sizes[this.size].diameter - sizes[this.size].stroke) * Math.PI;
      var dashOffset = dashArray - (dashArray * (val / 100));

      this.fillEl.setAttribute('style', 'stroke-dasharray: ' + dashArray + '; stroke-dashoffset: ' + dashOffset);

      this.lastDOMUpdateTime = updateTime;

      if (!this.notes) {
        return;
      }

      var i = 0;
      var len = this.notes.length;

      for (; i < len; i++) {
        this._toggleClass(this.notes[i].el, 'active', (this.notes[i].min <= val && this.notes[i].max >= val));
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onResizeBound = this._onResize.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      window.addEventListener('resize', this._onResizeBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      window.removeEventListener('resize', this._onResizeBound);
    },


    /**
     * When the window resizes, cache the dimensions.
     * @param {Object} e
     */
    _onResize: function() {
      this._cacheSize();
    }
  };

  Base.exportjQuery(ProgressIndicator, 'ProgressIndicator');

  return ProgressIndicator;
}));

/**
 * # Slider
 * A slider for number inputs.
 *
 * @example
 * new Slider(el, {
 *   // Optional. Slide along the x or y-axis?
 *   isX: true,
 *   // onChange callback
 *   onChange: function(inst, val){},
 * });
 *
 * @module components/slider.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Slider = factory(root.Spark.BaseComponent);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * Slider constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Slider = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
  };

  Slider.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _getElementOffset: Base.getElementOffset,
    _elementHasParent: Base.elementHasParent,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['isX', 'onChange', 'onWillChange'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      controlsEl: null,
      inputEl: null,
      handleEl: null,
      trackEl: null,
      trackFillEl: null,
      isActive: false,
      isX: true,
      onChange: null,
      onWillChange: null,
      position: 0,
      width: 0,
      height: 0,
      min: 0,
      max: 0,
      step: 1,
      value: 0,
      percentage: 0,
      offsetLeft: 0,
      offsetTop: 0,
      handleSizePercentage: 0,
      truncatePrecision: 0,
      truncateCharacters: ['k', 'm', 'b', 't'],
      _oldVal: null,
      _onTouchStartBound: null,
      _onTouchMoveBound: null,
      _onTouchEndBound: null,
      _onMouseDownBound: null,
      _onMouseMoveBound: null,
      _onMouseUpBound: null,
      _onMouseOutBound: null,
      _onFocusBound: null,
      _onKeydownBound: null,
      _onBlurBound: null,
      _onChangeBound: null,
      _onResizeBound: null,
      _onClickBound: null
    },


    /**
     * Start the slider moving.
     * @param {Number} position The position of the pointer.
     * @param {String} type Optional Which type of events to listen for.
     */
    start: function(position, type) {

      if (this.inputEl.getAttribute('disabled') !== null) {
        return;
      }

      this._addMoveEventListeners(type || 'mouse');
      this._cacheSize();
      this.isActive = true;
      this._updateActiveClasses();
      this._oldVal = this.value;
      this.move(position);
    },


    /**
     * Move the value to a given position
     * @param {Number} position
     * @param {Boolean} force Force the move Optional
     */
    move: function(position, force) {

      // Noop if we haven't yet started dragging
      if ((!position || !this.isActive) && !force) {
        return;
      }

      // Treat positions beyond the boundaries as the boundaries
      if (this.isX) {

        // Too far left
        if (position < this.offsetLeft) {
          position = this.offsetLeft;
        }
        // Too far right
        else if (position > this.offsetLeft + this.width) {
          position = this.offsetLeft + this.width;
        }
      } else {

        // Too far top
        if (position < this.offsetTop) {
          position = this.offsetTop;
        }
        // Too far bottom
        else if (position > this.offsetTop + this.height) {
          position = this.offsetTop + this.height;
        }
      }

      // The percentage of the new position relative to slider-container width or height.
      var percentage = this.isX ? (position - this.offsetLeft) / (this.width - this.handleSize) : (position - this.offsetTop) / this.height;

      // The value of the input as a percentage of the value range.
      this.setValue(Math.round((percentage - this.handleSizePercentage / 2) * (this.max - this.min)) + this.min);
    },


    /**
     * Stop listening to movements.
     * @param {String} type Optional Which type of events to listen for.
     */
    stop: function(type) {
      this.isActive = false;
      if(this._oldVal !== this.value) {
        (this.onChange || noop)(this, this.value);
      }
      this._updateActiveClasses();
      this._removeMoveEventListeners(type || 'mouse');
    },


    /**
     * Set the value of the handle.
     * @param {Number} value
     */
    setValue: function(value) {

      // Move in increments if we have a defined step size
      if (this.step) {
        value = value - (value % this.step);
      }

      // Check bounds of the new value
      if (value > this.max) {
        value = this.max;
      } else if (value < this.min) {
        value = this.min;
      }

      // If there is an onWillChange callback, run it. If it returns
      // false, then this new value should be considered invalid.
      if (typeof this.onWillChange === 'function') {
        var change = this.onWillChange(this, value);
        if (typeof change === 'number') {
          value = change;
        }
      }

      // Store value
      this.value = value;

      // Update elements
      this.inputEl.value = this.value;
      this.handleEl.setAttribute('data-value', this._truncateValueText(this.value));

      // Set the percentage
      this.percentage = (this.value - this.min) / (this.max - this.min);

      // Update the position of the handle
      this._updateHandlePosition();
    },


    /**
     * Increment the value by the step size.
     * @param {Boolean} useMultiplier Optional Increment by a multiplied version of the step
     */
    increment: function(useMultiplier) {
      this.setValue(this.value + this.step * (useMultiplier ? 10 : 1));
    },


    /**
     * Decrement the value by the step size.
     * @param {Boolean} useMultiplier Optional Increment by a multiplied version of the step
     */
    decrement: function(useMultiplier) {
      this.setValue(this.value - this.step * (useMultiplier ? 10 : 1));
    },


    /**
     * Remove the element from the DOM and prepare for garbage collection by dereferencing values.
     * @param {Boolean} leaveElement Leave the element intact.
     */
    remove: function(leaveElement) {
      this._removeMoveEventListeners('touch');
      this._removeMoveEventListeners('mouse');
      this._removeMoveEventListeners('keyboard');
      Base.remove.call(this, leaveElement);
    },


    /**
     * Store a reference to the whole slider, as well as the
     * input element. Also, get some default values from the input
     * element (min, max, steps).
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.controlsEl = this.el.querySelector('.spark-slider__controls');
      this.inputEl = this.el.querySelector('input[type="number"]');
      this.handleEl = this.el.querySelector('.spark-slider__handle');
      this.trackEl = this.el.querySelector('.spark-slider__track');
      this.trackFillEl = this.trackEl.querySelector('.spark-slider__track-fill');

      if (!this.inputEl) {
        throw new Error('Tried to create a slider instance without a number input.');
      }

      if (!this.handleEl) {
        throw new Error('Tried to create a slider instance without a handle button.');
      }

      this.min = parseInt(this.inputEl.getAttribute('min'), 10) || this.min;
      this.max = parseInt(this.inputEl.getAttribute('max'), 10) || this.max;
      this.step = parseInt(this.inputEl.getAttribute('step'), 10) || this.step;

      // If we have a default value, set it after we cache the size of the element.
      // We have to do that first so we know the bounds of the slider.
      var value = parseInt(this.inputEl.getAttribute('value'), 10) || parseInt(this.inputEl.getAttribute('min'), 10);
      if (!isNaN(value)) {
        this._cacheSize();
        this.setValue(value);
      }
    },


    /**
     * Save the element dimensions.
     */
    _cacheSize: function() {

      this.width = this.trackEl.offsetWidth;
      this.height = this.trackEl.offsetHeight;

      this.handleSize = this.isX ? this.handleEl.offsetWidth : this.handleEl.offsetHeight;
      this.handleSizePercentage = this.isX ? this.handleEl.offsetWidth / this.width : this.handleEl.offsetHeight / this.height;

      var offset = this._getElementOffset(this.controlsEl);
      this.offsetLeft = offset.left;
      this.offsetTop = offset.top;
    },


    /**
     * Set the position of the handle.
     */
    _updateHandlePosition: function() {

      // Adjust the percentage of the total to be the percentage of the total less the size of the handle.
      var handleSize = isNaN(this.handleSizePercentage) ? 0 : this.handleSizePercentage;
      var percentage = this.percentage * (1 - handleSize) + (handleSize / 2);

      percentage = Math.round(Math.min(percentage, 1) * 100);

      this.handleEl.setAttribute('style', 'left: ' + percentage + '%;');
      this.trackEl.setAttribute('data-percentage', percentage);
      this.trackFillEl.setAttribute('style', 'width: ' + percentage + '%;');
    },


    /**
     * Update the active class on the handle.
     */
    _updateActiveClasses: function() {
      this._toggleClass(this.handleEl, 'active', this.isActive);
      this._toggleClass(this.el, 'active', this.isActive);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {

      this._onTouchStartBound = this._onTouchStart.bind(this);
      this._onTouchMoveBound = this._onTouchMove.bind(this);
      this._onTouchEndBound = this._onTouchEnd.bind(this);

      this._onClickBound = this._onClick.bind(this);
      this._onMouseDownBound = this._onMouseDown.bind(this);
      this._onMouseMoveBound = this._onMouseMove.bind(this);
      this._onMouseUpBound = this._onMouseUp.bind(this);
      this._onMouseOutBound = this._onMouseOut.bind(this);

      this._onFocusBound = this._onFocus.bind(this);
      this._onKeydownBound = this._onKeydown.bind(this);
      this._onBlurBound = this._onBlur.bind(this);

      this._onChangeBound = this._onChange.bind(this);

      this._onResizeBound = this._onResize.bind(this);
    },


    /**
     * Add event listeners for touchstart and mouse click.
     */
    _addEventListeners: function() {

      this.controlsEl.addEventListener('touchstart', this._onTouchStartBound);
      this.controlsEl.addEventListener('mousedown', this._onMouseDownBound);

      this.inputEl.addEventListener('change', this._onChangeBound);

      this.handleEl.addEventListener('focus', this._onFocusBound);
      this.handleEl.addEventListener('click', this._onClickBound);
    },


    /**
     * Remove event listeners for touchstart and mouse click.
     */
    _removeEventListeners: function() {

      this.controlsEl.removeEventListener('touchstart', this._onTouchStartBound);
      this.controlsEl.removeEventListener('mousedown', this._onMouseDownBound);

      this.inputEl.removeEventListener('change', this._onChangeBound);

      this.handleEl.removeEventListener('focus', this._onFocusBound);
      this.handleEl.removeEventListener('click', this._onClickBound);
    },


    /**
     * Add event listeners for touchmove, touchend, mousemove and mouseup.
     * We add these to the window so that the user can move off of the element
     * but keep dragging the slider handle. Otherwise it's really hard to
     * use the slider unless it's massive.
     * @param {String} type Which type of listeners to add
     */
    _addMoveEventListeners: function(type) {

      // Only listen for events of the type we asked for.
      switch (type) {
        case 'mouse':
          window.addEventListener('mousemove', this._onMouseMoveBound);
          window.addEventListener('mouseout', this._onMouseOutBound);
          window.addEventListener('mouseup', this._onMouseUpBound);
          break;
        case 'touch':
          window.addEventListener('touchmove', this._onTouchMoveBound);
          window.addEventListener('touchend', this._onTouchEndBound);
          break;
        case 'keyboard':
          window.addEventListener('keydown', this._onKeydownBound);
          this.handleEl.addEventListener('blur', this._onBlurBound);
          break;
      }

      window.addEventListener('resize', this._onResizeBound);
      window.addEventListener('orientationchange', this._onResizeBound);
    },


    /**
     * Remove event listeners for move events.
     * @param {String} type Which type of listeners to remove
     */
    _removeMoveEventListeners: function(type) {

      // Only unbind events of the type we asked for.
      switch (type) {
        case 'mouse':
          window.removeEventListener('mousemove', this._onMouseMoveBound);
          window.removeEventListener('mouseup', this._onMouseUpBound);
          break;
        case 'touch':
          window.removeEventListener('touchmove', this._onTouchMoveBound);
          window.removeEventListener('touchend', this._onTouchEndBound);
          break;
        case 'keyboard':
          window.removeEventListener('keydown', this._onKeydownBound);
          this.handleEl.removeEventListener('blur', this._onBlurBound);
          break;
      }

      window.removeEventListener('resize', this._onResizeBound);
      window.removeEventListener('orientationchange', this._onResizeBound);
    },


    /**
     * Truncate value text to fit.
     * @param {String} value
     * @return {String}
     */
    _truncateValueText: function(value, precision, characters) {

      if (value.length < 5) {
        return value;
      }

      var isNegative = value < 0 ? true : false;

      value = Math.abs(value);
      precision = precision || this.truncatePrecision;
      characters = characters || this.truncateCharacters;

      var i = characters.length - 1;

      precision = Math.pow(10, precision);

      for (i; i >= 0; i--) {

        var size = Math.pow(10, (i + 1) * 3);

        if (size <= value) {

          value = Math.round(value * precision / size) / precision;

          // @todo: what is this doing? i can't remember how we would ever
          // end up inside this condition.
          if ((value === 1000) && (i < characters.length - 1)) {
            value = 1;
            i++;
          }

          value += characters[i];

          break;
        }
      }

      return (isNegative ? '-' : '') + value;
    },


    /**
     * When the touch starts, start the slider.
     * @param {Object} e
     */
    _onTouchStart: function(e) {

      if (this.inputEl.getAttribute('disabled') !== null) {
        return;
      }

      e.preventDefault();

      this.start(this.isX ? e.touches[0].pageX : e.touches[0].pageY, 'touch');
    },


    /**
     * When the window fires a touchmove event, adjust our value accordingly
     * @param {Object} e
     */
    _onTouchMove: function(e) {

      if (!this.isActive) {
        return;
      }

      e.preventDefault();

      this.move(this.isX ? e.touches[0].pageX : e.touches[0].pageY);
    },


    /**
     * When the window fires a touchend event, stop tracking touches
     * @param {Object} e
     */
    _onTouchEnd: function(e) {

      if (!this.isActive) {
        return;
      }

      e.preventDefault();

      this.stop('touch');
    },


    /**
     * When the mouse presses down, start the slider.
     * @param {Object} e
     */
    _onMouseDown: function(e) {

      if (this.inputEl.getAttribute('disabled') !== null) {
        return;
      }

      e.preventDefault();

      this.start(this.isX ? e.pageX : e.pageY, 'mouse');
    },


    /**
     * When the window fires a mousemove event, adjust our value accordingly
     * @param {Object} e
     */
    _onMouseMove: function(e) {

      if (!this.isActive) {
        return;
      }

      e.preventDefault();

      this.move(this.isX ? e.pageX : e.pageY);
    },


    /**
     * When the window fires a mouseup event, stop tracking
     * @param {Object} e
     */
    _onMouseUp: function() {

      if (!this.isActive) {
        return;
      }

      this.stop('mouse');
    },


    /**
     * When the window fires a mouseout event, stop tracking if it was the html element.
     * @param {Object} e
     */
    _onMouseOut: function() {
      // @todo: make this work
      // if (e.relatedTarget === doc.body.parentNode) {
      //   this.stop('mouse');
      // }
    },


    /**
     * When the window resizes, cache size values for the slider.
     * @param {Object} e
     */
    _onResize: function() {
      this._cacheSize();
    },


    /**
     * When the element receives focus, start listening for keyboard events
     * @param {Object} e
     */
    _onFocus: function() {
      this.start(null, 'keyboard');
    },


    /**
     * When a key is pressed, see if it's a left or right arrow and move the
     * handle accordingly. If the shift key is pressed, we'll increment and
     * decrement by bigger values.
     * @param {Object} e
     */
    _onKeydown: function(e) {

      // Left for x or up for y
      var increment = (this.isX && e.keyCode === 39) || (!this.isX && e.keyCode === 38);

      // Right for x or down for y
      var decrement = (this.isX && e.keyCode === 37) || (!this.isX && e.keyCode === 40);

      if (increment) {
        this.increment(e.shiftKey);
      } else if (decrement) {
        this.decrement(e.shiftKey);
      }
    },


    /**
     * When the element loses focus, stop listening for keyboard events
     * @param {Object} e
     */
    _onBlur: function() {
      this.stop('keyboard');
    },


    /**
     * When the input value changes, set our interal value if it's not already our value.
     * @param {Object} e
     */
    _onChange: function(e) {
      if (e.target.value !== this.value) {
        this.setValue(e.target.value);
      }
      (this.onChange || noop)(this, this.value);
    },


    /**
     * Prevent click events on the button. This way we don't accidentally submit the form.
     * @param {Object} e
     */
    _onClick: function(e) {
      e.preventDefault();
    },


    /**
     * When the window is clicked and the element isn't part of the slider, trigger a blur.
     * @param {Object} e
     */
    _onWindowClick: function(e) {

      if (this._elementHasParent(e.target, this.el)) {
        this._onBlur();
      }
    }
  };

  Base.exportjQuery(Slider, 'Slider');

  return Slider;
}));

/**
 * # Table
 *
 * @example
 * new Table(el);
 *
 * @module components/table.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/components/expand', 'spark/helpers/set-caret', 'spark/helpers/store-form-data', 'spark/helpers/each'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('./expand'), require('../helpers/set-caret'), require('../helpers/store-form-data'), require('../helpers/each'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Table = factory(root.Spark.BaseComponent, root.Spark.Expand, root.Spark.setCaret, root.Spark.storeFormData, root.Spark.each);
  }
}(this, function(Base, Expand, setCaret, formData, each) {

  'use strict';

  /**
   * Table constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Table = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._setParams(params || {});

    this._cacheElements(el);
    this._parseParams();

    this._bindEventListenerCallbacks();
    this._addEventListeners();

    this._initRows();

    if (this.isSpreadsheet || this.isEditRows) {
      this._deactivateAllInputs();
    }

    if (this.isResizable) {
      this._initResize();
    }

    this._disableRowsColumnsCells();

    this._initExpands();
  };

  var noop = function() {};

  Table.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _hasClass: Base.hasClass,
    _toggleClass: Base.toggleClass,
    _triggerEvent: Base.triggerEvent,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _getChildIndex: Base.getChildIndex,
    _getSiblingBefore: Base.getSiblingBefore,
    _getSiblingAfter: Base.getSiblingAfter,
    _elementMatches: Base.elementMatches,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['isSpreadsheet', 'isEditRows', 'isResizable', 'confirmDeleteCallback', 'onRowSave', 'onRowDelete'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      tableEl: null,
      isSpreadsheet: null,
      isEditRows: null,
      isResizable: null,
      onRowSave: null,
      onRowDelete: null,
      confirmDeleteCallback: null,
      _changePaused: false,
      _expands: null,
      _keyCodes: {
        ENTER: 13,
        UP: 38,
        DOWN: 40,
        LEFT: 37,
        RIGHT: 39,
        ESCAPE: 27
      },
      _editingCount: 0,
      _lastClickTime: 0,
      _lastClickEl: null,
      _lastScreenX: 0,
      _touchStartTime: 0,
      _touchStartEl: null,
      _resizeEls: null,
      _resizingEl: null,
      _sizeColumnsRun: false,
      _onClickBound: null,
      _onChangeBound: null,
      _onFocusBound: null,
      _onBlurBound: null,
      _onKeydownBound: null,
      _onTouchstartBound: null,
      _onTouchendBound: null,
      _onMouseDownBound: null,
      _onMouseMoveBound: null,
      _onMouseUpBound: null
    },


    /**
     * Disable the form field in a table cell.
     * @param {Element} el
     */
    disableCell: function(el) {
      el.disabled = true;
      this._addClass(this._getElementMatchingParent(el, 'td'), 'spark-table__disabled-cell');
    },


    /**
     * Enable the form field in a table cell.
     * @param {Element} el
     */
    enableCell: function(el) {
      el.disabled = false;
      this._removeClass(this._getElementMatchingParent(el, 'td'), 'spark-table__disabled-cell');
    },


    /**
     * Disable a row and all the cells inside of it.
     * @param {Element} el
     */
    disableRow: function(el) {
      this._addClass(el, 'spark-table__disabled-row');
      each(el.querySelectorAll('input, button, a'), function(i) {
        i.disabled = true;
      });
    },


    /**
     * Enable a row and all the cells inside of it.
     * @param {Element} el
     */
    enableRow: function(el) {
      this._removeClass(el, 'spark-table__disabled-row');
      each(el.querySelectorAll('input, button, a'), function(i) {
        i.disabled = false;
      });
    },


    /**
     * Disable a column and all the cells inside of it.
     * @param {Element} el
     */
    disableColumn: function(el) {

      var index = this._getChildIndex(el.parentNode.children, el);

      each(this.tableEl.querySelectorAll('tbody tr'), function(row) {
        this.disableCell(row.children[index].querySelector('input'));
      }.bind(this));

      this._addClass(el, 'spark-table__disabled-column');
    },


    /**
     * Enable a column and all the cells inside of it.
     * @param {Element} el
     */
    enableColumn: function(el) {

      var index = this._getChildIndex(el.parentNode.children, el);

      each(this.tableEl.querySelectorAll('tbody tr'), function(row) {
        this.enableCell(row.children[index].querySelector('input'));
      }.bind(this));

      this._removeClass(el, 'spark-table__disabled-column');
    },


    /**
     * Remove the table anc cleanup.
     * @param {Boolean} leaveElement Leave the element intact.
     */
    remove: function(leaveElement) {
      each(this._expands, function(e) {
        e.remove(leaveElement);
      });
      Base.remove.apply(this, arguments);
    },


    /**
     * Activate a row.
     * @param {Number|Element} row
     */
    activateRow: function(row) {
      row = typeof row === 'number' ? this.tableEl.querySelectorAll('tbody tr')[row] : row;
      if (!row) return;
      this._makeRowActive(row);
    },


    /**
     * Activate multiple rows.
     * @param {Array} rows
     */
    activateRows: function(rows) {
      each(rows, this.activateRow.bind(this));
    },


    /**
     * Deactivate a row.
     * @param {Number|Element} row
     */
    deactivateRow: function(row) {
      row = typeof row === 'number' ? this.tableEl.querySelectorAll('tbody tr')[row] : row;
      if (!row) return;
      this._makeRowInActive(row);
    },


    /**
     * Deactivate multiple rows.
     * @param {Array} rows
     */
    deactivateRows: function(rows) {
      each(rows, this.deactivateRow.bind(this));
    },


    /**
     * Get an array of currently active rows.
     * @return {Array}
     */
    getActiveRows: function() {

      var arr = [];

      each(this.el.querySelectorAll('tbody tr.active'), function(tr) {
        arr.push(tr);
      });

      return arr;
    },


    /**
     * Store a reference to the tabs list, each tab and each panel.
     * Set which tab is active, or use the first.
     * @param {Element} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.tableEl = el.querySelector('table');
    },


    /**
     * Parse parameters from the elements.
     */
    _parseParams: function() {

      if (!this.tableEl) {
        return;
      }

      this.isSpreadsheet = this.isSpreadsheet !== null ? this.isSpreadsheet : (this._hasClass(this.el, 'spark-table--spreadsheet') ? true : false);
      this.isEditRows = this.isEditRows !== null ? this.isEditRows : (this._hasClass(this.el, 'spark-table--edit-rows') ? true : false);
      this.isResizable = this.isResizable !== null ? this.isResizable : (this._hasClass(this.el, 'spark-table--resizable') ? true : false);
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {

      this._onClickBound = this._onClick.bind(this);
      this._onChangeBound = this._onChange.bind(this);
      this._onFocusBound = this._onFocus.bind(this);
      this._onBlurBound = this._onBlur.bind(this);

      this._onTouchstartBound = this._onTouchstart.bind(this);
      this._onTouchendBound = this._onTouchend.bind(this);
      this._onKeydownBound = this._onKeydown.bind(this);

      this._onMouseDownBound = this._onMouseDown.bind(this);
      this._onMouseMoveBound = this._onMouseMove.bind(this);
      this._onMouseUpBound = this._onMouseUp.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {

      this.el.addEventListener('click', this._onClickBound, false);
      this.el.addEventListener('change', this._onChangeBound, false);
      this.el.addEventListener('focus', this._onFocusBound, true);
      this.el.addEventListener('blur', this._onBlurBound, true);

      if (this.isSpreadsheet) {
        this.el.addEventListener('touchstart', this._onTouchstartBound, false);
        this.el.addEventListener('touchend', this._onTouchendBound, false);
        this.el.addEventListener('keydown', this._onKeydownBound, false);
      }

      if (this.isResizable) {
        this.tableEl.addEventListener('mousedown', this._onMouseDownBound, false);
      }
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {

      this.el.removeEventListener('click', this._onClickBound);
      this.el.removeEventListener('change', this._onChangeBound);
      this.el.removeEventListener('focus', this._onFocusBound);
      this.el.removeEventListener('blur', this._onBlurBound);

      this.el.removeEventListener('touchstart', this._onTouchstartBound);
      this.el.removeEventListener('touchend', this._onTouchendBound);
      this.el.removeEventListener('keydown', this._onKeydownBound);

      this.tableEl.removeEventListener('mousedown', this._onMouseDownBound);

      this._removeResizeListeners();
    },

    /**
     * Add listeners for mousemove and mouseup events.
     */
    _addResizeListeners: function() {
      window.addEventListener('mousemove', this._onMouseMoveBound, false);
      window.addEventListener('mouseup', this._onMouseUpBound, false);
    },


    /**
     * Remove listeners for mosuemove and mouseup.
     */
    _removeResizeListeners: function() {
      window.removeEventListener('mousemove', this._onMouseMoveBound);
      window.removeEventListener('mouseup', this._onMouseUpBound);
    },


    /**
     * Toggle the active state on a row.
     * @param {Object} row
     */
    _toggleRowActive: function(row) {

      if (this._hasClass(row, 'active')) {
        this._makeRowInActive(row);
        this._uncheckSelectAll();
      } else {
        this._makeRowActive(row);
      }
    },


    /**
     * Make a row active
     * @param {Element} row
     */
    _makeRowActive: function(row) {

      this._addClass(row, 'active');
      var checkbox = row.querySelector('input[type="checkbox"]:not([disabled])');

      if (checkbox && checkbox.checked !== true) {
        checkbox.checked = true;
        this._changePaused = true;
        this._triggerEvent(checkbox, 'change');
        this._changePaused = false;
      }
    },


    /**
     * Make a row active
     * @param  {Element} row
     */
    _makeRowInActive: function(row) {

      this._removeClass(row, 'active');
      var checkbox = row.querySelector('input[type="checkbox"]:not([disabled])');

      if (checkbox && checkbox.checked !== false) {
        checkbox.checked = false;
        this._changePaused = true;
        this._triggerEvent(checkbox, 'change');
        this._changePaused = false;
      }
    },


    /**
     * Toggle active on each row.
     * @param {NodeList} rows
     * @param {Boolean} active
     */
    _toggleRowsActive: function(rows, active) {

      var func = active ? '_makeRowActive' : '_makeRowInActive';
      var i = 0;
      var len = rows.length;

      for (; i < len; i++) {
        this[func](rows[i]);
      }
    },


    /**
     * Toggle whether everything should be selected. Find the checkbox input inside of the
     * given element and invert its state.
     * @param {Element} el
     */
    _toggleSelectAll: function(el) {

      var checkbox = el.querySelector('input[type="checkbox"]');

      if (!checkbox) {
        return;
      }

      this._toggleRowsActive(this.el.querySelectorAll('tbody tr'), !checkbox.checked);

      checkbox.checked = !checkbox.checked;
    },


    /**
     * Uncheck the select all checkboxes.
     */
    _uncheckSelectAll: function() {

      var checkboxes = this.el.querySelectorAll('.spark-table__select-all input[type="checkbox"]');
      var i = 0;
      var len = checkboxes.length;

      for (; i < len; i++) {
        checkboxes[i].checked = false;
      }
    },


    /**
     * Deactivate editing in all input fields.
     */
    _deactivateAllInputs: function() {

      if (!this.tableEl) {
        return;
      }

      this._deactivateInputs(this.tableEl);
    },


    /**
     * Deactivate all the inputs inside an element
     * @param {Element} el
     */
    _deactivateInputs: function(el) {

      var inputs = el.querySelectorAll('input:not([type="checkbox"])');
      var i = 0;

      var len = inputs.length;

      for (; i < len; i++) {
        this._deactivateInput(inputs[i]);
      }
    },


    /**
     * Make an input field readonly.
     * @param {Element} input
     */
    _deactivateInput: function(input) {
      input.setAttribute('readonly', '');
      this._removeClass(input.parentNode, 'editing');
    },


    /**
     * Activate all the inputs inside an element
     * @param {Element} el
     */
    _activateInputs: function(el) {

      var inputs = el.querySelectorAll('input:not([type="checkbox"])');
      var i = 0;

      var len = inputs.length;

      for (; i < len; i++) {
        this._activateInput(inputs[i]);
      }
    },


    /**
     * Make an input field readable.
     * @param {Element} input
     */
    _activateInput: function(input) {
      input.removeAttribute('readonly');
      this._addClass(input.parentNode, 'editing');
      if (input.type !== 'checkbox' && input.type !== 'radio') {
        setCaret(input, -1);
      }
    },


    /**
     * Activate an input, unless it's already enabled in which case
     * the focus should move down a row.
     * @param {Element} input
     */
    _activateInputOrFocusDown: function(input) {

      // Currently readonly
      if (input.getAttribute('readonly') === '') {
        this._activateInput(input);
        return;
      }

      this._focusDown(input, true);
    },


    /**
     * Find all the rows, columns and cells that should be disabled.
     */
    _disableRowsColumnsCells: function() {
      each(this.tableEl.querySelectorAll('td input[disabled]'), this.disableCell.bind(this));
      each(this.tableEl.querySelectorAll('.spark-table__disabled-row'), this.disableRow.bind(this));
      each(this.tableEl.querySelectorAll('.spark-table__disabled-column'), this.disableColumn.bind(this));
    },


    /**
     * Move our focus up a row from the given element.
     * @param {Element} input
     * @param {Boolean} force Force the move even if the element is active.
     */
    _focusUp: function(input, force) {
      return this._focusUpDown(input, 'up', force);
    },


    /**
     * Move our focus down a row from the given element.
     * @param {Element} input
     * @param {Boolean} force Force the move even if the element is active.
     */
    _focusDown: function(input, force) {
      return this._focusUpDown(input, 'down', force);
    },


    /**
     * Focus on a row up or down from the given element.
     * @param {Element} input
     * @param {String} direction up|down
     * @param {Boolean} force Force the move even if the element is active.
     */
    _focusUpDown: function(input, direction, force) {

      // If we're not being told to force and the item is not read only
      if (!force && input.getAttribute('readonly') === null) {
        return;
      }

      this._deactivateInput(input);
      var td = this._getElementMatchingParent(input, 'td', this.el);

      if (!td) {
        return;
      }

      var index = this._getChildIndex(td.parentNode.children, td);
      var nextRow = this[direction === 'up' ? '_getSiblingBefore' : '_getSiblingAfter'](td.parentNode, 'tr');

      if (!nextRow) {
        return;
      }

      var newTd = nextRow.children[index];

      if (!newTd) {
        return;
      }

      var newInput = newTd.querySelector('input:not([type="checkbox"]), select');

      if (newInput) {
        if (newInput.disabled) {
          this._focusUpDown(newInput, direction, force);
        } else {
          newInput.focus();
        }
      }
    },


    /**
     * Move our focus left a cell from the given element.
     * @param {Element} input
     * @param {Boolean} force Force the move even if the element is active.
     */
    _focusLeft: function(input, force) {
      return this._focusLeftRight(input, 'left', force);
    },


    /**
     * Move our focus right a cell from the given element.
     * @param {Element} input
     * @param {Boolean} force Force the move even if the element is active.
     */
    _focusRight: function(input, force) {
      return this._focusLeftRight(input, 'right', force);
    },


    /**
     * Focus on a cell left or down from the given element.
     * @param {Element} input
     * @param {String} direction up|down
     * @param {Boolean} force Force the move even if the element is active.
     */
    _focusLeftRight: function(input, direction, force) {

      // If we're not being told to force and the item is not read only
      if (!force && input.getAttribute('readonly') === null) {
        return;
      }

      this._deactivateInput(input);
      var td = this._getElementMatchingParent(input, 'td', this.el);

      if (!td) {
        return;
      }

      var newTd = this[direction === 'left' ? '_getSiblingBefore' : '_getSiblingAfter'](td, 'td');

      if (!newTd) {
        return;
      }

      var newInput = newTd.querySelector('input:not([type="checkbox"]), select');

      if (newInput) {
        if (newInput.disabled) {
          this._focusLeftRight(newInput, direction, force);
        } else {
          newInput.focus();
        }
      }
    },


    /**
     * Check for two click events on the same element in short succession.
     * @param {Element} el
     */
    _checkDoubleClick: function(el) {

      var now = Date.now();
      var lastTime = this._lastClickTime;
      var lastEl = this._lastClickEl;

      this._lastClickTime = now;
      this._lastClickEl = el;

      if (el === lastEl && now - 500 < lastTime) {
        return true;
      }

      return false;
    },


    /**
     * Unset the last clicked element.
     */
    _clearClicked: function() {
      this._lastClickEl = null;
    },


    /**
     * Enable editing on a row.
     * @param {Element} row
     */
    _editRow: function(row) {

      if (!row) {
        return;
      }

      this._editingCount++;

      this._activateInputs(row);
      formData.store(row);
      this._addClass(row, 'editing');
      this._updateBindings();
    },


    /**
     * Cancel editing a row.
     * @param {Element} row
     */
    _cancelRow: function(row) {

      if (!row) {
        return;
      }

      this._editingCount--;

      this._deactivateInputs(row);
      formData.restore(row);
      this._removeClass(row, 'editing');
      this._updateBindings();
    },


    /**
     * Save a row.
     * @param {Element} row
     */
    _saveRow: function(row) {

      if (!row) {
        return;
      }

      this._editingCount--;

      this._deactivateInputs(row);
      formData.clear(row);
      this._removeClass(row, 'editing');
      this._updateBindings();

      (this.onRowSave || noop)(this._getChildIndex(row.parentNode.children, row), row);
    },


    /**
     * Delete a row.
     * @param {Element} row
     */
    _deleteRow: function(row) {

      if (!row) {
        return;
      }

      (this.onRowDelete || noop)(this._getChildIndex(row.parentNode.children, row), row);
      row.parentNode.removeChild(row);
    },


    /**
     * Confirm the deletion of a row.
     * @param {Element} row
     */
    _confirmDelete: function(row) {

      if (!this.confirmDeleteCallback || typeof this.confirmDeleteCallback !== 'function') {
        this._deleteRow(row);
      } else {
        this.confirmDeleteCallback(row, this._deleteRow);
      }
    },


    /**
     * Update data bindings.
     */
    _updateBindings: function() {
      this._toggleClass(this.el, 'editing', this._editingCount);
    },


    /**
     * Add handles to the header that can be grabbed for resizing.
     */
    _initResize: function() {

      this._resizeEls = [];

      var ths = this.tableEl.querySelectorAll('thead th');

      each(ths, function(th) {
        th.innerHTML = '<span class="spark-table__resize spark-table__resize--left"></span>' + th.innerHTML + '<span class="spark-table__resize spark-table__resize--right"></span>';
        this._resizeEls.push(th);
      }.bind(this));
    },


    /**
     * Initialize rows active states.
     */
    _initRows: function() {

      each(this.tableEl.querySelectorAll('td.spark-table__checkbox input:checked'), function(c) {
        this._makeRowActive(this._getElementMatchingParent(c, 'tr'));
      }.bind(this));
    },


    /**
     * Set the size of each column as a percentage so it can be adjusted
     * while cells are resized.
     * @param {String} unit Optional
     * @param {Boolean} force Optional
     */
    _sizeColumns: function(unit, force) {

      unit = unit || '%';

      if (this._sizeColumnsRun && !force) {
        return;
      }

      var width = this.tableEl.offsetWidth;

      each(this.tableEl.querySelectorAll('thead th'), function(th) {
        if (unit === '%')
          th.style.width = (Math.round(th.offsetWidth / width * 100000) / 100000) * 100 + '%';
        else
          th.style.width = th.offsetWidth + 'px';
      }.bind(this));

      this._sizeColumnsRun = true;
    },


    /**
     * Initialize expand/collapse rows.
     */
    _initExpands: function() {

      var expands = this.tableEl.querySelectorAll('.spark-table-expand');

      this._expands = [];

      each(expands, function(e) {
        this._expands.push(new Expand(e, {
          onBeforeExpand: this._onBeforeExpand.bind(this)
        }));
      }.bind(this));
    },


    /**
     * Before an expand is called, size all the columns so that
     * the expand does cause width changes.
     */
    _onBeforeExpand: function() {
      this._sizeColumns();
    },


    /**
     * When we are clicked determine the proper action to take.
     * @param  {Object} e
     */
    _onClick: function(e) {

      var target = e.target || e.srcElement;
      var row;
      var selectAll;
      var actionTaken = false;
      var clearClicked = true;

      // Select all rows checkbox
      if ((selectAll = this._getElementMatchingParent(target, '.spark-table__select-all', this.el)) && !this._elementMatches(target, 'input[type="checkbox"]')) {
        this._toggleSelectAll(selectAll);
        actionTaken = true;
      }
      // Editable field
      else if (this._elementMatches(target, 'input:not([type="checkbox"]), select')) {

        if (!target.disabled) {

          // Listen for double clicks on a spreadsheet
          if (this.isSpreadsheet) {
            clearClicked = false;
            if (this._checkDoubleClick(target)) {
              clearClicked = true;
              this._activateInput(target);
            }
          }

          actionTaken = true;
        }
      }
      // Edit button
      else if (this._elementMatches(target, '.spark-table__edit-row')) {
        this._editRow(this._getElementMatchingParent(target, 'tr', this.el));
        actionTaken = true;
      }
      // Delete button
      else if (this._elementMatches(target, '.spark-table__delete-row')) {
        this._confirmDelete(this._getElementMatchingParent(target, 'tr', this.el));
        actionTaken = true;
      }
      // Save button
      else if (this._elementMatches(target, '.spark-table__edit-row-save')) {
        this._saveRow(this._getElementMatchingParent(target, 'tr', this.el));
        actionTaken = true;
      }
      // Cancel button
      else if (this._elementMatches(target, '.spark-table__edit-row-cancel')) {
        this._cancelRow(this._getElementMatchingParent(target, 'tr', this.el));
        actionTaken = true;
      }
      // Select a row
      else if (!this._getElementMatchingParent(target, 'button, a', this.el) && !this._elementMatches(target, 'input[type="checkbox"]') && (row = this._getElementMatchingParent(target, 'tbody tr', this.el))) {
        if (!(row.querySelector('input[type="checkbox"]') || {}).disabled) {
          this._toggleRowActive(row);
          actionTaken = true;
        }
      }

      if (clearClicked) {
        this._clearClicked();
      }

      if (actionTaken) {
        e.preventDefault();
      }
    },


    /**
     * When the change event fires on our element.
     * @param {Object} e
     */
    _onChange: function(e) {

      if (this._changePaused) {
        return;
      }

      var target = e.target || e.srcElement;
      var row;
      var selectAll;

      // Select all rows checkbox. We have to invert the checked value here because it
      // get toggled back in the select all call.
      if ((selectAll = this._getElementMatchingParent(target, '.spark-table__select-all', this.el))) {
        target.checked = !target.checked;
        this._toggleSelectAll(selectAll);
      }
      // Checkbox for a row
      else if (this._elementMatches(target, 'input[type="checkbox"]') && (row = this._getElementMatchingParent(target, 'tbody tr', this.el))) {
        this._toggleRowActive(row);
      }
    },


    /**
     * If this is a spreadsheet, whenever a field gains focus, highlight its parent.
     * @param {Object} e
     */
    _onFocus: function(e) {

      var target = e.target || e.srcElement;

      if (!this.isSpreadsheet || !this._elementMatches(target, 'input:not([type="checkbox"]), select')) {
        return;
      }

      var td = this._getElementMatchingParent(target, 'td', this.el);
      this._addClass(td, 'focus');
    },


    /**
     * If this is a spreadsheet, whenever a field gains focus, highlight its parent.
     * @param {Object} e
     */
    _onBlur: function(e) {

      if (!this.isSpreadsheet) {
        return;
      }

      var target = e.target || e.srcElement;
      var td = this._getElementMatchingParent(target, 'td', this.el);
      this._removeClass(td, 'focus');
      this._deactivateInput(target);
    },


    /**
     * When a key is pressed, if this is a spreadsheet then we should detect
     * enter or arrow keys.
     * @param {Object} e
     */
    _onKeydown: function(e) {

      var target = e.target || e.srcElement;

      if (!this.isSpreadsheet || !this._elementMatches(target, 'input:not([type="checkbox"]), select')) {
        return;
      }

      var code = e.keyCode || e.which;

      switch (code) {
        case this._keyCodes.ENTER:
          this._activateInputOrFocusDown(target);
          break;
        case this._keyCodes.ESCAPE:
          this._deactivateInput(target);
          break;
        case this._keyCodes.DOWN:
          this._focusDown(target);
          break;
        case this._keyCodes.UP:
          this._focusUp(target);
          break;
        case this._keyCodes.LEFT:
          this._focusLeft(target);
          break;
        case this._keyCodes.RIGHT:
          this._focusRight(target);
          break;
      }
    },


    /**
     * Listen for a touch and hold on an input.
     * @param {Object} e
     */
    _onTouchstart: function(e) {

      var target = e.target || e.srcElement;

      if (!this.isSpreadsheet || !this._elementMatches(target, 'input:not([type="checkbox"])')) {
        return;
      }

      this._touchStartEl = target;
      this._touchStartTime = Date.now();
      this._touchStartTimer = setTimeout(this._onTouchHold.bind(this), 1000);
    },


    /**
     * Listen for the end of a touch to cancel the hold timer.
     * @param {Object} e
     */
    _onTouchend: function(e) {

      var target = e.target || e.srcElement;

      if (!this._touchStartEl || target !== this._touchStartEl) {
        return;
      }

      this._touchStartEl = null;
      this._touchStartTime = null;
      clearTimeout(this._touchStartTimer);
    },


    /**
     * When the user has held on an input for the defined amount of time.
     */
    _onTouchHold: function() {

      this._activateInput(this._touchStartEl);

      this._touchStartEl = null;
      this._touchStartTime = null;
      clearTimeout(this._touchStartTimer);
    },


    /**
     * When the mouse is depressed.
     * @param {Object} e
     */
    _onMouseDown: function(e) {

      var target = e.target || e.srcElement;

      if (!this.isResizable || !this._elementMatches(target, '.spark-table__resize')) {
        return;
      }

      e.preventDefault();

      this._lastScreenX = e.screenX;

      this._sizeColumns('px', true);

      this._resizingEl = target.parentNode;
      var index = this._resizeEls.indexOf(this._resizingEl);

      if (this._hasClass(target, 'spark-table__resize--left')) {
        this._resizingEl = this._resizeEls[index - 1];
      }

      if (!this._resizingEl) {
        return;
      }

      this._addResizeListeners();
    },


    /**
     * When the mouse moves after being depressed, resize the columns.
     * @param {Object} e
     */
    _onMouseMove: function(e) {

      var x = e.screenX;
      var d = x - this._lastScreenX;

      // No delta change
      if (!d) {
        return;
      }

      e.preventDefault();

      var w = this._resizingEl.offsetWidth;
      var tW = this.tableEl.offsetWidth;
      var newW = w + d;
      var newTW = tW + d;

      this._resizingEl.style.width = newW + 'px';
      this.tableEl.style.width = newTW + 'px';

      // Size was not affected because we're too small
      if (this._resizingEl.offsetWidth === w || this.tableEl.offsetWidth < this.tableEl.parentNode.offsetWidth) {
        this._resizingEl.style.width = w + 'px';
        this.tableEl.style.width = tW + 'px';
      }

      this._lastScreenX = x;
    },


    /**
     * When the mouse is released, stop tracking mouse move events and
     * convert table sizes to percentages.
     * @param {Object} e
     */
    _onMouseUp: function() {
      this._sizeColumns('%', true);
      this.tableEl.style.width = (this.tableEl.offsetWidth / this.tableEl.parentNode.offsetWidth * 100) + '%';
      this._removeResizeListeners();
    }
  };

  Base.exportjQuery(Table, 'Table');

  return Table;
}));

/**
 * # Tabs
 * Tabbed navigation
 *
 * @example
 * new Tabs(el, {
 *   // Optional. Alternate breakpoint values.
 *   breakpoints: {
 *    xs: {
 *      min: 0,
 *      max: 639
 *    },
 *    // ...
 *  }
 * });
 *
 * @module components/tabs.js
 */

/**
 * @ignore
 */
(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/transform'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/transform'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.Tabs = factory(root.Spark.BaseComponent, root.Spark.transform);
  }
}(this, function(Base, transform) {

  'use strict';

  /**
   * Tabs constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Tabs = function(el, params) {

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params || {});
    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._determineSize();
  };

  Tabs.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _getChildIndex: Base.getChildIndex,
    _elementHasParent: Base.elementHasParent,
    _getElementMatchingChildren: Base.getElementMatchingChildren,
    _getBreakpoint: Base.getBreakpoint,


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['useHash', 'breakpoints'],


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      tabListEl: null,
      tabListScrollEl: null,
      tabEls: null,
      panelEls: null,
      activeTabEl: null,
      activePanelEl: null,
      navEl: null,
      leftEl: null,
      rightEl: null,
      useHash: false,
      isScrollable: false,
      isDragging: false,
      isAnimatable: false,
      scrollDirection: '',
      scrollDistance: 0,
      lastX: 0,
      lastY: 0,
      minX: 0,
      maxX: 0,
      x: 0,
      _onFocusBound: null,
      _onBlurBound: null,
      _onTabListClickBound: null,
      _onLeftClickBound: null,
      _onRightClickBound: null,
      _onResizeBound: null,
      _onTouchStartBound: null,
      _onTouchMoveBound: null,
      _onTouchEndBound: null,
      _onMouseDownBound: null,
      _onMouseMoveBound: null,
      _onMouseUpBound: null,
      _onScrollBound: null
    },


    /**
     * Set the active item.
     * @param {String|Number|Object} el
     */
    setActive: function(el) {

      var panel;

      // If we're passed a string instead of an element or number,
      // get the panel with that id.
      if (typeof el === 'string') {
        panel = this._findPanelByName(el);

        // If we've found a panel, find the corresponding tab.
        if (panel) {
          el = this._findTabByPanel(panel);
        }
      }

      // If we're passed a number instead of an element,
      // get that item from the tabEls NodeList
      if (typeof el === 'number') {
        el = this.tabEls.item(el);
      }

      // If we couldn't find the element or it's already active, stop.
      if (!el || typeof el !== 'object' || el === this.activeTabEl) {
        return;
      }

      // Remove the active class from the currently active tab
      if (this.activeTabEl) {
        this._toggleClass(this.activeTabEl, 'active', false);
        this.previousTabEl = this.activeTabEl;
      }

      // Add the active class and store.
      this._toggleClass(el, 'active', true);
      this.activeTabEl = el;

      // Focus the tab on the left side if it's to the left of the frame.
      if (-el.offsetLeft > this.x) {
        this.focus(el, 'left');
      }
      // Focus the tab on the right side if it's to the right of the frame.
      else if (el.offsetLeft + el.clientWidth > this.tabListScrollEl.clientWidth - this.x) {
        this.focus(el, 'right');
      }

      // If we don't already have a panel, find the panel that corresponds to this tab.
      if (!panel) {
        panel = this._findPanelByTab(el);
      }

      // Set the new panel to be active.
      this._toggleClass(panel, 'active', true);

      // Remove the active class from the currently active panel.
      if (this.activePanelEl) {
        this._toggleClass(this.activePanelEl, 'active', false);
      }

      // Store the new active panel
      this.activePanelEl = panel;

      // Set the hash
      if (this.useHash) {
        window.location.hash = this.activePanelEl.getAttribute('id') || '';
      }
      var e = document.createEvent('Event');
      e.initEvent('spark.visible-children', true, true);
      this.activePanelEl.dispatchEvent(e);
    },


    /**
     * Start the drag
     * @param {Object} params
     */
    start: function(params) {

      params = params || {};

      // Start dragging
      this.isDragging = true;

      // Stash the element and its position
      this.lastX = params.lastX;
      this.lastY = params.lastY;

      // Stash the min and max values
      this._determineMinMax();

      // Add listeners to the body so we can drag this thing anywhere and still get events
      this._addMoveEventListeners(params.type || 'mouse');
    },


    /**
     * Stop the drag
     * @param {Object} params
     */
    stop: function(params) {

      params = params || {};

      // Make sure we're in bounds
      this._checkX();

      // Stop dragging
      this.isDragging = false;
      this.scrollDistance = 0;

      // Reset the scroll direction
      this.scrollDirection = '';

      // Unbind event listeners on the body
      this._removeMoveEventListeners(params.type);
    },


    /**
     * Move the drag point
     * @param {Object} params
     */
    move: function(params) {

      // Make sure we're currently dragging
      if (!this.isDragging && !params.scroll && !params.force) {
        return;
      }

      // If we're beyond the bounds, add some resistance to the scroll.
      if (!params.force && (this.x + params.x > this.maxX || this.x + params.x < this.minX)) {
        this.x += params.x / 4;
      } else {
        this.x += params.x;
      }

      this.scrollDistance += Math.abs(params.x);

      if (params.scroll) {
        this._checkX();
      }

      this._updatePosition();
    },


    /**
     * Focus on a specific element by bringing it to the middle of the scroller.
     * @param {Element} el
     * @param {String} align Which side to align with.
     */
    focus: function(el, align) {

      align = align || 'left';

      this.x = align === 'left' ? -el.offsetLeft : -(el.offsetLeft - this.tabListScrollEl.clientWidth + el.offsetWidth);

      this._checkX();
      this._updatePosition();
    },


    /**
     * Remove the element from the DOM and prepare for garbage collection by dereferencing values.
     * @param {Boolean} leaveElement Leave the element intact.
     */
    remove: function(leaveElement) {
      this._removeMoveEventListeners('touch');
      this._removeMoveEventListeners('mouse');
      this._removeMoveEventListeners('keyboard');
      Base.remove.call(this, leaveElement);
    },


    /**
     * Update the elements used.
     */
    update: function() {

      this._removeEventListeners();
      this._cacheElements(this.el);
      this._addEventListeners();

      // Run on the next frame so sizes have updated
      setTimeout(function() {
        this._determineSize();
        this.focus(this.activeTabEl);
      }.bind(this), 0);
    },


    /**
     * Find a panel element by name.
     * @param {String} name
     * @return {Object|Null}
     */
    _findPanelByName: function(name) {

      name = name.replace('#', '');

      var i = 0;
      var len = this.panelEls.length;
      var found = null;
      var el;

      for (; i < len && !found; i++) {
        if (this.panelEls[i].getAttribute('id') === name) {
          el = this.panelEls[i];
          found = true;
        }
      }

      return found && el;
    },


    /**
     * Find a panel given its corresponding tab. Try to match based on the
     * id attribute, but fall back to matching based on index.
     * @param {Object} tab
     * @return {Object|Null}
     */
    _findPanelByTab: function(tab) {

      var anchorChild = tab.querySelector('a');
      var id = anchorChild && anchorChild.getAttribute('href');
      var index = this._getChildIndex(tab.parentNode.children, tab);
      var i = 0;
      var len = this.panelEls.length;
      var foundById = null;
      var idMatch = null;
      var indexMatch = null;

      id = id ? id.replace('#', '') : id;

      for (; i < len && !foundById; i++) {
        if (id && this.panelEls[i].getAttribute('id') === id) {
          foundById = true;
          idMatch = this.panelEls[i];
        } else if (i === index) {
          indexMatch = this.panelEls[i];
        }
      }

      return (foundById && idMatch) || indexMatch;
    },


    /**
     * Find a tab given its corresponding panel. Try to match based on the
     * [href] attribute, but fall back to matching based on index.
     * @param {Object} panel
     * @return {Object|Null}
     */
    _findTabByPanel: function(panel) {

      var id = panel.getAttribute('id');
      var index = this._getChildIndex(panel.parentNode.children, panel);
      var i = 0;
      var len = this.tabEls.length;
      var foundById = null;
      var idMatch = null;
      var indexMatch = null;

      for (; i < len && !foundById; i++) {
        if (id && (this.tabEls.item(i).querySelector('a').getAttribute('href') === ('#' + id) || this.tabEls.item(i).getAttribute('href') === ('#' + id))) {
          foundById = true;
          idMatch = this.tabEls.item(i);
        } else if (i === index) {
          indexMatch = this.tabEls.item(i);
        }
      }

      return (foundById && idMatch) || indexMatch;
    },


    /**
     * Find the tab which an element lives inside.
     * @param {Element} el
     * @return {Object}
     */
    _findTabByChildElement: function(el) {

      var i = 0;
      var len = this.tabEls.length;
      var found;
      var tab;

      for (; i < len && !found; i++) {

        // There is a chance that the element passed IS a tab. Or maybe a tab is its parent.
        if (this.tabEls.item(i) === el || this._elementHasParent(el, this.tabEls.item(i))) {
          found = true;
          tab = this.tabEls.item(i);
        }
      }

      return found && tab;
    },


    /**
     * Store a reference to the tabs list, each tab and each panel.
     * Set which tab is active, or use the first.
     * @param {Element} el
     */
    _cacheElements: function(el) {

      this.el = el;
      this.tabListEl = this.el.querySelector('.spark-tabs__list');
      this.tabListScrollEl = this.tabListEl.parentNode;
      this.tabEls = this.tabListEl.querySelectorAll('.spark-tabs__tab');
      this.panelEls = this._getElementMatchingChildren(this.el.querySelector('.spark-tabs__panels'), '[role="tabpanel"]');
      this.navEl = this.el.querySelector('.spark-tabs__nav');
      this.leftEl = this.navEl.querySelector('.spark-tabs__btn--left');
      this.rightEl = this.navEl.querySelector('.spark-tabs__btn--right');

      // Make sure we have the elements we need
      if (!this.tabListEl || !this.tabEls.length || !this.panelEls.length) {
        throw new Error('Tab element missing either a .spark-tabs__list, or elements with .spark-tabs__tab and .spark-tabs__panel!', this.el);
      }

      // If there is a hash set, use that to try and set the active panel
      var hasSet = window.location.hash && this.setActive(window.location.hash);

      // If we weren't able to set with a hash, find the tab marked active or default to the first tab
      if (!hasSet) {
        this.setActive(this.tabListEl.querySelector('.spark-tabs__tab.active') || 0);
      }
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {

      this._onResizeBound = this._onResize.bind(this);

      this._onTabListClickBound = this._onTabListClick.bind(this);

      this._onRightClickBound = this._onRightClick.bind(this);
      this._onLeftClickBound = this._onLeftClick.bind(this);

      this._onTouchStartBound = this._onTouchStart.bind(this);
      this._onTouchMoveBound = this._onTouchMove.bind(this);
      this._onTouchEndBound = this._onTouchEnd.bind(this);

      this._onMouseDownBound = this._onMouseDown.bind(this);
      this._onMouseMoveBound = this._onMouseMove.bind(this);
      this._onMouseUpBound = this._onMouseUp.bind(this);

      this._onScrollBound = this._onScroll.bind(this);

      this._onFocusBound = this._onFocus.bind(this);
      this._onBlurBound = this._onBlur.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {

      window.addEventListener('resize', this._onResizeBound);
      window.addEventListener('orientationchange', this._onResizeBound);

      this.tabListEl.addEventListener('click', this._onTabListClickBound);

      this.tabListEl.addEventListener('touchstart', this._onTouchStartBound);
      this.tabListEl.addEventListener('mousedown', this._onMouseDownBound);
      this.tabListEl.addEventListener('mousewheel', this._onScrollBound);
      this.tabListEl.addEventListener('DOMMouseScroll', this._onScrollBound);

      this.tabListEl.addEventListener('focus', this._onFocusBound, true);
      this.tabListEl.addEventListener('blur', this._onBlurBound, true);

      if (this.leftEl) {
        this.leftEl.addEventListener('click', this._onLeftClickBound);
      }

      if (this.rightEl) {
        this.rightEl.addEventListener('click', this._onRightClickBound);
      }
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {

      window.removeEventListener('resize', this._onResizeBound);
      window.removeEventListener('orientationchange', this._onResizeBound);

      this.tabListEl.removeEventListener('click', this._onTabListClickBound);

      this.tabListEl.removeEventListener('touchstart', this._onTouchStartBound);
      this.tabListEl.removeEventListener('mousedown', this._onMouseDownBound);
      this.tabListEl.removeEventListener('mousewheel', this._onScrollBound);
      this.tabListEl.removeEventListener('DOMMouseScroll', this._onScrollBound);

      this.tabListEl.removeEventListener('focus', this._onFocusBound);
      this.tabListEl.removeEventListener('blur', this._onBlurBound);

      if (this.leftEl) {
        this.leftEl.removeEventListener('click', this._onLeftClickBound);
      }

      if (this.rightEl) {
        this.rightEl.removeEventListener('click', this._onRightClickBound);
      }
    },


    /**
     * Add event listeners for touchmove, touchend, mousemove and mouseup.
     * We add these to the window so that the user can move off of the element
     * but keep dragging the tabs.
     * @param {String} type Which type of listeners to add
     */
    _addMoveEventListeners: function(type) {

      // Only listen for events of the type we asked for.
      switch (type) {
        case 'mouse':
          window.addEventListener('mousemove', this._onMouseMoveBound);
          window.addEventListener('mouseup', this._onMouseUpBound);
          break;
        case 'touch':
          window.addEventListener('touchmove', this._onTouchMoveBound);
          window.addEventListener('touchend', this._onTouchEndBound);
          break;
      }
    },


    /**
     * Remove event listeners for move events.
     * @param {String} type Which type of listeners to remove
     */
    _removeMoveEventListeners: function(type) {

      // Only unbind events of the type we asked for.
      switch (type) {
        case 'mouse':
          window.removeEventListener('mousemove', this._onMouseMoveBound);
          window.removeEventListener('mouseup', this._onMouseUpBound);
          break;
        case 'touch':
          window.removeEventListener('touchmove', this._onTouchMoveBound);
          window.removeEventListener('touchend', this._onTouchEndBound);
          break;
      }
    },


    /**
     * Determine which size class to set on the element. This is a way of using breakpoint-like
     * logic for the tabs. We can't rely on real breakpoints because there is no guarantee that
     * the tabs will be the width of the window.
     * Also determine if we should be showing navigation arrows.
     */
    _determineSize: function() {

      var width = this.el.clientWidth;
      var breakpoint = this._getBreakpoint(width, this.breakpoints);

      // If the found breakpoint is different than the current breakpoint, set the proper state.
      if (this.currentBreakpoint !== breakpoint) {
        this._toggleClass(this.el, this.currentBreakpoint, false);
        this.currentBreakpoint = breakpoint;
        this._toggleClass(this.el, this.currentBreakpoint, true);
      }

      // If the tab list is wider than the scroll container, set the scrollable class.
      this.isScrollable = this.tabListEl.clientWidth > this.tabListScrollEl.clientWidth;
      this._toggleClass(this.navEl, 'scrollable', this.isScrollable);
      this._determineMinMax();
    },


    /**
     * Determine the min and max values for the slider.
     */
    _determineMinMax: function() {

      if (!this.tabListEl || !this.tabListScrollEl) {
        return;
      }

      this.maxX = 0;
      this.minX = this.tabListScrollEl.clientWidth - this.tabListEl.clientWidth - this.maxX;
    },


    /**
     * Enable the animation state.
     */
    _enableAnimation: function() {
      this.isAnimatable = true;
      this._toggleClass(this.navEl, 'no-animation', !this.isAnimatable);
    },


    /**
     * Disable the animation state.
     */
    _disableAnimation: function() {
      this.isAnimatable = false;
      this._toggleClass(this.navEl, 'no-animation', !this.isAnimatable);
    },


    /**
     * Update the position of the tabs.
     */
    _updatePosition: function() {
      this.tabListEl.setAttribute('style', transform('translate', this.x + 'px'));
    },


    /**
     * Check the x position
     */
    _checkX: function() {

      if (this.x < this.minX) {
        this.x = this.minX;
        this._updatePosition();
      }

      if (this.x > 0) {
        this.x = 0;
        this._updatePosition();
      }
    },


    /**
     * When the user clicks on a tab, make it active.
     * @param {Object} e
     */
    _onTabListClick: function(e) {

      // Make sure we haven't scrolled.
      if (this.scrollDistance > 5) {
        e.preventDefault();
        return;
      }

      var tab;

      // Find if one of our tab elements is in the path
      if ((tab = this._findTabByChildElement(e.target))) {
        e.preventDefault();
        this.setActive(tab);
      }
    },


    /**
     * When the window resizes, determine the size we should be using for tabs.
     * @param {Object} e
     */
    _onResize: function() {
      this._determineSize();
      this.focus(this.activeTabEl);
    },


    /**
     * When the touchstart event fires, start the scrolling process
     * @param {Object} e
     */
    _onTouchStart: function(e) {

      if (!this.isScrollable) {
        return;
      }

      // Disable the animation class so we scroll smoothly
      this._disableAnimation();

      this.start({
        lastX: e.touches[0].clientX,
        lastY: e.touches[0].clientY,
        type: 'touch'
      });
    },


    /**
     * As the user continues moving the touch, determine
     * if we should move.
     * @param {Object} e
     */
    _onTouchMove: function(e) {

      var xDistance = e.touches[0].clientX - this.lastX;
      var yDistance = e.touches[0].clientY - this.lastY;

      // If we haven't yet determined a scroll direction
      if (!this.scrollDirection) {

        // Moving up and down
        if (Math.abs(yDistance) > Math.abs(xDistance)) {
          this.scrollDirection = 'ns';
        }
        // Moving side to side
        else {
          this.scrollDirection = 'ew';
        }
      }

      // If We're moving left to right, start the move.
      if (this.scrollDirection === 'ew') {

        e.preventDefault();

        this.move({
          x: xDistance
        });
      }

      this.lastX = e.touches[0].clientX;
      this.lastY = e.touches[0].clientY;
    },


    /**
     * When the touch is over.
     * @param {Object} e
     */
    _onTouchEnd: function() {

      // Enable the animation class
      this._enableAnimation();

      // Stop after one frame so that animation is fully reenabled
      window.setTimeout(function() {
        this.stop({
          type: 'touch'
        });
      }.bind(this), 1);
    },


    /**
     * When the mousedown event fires, start the scrolling process
     * @param {Object} e
     */
    _onMouseDown: function(e) {

      if (!this.isScrollable) {
        return;
      }

      // Disable the animation class so we scroll smoothly
      this._disableAnimation();

      this.start({
        lastX: e.clientX,
        lastY: e.clientY,
        type: 'mouse'
      });
    },


    /**
     * As the user continues moving the mouse, determine
     * if we should move.
     * @param {Object} e
     */
    _onMouseMove: function(e) {


      var xDistance = e.clientX - this.lastX;
      var yDistance = e.clientY - this.lastY;

      // If we haven't yet determined a scroll direction
      if (!this.scrollDirection) {

        // Moving up and down
        if (Math.abs(yDistance) > Math.abs(xDistance)) {
          this.scrollDirection = 'ns';
        }
        // Moving side to side
        else {
          this.scrollDirection = 'ew';
        }
      }

      // If We're moving left to right, start the move.
      if (this.scrollDirection === 'ew') {

        e.preventDefault();

        this.move({
          x: xDistance
        });
      }

      this.lastX = e.clientX;
      this.lastY = e.clientY;
    },


    /**
     * When the mouse move is complete.
     * @param {Object} e
     */
    _onMouseUp: function() {

      // If we haven't been dragging, get outta here!
      if (!this.isDragging) {
        return;
      }

      // Enable the animation class
      this._enableAnimation();

      // Stop after one frame so that animation is fully reenabled
      window.setTimeout(function() {
        this.stop({
          type: 'mouse'
        });
      }.bind(this), 1);
    },


    /**
     * When the user scrolls horizontally on the tabs, slide.
     * @param {Object} e
     */
    _onScroll: function(e) {

      // Don't bother if we aren't scrollable
      if (!this.isScrollable) {
        return;
      }

      // Disable the animation class so we scroll smoothly
      this._disableAnimation();

      // Allow for Firefox's wheel detail
      var val = e.wheelDeltaX || (-e.detail * 40);

      // If the scroll has moved...
      if (val) {

        // Supress native
        e.preventDefault();

        // Move us to the new position
        this.move({
          x: val,
          scroll: true
        });
      }

      // Cancel an existing scroll timer
      if (this.scrollTimer) {
        window.clearTimeout(this.scrollTimer);
        this.scrollTimer = null;
      }

      // The scroll is considered "done" after 100ms
      this.scrollTimer = window.setTimeout(this._onScrollEnd.bind(this), 100);
    },


    /**
     * When the scrolling ends, reset the scrollTop
     */
    _onScrollEnd: function() {

      // Enable the animation class
      this._enableAnimation();

      // Stop after one frame so that animation is fully reenabled
      window.setTimeout(function() {
        this.stop({
          type: 'scroll'
        });
      }.bind(this), 1);
    },


    /**
     * When the left button is clicked, slide the tabs to the right.
     */
    _onLeftClick: function() {
      this.move({
        x: this.tabListScrollEl.clientWidth,
        force: true
      });
      this.stop({
        type: 'force'
      });
    },


    /**
     * When the right button is clicked, slide the tabs to the left.
     */
    _onRightClick: function() {
      this.move({
        x: -this.tabListScrollEl.clientWidth,
        force: true
      });
      this.stop({
        type: 'force'
      });
    },


    /**
     * When focus is gained on a tab.
     * @param {Object} e
     */
    _onFocus: function(e) {
      var target = e.target || e.srcElement;
      var parent = this._getElementMatchingParent(target, '.spark-tabs__tab', this.tabListEl);
      if (parent) this._addClass(parent, 'focus');
    },


    /**
     * When focus is lost on a tab.
     * @param {Object} e
     */
    _onBlur: function(e) {
      var target = e.target || e.srcElement;
      var parent = this._getElementMatchingParent(target, '.spark-tabs__tab', this.tabListEl);
      if (parent) this._removeClass(parent, 'focus');
    }
  };

  Base.exportjQuery(Tabs, 'Tabs');

  return Tabs;
}));

/**
 * # ToggleSwitch
 * ToggleSwitch and collapse an element.
 *
 * @example
 * new ToggleSwitch(el);
 *
 * @module components/ToggleSwitch.js
 */

/**
 * @ignore
 */
(function(root, factory) {

    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['spark/components/base'], factory);
    } else if (typeof module === 'object' && module.exports) {
        module.exports = factory(require('./base'));
    } else {
        root.Spark = root.Spark || {};
        root.Spark.ToggleSwitch = factory(root.Spark.BaseComponent);
    }
}(this, function(Base) {

    'use strict';

    /**
     * ToggleSwitch constructor.
     * @param {Element} el
     * @param {Object} params
     */
    var ToggleSwitch = function(el) {

        if (!el) {
            return;
        }

        this._cacheElements(el);
        this._bindEventListenerCallbacks();
        this._addEventListeners();
    };

    ToggleSwitch.prototype = {

        /**
         * Include common functionality.
         */
        _getElementMatchingParent: Base.getElementMatchingParent,
        _getMatchingChild: Base.getMatchingChild,
        remove: Base.remove,


        /**
         * Default values for internal properties we will be setting.
         * These are set on each construction so we don't leak properties
         * into the prototype chain.
         * @type {Object}
         */
        defaults: {
            el: null,
            activate: null,
            deactivate: null,
            toggle: null,
            input: null,
            _onKeydownBound: null
        },

        /**
         * Toggle the ToggleSwitch state.
         */
        toggle: function() {
            // Check if input exists
            if (!this.input) return;

            // toggle checked property
            if (this.input.checked)
                this.deactivate();
            else
                this.activate();
        },

        /**
         * Activate toggle state
         */
        activate: function() {
            // Check if input exists
            if (!this.input) return;
            this.input.checked = true;
        },

        /**
         * Deactivate toggle state
         */
        deactivate: function() {
            // Check if input exists
            if (!this.input) return;
            this.input.checked = false;
        },

        /**
         * Store a reference to the element.
         * @param {Element} el
         */
        _cacheElements: function(el) {
            this.el = el;
            this.input = el.querySelector('.spark-toggle__input');
        },


        /**
         * Create bound versions of event listener callbacks and store them.
         * Otherwise we can't unbind from these events later because the
         * function signatures won't match.
         */
        _bindEventListenerCallbacks: function() {
            this._onKeydownBound = this._onKeydown.bind(this);
        },


        /**
         * Add event listeners for DOM events.
         */
        _addEventListeners: function() {
            this.el.addEventListener('keydown', this._onKeydownBound);
        },


        /**
         * Remove event listeners for DOM events..
         */
        _removeEventListeners: function() {
            this.el.removeEventListener('keydown', this._onKeydownBound);
        },


        /**
         * When the space or enter key is pressed on the toggle, toggle!
         * @param {Object} e
         */
        _onKeydown: function(e) {

            if (!this._getElementMatchingParent(e.target, '.spark-toggle-switch, spark-toggle-switch__handle', this.el)) {
                return;
            }

            var code = e.keyCode || e.which;

            switch (code) {
                case 32:
                    // space
                    // Skip, native works as expected
                    break;
                case 13:
                    // enter
                    e.preventDefault();
                    this.toggle();
                    break;
                case 39:
                case 40:
                    // right
                    // down
                    e.preventDefault();
                    this.activate();
                    break;
                case 37:
                case 38:
                    // left
                    // up
                    e.preventDefault();
                    this.deactivate();
                    break;
            }

        }
    };

    Base.exportjQuery(ToggleSwitch, 'ToggleSwitch');

    return ToggleSwitch;
}));

/**
 * # Toolbar
 * Setup a responsive toolbar
 *
 * @example
 * new Toolbar(el);
 *
 * @module components/toolbar.js
 */

(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/debounce'], factory);
  }
  else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/debounce'));
  }
  else {
    root.Spark = root.Spark || {};
    root.Spark.Toolbar = factory(root.Spark.BaseComponent, root.Spark.debounce);
  }
}(this, function(Base, Debounce) {

  'use strict';

  /**
   * Toolbar constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Toolbar = function(el, params) {

    params = params || {};

    if (!el) {
      return;
    }
    this._init(el);
  };

  Toolbar.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _unsetParams: Base.unsetParams,
    _toggleClass: Base.toggleClass,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _debounce: Debounce,

    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: [],

    /**
     * Setup a toolbarItem Instance to track the state of individual toolbar items
     * @param {Element} parent Reference to parent toolbar
     * @param {Element} el Node to initalize as toolbarItem
     * @param {Number} order The original index of the item in list of toolbarItems (used for maintaining order when sorting)
     */
    toolbarItem: function(parent, el, order) {
      //Setup and cache the values for this item
      var a = {};
      a.parent = parent;
      a.el = el;
      //cache the priority value present on the toolbar element if it is present, else default to 0
      a.priority = a.el.attributes['data-priority'] ? a.el.attributes['data-priority'].value : 0;
      a.order = a.el.attributes['data-order'] ? a.el.attributes['data-order'].value : order;
      a.hasContent = a.el.querySelector('.spark-toolbar__item--content') ? true : false;
      a.helper = a.el.querySelector('.spark-toolbar__item-helper');
      a.label = a.el.attributes.label ? a.el.attributes.label.value : false;
      a.closeOnClick = a.parent._hasClass(a.el, 'spark-toolbar__item--close-more-on-click');
      a.width = a.el.offsetWidth;
      a.height = a.el.offsetHeight;
      a.dropdown = el.querySelector('.spark-toolbar__item--content');
      if (a.dropdown) {
        a.dropdown.sparktoolbardropdown = true;
      }
      /**
       * Call method to toggle the open state, optional param sets open state to value
       * Can get current state by referencing a.toggleDropdown.open
       * @param {Boolean} open Set state to this regardless of current state
       */
      a.toggleDropdown = function(open) {
        var o = typeof open !== 'undefined' ? !open : a.toggleDropdown.open;
        if (o) {
          a.toggleDropdown.open = false;
          a.parent._removeClass(a.el, 'animate');
          window.setTimeout(function() {
            a.parent._removeClass(a.el, 'open');
          }, 100);
        }
        else {
          if (a.hasContent) {
            a.toggleDropdown.open = true;
            a.parent._addClass(a.el, 'open');
            a.positionDropdown();
            var e = document.createEvent('Event');
            e.initEvent('spark.visible-children', true, true);
            a.dropdown.dispatchEvent(e);
            window.setTimeout(function() {
              a.parent._addClass(a.el, 'animate');
            }, 0);
          }
          else {
            a.parent._toggleShowMore(false);
          }
        }
      };
      /**
       * Click handler for local element - determines to close element
       * conditionally based on presence of spark-toolbar__item--close-on-click
       * closes parent's more dropdown conditionally as well
       * @param {Boolean} open Set state to this regardless of current state
       */
      a.handleClick = function(e) {
        if (!a.toggleDropdown.open) {
          a.toggleDropdown(true);
        }
        else {
          if (e.target === a.el || e.target === a.helper) {
            a.toggleDropdown();
          }
          else {
            var b = e.target;
            while (b !== a.el) {
              if (a.parent._hasClass(b, 'spark-toolbar__item--close-on-click')) {
                a.toggleDropdown(false);
                //close the mode section, as event originated inside a close-on-click area
                a.parent._toggleShowMore(false);
                break;
              }
              b = b.parentElement;
            }
          }
        }
        //e.preventDefault();
      };
      //perform bounds checking on dropdown open to position dropdown inside visual area
      //this is called each time a dropdown is opened, in case the state of the component has
      //changed since initialization
      a.positionDropdown = function() {
        if (a.dropdown) {
          a.dropdown.style.left = '';
          a.dropdown.style.right = '';
          var pos = a.dropdown.getBoundingClientRect();
          var left = window.pageXOffset;
          var right = window.pageXOffset + document.documentElement.clientWidth;
          if (pos.right > right) {
            a.dropdown.style.left = 'inherit';
            a.dropdown.style.right = 0;
          }
          if (pos.left < left) {
            a.dropdown.style.left = 0;
            a.dropdown.style.right = 'inherit';
          }
        }
      };
      a.remove = function() {
        if (a.el) {
          delete a.el.sparktoolbar;
        }
        if (a.dropdown) {
          delete a.dropdown.sparktoolbardropdown;
        }
      };
      a.el.sparktoolbar = a;
      return a;
    },
    /**
     * Close any open items, and more dropdown
     */
    _closeAll: function() {
      this._closeItems();
      this._toggleShowMore(false);
    },
    /**
     * Returns array of open toolbarItems
     */
    _getOpenItems: function() {
      var a = [];
      for (var i = 0; i < this.items.length; i++) {
        if (this.items[i].toggleDropdown.open) {
          a.push(this.items[i]);
        }
      }
      return a;
    },
    /**
     * Close any open items
     * @param {Array} a Optional array of toolbarItems to close, defaults to all open items
     */
    _closeItems: function(a) {
      a = typeof a === 'undefined' ? this._getOpenItems() : a;
      for (var i = 0; i < a.length; i++) {
        a[i].toggleDropdown(false);
      }
    },
    /**
     * Setup the toolbar element, cache properties, and initalize styling
     * when complete, show toolbar
     * @param {Element} el The node to initalize on
     */
    _init: function(el) {
      this.el = el;
      //store a reference to this on the node to expedite event handling
      this.el.sparktoolbarcon = this;
      this.visibleContainer = this.el.querySelector('.spark-toolbar__container--visible');
      this.hiddenContainer = this.el.querySelector('.spark-toolbar__container--hidden');
      this.showMoreButton = this.el.querySelector('.spark-toolbar__show-more');
      this.showMoreButton.sparktoolbarshowmore = true;
      this.isOpen = false;
      this.isFocus = false;
      this._setupListeners();
      this.el.style.width = '100%';
      this._initItems();
      this._addClass(this.el, 'measured');
      this._calculateStyles();
      this.tabindex = this.el.attributes.tabindex ? this.el.attributes.tabindex.value : 0;
      this._addClass(this.el, 'ready');
    },
    _initItems: function() {
      var items = this.el.querySelectorAll('.spark-toolbar__item');
      this.items = [];
      for (var i = 0; i < items.length; i++) {
        this.items[i] = new this.toolbarItem(this, items[i], i);
      }
    },
    /**
     * Setup event listeners for clicks and resize events
     */
    _setupListeners: function() {
      this._handleWindowClick = this._handleWindowClickH.bind(this);
      document.addEventListener('click', this._handleWindowClick);
      this._handleResize = this._debounce(this._handleResizeH.bind(this), 100);
      window.addEventListener('resize', this._handleResize);
      this._handleKeyDown = this._handleKeyDownH.bind(this);
      this.el.addEventListener('keydown', this._handleKeyDown);
      this._handleFocus = this._handleFocusH.bind(this);
      document.addEventListener('focus', this._handleFocus, true);
      this._handleBlur = this._handleBlurH.bind(this);
      document.addEventListener('blur', this._handleBlur, true);
      this._handleVisibleChildren = this._handleVisibleChildrenH.bind(this);
      document.addEventListener('spark.visible-children', this._handleVisibleChildren, true);
    },
    /**
     * Remove event listeners for clicks and resize events
     */
    _removeListeners: function() {
      document.removeEventListener('click', this._handleWindowClick);
      window.removeEventListener('resize', this._handleResize);
      this.el.removeEventListener('keydown', this._handleKeyDown);
      document.removeEventListener('blur', this._handleBlur, true);
      document.removeEventListener('focus', this._handleFocus, true);
      document.removeEventListener('spark.visible-children', this._handleVisibleChildren, true);
    },
    /**
     * Remove the element from the DOM and prepare for garbage collection by dereferencing values.
     * @param {Boolean} leaveElement Leave the element intact.
     */
    remove: function(leaveElement) {
      this._removeListeners();
      delete this.el.sparktoolbarcon;
      delete this.showMoreButton.sparktoolbarshowmore;
      for(var i = 0; i < this.items.length; i++) {
        this.items[i].remove();
      }
      Base.remove.call(this, leaveElement);
    },
    /**
     * reset our tab index when user focuses outside of element (gets immediately reset to -1 if focus is placed back inside element)
     * @param {Event} e The FocusEvent
     */
    _handleBlurH: function(e) {
      if (this.el.contains(e.target)) {
        this.el.attributes.tabindex.value = this.tabindex;
      }
    },
    /**
     * focus handler, works in conjunction with blur handler to set correct tabindex value
     * @param {Event} e The FocusEvent
     */
    _handleFocusH: function(e) {
      //if we're not being focused, reset our tabindex so we are accessible again, and close anything open
      if (!this.el.contains(e.target)) {
        this._closeAll();
        this.el.attributes.tabindex.value = this.tabindex;
      }
      else {
        //set our tabindex to -1 so the user can shift-tab out of our element
        this.el.attributes.tabindex.value = -1;
        if (e.target.sparktoolbarcon) {
          this._focusLast();
          return;
        }
        //handle focusing an item
        if (e.target.sparktoolbar) {
          e.target.sparktoolbar.el.focus();
          return;
        }
        var a = e.target;
        //harder case - look up the tree to find if we're focusing inside content
        while (!a.sparktoolbarcon) {
          if (a.sparktoolbar) {
            break;
          }
          //if we are - give our parent element a tabindex so the user can refocus the menu using shift-tab
          if (a.sparktoolbardropdown) {
            this.el.attributes.tabindex.value = this.tabindex;
            return;
          }
          a = a.parentElement;
        }
      }
    },
    /**
     * reset our focus to the last menu item that was focused
     */
    _focusLast: function() {
      if (!this._lastFocus) {
        var a = this.visibleContainer.querySelector('.spark-toolbar__item') || this.hiddenContainer.querySelector('.spark-toolbar__item');
        this._lastFocus = a.sparktoolbar;
      }
      if (this.hiddenContainer.contains(this._lastFocus.el)) {
        this._toggleShowMore(true);
      }
      this._lastFocus.el.focus();
    },
    /**
     * keydown handler, used for keyboard navigation
     * @param {Event} e The KeyDown Event
     */
    _handleKeyDownH: function(e) {
      var a = e.target;
      //find the nearest toolbaritem
      while (!a.sparktoolbarcon) {
        if (a.sparktoolbar) {
          break;
        }
        if (a.sparktoolbardropdown) {
          return;
        }
        a = a.parentElement;
      }
      if (a.sparktoolbar) {
        //handle keys
        switch (e.keyCode) {
          //left arrow
          case 37:
            //up arrow
          case 38:
            if (a.previousSibling && a.previousSibling.sparktoolbar) {
              this._lastFocus = a.previousSibling.sparktoolbar;
              a.previousSibling.focus();
            }
            else {
              if (this.visibleContainer.querySelector('.spark-toolbar__item') !== a.sparktoolbar.el) {
                a = this.visibleContainer.querySelector('.spark-toolbar__item:last-of-type');
                if (a) {
                  this._toggleShowMore(false);
                  this._lastFocus = a.sparktoolbar;
                  a.focus();
                }
              }
            }
            this._closeItems();
            e.preventDefault();
            break;
            //right arrow
          case 39:
            //down arrow
          case 40:
            if (a.nextSibling && a.nextSibling.sparktoolbar) {
              this._lastFocus = a.nextSibling.sparktoolbar;
              a.nextSibling.focus();
            }
            else {
              if (this.hiddenContainer.querySelector('.spark-toolbar__item:last-of-type') !== a.sparktoolbar.el) {
                a = this.hiddenContainer.querySelector('.spark-toolbar__item');
                if (a) {
                  this._toggleShowMore(true);
                  this._lastFocus = a.sparktoolbar;
                  a.focus();
                }
              }
            }
            this._closeItems();
            e.preventDefault();
            break;
            //spacebar
          case 32:
            e.preventDefault();
            //we only want to toggle the toolbar if we are actually focused directly on it;
            if (e.target.sparktoolbar) {
              e.target.sparktoolbar.el.click();
            }
            break;
            //enter
          case 13:
            //we only want to toggle the toolbar if we are actually focused directly on it;
            if (e.target.sparktoolbar) {
              e.target.sparktoolbar.el.click();
            }
            break;
        }
      }
    },
    /**
     * Hanldes the spark.visible-children event to resize the component when it is made visible.
     * @param {Event} e The spark.visible-children event
     */
    _handleVisibleChildrenH: function(e) {
      if(e.target.contains(this.el)) {
        window.setTimeout(function() {
          this.change();
        }.bind(this),0);
      }
    },
    /**
     * Event handler for click events, handles window clicks, control element clicks,
     * and forwards events to toolbarItem click handlers as needed
     * @param {Event} e The click event
     */
    _handleWindowClickH: function(e) {
      //Check to see if the click was outside of the toolbar
      if (!this.el.contains(e.target)) {
        this._closeItems();
        this._toggleShowMore(false);
      }
      else {
        var a = e.target;
        //traverse the dom node tree until we find an element that handles the event,
        //or we reach the toolbar root node
        if(a === this.visibleContainer || a === this.el) {
          e.stopPropagation();
          e.preventDefault();
          return;
        }
        while (a !== this.el) {
          if (a.sparktoolbar) {
            var c = this._getOpenItems();
            if (c.indexOf(a.sparktoolbar) >= 0) {
              c.splice(c.indexOf(a.sparktoolbar), 1);
            }
            this._closeItems(c);
            if (!this.hiddenContainer.contains(e.target)) {
              this._toggleShowMore(false);
            }
            return a.sparktoolbar.handleClick(e);
          }
          if (a.sparktoolbarshowmore) {
            this._closeItems();
            this._toggleShowMore();
            return;
          }
          a = a.parentElement;
        }
        this._closeAll();
      }
    },
    /**
     * Toggle the state of the show more dropdown, optional parameter overrides toggle and
     * sets state to passed value
     * @param {Boolean} open The new state of the show more dropdown
     */
    _toggleShowMore: function(open) {
      var o = typeof open !== 'undefined' ? !open : this.isOpen;
      if (o) {
        this._removeClass(this.el, 'animate');
        window.setTimeout(function() {
          this._removeClass(this.el, 'open');
          this.isOpen = false;
        }.bind(this), 100);
      }
      else {
        this.isOpen = true;
        this._addClass(this.el, 'open');
        this._positionShowMore();
        window.setTimeout(function() {
          this._addClass(this.el, 'animate');
        }.bind(this), 0);
      }
    },
    /**
    * Do bounds checking on show-more dropdown when it is opened, and position it accordingly
    */
    _positionShowMore: function() {
      this.hiddenContainer.style.right = '0px';
      var pos = this.hiddenContainer.getBoundingClientRect();
      var left = window.pageXOffset;
      var right = window.pageXOffset + document.documentElement.clientWidth;
      if (pos.right > right) {
        this.hiddenContainer.style.right = 'calc(' + (pos.right - right) + 'px + 1rem)';
      }
      if (pos.left < left) {
        this.hiddenContainer.style.right = 'calc(' + (pos.left - left) + 'px - 1rem)';
      }
    },
    /**
     * Resize event helper, closes items then triggers recalculation of styles
     */
    _handleResizeH: function() {
      this._closeAll();
      this._calculateStyles();
    },
    /**
     * Reevaluates the available area of the toolbar and places toolbarItems into
     * the hidden container, as necessary. Should not call with any specified value
     * for showMore (used internally)
     * @param {Boolean} showMore Used to conditionally evaluate styling when showMore area is used
     */
    _calculateStyles: function(showMore) {
      this.el.style.width = '100%';
      showMore = typeof showMore !== 'undefined' ? showMore : false;
      if (!showMore) {
        this._removeClass(this.el, 'show-more');
      }
      var visible = [];
      var hidden = [];
      var i;
      //sort items by their priority to ensure higher-priority items are always placed
      //into the visible area first
      this.items.sort(this._prioritySort);
      //get container width and start placing items into their containers
      var visibleWidth = this.visibleContainer.clientWidth;
      for (i = 0; i < this.items.length; i++) {
        if (visibleWidth - this.items[i].width >= 0) {
          visible.push(this.items[i]);
          visibleWidth -= this.items[i].width;
        }
        else {
          if (!showMore) {
            this._addClass(this.el, 'show-more');
            return this._calculateStyles(true);
          }
          hidden.push(this.items[i]);
        }
      }
      //sort items back into their original order before inserting them into the document
      visible.sort(this._orderSort);
      hidden.sort(this._orderSort);
      var v = document.createDocumentFragment();
      var h = document.createDocumentFragment();
      for (i = 0; i < visible.length; i++) {
        v.appendChild(visible[i].el);
      }
      for (i = 0; i < hidden.length; i++) {
        h.appendChild(hidden[i].el);
      }
      this.visibleContainer.appendChild(v);
      this.hiddenContainer.appendChild(h);
      this.el.style.width = '';
    },
    /**
     * Sorts toolbar items in descending order based on their priority value
     */
    _prioritySort: function(l, r) {
      return r.priority - l.priority;
    },
    /**
     * Sorts toolbar items in ascending order based on their order value
     */
    _orderSort: function(l, r) {
      return l.order - r.order;
    },
    /**
     * This function will update cached sizing when an element in the toolbar is changed
     * or, when toolbar items are added or removed
     */
    change: function() {
      this._closeAll();
      this._removeClass(this.el, ['ready', 'show-more', 'measured']);
      var v = document.createDocumentFragment();
      for (var i = 0; i < this.items.length; i++) {
        v.appendChild(this.items[i].el);
      }
      this.visibleContainer.appendChild(v);
      this._initItems();
      this._addClass(this.el, 'measured');
      this._calculateStyles();
      this._addClass(this.el, 'ready');
    }
  };

  Base.exportjQuery(Toolbar, 'Toolbar');
  return Toolbar;
}));

/**
 * # Carousel
 * Create a Carousel
 *
 * @example
 * new Carousel(el);
 *
 * @module components/carousel.js
 */

(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/debounce', 'spark/helpers/transform'], factory);
  }
  else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/debounce'), require('../helpers/transform'));
  }
  else {
    root.Spark = root.Spark || {};
    root.Spark.Carousel = factory(root.Spark.BaseComponent, root.Spark.debounce, root.Spark.transform);
  }
}(this, function(Base, Debounce, Transform) {

  'use strict';

  /**
   * Carousel constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var Carousel = function(el, params) {

    params = params || {};

    if (!el) {
      return;
    }
    this._init(el);

  };

  Carousel.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _getElementMatchingParent: Base.getElementMatchingParent,
    _debounce: Debounce,
    _transform: Transform,
    removeB: Base.remove,


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: [],

    /**
     * Carousel Item Contructor, exposes access to functions setTransform, setSelected, and currentPosition
     * @param {Element} el Node to initalize as carouselItem
     * @param {Object} parent reference to the parent carousel
     * @param {Element} dot Node to use as dot element
     * @param {Number} order original order in the markup
     */
    _carouselItem: function(el, parent, dot, order) {
      var a = {};
      a.el = el;
      a.parent = parent;
      a.dot = dot;
      a.order = order;
      a.addTransform = function(x) {
        a.setTransform(a.transform.x + x);
      };
      a.setTransform = function(x) {
        x = x ? x : 0;
        a.transform = {
          'x': x
        };
        a.el.setAttribute('style', a.parent._transform('translate3d', a.transform.x + 'px, 0px, 0px'));
      };
      a.setSelected = function(b) {
        if (a.setSelected.selected === b && typeof a.setSelected.selected !== 'undefined') {
          return;
        }
        if (b) {
          a.parent._addClass(a.el, 'selected');
          a.parent._addClass(a.dot, 'selected');
          a.setSelected.selected = true;
        }
        else {
          a.parent._removeClass(a.el, 'selected');
          a.parent._removeClass(a.dot, 'selected');
          a.setSelected.selected = false;
        }
      };
      a.currentPosition = function() {
        return parent.transform.x + a.transform.x + a.dims.left + (a.width / 2) - parent.dims.left;
      };
      a.setTransform();
      a.dims = el.getBoundingClientRect();
      a.width = a.dims.width;
      a.el.sparkcarouselitem = a;
      a.dot.sparkcarouselitemdot = a;
      return a;
    },
    /**
     * Scans element and sets up or resets the carousel based on configuration
     * @param {Element} el Node to initalize as the carousel
     */
    _init: function(el) {
      //cache elements and controls
      this.el = el;
      this.el.sparkcarousel = this;
      this.backe = this.el.querySelector('.spark-carousel__back');
      this.forwarde = this.el.querySelector('.spark-carousel__forward');
      this.outerContainer = this.el.querySelector('.spark-carousel__outer-container');
      this.containerMask = this.el.querySelector('.spark-carousel__container-mask');
      this.container = this.el.querySelector('.spark-carousel__container');
      this.dots = this.el.querySelector('.spark-carousel__dots');
      this.pauseEl = this.el.querySelector('.spark-carousel__pause');
      //get the options from the element
      this.opts = {};
      this.opts.wrapItems = this.el.attributes['data-spark-carousel-wrap-items'] ? true : false;
      this.opts.startingVelocity = this.el.attributes['data-spark-carousel-scroll-velocity'] ? this.el.attributes['data-spark-carousel-scroll-velocity'].value : 10;
      this.opts.smoothScroll = this.el.attributes['data-spark-carousel-smooth-scroll'] ? true : false;
      this.opts.smoothScrollCenterItems = this.el.attributes['data-spark-carousel-smooth-scroll-center'] ? true : false;
      this.opts.panelScroll = this.el.attributes['data-spark-carousel-panel'] ? true : false;
      this.opts.edgeScroll = this.opts.panelScroll || this.el.attributes['data-spark-carousel-edge'] ? true : false;
      this.opts.autoAdvance = this.el.attributes['data-spark-carousel-auto-advance'] ? this.el.attributes['data-spark-carousel-auto-advance'].value : false;
      //setup autoAdvance
      if (this.opts.autoAdvance && !this.autoAdvance && !this.pause) {
        this.autoAdvance = window.setTimeout(function() {
          this._autoAdvance();
        }.bind(this), this.opts.autoAdvance * 1000);
      }
      //conditionally bind pause handlers.
      //needs to be done here so that resetting the carousel will work correctly
      if (this.opts.autoAdvance && !this.pauseH) {
        this.pauseH = this._pause.bind(this);
        this.pauseEl.addEventListener('click', this.pauseH);
      }
      //collect items and cache sizing
      var a = this.el.querySelectorAll('.spark-carousel__item');
      this.items = [];
      this.totalItemWidth = 0;
      this.dims = this.containerMask.getBoundingClientRect();
      this.width = this.dims.width;
      this.height = this.dims.height;
      var dots = document.createDocumentFragment();
      var b;
      //create our carouselItems
      for (var i = 0; i < a.length; i++) {
        b = document.createElement('div');
        dots.appendChild(b);
        this.items.push(new this._carouselItem(a[i], this, b, i));
        this.totalItemWidth += this.items[i].width;
      }
      //if we're resetting we need to empty out the exisiting elements first
      while (this.dots.firstChild) {
        this.dots.removeChild(this.dots.firstChild);
      }
      this.dots.appendChild(dots);
      //this is to test if we're dealing with 2011 flexbox (IE10) and need to do an adjustment
      //this is because ms-flex-pack: center doesn't work like 2012 flexbox center-pack
      if (typeof this.container.style.msFlexAlign !== 'undefined') {
        this._transformItems(-(this.totalItemWidth - this.width) / 2);
      }
      //setup inital transform
      this._setTransform();
      //need to bind this event handler here as we are always going to need to be listening
      //for this event, in order to react to container visibility changing
      this._handleVisibleChildren = this._handleVisibleChildrenH.bind(this);
      document.addEventListener('spark.visible-children', this._handleVisibleChildren, true);
      if (!this._rafHandler) {
        this._rafHandler = this._rafHandlerH.bind(this);
      }
      //need to also listen to resize events, even if we don't have items overflowing
      if (!this.resizeH) {
        this.resizeH = this._debounce(this._resize.bind(this), 100);
        window.addEventListener('resize', this.resizeH);
      }
      //if we haven't already init'd event listerers, and we have items overflowing
      if (this.totalItemWidth > this.width) {
        if (!this.touchstartH) {
          this._removeClass(this.el, 'noscroll');
          this._setupListeners();
          //center the first item
          this._addTransform(-this.items[0].currentPosition() + (this.width / 2));
        }
      }
      //if we don't have overflowing items, then disable scrolling and remove listeners
      else {
        this._addClass(this.el, 'noscroll');
        this._removeListeners();
      }
      //set the new selected item
      this._updateSelected();
      //finally, display the element
      this._addClass(this.el, 'ready');
    },
    /**
     * Tears down the component, removes listeners, and conditionally delete the DOM element
     * @param {Boolean} leaveElement Falsey value will remove the DOM element as well as the component instance
     */
    remove: function(leaveElement) {
      window.removeEventListener('resize', this.resizeH);
      delete this.resizeH;
      document.removeEventListener('spark.visible-children', this._handleVisibleChildren, true);
      delete this._handleVisibleChildren;
      this.removeB(leaveElement);
    },
    /**
     * Pause/unpause the autoAdvance feature
     */
    _pause: function() {
      if (this.pause) {
        delete this.moves;
        this.autoAdvance = window.setTimeout(function() {
          this._autoAdvance();
        }.bind(this), this.opts.autoAdvance * 1000);
      }
      else {
        this._rafCancel();
        if (!this.opts.smoothScroll || this.opts.smoothScrollCenterItems) {
          this._scrollTo(this._selectedItem());
        }
        window.clearTimeout(this.autoAdvance);
        delete this.autoAdvance;
      }
      this._setPause(!this.pause);
    },
    /**
     * toggles the pause class on the element
     * @param {Boolean} b Truthy value will give the element the pause class
     */
    _setPause: function(b) {
      if (typeof b === 'undefined') {
        this.pause = typeof this.pause === 'undefined' ? false : this.pause;
        return this.pause;
      }
      else {
        if (b) {
          this._addClass(this.el, 'pause');
        }
        else {
          this._removeClass(this.el, 'pause');
        }
        this.pause = b;
        return this.pause;
      }
    },
    /**
     * function called by window.setTimeout, will check first to see if element is in use before triggering a slide advance
     */
    _autoAdvance: function() {
      if (!this.moves && !this._laststart && !this.paused) {
        this._rafCancel();
        var a = this.items.indexOf(this._selectedItem());
        this._scrollToItem = true;
        this._scrollTo(this.items[a === this.items.length - 1 ? 0 : a + 1]);
        this.autoAdvance = window.setTimeout(function() {
          this._autoAdvance();
        }.bind(this), this.opts.autoAdvance * 1000);
      }
    },
    /**
     * initalize and bind even listeners
     */
    _setupListeners: function() {
      this.touchstartH = this._touchstart.bind(this);
      this.container.addEventListener('touchstart', this.touchstartH);
      this.touchmoveH = this._touchmove.bind(this);
      window.addEventListener('touchmove', this.touchmoveH);
      this.touchendH = this._touchend.bind(this);
      window.addEventListener('touchend', this.touchendH);
      this.mousedownH = this._mousedown.bind(this);
      this.container.addEventListener('mousedown', this.mousedownH);
      this.mousemoveH = this._mousemove.bind(this);
      window.addEventListener('mousemove', this.mousemoveH);
      this.mouseupH = this._mouseup.bind(this);
      window.addEventListener('mouseup', this.mouseupH);
      this.forwardH = this._forward.bind(this);
      this.forwarde.addEventListener('click', this.forwardH);
      this.backH = this._back.bind(this);
      this.backe.addEventListener('click', this.backH);
      this.clickH = this._click.bind(this);
      this.el.addEventListener('click', this.clickH);
      this._focusHandler = this._scrollToClicked.bind(this);
      this.container.addEventListener('focus', this._focusHandler, true);
    },
    /**
     * Removes non-essential event listeners, called when tearing down the component, or our content
     * does not exceed the width of our element
     */
    _removeListeners: function() {
      this.el.removeEventListener('touchstart', this.touchstartH);
      delete this.touchstartH;
      window.removeEventListener('touchmove', this.touchmoveH);
      delete this.touchmoveH;
      window.removeEventListener('touchend', this.touchendH);
      delete this.touchendH;
      this.el.removeEventListener('mousedown', this.mousedownH);
      delete this.mousedownH;
      window.removeEventListener('mousemove', this.mousemoveH);
      delete this.mousemoveH;
      window.removeEventListener('mouseup', this.mouseupH);
      delete this.mouseupH;
      this.forwarde.removeEventListener('click', this.forwardH);
      delete this.forwardH;
      this.backe.removeEventListener('click', this.backH);
      delete this.backH;
      this.el.removeEventListener('click', this.clickH);
      delete this.clickH;
      this.container.removeEventListener('focus', this._focusHandler, true);
      delete this._focusHandler;
      if (this.pauseEl) {
        this.pauseEl.removeEventListener('click', this.pauseH);
        delete this.pauseH;
      }
    },
    /**
     * Event handler for the spark.visible-children event, just call the change function to handle any
     * visibility or sizing changes
     * @param {Event} e The spark.visible-children event
     */
    _handleVisibleChildrenH: function(e) {
      if (e.target.contains(this.el)) {
        window.setTimeout(function() {
          this.change();
        }.bind(this), 0);
      }
    },
    /**
     * Forward button click handler triggers a scrollTo to the "next" element
     * @param {Event} e The click event
     */
    _forward: function(e) {
      var s = this.items.indexOf(this._selectedItem());
      s++;
      if (s > this.items.length - 1) {
        if (this.opts.wrapItems) {
          s = 0;
        }
        else {
          s--;
        }
      }
      delete this.moves;
      var a = this._startingVelocity;
      a = a < -this.opts.startingVelocity ? a : -this.opts.startingVelocity;
      this._rafCancel();
      this._scrollToItem = true;
      this._scrollTo(this.items[s], a);
      if (e) {
        e.preventDefault();
      }
    },
    /**
     * Back button click handler triggers a scrollTo to the "previous" element
     * @param {Event} e The click event
     */
    _back: function(e) {
      var s = this.items.indexOf(this._selectedItem());
      s--;
      if (s < 0) {
        if (this.opts.wrapItems) {
          s = this.items.length - 1;
        }
        else {
          s++;
        }
      }
      delete this.moves;
      var a = this._startingVelocity;
      a = a > this.opts.startingVelocity ? a : this.opts.startingVelocity;
      this._rafCancel();
      this._scrollToItem = true;
      this._scrollTo(this.items[s], a);
      if (e) {
        e.preventDefault();
      }
    },
    /**
     * Move start handler, handles both touchstart and mousedown events
     * @param {Object} e The start event
     */
    _movestart: function(e) {
      this._rafCancel();
      this.moves = [];
      this.moves.push(e);
    },
    /**
     * Move handler, handles internal move event objects
     * @param {Object} e The move event
     */
    _move: function(e) {
      if (this.moves && this.moves.length > 1) {
        this._addTransform(e.pageX - this.moves[this.moves.length - 1].pageX);
        this.moves.push(e);
        e.preventDefault = true;
      }
      else {
        if (this.moves && this.moves[0]) {
          if (Math.abs(this.moves[0].pageX - e.pageX) > Math.abs(this.moves[0].pageY - e.pageY) && Math.abs(this.moves[0].pageX - e.pageX) > 5 && e.cancelable) {
            this._addTransform(e.pageX - this.moves[0].pageX);
            this.moves.push(e);
            e.preventDefault = true;
          }
          else {
            if (Math.abs(this.moves[0].pageX - e.pageX) > 5) {
              this.moves[0] = e;
            }
          }
        }
        if (e.type === 'touchend' || e.type === 'mouseup') {
          delete this.moves;
        }
      }
    },
    /**
     * Move end handler, handles both touchend and mouseup events
     * @param {Object} e The moveend event
     */
    _moveend: function(e) {
      this._move(e);
      if (!this.opts.smoothScroll) {
        this._settle(this.moves);
      }
      else {
        this._interiaScroll(this.moves);
      }
      return e;
    },
    /**
     * Resize event handler, calls change to handle any element dimension changes
     */
    _resize: function() {
      this.change();
    },
    /**
     * Calling the change function will handle updating the element to take into account
     * any styling, sizing, or visibility changes, and the addition or removal of any carouselItems
     */
    change: function() {
      var dims = this.el.getBoundingClientRect();
      if (dims.width !== this.width || dims.height !== this.height) {
        if (this.autoAdvance) {
          window.clearTimeout(this.autoAdvance);
          delete this.autoAdvance;
        }
        this._rafCancel();
        var c = this._selectedItem();
        this._setTransform(0);
        this._setTransformItems(0);
        this._init(this.el);
        if (this.items.indexOf(c.el.sparkcarouselitem) > -1 && this.totalItemWidth > this.width) {
          if (this.opts.wrapItems) {
            this._addTransform(-this.totalItemWidth + (-c.el.sparkcarouselitem.currentPosition() + (this.width / 2)));
          }
          else {
            this._addTransform(-c.el.sparkcarouselitem.currentPosition() + (this.width / 2));
          }
        }
      }
    },
    /**
     * Touchstart event handler, passes necessary data points to the movestart function
     * @param {Object} e The touchstart event
     */
    _touchstart: function(e) {
      var a = {
        'type': e.type,
        'pageX': e.touches[0].pageX,
        'pageY': e.touches[0].pageY,
        'timeStamp': e.timeStamp
      };
      this._movestart(a);
    },
    /**
     * Touchmove event handler, passes necessary data points to the move function
     * @param {Object} e The touchmove event
     */
    _touchmove: function(e) {
      var a = {
        'type': e.type,
        'pageX': e.touches[0].pageX,
        'pageY': e.touches[0].pageY,
        'timeStamp': e.timeStamp,
        'cancelable': e.cancelable
      };
      this._move(a);
      if (a.preventDefault) {
        e.preventDefault();
      }
    },
    /**
     * Touchend event handler, passes necessary data points to the moveend function
     * @param {Object} e The touchend event
     */
    _touchend: function(e) {
      if (this.moves && this.moves.length > 2 && e.cancelable) {
        var a = {
          'type': e.type,
          'pageX': this.moves[this.moves.length - 1].pageX,
          'pageY': this.moves[this.moves.length - 1].pageY,
          'timeStamp': e.timeStamp
        };
        this._moveend(a);
      } else {
        delete this.moves;
      }
    },
    /**
     * Mousedown event handler, passes necessary data points to the movestart function
     * @param {Object} e The mousedown event
     */
    _mousedown: function(e) {
      if (e.button !== 0) {
        return e;
      }
      this.isMouseDown = true;
      var a = {
        'type': e.type,
        'pageX': e.pageX,
        'pageY': e.pageY,
        'timeStamp': e.timeStamp
      };
      this._movestart(a);
      e.preventDefault();
    },
    /**
     * Mousemove event handler, passes necessary data points to the move function
     * @param {Object} e The mousemove event
     */
    _mousemove: function(e) {
      if (this.isMouseDown) {
        var a = {
          'type': e.type,
          'pageX': e.clientX,
          'pageY': e.clientY,
          'timeStamp': e.timeStamp,
          //this was changed to correct an issue in safari - it doesn't report cancelable correctly
          'cancelable': true
        };
        this._move(a);
        if (a.preventDefault) {
          e.preventDefault();
        }
      }
    },
    /**
     * Mouseup event handler, passes necessary data points to the moveend function
     * @param {Object} e The mouseup event
     */
    _mouseup: function(e) {
      if (this.moves && this.moves.length > 2) {
        var a = {
          'type': e.type,
          'pageX': e.pageX,
          'pageY': e.pageY,
          'timeStamp': e.timeStamp
        };
        this._moveend(a);
        this.mouseUpHandled = true;
      }
      else {
        delete this.moves;
        this._scrollToClicked(e);
      }
      this.isMouseDown = false;
    },
    /**
     * Click event handler
     * @param {Object} e The click event
     */
    _click: function(e) {
      //if we are already tracking moves, then this will be handled by the mouseend event handler and we should prevent the default action
      if (this.moves) {
        e.preventDefault();
      }
      //if it has already been handled by mouseup handler, prevent the action
      if (this.mouseUpHandled) {
        e.preventDefault();
      }
      //reset our handled state
      delete this.mouseUpHandled;
      //checking both this.moves and this.mouseUpHandled ensures we capture events correctly in all browsers, where the order of the mouseup/click events can vary
    },
    /**
     * Calculate the user's recent cursor/finger velocity
     * @param {Array} moves The array of cursor positions
     */
    _velocity: function(moves) {
      var avg = 0;
      var m = Math.min(6, moves.length - 1);
      for (var i = 1; i < m; i++) {
        if (moves[moves.length - i].timeStamp === moves[moves.length - i - 1].timeStamp) {
          avg += avg / i;
        }
        else {
          avg += (10 * (moves[moves.length - i].pageX - moves[moves.length - i - 1].pageX) / (moves[moves.length - i].timeStamp - moves[moves.length - i - 1].timeStamp)) / m;
        }
      }
      return avg;
    },
    /**
     * Handles click events on items and dots, scrolling to the clicked item
     * @param {Event} e The click event
     */
    _scrollToClicked: function(e) {
      var tar = e.target;
      if (this.el.contains(tar)) {
        while (!tar.sparkcarousel) {
          if (tar.sparkcarouselitem) {
            this.containerMask.scrollLeft = 0;
            delete this.moves;
            this._rafCancel();
            this._scrollTo(tar.sparkcarouselitem);
            e.preventDefault();
            break;
          }
          if (tar.sparkcarouselitemdot) {
            this.containerMask.scrollLeft = 0;
            delete this.moves;
            this._rafCancel();
            var v = tar.sparkcarouselitemdot.order < this._selectedItem().order ? this.opts.startingVelocity : -this.opts.startingVelocity;
            this._scrollTo(tar.sparkcarouselitemdot, v);
            e.preventDefault();
            break;
          }
          tar = tar.parentNode;
        }
      }
    },
    /**
     * Scroll to the carouselItem, with specified startingVelocity, auto determines default velocity if not specified
     * @param {Object} item The carouselItem to scroll to
     * @param {Number} startingVelocity The startingVelocity of the scroll animation
     */
    _scrollTo: function(item, startingVelocity) {
      var offset = this.width / 2;
      var currentPosition = item.currentPosition();
      if (!startingVelocity) {
        startingVelocity = offset - item.currentPosition() > 0 ? this.opts.startingVelocity : -this.opts.startingVelocity;
      }
      if (this.opts.wrapItems) {
        if (startingVelocity > 0) {
          //left
          if (currentPosition > offset) {
            this._totalDistance = offset + this.totalItemWidth - currentPosition;
          }
          else {
            this._totalDistance = offset - currentPosition;
          }
        }
        else {
          //right
          if (currentPosition < offset) {
            this._totalDistance = -(this.totalItemWidth + currentPosition - offset);
          }
          else {
            this._totalDistance = offset - currentPosition;
          }
        }
      }
      else {
        this._totalDistance = offset - currentPosition;
      }
      this._startingVelocity = startingVelocity;
      delete this.moves;
      this._scrollToItem = true;
      this._raf = window.requestAnimationFrame(this._rafHandler);
    },
    /**
     * This is the animator function, it examines the options set on the carousel object
     * and selectively adds transform and requests addtional animation frames if necesary
     * @param {Number} t The timestamp for the current animation frame
     */
    _rafHandlerH: function(t) {
      if (this.opts.autoAdvance && this.autoAdvance) {
        window.clearTimeout(this.autoAdvance);
        delete this.autoAdvance;
      }
      var frames;
      if (this.moves || !this._startingVelocity) {
        this._rafCancel();
        return;
      }
      if (!this._laststart) {
        this._laststart = t;
      }
      if (!this._remainingDistance) {
        this._remainingDistance = this._totalDistance;
      }
      if (!this._lastframe) {
        this._lastframe = t;
        frames = 1;
      }
      else {
        frames = (t - this._lastframe) / ((1 / 60) * 1000);
      }
      var d = this._startingVelocity * frames;
      if (this.opts.smoothScroll && !this._scrollToItem) {
        this._addTransform(d);
        this._startingVelocity *= Math.pow(0.97, frames);
        if (this.opts.smoothScrollCenterItems && Math.abs(this._startingVelocity) < 1) {
          this._scrollTo(this._selectedItem());
        }
        if (Math.abs(this._startingVelocity) < 0.5) {
          if ((this._startingVelocity > 0 && this.transform.x > ((this.totalItemWidth / 2) - (this.items[0].width / 2))) ||
            (this._startingVelocity < 0 && this.transform.x < (-((this.totalItemWidth / 2) - (this.items[this.items.length - 1].width / 2))))) {
            this._scrollToItem = true;
            this._scrollTo(this._selectedItem());
          }
          else {
            this._rafCancel();
          }
        }
        else {
          this._raf = window.requestAnimationFrame(this._rafHandler);
        }
      }
      else {
        if (this._startingVelocity > 0) {
          if (d < this._remainingDistance) {
            this._addTransform(d);
            this._remainingDistance -= d;
            if (this._remainingDistance > this._totalDistance / 2) {
              this._startingVelocity *= Math.pow(1.15, frames);
            }
            else {
              this._startingVelocity *= Math.pow(0.9, frames);
              this._startingVelocity = this._startingVelocity > 2 ? this._startingVelocity : 2;
            }
            this._raf = window.requestAnimationFrame(this._rafHandler);
          }
          else {
            this._addTransform(this._remainingDistance);
            this._rafCancel();
          }
        }
        else {
          if (d > this._remainingDistance) {
            this._addTransform(d);
            this._remainingDistance -= d;
            if (this._remainingDistance < this._totalDistance / 2) {
              this._startingVelocity *= Math.pow(1.15, frames);
            }
            else {
              this._startingVelocity *= Math.pow(0.9, frames);
              this._startingVelocity = this._startingVelocity < -2 ? this._startingVelocity : -2;
            }
            this._raf = window.requestAnimationFrame(this._rafHandler);
          }
          else {
            this._addTransform(this._remainingDistance);
            this._rafCancel();
          }
        }
      }
      this._lastframe = t;
    },
    /**
     * This is the animator clearing function
     * it clears values used during animation, and selectively enables autoAdvance
     */
    _rafCancel: function() {
      if (this.opts.autoAdvance && !this.autoAdvance && !this.pause) {
        this.autoAdvance = window.setTimeout(function() {
          this._autoAdvance();
        }.bind(this), this.opts.autoAdvance * 1000);
      }
      window.cancelAnimationFrame(this._raf);
      delete this._scrollToItem;
      delete this._laststart;
      delete this._startingVelocity;
      delete this._remainingDistance;
      delete this._totalDistance;
      delete this._lastframe;
    },
    /**
     * This computes values necessary to start an animation frame when the carousel is
     * configured to use smoothScroll
     * @param {Array} moves The captured move events
     */
    _interiaScroll: function(moves) {
      if (moves[moves.length - 1].timeStamp - moves[moves.length - 2].timeStamp > 100 || moves.length < 3) {
        if(this.opts.smoothScrollCenterItems) {
          return this._scrollTo(this._selectedItem());
        }
        return;
      }
      this._startingVelocity = this._velocity(moves);
      delete this.moves;
      this._raf = window.requestAnimationFrame(this._rafHandler);
    },
    /**
     * This determines which carousel item should be focused based on the previous moves
     * made by the user
     * @param {Array} moves The captured move events
     */
    _settle: function(moves) {
      if (moves && moves.length > 3) {
        if (moves[moves.length - 1].timeStamp - moves[moves.length - 2].timeStamp > 80) {
          return this._scrollTo(this._selectedItem());
        }
        var v1 = 10 * (moves[moves.length - 3].pageX - moves[moves.length - 4].pageX) / (moves[moves.length - 3].timeStamp - moves[moves.length - 4].timeStamp);
        var v2 = 10 * (moves[moves.length - 2].pageX - moves[moves.length - 3].pageX) / (moves[moves.length - 2].timeStamp - moves[moves.length - 3].timeStamp);
        if (Math.abs(v1) < Math.abs(v2) || Math.abs(v2) > 0.5 && Math.abs(v2) > 0.5) {
          //user is probably trying to go to next or prev item
          var s = this.items.indexOf(this._selectedItem());
          if (v2 > 0) {
            //prev
            if (s > 0) {
              this._scrollTo(this.items[s - 1], v2);
            }
            else {
              if (this.opts.wrapItems) {
                this._scrollTo(this.items[this.items.length - 1], v2);
              }
              else {
                this._scrollTo(this.items[0]);
              }
            }
          }
          else {
            //next
            if (s < this.items.length - 1) {
              this._scrollTo(this.items[s + 1], v2);
            }
            else {
              if (this.opts.wrapItems) {
                this._scrollTo(this.items[0], v2);
              }
              else {
                this._scrollTo(this.items[this.items.length - 1]);
              }
            }
          }
        }
        else {
          if (this._selectedItem().currentPosition() > this.width / 2) {
            this._scrollTo(this._selectedItem(), -this.opts.startingVelocity);
          }
          else {
            this._scrollTo(this._selectedItem(), this.opts.startingVelocity);
          }
        }
      }
    },
    /**
     * Transforms the position of all carouselItems
     * @param {Number} x The pixel value to transform
     */
    _transformItems: function(x) {
      for (var i = 0; i < this.items.length; i++) {
        this.items[i].addTransform(x);
      }
    },
    /**
     * Sets the transform position of all carouselItems
     * @param {Number} x The pixel value to transform
     */
    _setTransformItems: function(x) {
      for (var i = 0; i < this.items.length; i++) {
        this.items[i].setTransform(x);
      }
    },
    /**
     * Adds transform to the container element, does checking for bounds conditions and
     * wraps items if necessary and configured
     * @param {Number} x The pixel value to transform
     */
    _addTransform: function(x) {
      var a;
      if ((this.opts.smoothScrollCenterItems || !this.opts.smoothScroll) && !this.opts.wrapItems && !this.opts.edgeScroll) {
        var l = this.items.indexOf(this.selectedItem);
        if (l === this.items.length - 1) {
          this._leftbound(true);
        }
        else {
          this._leftbound(false);
        }
        if (l === 0) {
          this._rightbound(true);
        }
        else {
          this._rightbound(false);
        }
      }
      else {
        this._leftbound(false);
        this._rightbound(false);
      }
      if (this.transform.x + x < 0 && x < 0) {
        if (this.opts.wrapItems) {
          //wrap items until we have covered the visible area
          while (this.transform.x + x < -(this.totalItemWidth - this.width) / 2 && (this.totalItemWidth < this.width ? this.transform.x + x < -this.totalItemWidth / 2 : true)) {
            a = this.items.shift();
            this.items.push(a);
            a.addTransform(this.totalItemWidth);
            this._transformItems(-a.width);
            x += a.width;
          }
        }
        else {
          //there is a 1 pixel adjustment to account for some math rounding
          if (this.opts.edgeScroll && x < 0 && this.transform.x + x - 1 <= -(this.totalItemWidth - this.width) / 2) {
            this._leftbound(true);
            return this._setTransform(-(this.totalItemWidth - this.width) / 2);
          }
          //progressively reduce scrolling when no more items to the right
          if (x < 0 && (this.transform.x + x) < (-((this.totalItemWidth / 2) - (this.items[this.items.length - 1].width / 2)))) {
            x = x * (((this.totalItemWidth / 2) + (this.items[this.items.length - 1].width / 2) + (this.transform.x + x)) / this.items[this.items.length - 1].width);
            x = x > 0 ? 0 : x;
          }
        }
        return this._setTransform(this.transform.x + x);
      }
      else {
        if (this.transform.x + x > 0 && x > 0) {
          if (this.opts.wrapItems) {
            //wrap items until we have covered the visible area
            while (this.transform.x + x > -(this.width - this.totalItemWidth) / 2 && (this.totalItemWidth < this.width ? this.transform.x + x > this.totalItemWidth / 2 : true)) {
              a = this.items.pop();
              this.items.unshift(a);
              a.addTransform(-this.totalItemWidth);
              this._transformItems(a.width);
              x -= a.width;
            }
          }
          else {
            //there is a 1 pixel adjustment to account for some math rounding
            if (this.opts.edgeScroll && x > 0 && this.transform.x + x + 1 >= (this.totalItemWidth - this.width) / 2) {
              this._rightbound(true);
              return this._setTransform((this.totalItemWidth - this.width) / 2);
            }
            //progressively reduce scrolling when no more items to the left
            if (x > 0 && (this.transform.x + x) > ((this.totalItemWidth / 2) - (this.items[0].width / 2))) {
              x = x * (((this.totalItemWidth / 2) + (this.items[0].width / 2) - (this.transform.x + x)) / this.items[0].width);
              x = x < 0 ? 0 : x;
            }
          }
        }
        return this._setTransform(this.transform.x + x);
      }
    },
    /**
     * Sets the leftbound class
     * @param {Boolean} b Set or unset the leftbound class
     */
    _leftbound: function(b) {
      if (typeof b === 'undefined') {
        this.leftbound = typeof this.leftbound === 'undefined' ? false : this.leftbound;
        return this.leftbound;
      }
      else {
        if (b) {
          this._addClass(this.el, 'leftbound');
        }
        else {
          this._removeClass(this.el, 'leftbound');
        }
        this.leftbound = b;
        return this.leftbound;
      }
    },
    /**
     * Sets the rightbound class
     * @param {Boolean} b Set or unset the rightbound class
     */
    _rightbound: function(b) {
      if (typeof b === 'undefined') {
        this.rightbound = typeof this.rightbound === 'undefined' ? false : this.rightbound;
        return this.rightbound;
      }
      else {
        if (b) {
          this._addClass(this.el, 'rightbound');
        }
        else {
          this._removeClass(this.el, 'rightbound');
        }
        this.rightbound = b;
        return this.rightbound;
      }
    },
    /**
     * Updates the selected item, by seeing which item has its center closest
     * to the center of the carousel
     */
    _updateSelected: function() {
      var tar = this.width / 2;
      var i = -1;
      var a = 1;
      var b = 0;
      while (a > b) {
        i++;
        if (i > this.items.length - 2) {
          break;
        }
        a = Math.abs(tar - this.items[i].currentPosition());
        b = Math.abs(tar - this.items[i + 1].currentPosition());
      }
      return this._selectedItem(this.items[i]);
    },
    /**
     * Stores the selected item for the carousel, and updates the previously
     * selected item and newly selected item to have the correct states
     * Conditionally sets the leftbound/rightbound states depending on configuration
     * @param {Object} item Optional: the new item select, if omitted it will
     * return the currently selected item.
     */
    _selectedItem: function(item) {
      if (typeof item !== 'object') {
        if (this.selectedItem) {
          return this.selectedItem;
        }
        else {
          return this._updateSelected();
        }
      }
      else {
        if (this.selectedItem) {
          this.selectedItem.setSelected(false);
        }
        this.selectedItem = item;
        if ((this.opts.smoothScrollCenterItems || !this.opts.smoothScroll) && !this.opts.wrapItems && !this.opts.edgeScroll) {
          var l = this.items.indexOf(this.selectedItem);
          if (l === this.items.length - 1) {
            this._leftbound(true);
          }
          else {
            this._leftbound(false);
          }
          if (l === 0) {
            this._rightbound(true);
          }
          else {
            this._rightbound(false);
          }
        }
        this.selectedItem.setSelected(true);
      }
    },
    /**
     * Sets the transform for the carousel container
     * @param {Number} x The pixel value to transform
     */
    _setTransform: function(x) {
      x = x ? x : 0;
      this.transform = {
        'x': x
      };
      this.container.setAttribute('style', this._transform('translate3d', x + 'px, 0px, 0px'));
      this._updateSelected();
      return x;
    }
  };
  Base.exportjQuery(Carousel, 'Carousel');

  return Carousel;
}));

/**
 * # Scroll To Top
 * Provide a way to scroll back to the top of a component.
 *
 * @example
 * new ScrollToTop(el);
 *
 * @module components/scroll-to-top.js
 */

(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/debounce', 'spark/helpers/each'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/debounce'), require('../helpers/each'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.ScrollToTop = factory(root.Spark.BaseComponent, root.Spark.debounce, root.Spark.each);
  }
}(this, function(Base, debounce, each) {

  'use strict';

  var noop = function() {};
  var canObserve = typeof MutationObserver !== 'undefined' ? true : false;

  /**
   * ScrollToTop constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var ScrollToTop = function(el, params) {

    params = params || {};

    if (!el) {
      return;
    }

    this._setParams(this.defaults, true);
    this._cacheElements(el);
    this._setParams(params);
    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._cacheSizes();

    this._clearRecomputeDebounced = debounce(this._clearRecompute.bind(this), 50);
    this._recomputeDebounced = debounce(this._recompute.bind(this), 50);

    this._checkScrollPosition();
  };

  ScrollToTop.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _toggleClass: Base.toggleClass,
    _scrollWindowTo: Base.scrollWindowTo,
    _getElementOffset: Base.getElementOffset,
    remove: Base.remove,


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      containerEl: null,
      scrollToEl: null,
      _isVisible: false,
      _isAtBottom: false,
      _containerBottom: 0,
      _visibleThreshold: 0,
      _windowHeight: 0,
      _right: 0,
      _recomputeRun: false,
      _observer: null,
      _onClickBound: null,
      _onScrollBound: null,
      _onVisibleBound: null,
      _onResizeBound: null,
      _onMutateBound: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['containerEl', 'scrollToEl'],


    /**
     * Scroll to the top of the containing element.
     * @param {Function} callback
     */
    scrollToTop: function(callback) {
      this._scrollWindowTo(this.scrollToEl || 0, {
        callback: callback || noop
      });
    },


    /**
     * Store a reference to the element.
     * @param {Element} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.containerEl = this.containerEl || el.parentNode;
    },


    /**
     * Cache element sizes.
     */
    _cacheSizes: function() {

      var containerOffset = this._getElementOffset(this.containerEl).top;
      var windowWidth = document.body.clientWidth;
      var windowHeight = window.innerHeight;

      this._windowHeight = windowHeight;
      this._containerBottom = this.containerEl.offsetHeight + containerOffset;
      this._visibleThreshold = containerOffset + windowHeight * 1.5;

      // Reset button styles
      var isAtBottom = this._isAtBottom;
      this._isAtBottom = false;
      this._right = 0;
      this._updateBindings();

      // Default right position of the element plus the right edge of the container
      this._right = (windowWidth - this.el.offsetLeft - this.el.offsetWidth) + (windowWidth - this.containerEl.offsetWidth);
      this._isAtBottom = isAtBottom;
      this._updateBindings();
    },


    /**
     * Check the scroll position. If we're far enough from the top,
     * make visible. If the bottom of our container element is past the bottom
     * of the screen, make us fixed.
     */
    _checkScrollPosition: function() {

      var isVisible = this._isVisible;
      var isAtBottom = this._isAtBottom;
      var scrollTop = window.pageYOffset;
      var bottomThreshold = scrollTop + this._windowHeight;

      this._isVisible = scrollTop + this._windowHeight >= this._visibleThreshold ? true : false;
      this._isAtBottom = this._containerBottom <= bottomThreshold ? true : false;

      if (this._isAtBottom !== isAtBottom || this._isVisible !== isVisible) {
        this._updateBindings();
      }
    },


    /**
     * Update bindings.
     */
    _updateBindings: function() {
      this._toggleClass(this.el, 'visible', this._isVisible);
      this._toggleClass(this.el, 'at-bottom', this._isAtBottom);
      this.el.style.right = this._isAtBottom || !this._right ? '' : this._right + 'px';
    },


    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onClickBound = this._onClick.bind(this);
      this._onScrollBound = this._onScroll.bind(this);
      this._onResizeBound = this._onResize.bind(this);
      this._onVisibleBound = this._onVisible.bind(this);
      this._onMutateBound = this._onMutate.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {

      this.el.addEventListener('click', this._onClickBound);
      window.addEventListener('scroll', this._onScrollBound);
      window.addEventListener('orientationchange', this._onScrollBound);
      document.addEventListener('spark.visible-children', this._onVisibleBound, true);

      if (canObserve)
        this._addMutationObserver();
      else
        window.addEventListener('resize', this._onResizeBound, false);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {

      this.el.removeEventListener('click', this._onClickBound);

      window.removeEventListener('scroll', this._onScrollBound);
      window.removeEventListener('orientationchange', this._onScrollBound);
      document.removeEventListener('spark.visible-children', this._onVisibleBound, true);

      if (canObserve)
        this._removeMutationObserver();
      else
        window.removeEventListener('resize', this._onResizeBound);

    },


    /**
     * Setup a mutation observer to know when the DOM has changed so we can recache.
     */
    _addMutationObserver: function() {
      this._observer = new MutationObserver(this._onMutateBound);
      this._observer.observe(this.containerEl, {childList: true, attributes: true, characterData: true, subtree: true});
    },


    /**
     * Remove a mutation observer.
     */
    _removeMutationObserver: function() {
      if (this._observer) this._observer.disconnect();
    },


    /**
     * Recompute the position.
     */
    _recompute: function() {
      this._cacheSizes();
      this._checkScrollPosition();
    },


    /**
     * Recompute styles, but only so often.
     */
    _recomputeThrottled: function() {

      if (!this._recomputeRun) {
        this._recompute();
        this._recomputeRun = true;
      }

      this._clearRecomputeDebounced();
    },


    /**
     * Clear the recompute run state.
     */
    _clearRecompute: function() {
      this._recomputeRun = false;
    },


    /**
     * When the window is scrolled, compute the position of the scroll-to-top.
     * @param {Object} e
     */
    _onScroll: function() {
      this._recomputeThrottled();
      this._checkScrollPosition();
    },


    /**
     * When the window is resized, re-cache element sizes.
     * @param {Object} e
     */
    _onResize: function() {
      this._recomputeDebounced();
    },


    /**
     * When the button is clicked, scroll to the top.
     * @param {Object} e
     */
    _onClick: function() {
      this.scrollToTop();
    },


    /**
     * When a parent container shows its children and our element
     * is inside of it, resize
     * @param  {[type]} e [description]
     * @return {[type]}   [description]
     */
    _onVisible: function(e) {
      if(e.target.contains(this.el)) {
        window.setTimeout(function() {
          this._cacheSizes();
          this._checkScrollPosition();
        }.bind(this),0);
      }
    },


    /**
     * When the DOM changes, recache our values because we might be in the wrong spot.
     */
    _onMutate: function(mutations) {
      each(mutations, function(m) {
        if (m.target !== this.el) this._recomputeDebounced();
      }.bind(this));
    }
  };

  Base.exportjQuery(ScrollToTop, 'ScrollToTop');

  return ScrollToTop;
}));

/**
 * # Step Indicator
 * Set of indicators represent different steps
 *
 * @example
 * new StepIndicator(el, params);
 *
 * @module components/step-indicator.js
 */

(function(root, factory) {

  'use strict';

  if (typeof define === 'function' && define.amd) {
    define(['spark/components/base', 'spark/helpers/outer-height'], factory);
  } else if (typeof module === 'object' && module.exports) {
    module.exports = factory(require('./base'), require('../helpers/outer-height'));
  } else {
    root.Spark = root.Spark || {};
    root.Spark.StepIndicator = factory(root.Spark.BaseComponent, root.Spark.outerHeight);
  }
}(this, function(Base) {

  'use strict';

  var noop = function() {};

  /**
   * StepIndicator constructor.
   * @param {Element} el
   * @param {Object} params
   */
  var StepIndicator = function(el, params) {

    params = params || {};

    if (!el) {
      return;
    }
    this._setParams(this.defaults, true);
    this._setParams(params);
    this._cacheElements(el);
    this._bindEventListenerCallbacks();
    this._addEventListeners();
    this._calculateStyle();
  };

  StepIndicator.prototype = {

    /**
     * Include common functionality.
     */
    _setParams: Base.setParams,
    _addClass: Base.addClass,
    _removeClass: Base.removeClass,
    _hasClass: Base.hasClass,
    _toggleClass: Base.toggleClass,
    _getElementOffset: Base.getElementOffset,
    remove: Base.remove,


    /**
     * Default values for internal properties we will be setting.
     * These are set on each construction so we don't leak properties
     * into the prototype chain.
     * @type {Object}
     */
    defaults: {
      el: null,
      containerEl: null,
      header: null,
      subtitle: null,
      type: null,
      _body: null,
      _list: null,
      _listPaddingTotal: 0,
      _dropdownExpand: false,
      _dropdownHeader: null,
      _items: null,
      _onResizeBound: null,
      _onVisibleBound: null
    },


    /**
     * Whitelisted parameters which can be set on construction.
     * @type {Array}
     */
    _whitelistedParams: ['type', 'header', 'subtitle'],

    /**
     * Store a reference to the element.
     * @param {Element} el
     */
    _cacheElements: function(el) {
      this.el = el;
      this.containerEl = this.containerEl || el.parentNode;
      this._body = this.el.querySelector('.spark-step-indicator__body');
      this._list = this.el.querySelector('.spark-step-indicator__list');


      this.type = this.type !== null ? this.type : (this.el.attributes['data-type'] && this.el.attributes['data-type'].value);
      this.header = this.header !== null ? this.header : (this.el.attributes['data-header'] && true);
      this.subtitle = this.subtitle !== null ? this.subtitle : (this.el.attributes['data-subtitle'] && true);



      this._items = this.el.querySelectorAll('.spark-step-indicator__item');
    },

    /**
     * Create bound versions of event listener callbacks and store them.
     * Otherwise we can't unbind from these events later because the
     * function signatures won't match.
     */
    _bindEventListenerCallbacks: function() {
      this._onResizeBound = this._onResize.bind(this);
      this._onLoadBound = this._onLoad.bind(this);
      this._onVisibleBound = this._onVisible.bind(this);
      this._toggleDropdownBound = this._toggleDropdown.bind(this);
    },


    /**
     * Add event listeners for DOM events.
     */
    _addEventListeners: function() {
      window.addEventListener('resize', this._onResizeBound);
      document.addEventListener('spark.visible-children', this._onVisibleBound, true);
      window.addEventListener('load', this._onLoadBound);
    },


    /**
     * Remove event listeners for DOM events..
     */
    _removeEventListeners: function() {
      window.removeEventListener('resize', this._onResizeBound);
      document.removeEventListener('spark.visible-children', this._onVisibleBound, true);
      document.removeEventListener('click', this._toggleDropdownBound);
      window.removeEventListener('load', this._onLoadBound);
    },


    /**
     * Change selected step state into incomplete
     * @param {Int} index of selected step
     */
    _resetStepState: function(stepIndex) {
      if( stepIndex < 0 || stepIndex >= this._items.length ) {
        return;
      }
      this._removeClass(this._items[stepIndex], ['spark-step-indicator__item--completed', 'spark-step-indicator__item--current', 'spark-step-indicator__item--disabled']);
    },


    /**
     * Set selected step state: completed/current/disabled/incomplete
     * For current step, will remove current step from other steps
     * @param {Int} index of selected step
     * @param {String} state to be set
     */
    setStepState: function(stepIndex, state) {
      if( stepIndex < 0 || stepIndex >= this._items.length ) {
        return;
      }
      switch(state) {
        case 'completed':
          this._resetStepState(stepIndex);
          this._addClass(this._items[stepIndex], 'spark-step-indicator__item--completed');
          break;
        case 'current':
          var currentStep = this._list.querySelector('.spark-step-indicator__item--current');
          this._removeClass(currentStep, 'spark-step-indicator__item--current');
          this._addClass(this._items[stepIndex], 'spark-step-indicator__item--current');
          break;
        case 'disabled':
          this._resetStepState(stepIndex);
          this._addClass(this._items[stepIndex], 'spark-step-indicator__item--disabled');
          break;
        case 'incomplete':
          this._resetStepState(stepIndex);
          break;
      }
      this._calculateStyle();
    },


    /**
     * Public interface to let use manually update style of step indicator
     * including set different option
     */
    update: function() {
      this._cacheElements(this.el);
      this._calculateStyle();
    },

    /**
     * Provide a method to bind click callback function to certain step
     * @param {Array} a set of step indices represent the target of callback
     * @param {Function} callback function
     */
    bindStepClickCallback: function(indexArr, callback) {
      callback = callback || noop;
      for(var i=0; i < indexArr.length; i++) {
         this._items[indexArr[i]].addEventListener('click', callback);
      }
    },

    /**
     * Remove click callback from steps
     * @param {Array} a set of step indices represent the target of callback
     * @param {Function} callback function
     *
     */
    removeStepClickCallback: function(indexArr, callback) {
      callback = callback || noop;
      for(var i=0; i < indexArr.length; i++) {
         this._items[indexArr[i]].removeEventListener('click', callback);
      }
    },

    /**
     * Work for _calculateStyle.
     * Comparing the width of list and total items, including padding
     * Sometimes after calculating, even itemTotalWidth is bigger 1px than listWidth,
     * it also means items are fit in the list, not over width.
     * So add 2px to list width when comparing to avoid this.
     * @return {Boolean}
     */
    _isOverWidth: function() {
      var listWidth = this._list.offsetWidth;
      var itemTotalWidth = this._listPaddingTotal;
      for (var i = 0; i < this._items.length; i++) {
        itemTotalWidth += this._items[i].offsetWidth;
      }
      if( listWidth + 2 < itemTotalWidth ) {
        return true;
      }
      return false;
    },


    /**
     * Calcuate different style based settings
     */
    _calculateStyle: function() {
      if(!this.header) {
        var titleEl = this.el.querySelector('.spark-step-indicator__title');
        titleEl.style.display = 'none';
        this._addClass(this._body, 'spark-step-indicator__body--no-border');
      }
      if(!this.subtitle) {
        var subtitleEl = this.el.querySelector('.spark-step-indicator__subtitle');
        subtitleEl.style.display = 'none';
      }
      // Remove special DOM and Class of dropdown variation
      this._switchFromDropdown();
      switch(this.type) {
        case 'standard-dropdown':
          this._calculateRespStyle();
          break;
        case 'large':
          this._calculateLargeStyle();
          break;
        case 'condensed':
          this._calculateCondensedStyle();
          break;
        case 'dropdown':
          this._calculateDropdownStyle();
          break;
        default:
          this._calculateStandardStyle();
      }
      // Work for IE11, detect IE11 via userAgent
      // userAgent: Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv 11.0) like Gecko
      var isIE11 = !!(navigator.userAgent.match(/Trident/) && navigator.userAgent.match(/rv[ :]11/));
      if(isIE11) {
        this._addItemWidth();
      }
    },


    /**
     * For IE11 ONLY, Add width: 100% to work around IE11 bug
     * IE11 will have incorrect position with flex-grow max-width and justify-content when items hit max-width
     * Sometimes after calculating, even itemTotalWidth is less 1px than listWidth,
     * it also means items are fit in the list, not hit the max-width
     * So add 2px to list width when comparing to avoid this.
     */
    _addItemWidth: function() {
      if( !this._hasClass(this._list, 'spark-step-indicator__list--condensed') &&
         !this._hasClass(this._body, 'spark-step-indicator__body--dropdown') ) {
        var listWidth = this._list.offsetWidth;
        var itemTotalWidth = this._listPaddingTotal;
        for (var i = 0; i < this._items.length; i++) {
          itemTotalWidth += this._items[i].offsetWidth;
        }
        if(listWidth - itemTotalWidth > 2) {
          this._switchWidthForIe11(0, true);
        }
        else {
          this._switchWidthForIe11(0, false);
        }
      }
      else {
        this._switchWidthForIe11(0, false);
      }
    },


    /**
     * Switch spark-step-indicator__item style width: 100% to work around IE11 bug
     * @param {Number} start index of step
     * @param {Boolean} whether add width: 100%
     */
    _switchWidthForIe11: function(index, val) {
      for (var i = index; i < this._items.length; i++) {
        this._items[i].style.width = val ? '100%' : '';
      }
    },


    /**
     * Fix on large variation
     * Calculate node and list length remove connect line of each end
     * Sometimes after calculating, even itemTotalWidth is bigger 1px than listWidth,
     * it also means items are fit in the list, not over width.
     * So add 2px to list width when comparing to avoid this.
     * Dependency of _calculateCondensedStyle() and _calculateStandardStyle()
     *
     */
    _calculateLargeStyle: function() {
      this._list.style['flex-wrap'] = 'wrap';
      var listWidth = this._list.offsetWidth;
      var itemTotalWidth = this._listPaddingTotal;
      for (var i = 0; i < this._items.length; i++) {
        this._removeClass(this._items[i], 'noline');
        itemTotalWidth += this._items[i].offsetWidth;
        if( listWidth + 2 < itemTotalWidth && i > 0) {
          this._addClass(this._items[i-1], 'noline');
          itemTotalWidth = this._items[i].offsetWidth + this._listPaddingTotal;
        }
      }
    },


    /**
     * Fix on condensed variation
     * Same as fix large variation except CSS class
     */
    _calculateCondensedStyle: function(){
      this._addClass(this._list, 'spark-step-indicator__list--condensed');
      this._calculateLargeStyle();
    },

    /**
     * Fix on dropdown variation
     */
    _calculateDropdownStyle: function(){
      this._switchToDropdown();
    },


    /**
     * Change only between large and condensed variation
     * no dropdown variation
     */
    _calculateStandardStyle: function() {
      this._list.style.visibility = 'hidden';
      for (var i = 0; i < this._items.length; i++) {
        this._removeClass(this._items[i], 'noline');
      }
      this._removeClass(this._list, 'spark-step-indicator__list--condensed');
      if(this._isOverWidth()) {
        this._addClass(this._list, 'spark-step-indicator__list--condensed');
        this._calculateLargeStyle();
      }
      this._list.style.visibility = 'visible';
    },


    /**
     * Make list invisible and change it into large variation at first.
     * Switch variation from top to bottom to find the fit one
     */
    _calculateRespStyle: function() {
      this._list.style.visibility = 'hidden';
      this._removeClass(this._list, 'spark-step-indicator__list--condensed');
      if(this._isOverWidth()) {
        this._addClass(this._list, 'spark-step-indicator__list--condensed');
        if(this._isOverWidth()) {
          this._removeClass(this._list, 'spark-step-indicator__list--condensed');
          this._switchToDropdown();
        }
      }
      this._list.style.visibility = 'visible';
    },


    /**
     * Create header element for dropdown variation
     * Try to find the current step of indicator and show it on the header
     * if not found, show the first step
     * @return {Object} header element of dropdown variation
     */
    _createDropdownHeader: function() {
      var el = document.createElement('a');
      var currentStep = this._list.querySelector('.spark-step-indicator__item--current');
      // Can not find current step
      if(!currentStep) {
        currentStep = this._items[0];
      }
      el.innerHTML = currentStep.innerHTML;
      el.innerHTML += '<span class="spark-step-indicator__notice">Select a Step</span>';

      // Create label element like '2 of 7'
      var stepNotice = document.createElement('span');
      var currentIndex = Array.prototype.indexOf.call(this._items, currentStep);
      stepNotice.innerHTML = (currentIndex + 1) + ' of ' + this._items.length;
      stepNotice.className = 'spark-step-indicator__label--dropdown';

      el.appendChild(stepNotice);
      el.className = currentStep.className;
      this._addClass(el,'spark-step-indicator__item--dropdown__header');
      this._dropdownHeader = el;
    },


    /**
     * Some extra event binder and DOM of dropdown variation need to be set
     */
    _switchToDropdown: function() {
      this._addClass(this._body, 'spark-step-indicator__body--dropdown');
      this._createDropdownHeader();
      this._list.parentNode.insertBefore(this._dropdownHeader, this._list);
      document.removeEventListener('click', this._toggleDropdownBound);
      document.addEventListener('click', this._toggleDropdownBound);
      this._addClass(this._list, 'collapse');
    },


    /**
     * Remove extra event and DOM when switch variation from dropdown
     */
    _switchFromDropdown: function() {
      this._removeClass(this._body, 'spark-step-indicator__body--dropdown');
      var header = this._list.parentNode.querySelector('.spark-step-indicator__item--dropdown__header');
      document.removeEventListener('click', this._toggleDropdownBound);
      if(header) {
        this._list.parentNode.removeChild(header);
      }
      this._removeClass(this._list, 'collapse');
    },

    /**
     * Toggle collapse/expand state of step list in dropdown variation
     * If there is a scroll in dropdown variation, scroll to 'current' node
     * @param {Object} Reference of DOM obj
     * @param {Object} Event of click
     */
    _toggleDropdown: function(e) {
      var target = e.target || e.srcElement;
      if( this._dropdownHeader.contains(target) ) {
        if(!this._dropdownExpand) {
          this._removeClass(this._list, 'collapse');
          this._addClass(this._dropdownHeader , 'expand');
          this._dropdownExpand = true;
        }
        else {
          this._addClass(this._list, 'collapse');
          this._removeClass(this._dropdownHeader , 'expand');
          this._dropdownExpand = false;
        }
      }
      else {
        this._addClass(this._list, 'collapse');
        this._removeClass(this._dropdownHeader , 'expand');
        this._dropdownExpand = false;
      }
      if(this._list.offsetHeight > 0) {
        var scrollMove = 0;
        for( var i = 0; i < this._items.length; i ++) {
          if( this._hasClass(this._items[i], 'spark-step-indicator__item--current') ){
            this._list.scrollTop = scrollMove;
            break;
          }
          scrollMove += this._items[i].offsetHeight;
        }
      }
    },


    /**
     * When the window finish loading
     */
    _onLoad: function() {
      // Cache list left + right padding for width calculating
      var listStyles = getComputedStyle(this._list);
      var listPaddingLeft = parseInt(listStyles.getPropertyValue('padding-left'), 10);
      var listPaddingRight = parseInt(listStyles.getPropertyValue('padding-right'), 10);
      this._listPaddingTotal = listPaddingLeft + listPaddingRight;
      this._calculateStyle();
    },


    /**
     * When the window is resized, base on params make some reponsive change.
     */
    _onResize: function() {
      this._calculateStyle();
    },


    /**
     * When a parent container shows its children and our element
     * is inside of it, resize
     * @param  {[type]} e [description]
     * @return {[type]}   [description]
     */
    _onVisible: function(e) {
      if(e.target.contains(this.el)) {
        window.setTimeout(function() {
          this._calculateStyle();
        }.bind(this),0);
      }
    },
  };

  Base.exportjQuery(StepIndicator, 'StepIndicator');

  return StepIndicator;
}));
